{
 Example #2 - "A simple program playing music and effects."
 (c) Copyright 2000 Tecnic Software productions (TNSP)
 Programmed and designed by Matti "ccr" Hamalainen

 NOTICE! THIS EXAMPLE PROGRAM IS LICENSED UNDER
 "JSS EXAMPLE PROGRAM LICENSE" WHICH CAN BE FOUND FROM
 README FILE UNDER THIS (EXAMPLE PROGRAMS) DIRECTORY.

 DESCRIPTION:

 This example shows how you can
 1) Make your own functions/procs for JSS (load a sample :)
 2) How to play a module.
 3) How to allocate and access channels and play samples on them.

// Also a new procedure jssAutoInitDriver() is introduced here.
// It is used to do the sound driver initialization automatically,
// so we don't need to have any big case-construct of
// new(PJD_drivername, vmInit(aconfig)); things...


 The exitproc also shows how you do the cleanup of JSS
 (and libBaseAPI if necessary)
}
program Example2;

 uses Crt, BaseAPI, 
      JSS, JSSMod, JSSPlr, JSSSamp;

 var
    ASaveExitProc:Pointer;
    ADevice:PJSSSoundDevice;
    AModule:PJSSModule;
    APlayer:PJSSModulePlayer;
    ASamps:Array[1..3] of PJSSSample;
    AKey:Longint;
    f:Longint;


{------------------------------------------------------------------------------}

procedure vmExitProc;
 var AIndex:Longint;
begin
 {** Do some de-initializations **}
 write('o Disposing samples:');
 for AIndex:= 1 to vmMaxSamples do
 if (ASamps[AIndex] <> NIL) then
    begin
     dispose(ASamps[AIndex]);
     write('.');
    end;
 writeln;
 
 if (AModule <> NIL) then
    begin
     writeln('o Dispose module.');
     dispose(AModule, vmDispose);
     AModule:= NIL;
    end;

 if (APlayer <> NIL) then
    begin
     writeln('o Close player.');
     dispose(APlayer, mpClose);
     APlayer:= NIL;
    end;

 if (ADevice <> NIL) then
    begin
     writeln('o Close sound device.');
     dispose(ADevice, vmClose);
     ADevice:= NIL;
    end;

 writeln('o jssClose(): ', jssClose);

 { Closing libBaseAPI is not necessary, if it has not been initialized! So...
  
 if (ioClose <> ioerrOK) then
    writeln('o Error closing libBaseAPI!');
 }

 ExitProc:= ASaveExitProc;
end;


{------------------------------------------------------------------------------}

{ Here we add a little wrapper for JSS's sample-subsystem
  routines to save some code writing.
}
procedure vmLoadWAVSample(AFilename:String; var ASample:PJSSSample);
 var AS:PStream;
begin
 {** Open a buffered file-stream with 128kB buffer **}
 AS:= new(PBufferedFileStream, vmOpen(AFilename, ioOpen_RdOnly, 128*1024));
 if (AS = NIL) then
    ioInternalError('vmLoadWAVSample', 'Error opening stream for file "'+ AFilename +'".');

 {** Then call the JSS routine **}
 if (vmSampLoadWAV(AS, ASample) < 0) then
    begin
     dispose(AS, vmClose);
     ioInternalError('vmLoadWAVSample', 'Loading of "'+ AFilename +'" failed.');
    end;
     
 {** Finally, delete the stream **}
 dispose(AS, vmClose);
end;

{------------------------------------------------------------------------------}


begin
 {** First we do some inits **}
 ASaveExitProc:= ExitProc;
 ExitProc:= @vmExitProc;

 {** Then we show some nice messages **}
 writeln;
 writeln('Simple sample-playing test of J Sound System.');
 writeln('by ccr/TNSP in 2000 for JSS tutorial.');
 
 {** And now we load the samples **}
 writeln('* Loading samples...');
 vmLoadWAVSample('ccr_says.jss', ASamp1);
 vmLoadWAVSample('explosio.raw', ASamp2);
 vmLoadWAVSample('zap.raw', ASamp3);

 {** ..and the module **}
 writeln('* Loading "Millimooli" XM module...');
 jssLoadS3M('fg_milmo.xm', AModule);
 if (AModule = NIL) then
    ioInternalError(vmProgName, 'Error loading module!');

 {** Then we initialize the J Sound System **}
 writeln('* Initializing JSS...');

 { Load the configuration file }
 if (jssLoadConfig('jss.cfg', AConfig) <> jerrOK) then
    ioInternalError(vmProgName, 'Error loading JSS configuration file!');

 { Initialize JSS }
 if (jssInit <> jerrOK) then
    ioInternalError(vmProgName, 'Error initializing JSS!');

 { Here comes the NEW thing:
   You can either use this jssAutoInitDriver() or set up the driver by hand,
   like in the Example program #1...
 }  
 if (jssAutoInitDriver(AConfig, ADriver) <> jerrOK) then
    ioInternalError(vmProgName, 'Error auto-initializing JSS sound device driver');

 {** The main loop is here **}
 writeln;
 writeln('Use keys 1, 2 and 3 to play test-samples.');

 repeat
 until (vmQuitProg);
 
 writeln;
 writeln('Exit:');
end.
 
