{
  J Sound System - MIDAS look-a-like setup program
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  A setup program for JSS that looks somewhat like the widely
  known MIDAS "standard" setup.

  Not a total clone due to differences in JSS and MIDAS, but
  at least you can fool some people ;)
}
{$I ..\src\jssdef.pas}
program JSSSetup;

 uses BaseAPI, JSS;

{$I ..\src\jsscfg.pas}

{$I setup_i1.pas}

 const
      {** Yleisi vakioita **}
      SETUPMessage  = 'JSS Digital Audio System v'+ Version +' configuration';

      {** Menusystem definations **}
      MHelpColor    = $70;
      MColor        = $08;
      MTitleColor   = $5e;

      MTextColor    = $70;
      MBarColor     = $07;
      MDisTextColor = $78;
      MDisBarColor  = $08;

      ScrColor1     = (ioCLB_Blue + 3);
      ScrColor2     = $07;
      ScrChar       = #32;

      LToggle       = '@';

{$I setup_i2.pas}

      {** Menu definations **}
      MidasMainMenu:TMenu =
      (Title: 'Main Menu';
       Center: true;
       Items: 4;
       ItemEnabled: (true, true, true, true, true, true, true, true, true, true);
       ItemList: ('', '', '', '-- Accept', '','','','','','');

       ItemHelp: ('Select and setup sound device.',
                  'Select sound quality.',
                  'Select sound format settings.',
                  'Accept settings and exit/continue.',
                  '','','','','','')
       );

      MidasOutputModeMenu:TMenu =
      (Title: 'Select Output Mode';
       Center: true;
       Items: 4;
       ItemEnabled: (true, true, true, true, true, true, true, true, true, true);
       ItemList: ('8-bit mono', '8-bit stereo',
                  '16-bit mono', '16-bit stereo',
                  '','','','','','');

       ItemHelp: ('Select 8-bit, monophonic sound output. (Fastest)',
                  'Select 8-bit, stereophonic sound output.',
                  'Select 16-bit, monophonic sound output.',
                  'Select 16-bit, stereophonic sound output. (Slowest)',
                  '','','','','','')
       );


      MidasMixingModeMenu:TMenu =
      (Title: 'Select Mixing Mode';
       Center: true;
       Items: 2;
       ItemEnabled: (true, true, true, true, true, true, true, true, true, true);
       ItemList: ('Normal (low quality, faster)',
                  'Interpolating (high quality, slower)',
                  '','','','','','','','');
       ItemHelp: ('Select normal, non-interpolating sound mixing for faster mixing.',
                  'Select interpolating sound mixing for higher quality sound.',
                  '','','','','','','','')
       );


 var
    {** General variables **}
    setupConfig:TJSSConfigFile;
    setupSaveExitProc:Pointer;

    {** Windowing system variables **}
    win_x, win_y,
    win_x0, win_y0,
    win_x1, win_y1:Longint;


{****************************************************************************}
{*                                                                          *}
{* Sekalaisia yleisi funktioita                                            *}
{*                                                                          *}
{****************************************************************************}
procedure setupEBox(x0, y0, x1, y1:Longint; ATitle:String);
 var ATemp:Longint;
begin
 {** Draw the outer box **}
 ioRectFill(x0+1, y0+1, x1-1, y1-1, #32, $70);
 for ATemp:= x0 to x1 do
     begin
      ioPutR(ATemp, y0, '', $1f);
      ioPutR(ATemp, y1, '', $18);
     end;

 for ATemp:= (y0+1) to (y1-1) do
     begin
      ioPutR(x0, ATemp, '', $1f);
      ioPutR(x1, ATemp, '', $18);
     end;

 {** Draw the box title **}
 ATemp:= ((x1 - x0) shr 1) - (length(ATitle) shr 1);
 ioWriteCXY(x0 + ATemp, y0 + 1, ATitle);

 {** Draw the inner box **}
 for ATemp:= (x0+3) to (x1-3) do
     begin
      ioPutR(ATemp, y0+2, '', $78);
      ioPutR(ATemp, y1-2, '', $7f);
     end;

 for ATemp:= (y0+3) to (y1-3) do
     begin
      ioPutR(x0+2, ATemp, '', $78);
      ioPutR(x1-2, ATemp, '', $7f);
     end;

 ioPutR(x0+2, y0+2, '', $78);
 ioPutR(x1-2, y0+2, '', $7f);

 ioPutR(x0+2, y1-2, '', $78);
 ioPutR(x1-2, y1-2, '', $7f);

 ioFlush;
end;


procedure NextLine;
begin
 win_x:= 0;
 inc(win_y);
end;


procedure WriteWin(AStr:String);
 var AIndex:Longint;
begin
 {** Print line **}
 AIndex:= 1;
 repeat
   case AStr[AIndex] of
    LToggle: NextLine;

    else begin
          ioPutChar((win_x0 + 1 + win_x), (win_y0 + 1 + win_y), AStr[AIndex]);
          inc(win_x);
          if ((win_x0 + 1 + win_x) >= win_x1)
             then NextLine;
         end;
   end;
   inc(AIndex);
 until (AIndex > length(AStr)) or
       ((win_y0 + 1 + win_y) > win_y1);
end;


procedure setupWindow(x0, y0, x1, y1:Longint; ATitle:String);
begin
 win_x:= 0; win_y:= 0;
 win_x0:= x0; win_y0:= y0;
 win_x1:= x1; win_y1:= y1;
 setupEBox(x0-3, y0-3, x1+3, y1+3, ATitle);
end;


procedure setupStartAnimation;
 var ATemp, AIndex:Longint;
begin
 ioClrScr(ScrColor1);

 ATemp:= (ioScrWidth shr 1) - (length(SETUPMessage) shr 1);
 ioWriteCXY(ATemp, 0, SETUPMessage);
 for AIndex:= 1 to length(SetupMessage) do
     ioPutChar((ATemp + AIndex - 1), 1, '');

 ioFlush;
end;


{****************************************************************************}
{*                                                                          *}
{* Menujrjestelmn funktiot                                                *}
{*                                                                          *}
{****************************************************************************}
function setupExecuteMenu(AMenu:TMenu; AInitItem:Longint):Longint;
 var
    MXCoord, MYCoord,
    MWidth, MHeight:Longint;

    AOldItem,
    ACurrItem:Longint;
    AKey:Word;
    AEndMenu:Boolean;

procedure setupDrawMenu;
 var
    AItemStr:String;
    ATemp, AItem:Longint;
begin
 {** Draw the menu items **}
 for AItem:= 1 to AMenu.Items do
 begin
      {** Delete the old menubar **}
      if (AItem = AOldItem) then
      for ATemp:= 0 to MWidth do
          ioPutAttr(MXCoord + ATemp, (MYCoord + AItem - 1), MTextColor);

      {** Now draw the menuitem **}
      AItemStr:= AMenu.ItemList[AItem];
      if (AItemStr[1] = '~') then
      begin
       for ATemp:= 1 to (MWidth - 1) do
           ioPutChar(MXCoord + ATemp, (MYCoord + AItem - 1), '');
      end else
      begin
       {** Draw a menuitem **}
       if (AMenu.ItemEnabled[AItem])
          then ioWriteXY(MXCoord+1, MYCoord + AItem - 1, AItemStr, MTextColor)
          else ioWriteXY(MXCoord+1, MYCoord + AItem - 1, AItemStr, MDisTextColor);

       {** Draw menu selection bar **}
       if (AItem = ACurrItem) then
       if (AMenu.ItemEnabled[AItem]) then
       begin
        for ATemp:= 0 to MWidth do
            ioPutAttr(MXCoord + ATemp, (MYCoord + AItem - 1), MBarColor);
       end else
       begin
        for ATemp:= 0 to MWidth do
            ioPutAttr(MXCoord + ATemp, (MYCoord + AItem - 1), MDisBarColor);
       end;
      end;
 end;

 AOldItem:= ACurrItem;
 ioFlush;
end;


begin
 {** Initialize menu **}
 MXCoord := 9;
 MYCoord := 6;
 MWidth  := ioScrWidth - 19;
 MHeight := 12;

 setupEBox(MXCoord - 3, MYCoord - 3,
           MXCoord + MWidth + 3,
           MYCoord + MHeight + 3,
           AMenu.Title);

 ioWriteCenter(-1, ioScrHeight-2, 'Use arrows to select, Enter to confirm or Esc to exit', $13);
 ioFlush;

 {** Okay. Execute the menu **}
 AOldItem:= AInitItem;
 ACurrItem:= AInitItem;
 AEndMenu:= false;
 setupDrawMenu;

 repeat
  {** Wait for a key **}
  AKey:= ioGetKey;

  case AKey of
   kbUp   : begin
             repeat
              dec(ACurrItem);
              if (ACurrItem < 1) then
                 ACurrItem:= AMenu.Items;
             until (AMenu.ItemList[ACurrItem] <> '~');
             setupDrawMenu;
            end;

   kbDown : begin
             repeat
              inc(ACurrItem);
              if (ACurrItem > AMenu.Items) then
                 ACurrItem:= 1;
             until (AMenu.ItemList[ACurrItem] <> '~');
             setupDrawMenu;
            end;

   kbHome: begin
            ACurrItem:= 1;
            while (AMenu.ItemList[ACurrItem] = '~') do
            begin
             inc(ACurrItem);
             if (ACurrItem > AMenu.Items) then
                ACurrItem:= 1;
            end;
            setupDrawMenu;
           end;

   kbEnd: begin
            ACurrItem:= AMenu.Items;
            while (AMenu.ItemList[ACurrItem] = '~') do
            begin
             dec(ACurrItem);
             if (ACurrItem < 1) then
                ACurrItem:= AMenu.Items;
            end;
            setupDrawMenu;
           end;

   kbEnter: if (AMenu.ItemList[ACurrItem][1] <> '*') and
               (AMenu.ItemEnabled[ACurrItem])
               then AEndMenu:= true;

   kbEsc  : begin
             ACurrItem:= -1;
             AEndMenu:= true;
            end;
  end;
 until (AEndMenu);

 setupExecuteMenu:= ACurrItem;
end;


{****************************************************************************}
{*                                                                          *}
{* Device settings menu routines                                            *}
{*                                                                          *}
{****************************************************************************}
function setupGetDMA8:Boolean;
 var
    AMenu:TMenu;
    AValue:Word;
    AOption:Longint;
begin
 {** Initialize menu **}
 with AMenu do
 begin
  Title  := 'Select LOW 8-bit DMA';
  Items  := SDMA8Items;

  for AOption:= 1 to Items do
      begin
       AValue:= SDMA8List[AOption];
       ItemList[AOption]:= 'DMA '+ ioStrW(AValue);
       ItemHelp[AOption]:= 'Set LOW 8-bit DMA channel to '+ ioStrW(AValue);
       ItemEnabled[AOption]:= TRUE;
      end;
 end;

 {** Execute the menu **}
 setupGetDMA8:= true;
 AOption:= setupExecuteMenu(AMenu, 1);
 if (AOption <> -1)
    then setupConfig.SlDMA:= SDMA8List[AOption]
    else setupGetDMA8:= false;
end;


function setupGetDMA16:Boolean;
 var
    AMenu:TMenu;
    AValue:Word;
    AOption:Longint;
begin
 {** Initialize menu **}
 with AMenu do
 begin
  Title  := 'Select HIGH 16-bit DMA';
  Items  := SDMA16Items;

  for AOption:= 1 to Items do
      begin
       AValue:= SDMA16List[AOption];
       ItemList[AOption]:= 'DMA '+ ioStrW(AValue);
       ItemHelp[AOption]:= 'Set HIGH 16-bit DMA channel to '+ ioStrW(AValue);
       ItemEnabled[AOption]:= TRUE;
      end;
 end;

 {** Execute the menu **}
 setupGetDMA16:= true;
 AOption:= setupExecuteMenu(AMenu, 1);
 if (AOption <> -1)
    then setupConfig.ShDMA:= SDMA16List[AOption]
    else setupGetDMA16:= false;
end;


function setupGetDMA:Boolean;
 var
    AMenu:TMenu;
    AValue:Word;
    AOption:Longint;
begin
 {** Initialize menu **}
 with AMenu do
 begin
  Title  := 'Select DMA';
  Items  := SDMAItems;

  for AOption:= 1 to Items do
      begin
       AValue:= SDMAList[AOption];
       ItemList[AOption]:= 'DMA '+ ioStrW(AValue);
       ItemHelp[AOption]:= 'Set DMA channel to '+ ioStrW(AValue);
       ItemEnabled[AOption]:= TRUE;
      end;
 end;

 {** Execute the menu **}
 setupGetDMA:= true;
 AOption:= setupExecuteMenu(AMenu, 1);
 if (AOption <> -1) then
    begin
     setupConfig.ShDMA:= SDMAList[AOption];
     setupConfig.SlDMA:= setupConfig.ShDMA;
    end else
     setupGetDMA:= false;
end;


function setupGetBase:Boolean;
 var
    AMenu:TMenu;
    AValue:Word;
    AOption:Longint;
begin
 {** Initialize menu **}
 with AMenu do
 begin
  Title  := 'Select base I/O port';
  Items  := SBaseItems;

  for AOption:= 1 to Items do
      begin
       AValue:= SBaseList[AOption];
       ItemList[AOption]:= ioHexW(AValue) +'H';
       ItemHelp[AOption]:= 'Set base I/O port to '+ ioHexW(AValue) +'H';
       ItemEnabled[AOption]:= TRUE;
      end;
 end;

 {** Execute the menu **}
 setupGetBase:= true;
 AOption:= setupExecuteMenu(AMenu, 1);
 if (AOption <> -1)
    then setupConfig.SBase:= SBaseList[AOption]
    else setupGetBase:= false;
end;


function setupGetIRQ:Boolean;
 var
    AMenu:TMenu;
    AValue:Word;
    AOption:Longint;
begin
 {** Initialize menu **}
 with AMenu do
 begin
  Title  := 'Select IRQ-line';
  Items  := SIRQItems;

  for AOption:= 1 to Items do
      begin
       AValue:= SIRQList[AOption];
       ItemList[AOption]:= 'IRQ '+ ioStrW(AValue);
       ItemHelp[AOption]:= 'Set IRQ-line to '+ ioStrW(AValue);
       ItemEnabled[AOption]:= TRUE;
      end;
 end;

 {** Execute the menu **}
 setupGetIRQ:= true;
 AOption:= setupExecuteMenu(AMenu, 1);
 if (AOption <> -1)
    then setupConfig.SIRQ:= SIRQList[AOption]
    else setupGetIRQ:= false;
end;


function setupGetMixingRate:Boolean;
 var
    AMenu:TMenu;
    AValue:Word;
    AOption:Longint;
begin
 {** Initialize menu **}
 with AMenu do
 begin
  Title  := 'Select mixing rate';
  Items  := SMixSpeedItems;

  for AOption:= 1 to Items do
      begin
       AValue:= SMixSpeedList[AOption];
       ItemList[AOption]:= ioStrW(AValue) +' Hz';
       ItemHelp[AOption]:= 'Set mixing frequency to '+ ioStrL(AValue)
       +' Hz ('+ ioStrL(AValue div 1000) +' kHz)';
       ItemEnabled[AOption]:= TRUE;
      end;
 end;

 {** Execute the menu **}
 setupGetMixingRate:= true;
 AOption:= setupExecuteMenu(AMenu, 1);
 if (AOption <> -1)
    then setupConfig.SOutFreq:= SMixSpeedList[AOption]
    else setupGetMixingRate:= false;
end;


function setupGetMixingMode:Boolean;
 var AOption:Longint;
begin
 setupGetMixingMode:= true;
 AOption:= setupExecuteMenu(MidasMixingModeMenu, 1);
 case AOption of
  1: setupConfig.SMixInterpolate:= FALSE;
  2: setupConfig.SMixInterpolate:= TRUE;
  else setupGetMixingMode:= false;
 end;
end;


function setupGetOutputMode:Boolean;
 var AOption:Longint;
begin
 setupGetOutputMode:= true;
 AOption:= setupExecuteMenu(MidasOutputModeMenu, 1);
 case AOption of
  1,2: begin
        setupConfig.SOutMode16bit:= FALSE;
        setupConfig.SOutModeStereo:= (AOption = 2);
       end;
  3,4: begin
        setupConfig.SOutMode16bit:= TRUE;
        setupConfig.SOutModeStereo:= (AOption = 4);
       end;
  else setupGetOutputMode:= false;
 end;
end;


{****************************************************************************}
{*                                                                          *}
{* Setup Main program                                                       *}
{*                                                                          *}
{****************************************************************************}
function setupGetDeviceName:String;
 var AStr:String;
begin
 case setupConfig.SDevice of
  sdSoundBlaster      : AStr:= 'Sound Blaster';
  sdGravisUltrasound  : AStr:= 'Gravis Ultrasound';
  sdProAudioSpectrum  : AStr:= 'Pro Audio Spectrum';
  sdWindowsSoundSystem: AStr:= 'Windows Sound System';
  sdPCSpeaker         : AStr:= 'PC Speaker';
  sdDiskWriter        : AStr:= 'Disk Writer';
  sdOSS		      : AStr:= 'Open Sound System/UNIX';
  sdNoSound           : AStr:= 'No Sound';
  else                  AStr:= 'INTERNAL ERROR!';
 end;

 setupGetDeviceName:= AStr;
end;

function setupGetOutputModeName:String;
 var AStr:String;
begin
 with setupConfig do
 begin
  if (SOutMode16bit)
     then AStr:= '16-bit '
     else AStr:= '8-bit ';

  if (SOutModeStereo)
     then AStr:= AStr +'stereo'
     else AStr:= AStr +'mono';
 end;
 setupGetOutputModeName:= AStr;
end;


procedure setupSoundDevice;
 var AOption:Longint;
begin
 {** Select the sound device **}
 AOption:= setupExecuteMenu(DeviceMenu, 1);
 if (AOption = -1) then exit;
 setupConfig.SDevice:= DeviceMenuList[AOption];

 {** Setup the device **}
 case setupConfig.SDevice of
  sdSoundBlaster, sdProAudioSpectrum, sdWindowsSoundSystem:
    begin
     if (setupGetBase = FALSE) then exit;
     if (setupGetIRQ = FALSE) then exit;
     if (setupGetDMA8 = FALSE) then exit;
     if (setupGetDMA16 = FALSE) then exit;
    end;

 sdGravisUltrasound:
    begin
     if (setupGetBase = FALSE) then exit;
     if (setupGetIRQ = FALSE) then exit;
     if (setupGetDMA = FALSE) then exit;
    end;
 end;

 setupConfig.SDevice:= DeviceMenuList[AOption];
end;


{****************************************************************************}
{*                                                                          *}
{* Setup Main program                                                       *}
{*                                                                          *}
{****************************************************************************}
 var
    setupConfFilename:String;
    setupExitPrg:Boolean;
    setupOption:Longint;

function setupLoadConfig:Boolean;
 var AConfigF:File;
begin
 {** Open the file **}
 setupLoadConfig:= false;
 Assign(AConfigF, setupConfFilename);
 Reset(AConfigF, 1);
 if (IOResult <> 0) then exit;

 {** Read the configuration **}
 blockread(AConfigF, setupConfig, sizeof(TJSSConfigFile));
 if (IOResult <> 0) then exit;
 close(AConfigF);

 setupLoadConfig:= true;
end;


function setupSaveConfig:Boolean;
 var AConfigF:File;
begin
 {** Create the file **}
 setupSaveConfig:= false;
 Assign(AConfigF, setupConfFilename);
 Rewrite(AConfigF, 1);
 if (IOResult <> 0) then exit;

 {** Read the configuration **}
 blockwrite(AConfigF, setupConfig, sizeof(TJSSConfigFile));
 if (IOResult <> 0) then exit;
 close(AConfigF);

 setupSaveConfig:= true;
end;


procedure setupResetConfig;
begin
 with setupConfig do
 begin
  {** Initialize header **}
  AHeader:= jssCfgHeader;
  AVersion:= jssCfgVersion;

  {** Initialize information **}
  SDevice:= sdNoSound;

  SBase:= -1;
  SIRQ:= -1;
  SlDMA:= -1;
  ShDMA:= -1;

  SOutFreq:= 0;
  SOutMode16bit:= FALSE;
  SOutModeStereo:= TRUE;
  SMixInterpolate:= FALSE;
 end;
end;


procedure setupExitProc;
begin
 {** Exit the program with a neat message ;) **}
 ioClrScr(ioCLB_Black + ioCL_Gray);
 ioGotoXY(0, 1);
 ioFlush;

 if (ioClose <> ioerrOK) then
    writeln(ProgName+': Error closing libBaseAPI!');

 ExitProc:= setupSaveExitProc;
end;


begin
 {**** Yleinen initialisaatio ****}
 setupSaveExitProc:= ExitProc;
 ExitProc:= @setupExitProc;
 writeln;
 writeln(SETUPMessage);
 writeln(ioCopyright);

 if (ioInit <> ioerrOK) then
    begin
     writeln('libBaseAPI initialization failed!');
     halt;
    end;

 setupConfFilename:= SETUPFilename;

 {** Load configuration or defaults **}
 if (setupLoadConfig = false)
    then setupResetConfig;

 {**** SETUP Main Program loop ****}
 setupOption:= 1;
 setupExitPrg:= false;
 setupStartAnimation;
 repeat
  {** Update the main menu **}
  with MidasMainMenu do
  begin
   ItemList[1]:= 'Sound Card: '+ setupGetDeviceName;
   ItemList[2]:= 'Sound quality: High';
   ItemList[3]:= 'Output mode: '+ setupGetOutputModeName;
  end;

  {** Execute the main menu **}
  setupOption:= setupExecuteMenu(MidasMainMenu, setupOption);
  case setupOption of
    1: setupSoundDevice;
    2: begin
        setupGetMixingRate;
        setupGetMixingMode;
       end;
    3: setupGetOutputMode;
   -1, 4: setupExitPrg:= true;
  end;
 until (setupExitPrg);

 {** Save the configuration file **}
 if (setupOption <> -1) then
 begin
  setupStartAnimation;
  setupWindow(9, 5, ioScrWidth-18, 10, 'Saving settings');
  if (setupSaveConfig)
     then WriteWin('Settings saved to file@')
     else WriteWin('Error saving settings to file@');

  WriteWin('"'+ setupConfFilename +'"@');
  WriteWin('<<Hit any key to quit>>');
  ioFlush;
  ioGetKey;
 end;

end.
