{
  J Sound System - Driver for Sound Blaster series
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  JSS driver for Sound Blaster series and "100% compatibles".

  TO DO:
  * Fix the general stuffos
  * Add autoinit DMA mode support
}

 const
     {** SB I/O port defines **}
     SB_DSP_RESET       = $06;
     SB_DSP_READ        = $0A;
     SB_DSP_WRITE       = $0C;
     SB_DSP_COMMAND     = $0C;
     SB_DSP_STATUS      = $0C;
     SB_DSP_DATA_AVAIL  = $0E;
     SB_DSP_IRQACK8     = $0E;
     SB_DSP_IRQACK16    = $0F;

     {** SB's DSP Commands **}
     SB_DSPCMD_SPKON    = $D1;
     SB_DSPCMD_SPKOFF   = $D3;
     SB_DSPCMD_DMAON    = $D0;
     SB_DSPCMD_DMAOFF   = $D4;
     SB_DSPCMD_GETVER   = $E1;
     SB_DSPCMD_GETCSTR  = $E3;


procedure TJD_SoundBlaster.vmSB_IRQHandler_Pascal;
 var ATempP:Pointer;
begin
 {** Now do the magic: swap pointers & set new DMA block **}
 ATempP:= SMixBuffer1;
 SMixBuffer1:= SMixBuffer2;
 SMixBuffer2:= ATempP;

 {** Set the new DMA-block **}
 vmDMA_SetData(SDMixBuffer, SMixBufSize);
 vmSB_SetDSP;

 {** Mix the audio **}
{$ifdef sexy_girls_on_the_house}
 { !! FIX ME !! We should do a custom optimized routine for this }
 DOSMemPut(SDMixBuffer.ARealSeg, 0, SMixBuffer1^, SMixBufSize);
{$endif}

 {** Notify SB's DSP about the irq **}
 if (SOutMode16bit)
    then inportb(SBase + SB_DSP_IRQACK16)
    else inportb(SBase + SB_DSP_IRQACK8);
end;


function TJD_SoundBlaster.vmSB_DSPRead:Byte;
 var ACounter:Longint;
begin
 {** Wait until we can read **}
{ FIXME FIX ME: We should use a system timer here to maintain
  compability with future 3GHz computers ;-)
}
 ACounter:= 200000;
 while ((inportb(SBase + SB_DSP_DATA_AVAIL) AND $80) = 0) and
       (ACounter > 0) do
       dec(ACounter);

 {** Read the data **}
 vmSB_DSPRead:= inportb(SBase + SB_DSP_READ);
end;


procedure TJD_SoundBlaster.vmSB_DSPWrite(AData:Byte);
 var ACounter:Longint;
begin
 {** Wait until we can write **}
{ FIXME FIX ME: We should use a system timer here to maintain
  compability with future 3GHz computers ;-)
}
 ACounter:= 200000;
 while ((inportb(SBase + SB_DSP_WRITE) AND $80) <> 0) and
       (ACounter > 0) do
       dec(ACounter);

 {** Clear to write... **}
 outportb(SBase + SB_DSP_WRITE, AData);
end;


function  TJD_SoundBlaster.vmSB_DetectDSP:Boolean;
begin
 { To detect the presence of SB compatible DSP chip,
   we just simply perform a DSP reset operation and
   check if we get the $AA response byte.
   If not, then we don't (most possibly) have a SB.
 }

 {** #1: Write $01 to [SBase + DSP_RESET] **}
 outportb(SBase + SB_DSP_RESET, $01);

 {** #2: Wait > ~3.3 microseconds **}
{ FIXME FIX ME: We should use a system timer here to maintain
  compability with future 3GHz computers ;-)
}
 asm
        mov ecx, $10000;
@vlp:   dec ecx;
        jnz @vlp;
 end;

 {** #3: Write $00 to [SBase + DSP_RESET] **}
 outportb(SBase + SB_DSP_RESET, $00);

 {** #4: Read the response byte $AA? *}
 vmSB_DetectDSP:= (vmSB_DSPRead = $AA);
end;


procedure TJD_SoundBlaster.vmSB_ResetDSP;
 var ACounter:Longint;
begin
 { We just call vmSB_DetectDSP, as it performs DSP reset.
   Then we discard the resulting 'was-dsp-detected' value. }

 {** #1: Write "1" to Reset Port. **}
 outportb(SBase + SB_DSP_RESET, 1);

 {** #2: Wait > ~3.3 microseconds **}
{ FIXME FIX ME: We should use a system timer here to maintain
  compability with future 3GHz computers ;-)
}
 asm
        mov ecx, $ffffff;
@vlp:   dec ecx;
        jnz @vlp;
 end;


 {** #3: Write "0" to Reset port **}
 outportb(SBase + SB_DSP_RESET, 0);

 {** #4: Check that it is O.K. **}
 vmSB_DSPRead;
end;



function TJD_SoundBlaster.vmSB_GetDSPVersion:Word;
 var AResult:Word; ATempB:Byte;
begin
 {** Now start the version detection **}
 { DSP-command: Get DSP version }
 vmSB_DSPWrite(SB_DSPCMD_GETVER);

 { Get LO-byte of version number }
 ATempB:= vmSB_DSPRead;
 AResult:= (ATempB SHL 8);

 { Get HI-byte of version number }
 ATempB:= vmSB_DSPRead;
 AResult:= (AResult OR ATempB);

 vmSB_GetDSPVersion:= AResult;
end;


function TJD_SoundBlaster.vmSB_GetCopyrightStr:String;
 var
    AIndex:Longint;
    AData:Byte;
    AResult:String;
begin
 {** Get copyright string **}
 vmSB_DSPWrite(SB_DSPCMD_GETCSTR);

 AIndex:= 1;
 repeat
  AData:= vmSB_DSPRead;
  AResult[AIndex]:= char(AData);
  inc(AIndex);
 until (AData = $00) or (AIndex >= 253);

 AResult[0]:= char(AIndex - 2);

 vmSB_GetCopyrightStr:= AResult;
end;


procedure TJD_SoundBlaster.vmSB_SetSampleRate;
begin
 if (SB_DSPVersion < $0400) then
 begin
  {** DSP v3.x and older use the 'TimeC' styled freq setting **}
  { The actual formula for this is:

    TimeConstant = 256 - (1000000 / (SampleChannels * SampleRate))

    It has been checked from several documents like the official
    SB programmer's guide from Creative, "Sound Blaster
    programming information v0.90" by Andr Baresel and Craig Jackson,
    and I also checked it against OSS/Free SB driver's source.
  }
  if (SOutModeStereo)
     then ATimeC:= (256 - (1000000 DIV (SOutFreq*2)) )
     else ATimeC:= (256 - (1000000 DIV SOutFreq) );

  vmSB_DSPWrite($40);
  vmSB_DSPWrite(ATimeC);
 end else
 begin
  {** DSP v4.x and newer support direct freq setting **}
  { And is much easier to program anyway :) -ccr }
  vmSB_DSPWrite($41);
  vmSB_DSPWrite(HI(word(SOutFreq)));
  vmSB_DSPWrite(LO(word(SOutFreq)));
 end;
end;


procedure TJD_SoundBlaster.vmSB_SetDSP;
 var ATimeC:Byte; ADataSize:Word;
begin
 ADataSize:= SMixBufSize;
 ADataSize:= (ADataSize - 1);

 if (SB_DSPVersion < $0400) then
 begin
  {** Set the DSP transfer mode and size **}
  if (SOutFreq < 22050) then
     begin
      { Output 8-bit sound, LOW speed, single-cycle }
      vmSB_DSPWrite($14);
      vmSB_DSPWrite(LO(ADataSize));
      vmSB_DSPWrite(HI(ADataSize));
     end else
     begin
      { Output 8-bit sound, HIGH speed, single-cycle }
      vmSB_DSPWrite($48);
      vmSB_DSPWrite(LO(ADataSize));
      vmSB_DSPWrite(HI(ADataSize));
      vmSB_DSPWrite($91);
     end;
 end else
 begin
  {** Set the DSP transfer mode and size **}
  { Set command: Output 8bit sound, single-cycle }
  vmSB_DSPWrite($c0);

  { Set mode: Unsigned stereo/mono }
  if (SOutModeStereo)
     then vmSB_DSPWrite($20)
     else vmSB_DSPWrite($00);

  vmSB_DSPWrite(LO(ADataSize));
  vmSB_DSPWrite(HI(ADataSize));
 end;
end;


procedure TJD_SoundBlaster.vmSB_SetSpeaker(AState:Boolean);
begin
 if (AState) then
    begin
     {** Set the speaker on! **}
     vmSB_DSPWrite(SB_DSPCMD_SPKON);

     if (SB_DSPVersion >= $0300) then
        begin
         {** Save the mixer settings **}
         outportb(SBase + 4, $22);
         SB_Saved_MasterVol:= inportb(SBase + 5);
         outportb(SBase + 4, 4);
         SB_Saved_PCMVol:= inportb(SBase + 5);
        end;
    end else
    begin
     {** Set the speaker off! **}
     vmSB_DSPWrite(SB_DSPCMD_SPKOFF);

     if (SB_DSPVersion >= $0300) then
        begin
         {** Restore the mixer settings **}
         outportb(SBase + 4, $22);  { Restore the master volume }
         outportb(SBase + 5, SB_Saved_MasterVol);
         outportb(SBase + 4, 4);    { Restore DAC (PCM volume level) }
         outportb(SBase + 5, SB_Saved_PCMVol);
        end;
    end;
end;


constructor TJD_SoundBlaster.vmInit(AConfig:TJSSConfigFile);
 var
    AStr:String;
    ATempB1, ATempB2:Byte;
    ATempPP:Longint;
begin
 {** Call inherited vmInit **}
 jssDev(jProcName_SB, jDevName_SB, jDevVer_SB);
 inherited vmInit(AConfig);

 {** Detect Sound Blaster **}
 jssMsgS(jProcName_SB, 'Detecting DSP: ');
 vmSB_ResetDSP;
 if (vmSB_DetectDSP)
    then jssMsgP('OK')
    else jssMsgP('FAILED! (We still continue for some bad clones)');

 {**** Init Sound Blaster ****}
 SB_DSPVersion:= vmSB_GetDSPVersion;
 if (SB_DSPVersion = 0) then
    begin
     vmSB_ResetDSP;
     vmSB_GetDSPVersion;
     if (SB_DSPVersion = 0) then
        jssError(jProcName_SB, 'Could not get SB''s DSP version. Maybe a I/O conflict?');
    end;

 {** Determine SB model from DSP version **}
 case HI(SB_DSPVersion) of
  1: { SB v1.0 or v1.5 }
     AStr:= 'SB Basic';

  2: { SB v2.x }
     if (SB_DSPVersion > $0200)
        then AStr:= 'SB 2.01'
        else AStr:= 'SB 2.0';

  3: { SB Pro and most of SB "compatible" clones }
     AStr:= 'SB Pro/clone';

  4: { SB16, SB32, AWE, etc. }
     case LO(SB_DSPVersion) of
      0..11:  AStr:= 'SB16';
      12..13: AStr:= 'SB32/AWE32';
      else    AStr:= 'SB16 compatible?';
     end;

  else AStr:= 'Unknown SB DSP';
 end;


 {** Show information **}
 jssMsg(jProcName_SB,
        AStr + ', DSP v'+ ioStrW(HI(SB_DSPVersion)) +'.'+ ioStrW0(LO(SB_DSPVersion),2)+
        ' at port=0x'+ ioHexW(SBase)+
        ', IRQ='+ ioStrW(SIRQ)+
        ', DMA='+ ioStrW(SDMA));


 {** Check DSP version specific stuff **}
 { With DSP > 2.0 we can get some additional manufacturer info: }
 if (SB_DSPVersion > $0200) then
    jssMsg(jProcName_SB, 'DSP info "'+ vmSB_GetCopyrightStr +'"');

 if (SOutMode16bit) and (SB_DSPVersion < $0400) then
    begin
     jssWarning(jProcName_SB, 'Cannot use 16-bit mode with SB DSP version older than 4.00');
     SOutMode16bit:= FALSE
    end;

 if (SOutFreq > 22050) then
 begin
  if (SB_DSPVersion <= $0200) then
     begin
      jssWarning(jProcName_SB, 'Too big mixing frequency on SB with DSP version < 2.0!');
      SOutFreq:= 22050;
     end;

  if (SB_DSPVersion < $0400) and
     (SOutModeStereo) then
     begin
      jssWarning(jProcName_SB, 'Too big mixing frequency with Stereo set on with DSP < 4.0!');
      SOutModeStereo:= FALSE;
     end;
 end;


 {** Re-init mixing routine tables, etc. **}
 vmDeallocMixBuffers;
 vmAllocMixBuffers(TRUE, 0);

 
 {** Initialize general SB-specific stuff **}
 SB_CallBackCount:= SCallBackTime;
 FillChar(SB_NewHandler, SizeOf(TSegInfo), 0);
 FillChar(SB_OldHandler, SizeOf(TSegInfo), 0);

 {** Initialize Sound Blaster DMA-channel **}
 jssMsg(jProcName_SB, 'Initializing DMA channel...');
 vmSB_ResetDSP;
 vmDMA_Initialize;

 {** Initialize IRQ handler **}
 jssMsg(jProcName_SB, 'Initializing IRQ handler...');

 { Install handler to the @Self table }
 if (jssIRQHandlerSelfTable[SIRQ] <> NIL) then
    begin
     jssErrorN:= jerrInvalidIRQ;
     jssError(jProcName_SB, 'IRQ self table entry already reserved! Two devices are possibly using same IRQ!');
     exit;
    end;

 jssIRQHandlerSelfTable[SIRQ]:= @Self;

 { Set the IRQ handler routine and initialize PIC }
 SB_NewHandler.Offset        := @vmSB_IRQHandler;
 SB_NewHandler.Segment       := Get_CS;
 if (SIRQ < 8) then
    begin
     {** Set IRQ channel 0..7 **}
     Get_PM_Interrupt(SIRQ + 8, SB_OldHandler);
     Set_PM_Interrupt(SIRQ + 8, SB_NewHandler);

     SB_SavedPICMask:= inportb($21);
     ATempB2:= NOT (1 SHL SIRQ);
     ATempB1:= (SB_SavedPICMask AND ATempB2);
     outportb($21, ATempB1);
    end else
    begin
     {** Set IRQ channel 8..15 **}
     Get_PM_Interrupt(SIRQ + $68, SB_OldHandler);
     Set_PM_Interrupt(SIRQ + $68, SB_NewHandler);

     SB_SavedPICMask:= inportb($A1);
     ATempB2:= NOT (1 SHL (SIRQ AND 7));
     ATempB1:= (SB_SavedPICMask AND ATempB2);
     outportb($A1, ATempB1);
    end;

 {** Initialize sound output **}
 jssMsg(jProcName_SB, 'Initializing sound output...');
 vmSB_SetSpeaker(TRUE);

 { Set DMA data block }
 vmDMA_SetData(SDMixBuffer, SMixBufSize);

 { Start sound output }
 vmSB_SetDSP;
end;


destructor TJD_SoundBlaster.vmClose;
 var ATempB1, ATempB2:Byte;
begin
 {** Shutdown Sound Blaster **}
 jssMsg(jProcName_SB, 'Sound Blaster shutdown...');

 { Stop the current DMA/IRQ }
 vmSB_SetSpeaker(FALSE);
 vmDMA_Stop;
 vmSB_ResetDSP;

 { Remove the IRQ handler:
   We need to restore the used IRQ's setting to the PIC mask ...
   NOTICE: We must NOT touch other bits, (i.e. cant just put
   the whole byte in unmasked) since they may have been changed
   between init and close!!
 }
 ATempB1:= (1 SHL (SIRQ AND 7));
 ATempB1:= (SB_SavedPICMask and ATempB1);
 
 if (SIRQ < 8) then
    begin
     {** Set IRQ channel 0..7 **}
     Set_PM_Interrupt(SIRQ + $8, SB_OldHandler);

     ATempB2:= inportb($21);
     outportb($21, (ATempB1 OR ATempB2));
    end else
    begin
     {** Set IRQ channel 8..15 **}
     Set_PM_Interrupt(SIRQ + $68, SB_OldHandler);

     ATempB2:= inportb($A1);
     outportb($A1, (ATempB1 OR ATempB2));
    end;

 { Finally, reset the DSP again.. }
 vmSB_ResetDSP;

 { Un-install handler from the @Self table }
 if (jssIRQHandlerSelfTable[SIRQ] <> @Self) then
    jssWarning(jProcName_SB, 'IRQ self table contained different @Self value!'+
                             ' Two devices are possibly using same IRQ or table corrupted!');

 jssIRQHandlerSelfTable[SIRQ]:= NIL;

 {** Call inherited vmClose **}
 inherited vmClose;
end;

