{
  J Sound System - Disk Writer device
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  A general (works on all platforms) disk-writer "device" driver.

  This driver outputs the sound to a file. It works differently
  from other drivers and therefore cannot be used in similar way.

  Although working well, this driver is mainly meant to be used
  for testing purposes only. If you need a disk-writer on your
  program, you're better off writing one by yourself. Of course
  you can use this code as an example.
}

 const
      chkRIFF = 'RIFF';
      chkWAVE = 'WAVE';
      chkFmt  = 'fmt'#0;
      chkData = 'data';

 type
     TRIFFChunk = record
      rID:String4;
      rSize:Longint;
     end;

     TFormatChunk = record
      wFormatTag,
      nChannels,
      nSamplesPerSec,
      nAvgBytesPerSec,
      nBlockAlign,
      RESERVED:Word;
     end;

constructor TJD_DiskWriter.vmInit(AConfig:TJSSConfigFile; AFilename:String);
begin
 {** Call inherited vmInit **}
 SOutFileOpen:= FALSE;
 inherited vmInit(AConfig);

 {** Open sound output file **}
 SOutFilename:= AFilename;
 Assign(SOutFile, SOutFilename);
 Rewrite(SOutFile, 1);
 if (IOResult <> 0) then
    begin
     jssErrorN:= jerrFileCreateError;
     jssError('TJD_DiskWriter.vmInit', 'Error creating output file "'+ SOutFilename +'".');
     exit;
    end;

 SOutFileOpen:= TRUE;

 {** Write WAV header **}

end;


destructor TJD_DiskWriter.vmClose;
begin
 {** Update the header **}

 {** Close the output file **}
 if (SOutFileOpen) then
    close(SOutFile);

 SOutFilename:= '';
 SOutFileOpen:= FALSE;

 {** Call inherited vmClose **}
 inherited vmClose;
end;


procedure TJD_DiskWriter.vmExecute(APlayTime:Cardinal);
begin
 {** Do the sound output **}
 STimer:= 0;
 repeat
  {** Mix the sound **}
  if (SOutMode16bit) then
  begin
{$ifdef SUP_MIXING_PREC_16BIT}
  vmMix16bitMono(SMixBuffer);
{$endif}
  end else
  begin
{$ifdef SUP_MIXING_PREC_8BIT}
  vmMix8bitMono(SMixBuffer);
{$endif}
  end;

  {** Write to file **}
  blockwrite(SOutFile, SMixBuffer^, SMixBufSize);
  if (IOResult <> 0) then
     begin
      jssErrorN:= jerrFileWriteError;
      jssError('TJD_DiskWriter.vmPlay', 'Error writing audio data to output file!');
      exit;
     end;

  {** Increase playing timer **}
  STimer:= (STimer + SUMixBufSize);
 until (STimer > (APlayTime * SOutFreq));
end;

