{
  J Sound System - Open Sound System (OSS) driver
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This meta-driver is used as interface for the UNIX's
  Open Sound System (shortly OSS - which is formerly also
  known as Unix Sound System and Universal Sound System etc).
}

 const
      {** IOCTL and format constants for /dev/dsp **}
      SNDCTL_DSP_RESET			= $00005000;
      SNDCTL_DSP_SYNC			= $00005001;
      SNDCTL_DSP_SPEED			= $C0045002;
      SNDCTL_DSP_STEREO			= $C0045003;
      SNDCTL_DSP_GETBLKSIZE		= $C0045004;
      SNDCTL_DSP_POST			= $00005008;
      SNDCTL_DSP_SUBDIVIDE		= $C0045009;
      SNDCTL_DSP_SETFRAGMENT		= $C004500A;
      SNDCTL_DSP_GETFMTS		= $8004500B;
      SNDCTL_DSP_SETFMT			= $C0045005;
      SNDCTL_DSP_GETOSPACE		= $800C500C;
      SNDCTL_DSP_GETISPACE		= $800C500D;
      SNDCTL_DSP_NONBLOCK		= $0000500E;
      SNDCTL_SOUND_PCM_WRITE_CHANNELS	= $C0045006;
      SNDCTL_SOUND_PCM_WRITE_FILTER	= $C0045007;
      SNDCTL_SOUND_PCM_READ_RATE	= $80045002;
      SNDCTL_SOUND_PCM_READ_CHANNELS	= $80045006;
      SNDCTL_SOUND_PCM_READ_BITS	= $80045005;
      SNDCTL_SOUND_PCM_READ_FILTER	= $80045007;
      AFMT_MU_LAW			= $001;
      AFMT_A_LAW			= $002;
      AFMT_IMA_ADPCM			= $004;
      AFMT_U8				= $008;
      AFMT_S16_LE			= $010;
      AFMT_S16_BE			= $020;
      AFMT_S8				= $040;
      AFMT_U16_LE			= $080;
      AFMT_U16_BE			= $100;
      AFMT_MPEG				= $200;

 type
     PPThread_T = ^PThread_T;
     PPointer = ^Pointer;
     TThreadFunction = Procedure(arg:Pointer); cdecl;

{ Short list of libpthread-functions }
{$LinkLib pthread}
function pthread_create(th:PPThread_T; attr:Pointer;
                        start_routine:TThreadFunction;
                        arg:Pointer):Longint;
                        cdecl; external;

function pthread_join(th:PThread_T; thread_return:PPointer):Longint;
                        cdecl; external;

procedure pthread_exit(retval:Pointer);
                        cdecl; external;


procedure vmThreadFunc(arg:Pointer); cdecl;
 var ASigSet, AOldSigSet:Longint;
begin
 {** Set the signal masking **}
 AOldSigSet:= 0;
 ASigSet:= (SIGINT or SIGTERM);
 SigProcMask(SIG_BLOCK, @ASigSet, @AOldSigSet);

 {** Start the actual execution **}
 PJD_OSS(arg)^.vmExecute;

 {** Exit **}
 pthread_exit(arg);
end;


constructor TJD_OSS.vmInit(AConfig:TJSSConfigFile; ADev_DSP:String);

function vmCheckFmt(AFmts, AMask:Word):Boolean;
begin
 vmCheckFmt:= ((AFmts AND AMask) = AMask);
end;

function vmCSFmt(var AFmts:Longint; AFmt:Word):Boolean;
begin
 if ((AFmts AND AFmt) = AFmt) then
	begin
	 AFmts:= AFmt;
	 vmCSFmt:= TRUE;
	end else
	 vmCSFmt:= FALSE;
end;

 var 
    ATempInt, ATempFreq, ATempFrm:Longint;

begin
 {** Call inherited vmInit **}
 jssDev(jProcName_OSS, jDevName_OSS, jDevVer_OSS);
 inherited vmInit(AConfig);
 SOutDeviceOpen:= FALSE;
 SOutDevice:= -1;
 SStopSignal:= FALSE;
 SChildStopped:= FALSE;
 SThread:= 0;
 SDev_DSP:= '/dev/dsp';

 {** Check the given device names **}
 if (ADev_DSP <> '') then
    SDev_DSP:= ADev_DSP;

 {** Open and initalize OSS sound device **}
 SOutDevice:= fdOpen(SDev_DSP, open_RdWr);
 if (SOutDevice < 0) then
    begin
     jssErrorN:= jerrFileOpenError;
     jssError(jProcName_OSS, 'Error opening sound device '+ SDev_DSP);
     exit;
    end;

 SOutDeviceOpen:= TRUE;

 {** Initialize device parameters **}
 ATempInt:= 0;
 if (ioctl(SOutDevice, SNDCTL_DSP_RESET, @ATempInt) = FALSE) then
    begin
     jssErrorN:= jerrIOCTLerror;
     jssError(jProcName_OSS, 'Error doing DSP reset!');
     exit;
    end;


 {** Check output modes **}
 if (ioctl(SOutDevice, SNDCTL_DSP_GETFMTS, @ATempFrm) = FALSE) then
    begin
     jssErrorN:= jerrIOCTLerror;
     jssError(jProcName_OSS, 'Error getting supported DSP formats.');
     exit;
    end;


 {** Determine the final output mode based on available formats **}
 if (SOutMode16bit) then
    begin
     { Check what we have, in order of desire ;) .. }
     if (vmCSFmt(ATempFrm, AFMT_S16_LE)) then
        begin
         SOutModeSigned:= TRUE;
        end else
     if (vmCSFmt(ATempFrm, AFMT_U16_LE)) then
        begin
         SOutModeSigned:= FALSE;
        end else
        begin
         jssErrorN:= jerrUnsupResolution;
         jssError(jProcName_OSS, 'No known output formats supported by this OSS device.');
         exit;
        end;
    end else
    begin
     { Same here, except for 8bit }
     if (vmCSFmt(ATempFrm, AFMT_S8)) then
        begin
         SOutModeSigned:= TRUE;
        end else
     if (vmCSFmt(ATempFrm, AFMT_U8)) then
        begin
         SOutModeSigned:= FALSE;
        end else
        begin
         jssErrorN:= jerrUnsupResolution;
         jssError(jProcName_OSS, 'No known output formats supported by this OSS device.');
         exit;
        end;
    end;


 {** Initialize device parameters **}
 { Then, we set the device parameters in order that
   is mandatory, as stated by OSS documentation:

   1) Format
   2) Channels
   3) Frequency
 }

 {** Set the output format **}
 ATempInt:= ATempFrm;
 if (ioctl(SOutDevice, SNDCTL_DSP_SETFMT, @ATempInt) = FALSE) then
    begin
     jssErrorN:= jerrIOCTLerror;
     jssError(jProcName_OSS, 'Error setting DSP output mode/format.');
     exit;
    end;

 {** Set Stereo/Mono **}
 if (SOutModeStereo)
    then ATempInt:= 1
    else ATempInt:= 0;

 if (ioctl(SOutDevice, SNDCTL_DSP_STEREO, @ATempInt) = FALSE) then
    begin
     jssErrorN:= jerrIOCTLerror;
     jssError(jProcName_OSS, 'Error setting DSP channels (stereo/mono)');
     exit;
    end;

 {** Set output frequency **}
 ATempFreq:= SOutFreq;
 if (ioctl(SOutDevice, SNDCTL_DSP_SPEED, @ATempFreq) = FALSE) then
    begin
     jssErrorN:= jerrIOCTLerror;
     jssError(jProcName_OSS, 'Error setting DSP output frequency.');
     exit;
    end;

 {**** Calculate fragments, etc... ****}
 if (ioctl(SOutDevice, SNDCTL_DSP_GETBLKSIZE, @ATempInt) = FALSE) then
    begin
     jssErrorN:= jerrIOCTLerror;
     jssError(jProcName_OSS, 'Could not get DSP block fragment size!');
     exit;
    end;
    
 jssMsg(jProcName_OSS, 'Using output format: 0x'+ ioHexD(ATempFrm)
                        +' in '+ jboolStereo[SOutModeStereo] 
                        +', '+ ioStrL(SOutFreq) +' Hz, fragsize='+ ioStrL(ATempInt) +' bytes');

 { If stereo, divide by 2 }
 if (SOutModeStereo)
    then ATempInt:= (ATempInt DIV 2);
 
 { If 16-bit, divide by 2 }
 if (SOutMode16bit)
    then ATempInt:= (ATempInt DIV 2);

 { We should now have the size of buffer in "units". } 

 { Re-init mixing routine tables, etc. }
 vmDeallocMixBuffers;
 vmAllocMixBuffers(FALSE, ATempInt);


 {**** Initialize the audio output thread ****}
 { Allocate stack memory area for child thread }
 jssMsg(jProcName_OSS, 'Initializing audio output thread...');
 SChildStopped:= TRUE;

 if (pthread_create(@SThread, NIL, @vmThreadFunc, @Self) <> 0) then
    begin
     jssErrorN:= jerrInitFailure;
     jssError(jProcName_OSS, 'pthread_create() failed. Could not create audio output thread!');
     exit;
    end;

 { And finally, we wait that the child thread gets started }
 repeat until (SChildStopped = FALSE);
end;


destructor TJD_OSS.vmClose;
 var
    AExitStatus:Longint;
    ATempI:Integer;
    ACount, ARetCode:Longint;
    ARetVal:Pointer;
begin
 {** De-initialize the driver **}
 jssMsg(jProcName_OSS, 'Closing down audio output thread...');
 SStopSignal:= TRUE;

 ACount:= 64000;
 while ((SChildStopped = FALSE) and (ACount > 0)) do
       dec(ACount);

 ARetCode:= pthread_join(SThread, @ARetVal);
 if (ARetCode <> 0) then
    jssWarning(jProcName_OSS, 'Error #'+ ioStrL(ARetCode) +' joining the player thread...');

 if (ARetVal <> @Self) then
    jssWarning(jProcName_OSS, 'pthread_join() retval != @Self, bug?');

 {** Reset the driver **}
 ATempI:= 0;
 if (ioctl(SOutDevice, SNDCTL_DSP_RESET, @ATempI) = FALSE) then
    jssWarning(jProcName_OSS, 'Error doing DSP reset!');

 {** Close the output device **}
 if (SOutDeviceOpen) then
    begin
     jssMsg(jProcName_OSS, 'Closing OSS audio device '+ SDev_DSP);
     SOutDeviceOpen:= FALSE;
     if (fdClose(SOutDevice) = FALSE) then
        jssWarning(jProcName_OSS, 'Error occured when closing '+ SDev_DSP);
    end;

 {** Call inherited vmClose **}
 inherited vmClose;
end;



procedure TJD_OSS.vmExecute;
 var AIndex:Longint;
begin
 {** Do the sound output **}
 jssMsg(jProcName_OSS, 'Child started as PID '+ ioStrL(GetPid));
 SChildStopped:= FALSE;

 repeat
  {** Mix the sound **}
  if (SOutMode16bit) then
  begin
{$ifdef SUP_MIXING_PREC_16BIT}
  vmMix16bitMono(SMixBuffer);
{$endif}
  end else
  begin
{$ifdef SUP_MIXING_PREC_8BIT}
  vmMix8bitMono(SMixBuffer);
{$endif}
  end;
  
  {** Write to device **}
  if (fdWrite(SOutDevice, SMixBuffer^, SMixBufSize) < 0) then
     begin
      jssErrorN:= jerrFileWriteError;
      exit;
     end;

 until (SStopSignal);

 jssMsg(jProcName_OSS, 'Child thread execution ended.');
 SChildStopped:= TRUE;
end;

