{
  J Sound System - Main Engine
  (C) Copyright 1999-2001 Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  J Sound System main engine source file.
}
{$I jssdef.pas}
unit JSS;

interface

 uses
{$ifdef PLATFORM_LINUX}
      Linux,
{$endif}
{$ifdef PLATFORM_WIN32}
      Windows,
{$endif}
{$ifdef PLATFORM_DOS}
      GO32,
      DOS,
{$endif}
      BaseAPI;

 const
      {** Global constants **}
      { Include the version number }
      jssVersion        =
{$I version}
;
      jssMessage        = 'J Sound System v'+ jssVersion + ioCopyright;
      jssName           = 'JSS';

      {** Global max-value definations **}
      jsetNotSet        = -1;   { A global "not set" value }

      jsetMinChannel    = 1;    { Minimum channel number }
      jsetMaxChannel    = 64;   { Maximum channel number }
      jsetMaxNChannels  = ((jsetMaxChannel - jsetMinChannel) + 1);

      jsetMaxIRQ        = 16;   { Number of last IRQ channel in tables }
      jsetMaxCallBacks  = 32;   { Number of maximum callbacks in device }

      jsetMinVol        = 0;    { Minimum volume }
      jsetMaxVol        = 255;  { Maximum volume }

      {** Some output stuff **}
      jboolStereo:Array[Boolean] of String[6] = ('Mono', 'Stereo');
      jboolYesNo:Array[Boolean] of String[3] = ('No', 'Yes');

      {** Sample instrument flags **}
      jsfLooped         = $01;  { Looped instrument? }
      jsfBiDi           = $02;  { Bi-directional-looping? }
      jsf16bit          = $04;  { 16-bit instrument? }

      {** Panning position constants **}
      jchPanLeft        = $00;  { Leftmost panning value }
      jchPanMiddle      = $80;  { Center/middle panning value }
      jchPanRight       = $ff;  { Rightmost panning value }

      {** JSS error codes **}
      jerrOK                   = 0;
      jerrError                = 1;
      jerrInvalidArgument      = 2;
      jerrDivisionByZero       = 3;
      jerrShutdownFailure      = 4;
      jerrInitFailure          = 5;
      jerrFileOpenError        = 6;
      jerrFileCreateError      = 7;
      jerrFileReadError        = 8;
      jerrFileWriteError       = 9;
      jerrDMAchannelNotDefined = 10;
      jerrInvalidDMAChannel    = 11;
      jerrNULLPointer          = 13;
      jerrNonNULLPointer       = 14;
      jerrMemAllocError        = 15;
      jerrOrigObject           = 16;
      jerrInternalError        = 17;
      jerrUnsupportedOutMode   = 18;
      jerrInvalidDMADataSize   = 19;
      jerrInvalidDMADataBlock  = 20;
      jerrOutOfDOSMemory       = 21;
      jerrOutOfChannels        = 22;
      jerrInvalidIRQ           = 23;
      jerrIOCTLerror           = 24;
      
      jerrInvalidMixingFreq    = 25;
      jerrUnsupResolution      = 26;

{$ifdef SUP_PLATFORM_USEDMA}
      {** General DMA constants **}
      jDMAPageReg:Array[0..7] of Word = ($87, $83, $81, $82,  $8f, $8b, $89, $8a);
      jDMAAddrReg:Array[0..7] of Word = ($00, $02, $04, $06,  $c0, $c4, $c8, $cc);
      jDMASizeReg:Array[0..7] of Word = ($01, $03, $05, $07,  $c2, $c6, $ca, $ce);
{$endif}

      {** Definations for device names and info strings **}
{$ifdef SUP_DEV_SB}
      jProcName_SB      = 'SB';
      jDevName_SB       = 'Sound Blaster/DOS';
      jDevVer_SB        = '0.20';
      jDevDesc_SB       = 'Driver for Creative Labs Sound Blaster and 100% compatibles';
{$endif}

{$ifdef SUP_DEV_GUS}
      jProcName_GUS     = 'GUS';
      jDevName_GUS      = 'Gravis Ultrasound/DOS';
      jDevVer_GUS       = '0.0';
      jDevDesc_GUS      = 'Driver for Gravis UltraSound, MAX and 100% compatibles';
{$endif}

{$ifdef SUP_DEV_PCSPEAKER}
      jProcName_PcSpk   = 'PCSPK';
      jDevName_PcSpk    = 'PC Internal Speaker/DOS';
      jDevVer_PcSpk     = '0.10';
      jDevDesc_PcSpk    = 'Driver for IBM PC internal speaker';
{$endif}

{$ifdef SUP_DEV_DSOUND}
      jProcName_DSound  = 'DSound';
      jDevName_DSound   = 'DirectSound/Win32';
      jDevVer_DSound    = '0.00';
      jDevDesc_DSound   = 'Meta-driver for using Win32 DirectSound drivers';
{$endif}

{$ifdef SUP_DEV_OSS}
      jProcName_OSS     = 'OSS';
      jDevName_OSS      = 'Open Sound System/UNIX';
      jDevVer_OSS       = '0.70';
      jDevDesc_OSS      = 'Meta-driver for using UNIX Open Sound System drivers';
{$endif}

{$ifdef SUP_DEV_ALSA}
      jProcName_ALSA    = 'ALSA';
      jDevName_ALSA     = 'ALSA/Linux';
      jDevVer_ALSA      = '0.0';
      jDevDesc_ALSA     = 'Meta-driver for using Advanced Linux Sound Architechture drivers';
{$endif}

{$ifdef SUP_DEV_DISKW}
      jProcName_DiskW   = 'DiskW';
      jDevName_DiskW    = 'Disk Writer';
      jDevVer_DiskW     = '0.51';
      jDevDesc_DiskW    = 'Disk-Writing device for outputting sound to files';
{$endif}


{$ifdef SUP_DEV_NOSOUND}
      jProcName_NoSound = 'NoSnd';
      jDevDesc_NoSound  = 'Driver for No Sound';

{$ifdef PLATFORM_UNIX}
      jDevName_NoSound  = 'No Sound/Unix';
      jDevVer_NoSound   = '0.00';
{$endif}
{$ifdef PLATFORM_DOS}
      jDevName_NoSound  = 'No Sound/DOS';
      jDevVer_NoSound   = '0.00';
{$endif}
{$ifdef PLATFORM_WIN32}
      jDevName_NoSound  = 'No Sound/Win32';
      jDevVer_NoSound   = '0.00';
{$endif}

{$endif} { SUP_DEV_NOSOUND }


{** Include JSS configuration file typedef **}
{$I jsscfg.pas}


 type
     {** General type definations **}
     TUByte     = Array[jsetMinChannel..jsetMaxChannel] of Byte;
     TUWord     = Array[jsetMinChannel..jsetMaxChannel] of Word;
     TULongint  = Array[jsetMinChannel..jsetMaxChannel] of Longint;
     TUCardinal = Array[jsetMinChannel..jsetMaxChannel] of Cardinal;
     TUShortInt = Array[jsetMinChannel..jsetMaxChannel] of Shortint;
     TUBoolean  = Array[jsetMinChannel..jsetMaxChannel] of Boolean;

     {** 64-bit (32:32) fixed point integer record **}
     TJFPRec = record
      Frac, Int:Cardinal;
     end;

     {** 8-bit mixing tables **}
     PJSSVolTable8 = ^TJSSVolTable8;
     TJSSVolTable8 = array[jsetMinVol..jsetMaxVol, 0..255] of Longint;


     {** JSS internal channel structure **}
     PJSSChannel = ^TJSSChannel;
     TJSSChannel = record
{000} ChPos,                   { Sample position in ELEMENTS }
{008} ChDelta:TJFPRec;         { Sample position delta ELEMENTS }
{016} ChFreq,                  { Sample frequency Hz }
{020} ChSize,                  { Sample size in ELEMENTS }
{024} ChLoopS,                 { Loop start in ELEMENTS }
{028} ChLoopE:Cardinal;        { Loop end in ELEMENTS }
{032} ChData:Pointer;          { Pointer to sampledata }

{036} ChPlaying,               { Is channel being played? }
{037} ChMute,                  { Is channel OUTPUT muted? }
{038} ChDirection:Boolean;     { TRUE: Position is INCREASING by ChDelta.
                                 (In english this means FORWARDS)
                                 FALSE: Position is DECREASING by ChDelta.
                                 (Which means BACKWARDS) }

{039} ChFlags,                 { Sample flags in channel, jsfXXXX }
{040} ChVolume,                { Volume of channel }
{041} ChPanning:Byte;          { Channel panning position }
     end;

{$ifdef SUP_PLATFORM_USEDMA}
     TJSSDMAAddr = record
      AUsed:Boolean;           { Is this memory block used? }
      APage:Byte;              { DMA Page Number }
      APageOffs:Word;          { DMA Page offset }
      ARealSeg,                { Realmode segment of block }
      ASelector:Word;          { Selector of memory block }
      ALinearAddr:Cardinal;    { Linear memory address of block }
     end;
{$endif}

     TJSSChannelTable = Array[jsetMinChannel..jsetMaxChannel] of Longint;


     {***************** General virtual callback object *********************}
     PJSSSoundDevice = ^TJSSSoundDevice;
     PJSSCallBack = ^TJSSCallBack;
     TJSSCallBack = object
      SID,                           { CallBack ID from device's vmInstall... }
      SLocked:Longint;
      SOwner:PJSSSoundDevice;        { Owner device of this callback }
      SCallBackSet:Boolean;          { Internal bool which the descendant
                                       objects can use to mark whether they
                                       have installed the callback into the
                                       device driver's callback-table. }

      SFreq,                         { Frequency of the callback, a saved
                                       value for internal use only }
      SCounter,                      { Counter for callback, used for
                                       computing when the next poll occurs }
      SCounterInit:Single;           { Time between polls (initial counter
                                       value) }

      {** Overloadable methods **}
      procedure vmExec; virtual;     { This procedure is the callback function }
      procedure vmNotifyFreqChange; virtual;

      {** Non-overloadable **}
{
      function vmLock:Longint;
      function vmUnlock:Longint;
      function vmGetLock:Longint;
}
      procedure vmSetFreq(AFreq:Single);
      procedure vmSetTime(ATime:Single);

      constructor vmInit;
      destructor vmClose; virtual;
     end;


     {***************** General soundcard driver object *********************}
     TJSSSoundDevice = object
      private
      {** General sound device variables (All device drivers) **}
      SDevice:ShortInt;        { Sound device type number }
      SBase:Word;              { Base I/O port address }
      SIRQ,                    { IRQ-line number }
      SDMA:ShortInt;           { DMA channel number }
      SDMAPageReg,             { DMA page register }
      SDMAAddrReg,             { DMA address register }
      SDMASizeReg:Word;        { DMA transfer length register }

      SOutModeSigned,          { Output SIGNED/UNSIGNED data? }
      SOutModeStereo,          { Should we use a STEREO or MONO mode? }
      SOutMode16bit:Boolean;   { Should we use 16-bit output mode }
      SOutFreq:Cardinal;       { Mixing/output frequency }
      SGlobalVol,              { Global volume (jsetMinVol...jsetMaxVol) }
      SNChannels:Longint;      { Number of channels open (nothing to do with SChannelTable[]) }
      SChannelTable:Array[jsetMinChannel..jsetMaxChannel] of Longint;
      { List of reserved/used channels. Every entry for a channel
        contains a value which is either -1 for "not reserved/used".
        OR > 0 if it is used. The value is the "handle" number for
        the end-user program to use the channels.
      }

{     FIXME FIX ME :  channel locking on driver level

      SChannelLocked:Array[jsetMinChannel..jsetMaxChannel] of Longint;
                                 x = 0 -- Not locked
                                 x > 0 -- Data is being locked by some thread 
                                 x < 0 -- Error
}

      NCallBacks:Longint;
      SCallBacks:Array[1..jsetMaxCallBacks] of PJSSCallBack;
      { Pointers of callback routines }


      {** Low-level routines **}
{$ifdef SUP_PLATFORM_USEDMA}
      function  vmDMA_Initialize:Boolean;
      function  vmDMA_AllocBuffer(ABufSize:Cardinal; var ABuffer:TJSSDMAAddr):Boolean;
      function  vmDMA_DeAllocBuffer(var ABuffer:TJSSDMAAddr):Boolean;
      function  vmDMA_SetData(ABuffer:TJSSDMAAddr; ABufSize:Word):Boolean;
      function  vmDMA_Stop:Boolean;
{$endif}

      public
      
      {** Callback system **}
      function  vmInstallCallBack(ACB:PJSSCallBack):Longint; virtual;
      procedure vmRemoveCallBack(ACB:Longint); virtual;

      {** Channel manipulation routines **}
      function  vmAllocChannels(ANChannels:Longint):Longint; virtual;
      procedure vmDeAllocChannels(AHandle:Longint); virtual;
      procedure vmGetChannelTransTable(AHandle:Longint; var ATable:TJSSChannelTable); virtual;

      procedure vmPlay(AChannel:Longint); virtual;
      procedure vmStop(AChannel:Longint); virtual;
      procedure vmSetSample(AChannel:Longint; AData:Pointer; ASize, ALoopS,
                            ALoopE:Cardinal; AFlags:Byte); virtual;
      procedure vmSetFreq(AChannel:Longint; AFreq:Cardinal); virtual;
      procedure vmGetFreq(AChannel:Longint; var AFreq:Cardinal); virtual;
      procedure vmSetVolume(AChannel:Longint; AVolume:Byte); virtual;
      procedure vmGetVolume(AChannel:Longint; var AVolume:Byte); virtual;
      procedure vmSetPosition(AChannel:Longint; APos:Cardinal); virtual;
      procedure vmGetPosition(AChannel:Longint; var APos:Cardinal); virtual;
      procedure vmSetPanning(AChannel:Longint; APanning:Byte); virtual;
      procedure vmGetPanning(AChannel:Longint; var APanning:Byte); virtual;
      procedure vmMuteChannel(AChannel:Longint; AMute:Boolean); virtual;
      procedure vmClearChannel(AChannel:Longint); virtual;
      procedure vmClearChannels; virtual;
      procedure vmSetGlobalVol(AGlobalVol:Longint); virtual;
      procedure vmGetGlobalVol(var AGlobalVol:Longint); virtual;

      procedure vmNotifyFreqChange; virtual;

      {** High level driver routines **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;


     PJSSMixingDevice = ^TJSSMixingDevice;
     TJSSMixingDevice = object(TJSSSoundDevice)
      {** Software mixing device specific variables **}
      SMixBufSize,                  { Mixing buffer size in BYTES }
      SUMixBufSize,                 { Mixing buffer size in UNITS }
      SAddBufSize:Cardinal;         { Addition buffer size in BYTES }

      SAddBuffer,                   { Addition buffer }
      SMixBuffer:Pointer;           { Mixing buffer }
{$ifdef SUP_PLATFORM_USEDMA}
      SDMixBuffer:TJSSDMAAddr;      { DOS base-memory DMA-buffer }
{$endif}

      SVolTable8:PJSSVolTable8;     { 8-bit volumetable }

      SChannel:Array[jsetMinChannel..jsetMaxChannel] of TJSSChannel;

      {** Software mixing routines **}
      procedure vmAllocMixBuffers(ADetermine:Boolean; AUSize:Cardinal); virtual;
      procedure vmDeallocMixBuffers; virtual;
      procedure vmInitVolTable; virtual;

      procedure vmClearAddBuffer;

  {$ifdef SUP_MIXING_PREC_8BIT}
      procedure vmDoPostProcessing8Signed(AMixDestBuffer:Pointer; AMixDestLength:Cardinal); virtual;
      procedure vmDoPostProcessing8Unsigned(AMixDestBuffer:Pointer; AMixDestLength:Cardinal); virtual;
   {$ifdef SUP_MIXING_PREC_MONO}
      procedure vmMix8bitMono(AMixDestBuffer:Pointer); virtual;
   {$endif}
   {$ifdef SUP_MIXING_PREC_STEREO}
      procedure vmMix8bitStereo(AMixDestBuffer:Pointer); virtual;
   {$endif}
  {$endif}

  {$ifdef SUP_MIXING_PREC_16BIT}
      procedure vmDoPostProcessing16Signed(AMixDestBuffer:Pointer; AMixDestLength:Cardinal); virtual;
      procedure vmDoPostProcessing16Unsigned(AMixDestBuffer:Pointer; AMixDestLength:Cardinal); virtual;
   {$ifdef SUP_MIXING_PREC_MONO}
      procedure vmMix16bitMono(AMixDestBuffer:Pointer); virtual;
   {$endif}
   {$ifdef SUP_MIXING_PREC_STEREO}
      procedure vmMix16bitStereo(AMixDestBuffer:Pointer); virtual;
   {$endif}
  {$endif}


      {** Channel manipulation routines **}
      procedure vmPlay(AChannel:Longint); virtual;
      procedure vmStop(AChannel:Longint); virtual;
      procedure vmSetSample(AChannel:Longint; AData:Pointer; ASize, ALoopS,
                            ALoopE:Cardinal; AFlags:Byte); virtual;
      procedure vmSetFreq(AChannel:Longint; AFreq:Cardinal); virtual;
      procedure vmGetFreq(AChannel:Longint; var AFreq:Cardinal); virtual;
      procedure vmSetVolume(AChannel:Longint; AVolume:Byte); virtual;
      procedure vmGetVolume(AChannel:Longint; var AVolume:Byte); virtual;
      procedure vmSetPosition(AChannel:Longint; APos:Cardinal); virtual;
      procedure vmGetPosition(AChannel:Longint; var APos:Cardinal); virtual;
      procedure vmSetPanning(AChannel:Longint; APanning:Byte); virtual;
      procedure vmGetPanning(AChannel:Longint; var APanning:Byte); virtual;
      procedure vmMuteChannel(AChannel:Longint; AMute:Boolean); virtual;
      procedure vmClearChannel(AChannel:Longint); virtual;

      {** High level driver routines **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;


{$ifdef SUP_MIXING_MMX_OPT}
     PJSSMMXMixingDevice = ^TJSSMMXMixingDevice;
     TJSSMMXMixingDevice = object(TJSSMixingDevice)
 
      constructor vmInit(AConfig:TJSSConfigFile);
     end;
{$endif}


{$ifdef SUP_MIXING_3DNOW_OPT}
     PJSS3DNOWMixingDevice = ^TJSS3DNOWMixingDevice;
     TJSS3DNOWMixingDevice = object(TJSSMixingDevice)

      constructor vmInit(AConfig:TJSSConfigFile);
     end;
{$endif}


     {** Objects for individual sound device drivers **}
{$ifdef SUP_DEV_SB}
     PJD_SoundBlaster = ^TJD_SoundBlaster;
     TJD_SoundBlaster = object(TJSSMixingDevice)
      {** Driver specific variables **}
      SB_DSPVersion:Word;       { Version of the DSP chip }
      SB_CallBackCount:Longint;
      SB_NewHandler,
      SB_OldHandler:TSegInfo;

      { Feeds audiodata to SB via DMA }
      procedure vmSB_IRQHandler_Pascal; virtual;

      { Handles the IRQ, sets up dataseg, @Self and calls the _Pascal routine }
      procedure vmSB_IRQHandler;
        
      { Utility functions for SB handling }
      function  vmSB_DSPRead:Byte; virtual;
      procedure vmSB_DSPWrite(AData:Byte); virtual;

      function  vmSB_DetectDSP:Boolean; virtual;
      function  vmSB_GetDSPVersion:Word; virtual;
      function  vmSB_GetCopyrightStr:String; virtual;
      procedure vmSB_SetDSP; virtual;
      procedure vmSB_ResetDSP; virtual;
      procedure vmSB_SetSpeaker(ASetting:Boolean); virtual;

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;
{$endif}


{$ifdef SUP_DEV_GUS}
     PJD_GravisUltrasound = ^TJD_GravisUltrasound;
     TJD_GravisUltrasound = object(TJSSMixingDevice)
      {** Driver specific variables **}
      GUS_CallBackCount:Longint;
      GUS_NewHandler,
      GUS_OldHandler:TSegInfo;

      { Feeds audiodata to GUS via DMA }
      procedure vmGUS_IRQHandler; virtual;

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;
{$endif}


{$ifdef PLATFORM_UNIX}
     PThread_T = Cardinal;
{$endif}

{$ifdef SUP_DEV_OSS}
     PJD_OSS = ^TJD_OSS;
     TJD_OSS = object(TJSSMixingDevice)
      {** Driver specific variables **}
      SThread:PThread_T;
      SOutDevice:Longint;
      SChildStopped,
      SStopSignal,
      SOutDeviceOpen:Boolean;

      SDev_DSP:String;

      {** Public items **}
      procedure vmExecute; virtual;

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile; ADev_DSP:String);
      destructor  vmClose; virtual;
     end;
{$endif}


{$ifdef SUP_DEV_ALSA}
     PJD_ALSA = ^TJD_ALSA;
     TJD_ALSA = object(TJSSMixingDevice)
      {** Driver specific variables **}
      SChildStack:Pointer;
      SPID,
      SOutDevice:Longint;
      SChildStopped,
      SStopSignal,
      SOutDeviceOpen:Boolean;

      SDev_DSP, SDev_SndStat:String;

      {** Public items **}
      procedure vmExecute; virtual;

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;
{$endif}


{$ifdef SUP_DEV_DISKW}
     PJD_DiskWriter = ^TJD_DiskWriter;
     TJD_DiskWriter = object(TJSSMixingDevice)
      {** Sum' variables **}
      STimer:Cardinal;
      SOutFile:File;
      SOutFileOpen:Boolean;
      SOutFilename:String;

      {** Public items **}
      procedure vmExecute(APlayTime:Cardinal); virtual;

      constructor vmInit(AConfig:TJSSConfigFile; AFilename:String);
      destructor  vmClose; virtual;
     end;
{$endif}


{$ifdef SUP_DEV_PCSPEAKER}
     PJD_PCSpeaker = ^TJD_PCSpeaker;
     TJD_PCSpeaker = object(TJSSMixingDevice)
      {** Driver specific variables **}

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;
{$endif}

{$ifdef SUP_DEV_NOSOUND}
{$ifdef PLATFORM_UNIX}
     PJD_NoSound = ^TJD_NoSound;
     TJD_NoSound = object(TJSSMixingDevice)
      {** Driver specific variables **}

{ FIX ME FIXME!!!! }

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;
{$endif}

{$ifdef PLATFORM_DOS}
     PJD_NoSound = ^TJD_NoSound;
     TJD_NoSound = object(TJSSMixingDevice)
      {** Driver specific variables **}
      STimerValue:Longint;
      SIRQ0Save,
      SIRQ0Handler:TSegInfo;

      procedure vmIRQ0Handler; virtual;
      procedure vmSetIRQ0TimerValue(ATimerValue:Word); virtual;
      function  vmRestoreDOSClock:Boolean; virtual;

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile);
      destructor  vmClose; virtual;
     end;
{$endif}
{$endif}


{$ifdef SUP_DEV_DSOUND}
     PJD_DirectSound = ^TJD_DirectSound;
     TJD_DirectSound = object(TJSSMixingDevice)
      {** Driver specific variables **}

      procedure vmExecute; virtual;
      function vmGetConsoleHWND:HWND;

      {** Public items **}
      constructor vmInit(AConfig:TJSSConfigFile; hwnd:HWND);
      destructor  vmClose; virtual;
     end;
{$endif}


     TJSSErrorProc   = Procedure(AFrom:String; AMessage:String);
     TJSSMsgProc     = Procedure(AFrom, AMessage:String);
     TJSSPMsgProc    = Procedure(AMessage:String);


 var
{$ifdef PLATFORM_DOS}
    jssLOCKDATA_START:Longint;
{$endif}

    {** Error routines **}
    jssErrorN:Longint;          { Last JSS error code }
    jssWarning:TJSSErrorProc;   { JSS warning procedure }
    jssError:TJSSErrorProc;     { JSS error procedure }
    jssMsg,                     { JSS message procedures }
    jssMsgS:TJSSMsgProc;
    jssMsgP:TJSSPMsgProc;

{$ifdef PLATFORM_DOS}
    {** IRQ-handler objects @Self-table **}
    jssIRQHandlerSelfTable:Array[0..jsetMaxIRQ] of Pointer;

    jssBackUpDS:Word; external name '___v2prt0_ds_alias';
    jssMM_OldManager:TMemoryManager;

    jssLOCKDATA_END:Longint;
{$endif}

 {** Sound system Routines **}
 function  jssLoadConfig(AFilename:String; var AConfig:TJSSConfigFile):Longint;
 function  jssSaveConfig(AFilename:String; AConfig:TJSSConfigFile):Longint;
 function  jssInit:Longint;
 function  jssClose:Longint;
 procedure jssSetANYFlag(var AValue:Cardinal; AFlag:Cardinal);
 function  jssGetANYFlag(AValue, AFlag:Cardinal):Boolean;


implementation

{ DOS/GO32V2: Start of code to be locked }
{$ifdef PLATFORM_DOS}
procedure jssLOCKCODE_START; begin end;
{$endif}


{ Show a device initialization message
}
procedure jssDev(AProcName, ADevName, ADevVer:String);
begin
 jssMsg(AProcName, 'Initializing '+ ADevName +' sound device v'+ ADevVer);
end;


{****************************************************************************}
{*                                                                          *}
{* Device Drivers                                                           *}
{*                                                                          *}
{****************************************************************************}
{$I jssdev.pas}


{****************************************************************************}
{*                                                                          *}
{* TJSSSoundDevice routines                                                 *}
{*                                                                          *}
{****************************************************************************}
{ Initializes the device object using given configuration as basis.
}
constructor TJSSSoundDevice.vmInit(AConfig:TJSSConfigFile);
 const jProcName = 'TJSSSoundDevice.vmInit';
 var AIndex:Longint;
begin
 {** Initialize configuration variables **}
 SDevice         := AConfig.SDevice;
 SBase           := AConfig.SBase;
 SIRQ            := AConfig.SIRQ;
 SOutModeSigned  := FALSE;
 SOutModeStereo  := AConfig.SOutModeStereo;
 SOutMode16bit   := AConfig.SOutMode16bit;
 SOutFreq        := AConfig.SOutFreq;
 SGlobalVol      := jsetMaxVol;
 SDMAAddrReg     := 0;
 SDMAPageReg     := 0;
 SDMASizeReg     := 0;

 {** Initialize device channel allocation table **}
 for SNChannels:= jsetMinChannel to jsetMaxChannel do
     SChannelTable[SNChannels]:= jsetNotSet;

 SNChannels      := 0;

 {** Check sound device type **}
 if (AConfig.SDevice = jsetNotSet) then
    begin
     jssErrorN:= jerrInternalError;
     jssError(jProcName, 'Sound device type not set!');
     exit;
    end;

 {** Set the _real_ DMA channel **}
 if (SOutMode16bit)
    then SDMA:= AConfig.ShDMA
    else SDMA:= AConfig.SlDMA;

 {** Initialize callbacks **}
 NCallBacks:= 0;
 for AIndex:= 0 to jsetMaxCallBacks do
     SCallBacks[AIndex]:= NIL;
end;


{ Closes the sound device, freeing dangling resources.
}
destructor TJSSSoundDevice.vmClose;
 var AIndex:Longint;
begin
 {** De-initialize callbacks **}
 { NOTICE! It is potentially "dangerous" to leave callbacks running
   if the sound device is about to be closed. Because the device closes
   the callbacks here, the program probably does not notice it (if they
   use their own variables -- which is just normal and not an error!)
   ... this can cause a nice segfault when the callback routines are
   called two times...

   So that is why the USER should remove the callbacks before the
   TJSSSoundDevice.vmClose() gets called.
 }
 if (NCallBacks > 0) then
 begin
  for AIndex:= 0 to jsetMaxCallBacks do
  if (SCallBacks[AIndex] <> NIL) then
     Dispose(SCallBacks[AIndex], vmClose);
 end;
 NCallBacks:= 0;
end;


{ Notifies all device's subsystems (callbacks, etc) that
  frequency has changed in the sound device's settings.
}
procedure TJSSSoundDevice.vmNotifyFreqChange;
 var AIndex:Longint;
begin
 if (NCallBacks > 0) then
 begin
  for AIndex:= 0 to jsetMaxCallBacks do
  if (SCallBacks[AIndex] <> NIL) then
     SCallBacks[AIndex]^.vmNotifyFreqChange;
 end;
end;


{ Allocate virtual channels from sound device. Number of
  channels as argument, returns a handle which can used
  with vmGetChannelTransTable().
}
function  TJSSSoundDevice.vmAllocChannels(ANChannels:Longint):Longint;
 var AIndex, ACount, AHandle:Longint;
begin
 {** Search for the next handle **}
 AHandle:= jsetNotSet;
 for AIndex:= jsetMinChannel to jsetMaxChannel do
 if (SChannelTable[AIndex] > AHandle) then
    AHandle:= SChannelTable[AIndex];

 if (AHandle = jsetNotSet)
    then AHandle:= 100;

 inc(AHandle);

 {** Allocate the channels **}
 AIndex:= jsetMinChannel;
 ACount:= 0;
 while (ACount < ANChannels) and
       (SNChannels <= jsetMaxNChannels) and
       (AIndex <= jsetMaxChannel) do
 begin
  if (SChannelTable[AIndex] = jsetNotSet) then
     begin
      SChannelTable[AIndex]:= AHandle;
      inc(ACount);
      inc(SNChannels);
     end;

  inc(AIndex);
 end;

 {** Check if too many channels have been allocated **}
 if (SNChannels > jsetMaxNChannels) then
    begin
     jssErrorN:= jerrOutOfChannels;
     jssError('TJSSSoundDevice.vmAllocChannels', 'Too many total used channels! SNChannels > jsetMaxNChannels!!');
     exit;
    end;

 {** Check if we got enough channels allocated **}
 if (ACount < ANChannels) then
    begin
     jssErrorN:= jerrOutOfChannels;
     jssError('TJSSSoundDevice.vmAllocChannels', 'Out of channels. All '+ ioStrL(jsetMaxChannel - jsetMinChannel + 1) +' channels are reserved!');
     exit;
    end;

 vmAllocChannels:= AHandle;
end;


{ De-allocates channels allocated for a given handle.
}
procedure TJSSSoundDevice.vmDeAllocChannels(AHandle:Longint);
 var AIndex, ACount:Longint;
begin
 {** De-allocate the channels **}
 ACount:= 0;
 for AIndex:= jsetMinChannel to jsetMaxChannel do
 if (SChannelTable[AIndex] = AHandle) then
    begin
     SChannelTable[AIndex]:= jsetNotSet;
     inc(ACount);
     dec(SNChannels);     
    end;     

 {** Check for internal inconsistency **}
 if (SNChannels < 0) then
    begin
     jssErrorN:= jerrInternalError;
     jssError('TJSSSoundDevice.vmDeAllocChannels', 'Number of total used channels decreased below 0!');
     exit;
    end;
end;


{ Returns a special channel translation table, which is used
  when accessing the driver's sound channels. Before using this
  function you must allocate some channels and use handle from
  the allocation function here.
}
procedure TJSSSoundDevice.vmGetChannelTransTable(AHandle:Longint; var ATable:TJSSChannelTable);
 var AIndex, ACount:Longint;
begin
 {** Get the channel translation table **}
 ACount:= jsetMinChannel;
 for AIndex:= jsetMinChannel to jsetMaxChannel do
 if (SChannelTable[AIndex] = AHandle) then
    begin
     ATable[ACount]:= AIndex;
     inc(ACount);
    end;
end;


procedure TJSSSoundDevice.vmPlay(AChannel:Longint);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmPlay', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmStop(AChannel:Longint);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmStop', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmSetSample(AChannel:Longint; AData:Pointer; ASize, ALoopS,
                            ALoopE:Cardinal; AFlags:Byte);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmSetSample', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmSetFreq(AChannel:Longint; AFreq:Cardinal);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmFreq', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmGetFreq(AChannel:Longint; var AFreq:Cardinal);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmGetFreq', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmSetVolume(AChannel:Longint; AVolume:Byte);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmSetVolume', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmGetVolume(AChannel:Longint; var AVolume:Byte);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmGetVolume', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmSetPosition(AChannel:Longint; APos:Cardinal);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmSetPosition', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmGetPosition(AChannel:Longint; var APos:Cardinal);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmGetPosition', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmSetPanning(AChannel:Longint; APanning:Byte);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmSetPanning', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmGetPanning(AChannel:Longint; var APanning:Byte);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmGetPanning', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmMuteChannel(AChannel:Longint; AMute:Boolean);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmMuteChannel', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmClearChannel(AChannel:Longint);
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSSoundDevice.vmClearChannel', 'Originator object called.');
end;


procedure TJSSSoundDevice.vmClearChannels;
 var AChannel:Longint;
begin
 {** Clear all channels of the device **}
 for AChannel:= jsetMinChannel to jsetMaxChannel do
     vmClearChannel(AChannel);
end;


procedure TJSSSoundDevice.vmSetGlobalVol(AGlobalVol:Longint);
begin
 SGlobalVol:= AGlobalVol;

 if (SGlobalVol < jsetMinVol) then
    SGlobalVol:= jsetMinVol else

 if (SGlobalVol > jsetMaxVol) then
    SGlobalVol:= jsetMaxVol;
end;


procedure TJSSSoundDevice.vmGetGlobalVol(var AGlobalVol:Longint);
begin
 AGlobalVol:= SGlobalVol;
end;



{****************************************************************************}
{*                                                                          *}
{* Callback handling routines                                               *}
{*                                                                          *}
{****************************************************************************}
function TJSSSoundDevice.vmInstallCallBack(ACB:PJSSCallBack):Longint;
 var AIndex:Longint;
begin
 {** Check if we can have more callbacks **}
 AIndex:= 1;
 while (AIndex <= jsetMaxCallBacks) AND
       (SCallBacks[AIndex] <> NIL) do
       inc(AIndex);

 if (AIndex > jsetMaxCallBacks) OR
    (SCallBacks[AIndex] <> NIL) then
    jssError('TJSSSoundDevice.vmInstallCallBack', 'Out of callbacks. Maximum number of callbacks for this device exceeded.');

 {** Add a new callback **}
 inc(NCallBacks);
 if (NCallBacks > jsetMaxCallBacks) then
    jssWarning('TJSSSoundDevice.vmInstallCallBack', 'Internal inconsistency: number of callbacks exceeded but callback space OK?');

 with ACB^ do
 begin
  SID:= AIndex;
  SOwner:= @Self;
  SCallBackSet:= TRUE;
 end;

 SCallBacks[AIndex]:= ACB;
 vmInstallCallBack:= AIndex;
end;


procedure TJSSSoundDevice.vmRemoveCallBack(ACB:Longint);
begin
 {** Delete the callback **}
 if (ACB < 1) or (ACB > jsetMaxCallBacks) then
    jssError('TJSSSoundDevice.vmRemoveCallBack', 'Tried to remove unexisting callback.');

 dec(NCallBacks);
 if (NCallBacks < 0) then
    jssWarning('TJSSSoundDevice.vmRemoveCallBack', 'Internal inconsistency: number of callbacks below zero but callback space OK?');

 with SCallBacks[ACB]^ do
 begin
  SID:= -1;
  SOwner:= NIL;
  SCallBackSet:= FALSE;
 end;

 SCallBacks[ACB]:= NIL;
end;


constructor TJSSCallBack.vmInit;
begin
 {** Initialize callback values **}
 SID:= -1;
 SLocked:= 0;
 SOwner:= NIL;
 SCallBackSet:= FALSE;
 SCounter:= 0;
 SCounterInit:= 0;
 SFreq:= 0;
end;


destructor TJSSCallBack.vmClose;
begin
 {** De-initialized callback values **}
 if (SCallBackSet) AND 
    (SOwner <> NIL) AND
    (SID <> -1) then
    SOwner^.vmRemoveCallBack(SID);
end;


procedure TJSSCallBack.vmExec;
begin
 jssErrorN:= jerrOrigObject;
 jssError('TJSSCallBack.vmExec', 'Originator object called.');
end;


procedure TJSSCallBack.vmSetFreq(AFreq:Single);
begin
 if (SOwner = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     jssError('TJSSCallBack.vmSetFreq', 'Owner device pointer was NULL.');
    end;

 SFreq:= AFreq;

 if (SFreq > 0)
    then SCounterInit:= (SOwner^.SOutFreq / SFreq)
    else SCounterInit:= 0;
 
 SCounter:= SCounterInit;
end;


procedure TJSSCallBack.vmSetTime(ATime:Single);
begin
 if (SOwner = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     jssError('TJSSCallBack.vmSetTime', 'Owner device pointer was NULL.');
    end;

 SCounter:= ATime;
 SCounterInit:= ATime;
 if (ATime <> 0)
    then SFreq:= (SOwner^.SOutFreq / ATime)
    else SFreq:= 0;
end;


procedure TJSSCallBack.vmNotifyFreqChange;
begin
 { By default we handle owner's frequency changes
   simply by just adjusting our own variables
 }
  
 if (SFreq > 0)
    then SCounterInit:= (SOwner^.SOutFreq / SFreq)
    else SCounterInit:= 0;
 
 SCounter:= SCounterInit;
end;


{****************************************************************************}
{*                                                                          *}
{* TJSSMixingDevice routines                                                *}
{*                                                                          *}
{****************************************************************************}

{ Include the mixing routines }
{$I mixers\jmixer.pas}
{$I mixers\jmixpost.inc}

{$ifdef SUP_MIXING_MMX_OPT}
 {$I mixers\mmx\jmixer.pas}
{$endif}

{$ifdef SUP_MIXING_3DNOW_OPT}
 {$I mixers\3dnow\jmixer.pas}
{$endif}


constructor TJSSMixingDevice.vmInit(AConfig:TJSSConfigFile);
 const jProcName = 'TJSSMixingDevice.vmInit';

begin
 {** Call inherited InitDriver **}
 jssMsg(jProcName, 'Calling inherited TJSSSoundDevice.vmInit()');
 TJSSSoundDevice.vmInit(AConfig);

 SMixBufSize  := 0;
 SUMixBufSize := 0;
 SAddBufSize  := 0;
 SAddBuffer   := NIL;
 SMixBuffer   := NIL;
{$ifdef SUP_PLATFORM_USEDMA}
 FillChar(SDMixBuffer, SizeOf(TJSSDMAAddr), 0);
{$endif}
 SVolTable8 := NIL;

 {** Check the parameters **}
{$ifndef SUP_MIXING_PREC_16bit}
 if (SOutMode16bit) then
    begin
     jssErrorN:= jerrUnsupResolution;
     jssError(jProcName, '16-bit mixing routines not compiled in.');
     exit;
    end;
{$endif}

{$ifndef SUP_MIXING_PREC_8bit}
 if (SOutMode16bit = FALSE) then
    begin
     jssErrorN:= jerrUnsupResolution;
     jssError(jProcName, '8-bit mixing routines not compiled in.');
     exit;
    end;
{$endif}
 
 {** Pre-allocate mixing buffers **}
 { The actual device drivers may change these afterwards,
   but let's allocate some memory to be sure...
 }
 vmAllocMixBuffers(TRUE, 0);

 {** Allocate volume table **}
 jssMsg(jProcName, 'Initializing volume table.');
 New(SVolTable8);
 if (SVolTable8 = NIL) then
    begin
     jssErrorN:= jerrMemAllocError;
     jssError(jProcName, 'Error allocating volume table memory.');
     exit;
    end;

 vmInitVolTable;

 {** Initialize channels **}
 jssMsg(jProcName, 'Initializing channels data.');
 vmClearChannels;
end;


destructor TJSSMixingDevice.vmClose;
 const jProcName = 'TJSSMixingDevice.vmClose';
begin
 {** Call inherited CloseDriver **}
 jssMsg(jProcName, 'Calling inherited TJSSSoundDevice.vmClose()');
 TJSSSoundDevice.vmClose;

 {** Free the mixing buffers, etc. **}
 vmDeallocMixBuffers;

 if (SVolTable8 <> NIL) then
    Dispose(SVolTable8);
end;


procedure TJSSMixingDevice.vmAllocMixBuffers(ADetermine:Boolean; AUSize:Cardinal);
 const jProcName = 'vmAllocMixBuffers';
begin
 if (ADetermine) then
 begin
  {** Determine mixing buffer size automagically **}
  case SOutFreq of
   5000..11024:  SUMixBufSize:= 768;
   11025..22050: SUMixBufSize:= 1024;
   22051..33075: SUMixBufSize:= 1536;
   33076..44100: SUMixBufSize:= 2048;
   else begin
         jssErrorN:= jerrInvalidMixingFreq;
         jssError(jProcName, 'Invalid mixing frequency Hz (44100 < '+ ioStrL(SOutFreq) +' < 5000)');
         exit;
        end;
  end;
 end else
  {** Set the mixing buffer size to given value **}
  SUMixBufSize:= AUSize;

 SMixBufSize:= SUMixBufSize;

 { Stereo mode requires 2x sized mixing buffer }
 if (SOutModeStereo)
    then SMixBufSize:= (SMixBufSize * 2);

 { 16-bit mode requires also 2x sized mixing buffer }
 if (SOutMode16bit)
    then SMixBufSize:= (SMixBufSize * 2);

 { Compute the add-buffer size from mixing buffer size }
 SAddBufSize:= (SMixBufSize * SizeOf(Longint));

 {** Allocate mixing and addition buffers **}
 jssMsg(jProcName, 'Initializing mixing buffers.');
{$ifdef SUP_PLATFORM_USEDMA}
 vmDMA_AllocBuffer(SMixBufSize, SDMixBuffer);
{$endif}

 GetMem(SAddBuffer, SAddBufSize);
 if (SAddBuffer = NIL) then
    begin
     jssErrorN:= jerrMemAllocError;
     jssError(jProcName, 'Error allocating mixer''s addition-buffer memory.');
     exit;
    end;

 GetMem(SMixBuffer, SMixBufSize);
 if (SMixBuffer = NIL) then
    begin
     jssErrorN:= jerrMemAllocError;
     jssError(jProcName, 'Error allocating mixing buffer memory.');
     exit;
    end;

 {** Finally, clear the buffers **}
 FillChar(SMixBuffer^, SMixBufSize, 0);
 FillChar(SAddBuffer^, SAddBufSize, 0);
end;


procedure TJSSMixingDevice.vmDeallocMixBuffers;
 const jProcName = 'vmDeallocMixBuffers';
begin
 {** De-initialize Mixer **}
 jssMsg(jProcName, 'Deallocating mixer data-areas.');
{$ifdef SUP_PLATFORM_USEDMA}
 vmDMA_DeAllocBuffer(SDMixBuffer);
{$endif}

 if (SMixBuffer <> NIL) then
    FreeMem(SMixBuffer, SMixBufSize);

 if (SAddBuffer <> NIL) then
    FreeMem(SAddBuffer, SAddBufSize);
end;


{ Initializes the 8-bit Volume Table.
}
procedure TJSSMixingDevice.vmInitVolTable;
 var AVolume, ASample, AValue:Longint;
begin
 {** Check the existence of voltable **}
 if (SVolTable8 = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     jssError('TJSSMixingDevice.vmInitVolTable8', 'No 8-bit volume table allocated!');
     exit;
    end;

 {** Initialize volume table **}
 for AVolume:= jsetMinVol to jsetMaxVol do
 for ASample:= 0 to 255 do
     SVolTable8^[AVolume, ASample]:= (AVolume * (ASample-128));
end;


procedure TJSSMixingDevice.vmPlay(AChannel:Longint);
begin
 SChannel[AChannel].ChPlaying:= TRUE;
end;


procedure TJSSMixingDevice.vmStop(AChannel:Longint);
begin
 SChannel[AChannel].ChPlaying:= FALSE;
end;


procedure TJSSMixingDevice.vmSetSample(AChannel:Longint; AData:Pointer; ASize, ALoopS,
                                    ALoopE:Cardinal; AFlags:Byte);
begin
 with SChannel[AChannel] do
 begin
  ChSize:= ASize;
  ChLoopS:= ALoopS;
  ChLoopE:= ALoopE;
  ChFlags:= AFlags;
  ChData:= AData;
  ChDirection:= TRUE;

  ChPos.Int:= 0;
  ChPos.Frac:= 0;
  ChDelta.Int:= 0;
  ChDelta.Frac:= 0;
 end;
end;


procedure vmGetScale(AVal, BVal:Cardinal; AResult:Pointer); assembler; inline;
asm
         mov esi, [AResult];
         xor edx, edx;    { EDX:EAX = AVal }
         mov eax, [AVal]
         mov ebx, [BVal]  {     EBX = BVal }
         or  ebx, ebx;
         jnz @notzero;

         { Division by zero }
         xor eax, eax;
         mov [esi + 4], eax;
         jmp @ok;

@notzero:div ebx          { EDX:EAX / EBX }

         mov [esi + 4], eax;

         xor eax, eax;    { Divide again since EDX:EAX = Remainder SHL 32 ! }
         div ebx;         { And EBX = BVal .. ingenious ;-) }

@ok:     mov [esi], eax;
end;


procedure TJSSMixingDevice.vmSetFreq(AChannel:Longint; AFreq:Cardinal);
begin
 with SChannel[AChannel] do
 begin
  ChFreq      := AFreq;
  vmGetScale(AFreq, SOutFreq, @ChDelta);
 end;
end;


procedure TJSSMixingDevice.vmGetFreq(AChannel:Longint; var AFreq:Cardinal);
begin
 AFreq:= SChannel[AChannel].ChFreq;
end;


procedure TJSSMixingDevice.vmSetVolume(AChannel:Longint; AVolume:Byte);
begin
 SChannel[AChannel].ChVolume:= AVolume;
end;


procedure TJSSMixingDevice.vmGetVolume(AChannel:Longint; var AVolume:Byte);
begin
 AVolume:= SChannel[AChannel].ChVolume;
end;


procedure TJSSMixingDevice.vmSetPosition(AChannel:Longint; APos:Cardinal);
begin
 with SChannel[AChannel] do
 begin
  ChPos.Int:= APos;
  ChPos.Frac:= 0;
 end;
end;


procedure TJSSMixingDevice.vmGetPosition(AChannel:Longint; var APos:Cardinal);
begin
 with SChannel[AChannel] do
 APos:= ChPos.Int;
end;


procedure TJSSMixingDevice.vmSetPanning(AChannel:Longint; APanning:Byte);
begin
 SChannel[AChannel].ChPanning:= APanning;
end;


procedure TJSSMixingDevice.vmGetPanning(AChannel:Longint; var APanning:Byte);
begin
 APanning:= SChannel[AChannel].ChPanning;
end;


procedure TJSSMixingDevice.vmMuteChannel(AChannel:Longint; AMute:Boolean);
begin
 SChannel[AChannel].ChMute:= AMute;
end;


procedure TJSSMixingDevice.vmClearChannel(AChannel:Longint);
begin
 {** Clear one mixing device channel **}
 with SChannel[AChannel] do
 begin
  ChPos.Int    := 0;
  ChPos.Frac   := 0;
  ChDelta.Int  := 0;
  ChDelta.Frac := 0;
  ChFreq       := 0;
  ChSize       := 0;
  ChLoopS      := 0;
  ChLoopE      := 0;
  ChData       := NIL;
  
  ChPlaying    := FALSE;
  ChMute       := FALSE;
  ChDirection  := TRUE;
  ChFlags      := 0;
  ChVolume     := 0;
  ChPanning    := jchPanMiddle;
 end;
end;


{****************************************************************************}
{*                                                                          *}
{* J Sound System initialization and shutdown routines                      *}
{*                                                                          *}
{****************************************************************************}
procedure jssWarningProc(AFrom, AMessage:String);
begin
 writeln(AFrom, ': ', AMessage);
end;

procedure jssErrorProc(AFrom, AMessage:String);
begin
 writeln(AFrom, ': ', AMessage);
end;

{ General messages routines }
procedure jssMsgProc(AFrom, AWhat:String);
begin
 writeln(jssName +'_', AFrom, ': ', AWhat);
end;


procedure jssSMsgProc(AFrom, AWhat:String);
begin
 write(jssName +'_', AFrom, ': ', AWhat);
end;


procedure jssPMsgProc(AMessage:String);
begin
 writeln(AMessage);
end;


{ Loads a JSS configuration file to a given structure.
}
function jssLoadConfig(AFilename:String; var AConfig:TJSSConfigFile):Longint;
 var AConfigF:File;
begin
 {** Open the file **}
 Assign(AConfigF, AFilename);
 Reset(AConfigF, 1);
 if (IOResult <> 0) then
    begin
     jssErrorN:= jerrFileOpenError;
     exit(jssErrorN);
    end;

 {** Read the configuration **}
 blockread(AConfigF, AConfig, sizeof(TJSSConfigFile));
 if (IOResult <> 0) then
    begin
     jssErrorN:= jerrFileReadError;
     exit(jssErrorN);
    end;

 close(AConfigF);
 jssLoadConfig:= jerrOK;
end;


{ Saves a JSS configuration file from a given structure.
}
function jssSaveConfig(AFilename:String; AConfig:TJSSConfigFile):Longint;
 var AConfigF:File;
begin
 {** Create the file **}
 Assign(AConfigF, AFilename);
 Rewrite(AConfigF, 1);
 if (IOResult <> 0) then
    begin
     jssErrorN:= jerrFileCreateError;
     exit(jssErrorN);
    end;

 {** Write the configuration **}
 blockwrite(AConfigF, AConfig, sizeof(TJSSConfigFile));
 if (IOResult <> 0) then
    begin
     jssErrorN:= jerrFileWriteError;
     exit(jssErrorN);
    end;

 close(AConfigF);
 jssSaveConfig:= jerrOK;
end;


{ Sets ANY flags
}
procedure jssSetANYFlag(var AValue:Cardinal; AFlag:Cardinal);
begin
 AValue:= (AValue OR AFlag);
end;


{ Gets ANY flags.
}
function jssGetANYFlag(AValue, AFlag:Cardinal):Boolean;
begin
 jssGetANYFlag:= ((AValue AND AFlag) = AFlag);
end;


{****************************************************************************}
{*                                                                          *}
{* Internal Memory manager routines                                         *}
{*                                                                          *}
{****************************************************************************}
{$ifdef PLATFORM_DOS}
function jssMM_GetMem(Size:Longint):Pointer;
 var AResult:Pointer;
begin
 AResult:= jssMM_OldManager.GetMem(Size);

 if (AResult <> NIL) then
 if (Lock_Data(AResult^, ASize) = FALSE) then
    jssError('jssMM_GetMem', 'Could not LOCK allocated region, DPMI error 0x'+ ioStrL(Int31Error));

 jssMM_GetMem:= AResult;
end;


function jssMM_FreeMem(var P:Pointer):Longint;
begin
 if (P <> NIL) then
 if (Lock_Data(P^, ASize) = FALSE) then
    jssError('jssMM_GetMem', 'Could not UNLOCK allocated region, DPMI error 0x'+ ioStrL(Int31Error));

 jssMM_FreeMem:= jssMM_OldManager.FreeMem(P); 
end;


function jssMM_FreeMemSize(var P:Pointer; Size:Longint):Longint;
begin
end;


function jssMM_AllocMem(Size:Longint):Pointer;
begin
end;


function jssMM_ReAllocMem(var P:Pointer; Size:Longint):Pointer;
begin
end;


 const
      jssMM_Manager:TMemoryManager =
      (
       GetMem      : jssMM_GetMem; 
       FreeMem     : jssMM_FreeMem;
       FreeMemSize : jssMM_FreeMemSize;
       AllocMem    : jssMM_AllocMem;
       ReAllocMem  : jssMM_ReAllocMem;
       MemSize     : MemSize;
       MemAvail    : MemAvail;
       MaxAvail    : MaxAvail;
       HeapSize    : HeapSize;
      );
      

{ DOS/GO32V2: End of code to be locked }
procedure jssLOCKCODE_END; begin end;
{$endif}


{ General J Sound System initialization function.
}
function jssInit:Longint;
 var AIndex:Longint;
begin
{$ifdef PLATFORM_DOS}
 {** Lock the data and code if under DOS/DPMI **}
 Lock_Data(DOSMemSelector, SizeOf(DOSMemSelector));
 
 Lock_Data(jssLOCKDATA_START,
 longint(@jssLOCKDATA_END) - longint(@jssLOCKDATA_START));
 
 Lock_Code(jssLOCKCODE_START,
 longint(@jssLOCKCODE_END) - longint(@jssLOCKCODE_START));

 {** Initialize our internal memory manager **}
 GetMemoryManager(jssMM_OldManager);
 SetMemoryManager(jssMM_Manager);
 
 {** Initialize the IRQ handler @SELF table **}
 for AIndex:= 0 to jsetMaxIRQ do
     jssIRQHandlerSelfTable[AIndex]:= NIL;
{$endif}

 {** Some pointer & routine inits **}
 jssWarning := @jssWarningProc;
 jssError   := @jssErrorProc;
 jssMsg     := @jssMsgProc;
 jssMsgS    := @jssSMsgProc;
 jssMsgP    := @jssPMsgProc;

 jssInit:= jerrOK;
end;


{ General J Sound System shutdown function.
}
function jssClose:Longint;
begin
{$ifdef PLATFORM_DOS}
 {** UnLock the data and code if under DOS/DPMI **}
 UnLock_Data(DOSMemSelector, SizeOf(DOSMemSelector));
 
 UnLock_Data(jssLOCKDATA_START,
 longint(@jssLOCKDATA_END) - longint(@jssLOCKDATA_START));
 
 UnLock_Code(jssLOCKCODE_START,
 longint(@jssLOCKCODE_END) - longint(@jssLOCKCODE_START));

 {** Uninstall our memory manager **}
 SetMemoryManager(jssMM_OldManager);
{$endif}

 jssClose:= jerrOK;
end;


end.
