{
  J Sound System - DMA (Direct Memory Access) Handling Routines
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.
}


{ Allocates a DMA-block bound buffer for DMA transfer.
}
function TJSSSoundDevice.vmDMA_AllocBuffer(ABufSize:Cardinal; var ABuffer:TJSSDMAAddr):Boolean;
 var AResult:Longint;
begin
 {** Try to allocate DOS memory from lowest 1MB area **}
 vmDMA_AllocBuffer:= FALSE;
 AResult:= Global_DOS_Alloc(ABufSize);
 if (Int31Error <> 0) then
    begin
     jssErrorN:= -1;
     jssError('vmDMA_AllocBuffer', 'Error allocating DOS memory block of '+ ioStrL(ABufSize) +' bytes.');
     exit;
    end;

 {** Initialize the information **}
 ABuffer.ASelector:= word(AResult);
 ABuffer.ARealSeg:= (AResult SHR 16);
 ABuffer.ALinearAddr:= (longint(ABuffer.ARealSeg) SHL 4);
 ABuffer.AUsed:= TRUE;

 vmDMA_AllocBuffer:= TRUE;
end;


{ De-Allocates a DMA-block bound buffer.
}
function TJSSSoundDevice.vmDMA_DeAllocBuffer(var ABuffer:TJSSDMAAddr):Boolean;
begin
 {** Try to de-allocate the given block **}
 vmDMA_DeAllocBuffer:= FALSE;
 if (ABuffer.AUsed) then
 begin
  Global_DOS_Free(ABuffer.ASelector);
  if (Int31Error <> 0) then
     begin
      jssErrorN:= -1;
      jssError('vmDMA_DeAllocBuffer', 'Error de-allocating DOS memory! (sel='+ ioHexW(ABuffer.ASelector) +')');
      exit;
     end;
 end;

 {** Reset the information **}
 ABuffer.AUsed:= FALSE;
 ABuffer.ASelector:= 0;
 ABuffer.ARealSeg:= 0;
 ABuffer.ALinearAddr:= 0;

 vmDMA_DeAllocBuffer:= TRUE;
end;


{ Initializes DMA channel & all relating data according to SOutMode16.
  Exits with error if DMA settings are invalid.
}
function TJSSSoundDevice.vmDMA_Initialize:Boolean;
 const jProcName = 'vmDMA_Initialize';
begin
 {** Initialize DMA-channel data **}
 vmDMA_Initialize:= FALSE;

 if (SOutMode16bit) and
    (SDMA > 4) then
 begin
  {** We SHOULD have a 16-bit channel **}
  jssMsg(jProcName, 'Initializing 16-bit DMA channel #'+ ioStrL(SDMA));

  if (SDMA = jsetNotSet) then
     begin
      jssErrorN:= jerrDMAchannelNotDefined;
      jssError(jProcName, '16-bit DMA channel not selected!');
      exit;
     end;

  if (SDMA > 7) then
     begin
      jssErrorN:= jerrInvalidDMAchannel;
      jssError(jProcName, 'Invalid 16-bit DMA channel selected (#'+ ioStrL(SDMA) +').');
      exit;
     end;

 end else
 if (SOutMode16bit = FALSE) or
    ((SOutMode16bit) and (SDMA <= 4)) then
 begin
  {** We SHOULD have a 8-bit channel **}
  SOutMode16bit:= FALSE;
  jssMsg(jProcName, 'Initializing 8-bit DMA channel #'+ioStrL(SDMA));

  if (SDMA = jsetNotSet) then
     begin
      jssErrorN:= jerrDMAchannelNotDefined;
      jssError(jProcName, '8-bit DMA channel not selected.');
      exit;
     end;

  if (SDMA > 3) or (SDMA < 0) then
     begin
      jssErrorN:= jerrInvalidDMAchannel;
      jssError(jProcName, 'Invalid 8-bit DMA channel selected (#'+ ioStrL(SDMA) +').');
      exit;
     end;

 end else
     begin
      jssErrorN:= jerrUnsupportedOutMode;
      jssError(jProcName, 'Unsupported output mode!');
      exit;
     end;

 {** Initialize rest of data **}
 SDMAAddrReg:= jDMAAddrReg[SDMA];
 SDMAPageReg:= jDMAPageReg[SDMA];
 SDMASizeReg:= jDMASizeReg[SDMA];

 jssMsg(jProcName, 'AddrReg=0x'+ioHexW(SDMAAddrReg)
                    +', PageReg=0x'+ioHexW(SDMAPageReg)
                    +', SizeReg=0x'+ ioHexW(SDMASizeReg));

 vmDMA_Initialize:= TRUE;
end;


function TJSSSoundDevice.vmDMA_Stop:Boolean;
 const jProcName = 'vmDMA_Stop';
begin
 {** Stop current DMA transfer **}
 vmDMA_Stop:= FALSE;

 if (SOutMode16bit) then
    begin
     {** Stop 16-bit DMA transfer **}
     outportb($D4, (SDMA OR $04));
     outportb($D8, 0);
     outportb($D4, SDMA);
    end else
    begin
     {** Stop 8-bit DMA transfer **}
     outportb($0A, (SDMA OR $04));
     outportb($0C, 0);
     outportb($0A, SDMA);
    end;

 vmDMA_Stop:= TRUE;
end;


function TJSSSoundDevice.vmDMA_SetData(ABuffer:TJSSDMAAddr; ABufSize:Word):Boolean;
 var ASize, APageOffs:Word;
begin
 {** Check the transfer parameters **}
 ASize:= (ABufSize - 1);

 vmDMA_SetData:= FALSE;
 if (ASize <= 0) then
    begin
     jssErrorN:= jerrInvalidDMADataSize;
     jssError('vmDMA_SetData', 'Invalid DMA data block size '+ ioStrL(ASize) +'.');
     exit;
    end;

 if (ABuffer.AUsed = FALSE) or
    (ABuffer.ASelector = 0) then
    begin
     jssErrorN:= jerrInvalidDMADataBlock;
     jssError('vmDMA_SetData', 'Invalid DMA data block!');
     exit;
    end;

 {** Set up the transfer **}
 if (SDMA < 4) then
    begin
     {** Program DMA channels 0-3 **}
     { Select the dma channel by masking right bits }
     outportb($0A, (4 OR (SDMA AND 3)));

     { Clear the DMA-flipflop for reprogramming }
     outportb($0C, 0);

     {** Set DMA transfer mode **}
     outportb($0B, ($48 OR (SDMA AND 3)));

     {** Set DMA-data memory address **}
     APageOffs:= (ABuffer.ALinearAddr MOD 65536);
     outportb(SDMAAddrReg, LO(APageOffs));
     outportb(SDMAAddrReg, HI(APageOffs));

     {** Set DMA data transfer length **}
     outportb(SDMASizeReg, LO(ASize));
     outportb(SDMASizeReg, HI(ASize));

     outportb(SDMAPageReg, (ABuffer.ALinearAddr SHR 16));

     {** Activate the DMA channel **}
     outportb($0A, (SDMA AND 3));
    end else
    begin
     {** Program DMA channels 4-7 **}
     { Select the dma channel by masking right bits }
     outportb($D4, (4 OR (SDMA AND 3)));

     { Clear the DMA-flipflop for reprogramming }
     outportb($D8, 0);

     {** Set DMA transfer mode **}
     outportb($D6, ($48 OR (SDMA AND 3)));

     {** Set DMA-data memory address **}
     APageOffs:= (ABuffer.ALinearAddr MOD 65536);
     outportb(SDMAAddrReg, LO(APageOffs));
     outportb(SDMAAddrReg, HI(APageOffs));

     {** Set DMA data transfer length **}
     outportb(SDMASizeReg, LO(ASize));
     outportb(SDMASizeReg, HI(ASize));

     outportb(SDMAPageReg, (ABuffer.ALinearAddr SHR 16));

     {** Activate the DMA channel **}
     outportb($D4, (SDMA AND 3));
    end;

 vmDMA_SetData:= TRUE;
end;

