{
  J Sound System - DOS/Go32v2 general IRQ handler routine
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.
}

{ This routine handles all incoming interrupt requests (for
  the IRQs that JSS drivers have set up), looks up the responsible
  driver and passes control to it. After the driver's handler
  routine returns, we acknowledge PIC and return from isr.
}

procedure jssGeneric_IRQ_Handler; assembler;
asm
        {** Save registers to stack **}
        cli;
        pushfd;    { 32-bit flags }
        pushad;    { EAX, EBX, ECX, EDX, ESI, EDI, EBP, [and ESP] }
        push ds;   { And the segment-registers }
        push es;
        push fs;
        push gs;

        {** Initialize datasegment **}
        mov ax, cs:[engBackupDS];
        mov ds, ax;
        mov es, ax;

        mov ax, DOSMemSelector;
        mov fs, ax;

        {** Call the IRQ-handler routine **}
        { First, we determine the IRQ we are serving...
          Let's ask the #1 PIC if it knows.
        }
        mov al, $0B;     { 00001011 B = Read interrupt-in-service }
        out $20, al;
        in  al, $20;     { Read the service bitfield }

        xor ebx, ebx;    { Adjust result by 0 }

        test al, $08;    { Is it the PIC2 hook-IRQ? }
        jz   @bgetb;

        { If PIC1 said it was PIC2, let's ask him! }
        mov al, $0B;     { 00001011 B = Read interrupt-in-service }
        out $a0, al;
        in  al, $a0;     { Read the service bitfield }

        mov ebx, 8;      { Adjust result by +8 }

        { Determine IRQ # from bits }
@bgetb: mov ecx, 8;      { Max. number of bits }
@bloop: test al, 1;      { Test bit, if 1 then exit }
        jnz  @bfound;
        shr  al, 1;      { If not, continue to next bit }
        inc  ebx;        { Increase our result counter }
        dec  ecx;        { Let's not loop forever! }
        jnz  @bloop;

        { Error! No bit was set!!! This should not happen?
          But if it happens, we acknowledge both PICs and
          exit as fast as we can to avoid deadlocking the IRQs
        }
        mov al, $20;
        out $20, al;
        out $a0, al;
        jmp  @bExit;

@bfound:{ Now we have the IRQ's number }
        push ebx;               { Let's save it for later use... }

        lea edi, jssIRQHandlerTable;    { Get the address of handler }
        mov edi, [edi + ebx*4]; { To EDI }
        or  edi, edi;           { Test for NIL }
        jz  @bSkip;

        lea esi, jssIRQHandlerSelfTable;{ Get the value of @Self }
        mov esi, [esi + ebx*4]; { To ESI (where FPC expects it to be) }

        push esi;               { Put it also to stack }


        {** Call the handler method! **}
        call cs:[edi];



@bSkip: {** Acknowledge PIC(s) that IRQ is about to end **}
        mov al, $20;

        pop ebx;         { Get IRQ number from stack }
        cmp ebx, 8;      { Check it }
        jb  @bThis;
        
        out $a0, al;     { if >= 8, acknowledge both! }
@bThis: out $20, al;     { if <  8, only PIC#1 }

@bExit: {** Restore the registers and exit IRQ **}
        pop gs;
        pop fs;
        pop es;
        pop ds;
        popad;
        popfd;
        sti;
        iret;
end;
