{
  J Sound System - Mono channel mixing routine
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "cCr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  Mixes one channel into addition buffer with given length.
}

procedure vmMixChannel(AChannel:PJSSChannel; AAddBuffer:Pointer; SMixLengthUnits:Cardinal);
 var
    AGlobalVol,
    AFinalVol:Byte;
    AVolTable8,
    ACurrMixPos:Pointer;
    AMixEndPos,
    ACurrMixLeft:Cardinal;
    AChDelta:TJFPRec;
begin
 {** Initialize mixing of one channel **}
 ACurrMixPos:= AAddBuffer;        { Position in ADD-BUFFER }
 ACurrMixLeft:= SMixLengthUnits;  { Length of block to mix, in UNITS }
 AVolTable8:= SVolTable8;
 AGlobalVol:= SGlobalVol;

 asm
          {** Mix the channel data to add-buffer **}
@jMixLp:  mov  eax, [ACurrMixLeft];
          cmp  eax, 0
          jle  @jExit;


          {** Now, after that we check if we are playing .. **}
          mov esi, [AChannel];
          mov  al, TJSSChannel[esi].ChPlaying;
          or   al, al;
          jz   @jExit;


          {** Now check for MUTE **}
          mov  al, TJSSChannel[esi].ChMute;
          or   al, al;
          jnz  @jExit;  { !FIXME FIX ME !! We should actually implement
                          the muted channel update process! }


          {**** Check and update the channel data ****}
@jMixOk:  mov cl, TJSSChannel[esi].ChFlags;

          { Check the direction of sample playing }
          mov al, TJSSChannel[esi].ChDirection;
          or  al, al;
          jz  @bwPlay; { "FALSE" = 0 = Play backwards }



          {******** Channel is playing FORWARDS ********}
@fwPlay:  test cl, jsfLooped;
          jz   @fwNoLP;


          {** SAMPLE: LOOPED **}
          mov eax, TJSSChannel[esi].ChPos.INT;
          mov ebx, TJSSChannel[esi].ChLoopE;
          cmp eax, ebx;
          jl  @fwJLP;  { ?? JLE ?? }

          {** Check for BIDI loop **}
          test cl, jsfBiDi;
          jz   @fwNorm;

          {:: BiDi loop ::}
          mov TJSSChannel[esi].ChDirection, 0; { Change to "0" = Backwards play }

          mov TJSSChannel[esi].ChPos.Int, ebx; { ChPos --> ChLoopE:0 }
          xor ebx, ebx;
          mov TJSSChannel[esi].ChPos.Frac, ebx;
          
          mov ebx, TJSSChannel[esi].ChLoopS;   { AMixEndPos --> ChLoopS }
          mov [AMixEndPos], ebx;
          jmp @jMixStart;


@fwNorm:  {:: Normal loop ::}
          sub eax, ebx;

          mov edx, TJSSChannel[esi].ChLoopS;
          add edx, eax;
          mov TJSSChannel[esi].ChPos.Int, edx;

          xor eax, eax;
          mov TJSSChannel[esi].ChPos.Frac, eax;

@fwJLP:   mov [AMixEndPos], ebx;
          jmp @jMixStart;


          {** SAMPLE: NOT LOOPED **}
@fwNoLP:  mov eax, TJSSChannel[esi].ChPos.INT;
          mov ebx, TJSSChannel[esi].ChSize;
          cmp eax, ebx;
          jl  @fwJLP;    { ?? JLE ?? }

          { Playing stops here, so exit. }
          mov TJSSChannel[esi].ChPlaying, 0;
          jmp @jExit;


          {******** Channel is playing BACKWARDS ********}
@bwPlay:  test cl, jsfLooped;
          jz   @bwNoLP;


          {** SAMPLE: LOOPED **}
          mov eax, TJSSChannel[esi].ChPos.INT;
          mov ebx, TJSSChannel[esi].ChLoopS;
          cmp eax, ebx;
          jg  @bwJLP;    { ?? JGE ?? }

          {** Check for BIDI loop **}
          test cl, jsfBiDi;
          jz   @bwNorm;

          {:: BiDi loop ::}
          mov TJSSChannel[esi].ChDirection, 1; { Change to "1" = Forwards play }

          mov TJSSChannel[esi].ChPos.Int, ebx; { ChPos --> ChLoopS:0 }
          xor ebx, ebx;
          mov TJSSChannel[esi].ChPos.Frac, ebx;
          
          mov ebx, TJSSChannel[esi].ChLoopS;   { AMixEndPos --> ChLoopE }
          mov [AMixEndPos], ebx;
          jmp @jMixStart;


@bwNorm:  {:: Normal loop ::}
          mov edx, ebx;
          
          sub ebx, eax;

          mov eax, TJSSChannel[esi].ChLoopE;
          sub eax, ebx;
          mov TJSSChannel[esi].ChPos.Int, eax;

          xor eax, eax;
          mov TJSSChannel[esi].ChPos.Frac, eax;

          mov ebx, edx;

@bwJLP:   mov [AMixEndPos], ebx;
          jmp @jMixStart;


          {** SAMPLE: NOT LOOPED **}
@bwNoLP:  mov eax, TJSSChannel[esi].ChPos.Int;
          mov ebx, TJSSChannel[esi].ChDelta.Int;
          cmp eax, ebx;
          jg  @bwJLP;   { ?? JGE ?? }

          { Playing stops here, so exit. }
          mov TJSSChannel[esi].ChPlaying, 0;
          jmp @jExit;


{****************************************************************************}
{*                                                                          *}
{* Actual start of mixing ..                                                *}
{*                                                                          *}
{****************************************************************************}
@jMixStart:
          {** Compute "finalvol" **}
          { AFinalVol:= ((ChVolume * SGlobalVol) SHR 8); }
          xor eax, eax;
          mov al, TJSSChannel[esi].ChVolume;
          mov dl, [AGlobalVol];
          mul dl;
          shr eax, 8;
          mov [AFinalVol], al;


          {** Initialize registers for mixing **}
          mov ebx, TJSSChannel[esi].ChPos.Frac; { EDX:EBX = ChPos }
          mov edx, TJSSChannel[esi].ChPos.Int;

          mov edi, TJSSChannel[esi].ChData;     { DS:EDI = ChData }

          { Copy the delta values to local vars for optimizing the inner loop }
          mov eax, TJSSChannel[esi].ChDelta.Frac;
          mov [AChDelta.Frac], eax;
          mov eax, TJSSChannel[esi].ChDelta.Int;
          mov [AChDelta.Int], eax;

          {
           Now we re-calculate the ACurrMixLeft because this way
           we can optimize the innerloop of mixing. (By removing
           the check for end of sample)

           ASampleDataNeededToMixLeft = abs(ChPos.Int - AMixEndPos) / ChDelta;
                    
          push edx;
          push ecx;

	  mov eax, edx;
          sub eax, [AMixEndPos];
 
          mov ecx, eax;          { Compute abs(eax) }
          sar ecx, 31;
          xor eax, ecx;
          sub eax, ecx;

          xor edx, edx;          { EDX:EAX = abs(ChPos.Int - AMixEndPos) } 
          mov ecx, TJSSChannel[esi].ChDelta.Int;
          div ecx;               { EAX = Quotient, EDX = Remainder }

          pop ecx;
          pop edx;
          }
        
          { Get the left units to mix to ECX }
          mov ecx, [ACurrMixLeft];


          {** Then we check the type of the sample **}
          mov  al, TJSSChannel[esi].ChFlags;
          test al, jsf16bit;
          jz   @lmix8;

{****************************************************************************}

	  {** Mix a 16-bit sample to the addition buffer **}
          { Check the direction of mixing }
@lmix16:  mov  al, TJSSChannel[esi].ChDirection;
          or   al, al;
          jz   @lmixBWi16;

          {**** MIX CHANNEL FORWARDS ****}
          mov esi, [ACurrMixPos];           { Current mixing position }

@lmixFW16:
          push edx
          push ebx
          xor ebx, ebx;
          xor eax, eax;
          mov bl, [AFinalVol];              { AH = ChVolume }
          mov ax, [edi + edx*2];        { Get byte of data }
          xor edx, edx;
          sub eax, 32768;
          imul ebx;
          pop ebx;
          sar eax, 8;
          pop edx;

          add dword ptr [esi], eax;         { And add to add-buffer! }
          add esi, 4;

          xor eax, eax;

          { Change sample position }
          add ebx, [AChDelta.Frac];
          adc edx, [AChDelta.Int];

          { Check if we have mixed enough data? }
          dec ecx;
          jz  @lmmOK0;

          cmp edx, [AMixEndPos];
          jl  @lmixFW16;

@lmmOK0:  {** MIX OK, CONTINUE **}
          jmp @lmmOK;


          {**** MIX CHANNEL BACKWARDS ****}
@lmixBWi16:
          mov esi, [ACurrMixPos];           { Current mixing position }

@lmixBW16:
          push edx
          push ebx
          xor ebx, ebx;
          xor eax, eax;
          mov bl, [AFinalVol];              { AH = ChVolume }
          mov ax, [edi + edx*2];        { Get byte of data }
          xor edx, edx;
          sub eax, 32768;
          imul ebx;
          pop ebx;
          sar eax, 8;
          pop edx;

          add dword ptr [esi], eax;         { And add to add-buffer! }
          add esi, 4;

          xor eax, eax;

          { Change sample position }
          sub ebx, [AChDelta.Frac];
          sbb edx, [AChDelta.Int];

          { Check if we have mixed enough data? }
          dec ecx;
          jz  @lmmOK1;

          cmp edx, [AMixEndPos];
          jg  @lmixBW16;

@lmmOK1:  {** MIX OK, CONTINUE **}
          jmp @lmmOK;

{****************************************************************************}

          {** Mix a 8-bit sample to the addition buffer **}
          { Check the direction of mixing }
@lmix8:   mov  al, TJSSChannel[esi].ChDirection;
          or   al, al;
          jz   @lmixBWi8;

          {**** MIX CHANNEL FORWARDS ****}
          xor eax, eax;                     { Reset EAX }
          mov esi, [ACurrMixPos];           { Current mixing position }

@lmixFW8:
          mov ah, [AFinalVol];              { AH = ChVolume }
          push ebp;
          mov al, [edi + edx];              { Get byte of data }
          mov ebp, [AVolTable8];            { Volume table pointer }
          mov eax, dword ptr [ebp + eax*4]; { Convert it through voltable }
          pop ebp;

          add dword ptr [esi], eax;         { And add to add-buffer! }
          add esi, 4;

          xor eax, eax;

          { Change sample position }
          add ebx, [AChDelta.Frac];
          adc edx, [AChDelta.Int];

          { Check if we have mixed enough data? }
          dec ecx;
          jz  @lmmOK2;

          cmp edx, [AMixEndPos];
          jl  @lmixFW8;

@lmmOK2:  {** MIX OK, CONTINUE **}
          jmp @lmmOK;

          {**** MIX CHANNEL BACKWARDS ****}
@lmixBWi8:xor eax, eax;                     { Reset EAX }
          mov esi, [ACurrMixPos];           { Current mixing position }

@lmixBW8:
          mov ah, [AFinalVol];              { AH = ChVolume }
          push ebp;
          mov al, [edi + edx];              { Get byte of data }
          mov ebp, [AVolTable8];            { Volume table pointer }
          mov eax, dword ptr [ebp + eax*4]; { Convert it through voltable }
          pop ebp;

          add dword ptr [esi], eax;         { And add to add-buffer! }
          add esi, 4;

          xor eax, eax;

          { Change sample position }
          sub ebx, [AChDelta.Frac];
          sbb edx, [AChDelta.Int];

          { Check if we have mixed enough data? }
          dec ecx;
          jz  @lmmOK;

          cmp edx, [AMixEndPos];
          jg  @lmixBW8;

          {** MIX OK, CONTINUE **}

{****************************************************************************}

          {** End of block mixing **}
@lmmOK:   { Save the new sample position }
          mov [ACurrMixPos], esi;
          
          mov esi, [AChannel];
          mov TJSSChannel[esi].ChPos.Frac, ebx;
          mov TJSSChannel[esi].ChPos.Int, edx;

          {** Restore the registers **}
          mov [ACurrMixLeft], ecx;

          { Is there something to mix left? }
          or   ecx, ecx;
          jnz  @jMixOk;

@jExit:   {** Mixing of the block is now done **}
 end;
end;

