{
  J Sound System - Impulse Tracker (IT) module loader
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  Self explanatory.

  Bugs: Countless. Too many. Does not do anything. Just can't stand it ;)
}

{$ifdef GET_IMPLEMENTATION}
 const
      {** IT data string **}
      IT_DataString = 'IT loader v0.05 (updated 17/8/2000)';
{$endif}

{$ifdef GET_REGISTRATION}
 jssRegisterLoader(jmdtIT, @jssLoadIT, @jssDetectIT, 'IT', IT_DataString);
{$endif}


{$ifdef GET_IMPLEMENTATION}
procedure jssLoadIT(var AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);

 const
      jProcName = 'jssLoadIT';

      {** IT definations **}
      IT_MaxChannels    = 64;
      IT_MaxPatterns    = 100;
      IT_MaxOrders      = 255;
      IT_MaxInstruments = 100;
      IT_MaxSamples     = 100;

      {** IT module flag constants **}
      itMFStereo        = $0001; { Use Stereo output (if possible) }
      itMFVol0Opt       = $0002; { Use Volume 0 Optimizations }
      itMFUseInstruments= $0004; { TRUE=Instruments Mode, FALSE=Sample Mode }
      itMFLinearSlides  = $0008; { TRUE=Use Linear Slides, FALSE=Amiga sld }
      itMFUseOldEffects = $0010; { Use Old Effects (see IT docs) }
      itMFLinkGxxToFE   = $0020; { Link Gxx effect data with Fxx and Exx,
                                   see IT format docs for more info! }

      {** IT sample flag constants **}
      itSFUsed          = $01;  { Sample is used }
      itSF16bit         = $02;  { 16-bit sample }
      itSFStereo        = $04;  { Stereo sample (not supported in IT) }
      itSFLooped        = $10;  { Looped sample }
      itSFSustain       = $20;  { Loop is a sustain loop }
      itSFBiDi          = $40;  { Bi-directional loop }
      itSFBiDiSustain   = $80;  { Bi-directional sustain loop }

 type
     {** IT-module structure defs **}
     {** IT-header **}
     TITHeader = record
      AIdent:String4;            { IT signature/identificator "IMPM" }
      ASongName:String26;        { Module songname in ASCIIZ }
      AUnused1, AUnused2:Byte;   { RESERVED }
      NOrders,                   { Number of orders }
      NInstruments,              { Number of instruments }
      NSamples,                  { Number of samples }
      NPatterns,                 { Number of patterns }
      CWTVersion,                { Created With Tracker Version }
      CMWTVersion,               { Compatible With Tracker Version }
      AFlags,                    { Module flags }
      ASpecial:Word;             { Special }

      AGlobalVol,                { Global volume }
      AMasterVol,                { Master volume }
      ADefSpeed,                 { Default (initial) speed }
      ADefTempo,                 { Default tempo }
      ADefPanningSep,            { Panning separation }
      AUnused3:Byte;             { RESERVED }

      AMsgLength:Word;
      AMsgOffset,
      AUnused4:Longint;

      ADefPanning,               { Default channel panning and volume }
      ADefVolume:Array[1..IT_MaxChannels] of Byte;
     end;


     {** IT Instrument formats **}
     { First we have this new version format (CMWT >= v2.00) }
     TITNewInstrument = record
      AIdent:String4;            { Instrument header signature "IMPI" }
      AFilename:String12;        { Filename of instrument in ASCIIZ }
      ARESERVED1,                { Always 0x00 }
      NNAType,                   { Instrument flags }
      DNCType,                   { Duplicate Note Check Type }
      DNCAction:Byte;            { Duplicate Note Check Action }
      AFadeOut:Word;             { Fadeout value }
      APitchPanSep,              { Pitch Panning Separation }
      APitchPanCenter,           { Pitch Panning Center }
      AGlobalVol,                { Global Volume }
      ADefPan,                   { Default Panning Position }
      ARandVolVar,               { Random Volume Variation % }
      ARandPanVar:Byte;          { Random Panning Pos Variation % }
      CWTVersion:Word;           { Tracker version [only in inst-files] }
      ARESERVED2,                { Number of samples [only in inst-files] }
      ARESERVED3:Byte;           { RESERVED }
      ADescription:String26;     { Instrument description }
      ARESERVED4:Word;           { RESERVED }
      AMidiChan,                 { MIDI channel }
      AMidiProg:Byte;            { MIDI program }
      ARESERVED5:Word;           { RESERVED }
     end;

     { And then the old instrument format (CMWT < v2.00) }
     TITOldInstrument = record
      AIdent:String4;            { Instrument header signature "IMPI" }
      AFilename:String12;        { Filename of instrument in ASCIIZ }
      ARESERVED1,                { Always 0x00 }
      AFlags,
      AVolumeLoopS,
      AVolumeLoopE,
      ASustainLoopS,
      ASustainLoopE:Byte;
      ARESERVED2,
      AFadeOut:Word;
      NNAType,
      DNCType:Byte;
      ATrkVersion:Word;
      ANSamples,
      ARESERVED3:Byte;
      ADescription:String26;     { Instrument description }
     end;

     {** IT Sample format **}
     TITSample = record
      AIdent:String4;            { Sample header signature "IMPS" }
      AFilename:String12;        { Filename of the sample (11 chars + #0) }
      AGlobalVol,                { Sample's global volume }
      AFlags,                    { Sample flags, see itSFXXXX constants }
      ADefVol:Byte;              { Default sample volume }
      ADescription:String26;     { Description }
      CWTVersion,                { Created With Tracker Version }
      ADefPan:Byte;              { Default sample panning position }
      ALength,                   { Sample Data Length }
      ALoopS,                    { Sample Loop Start offset }
      ALoopE,                    { Sample Loop End offset }
      C5BaseSpeed,               { C-5 Base Speed frequency }
      ASustainLoopS,             { Sustain Loop Start offset }
      ASustainLoopE,             { Sustain Loop End offset }
      ADataFileOffs:Longint;     { Sample Data's File Offset }
      AVibratoSpeed,             { Vibrato Speed }
      AVibratoDepth,             { Vibrato Depth }
      AVibratoRate,              { Vibrato Rate }
      AVibratoType:Byte;         { Vibrato Type }
     end;


     {** IT Pattern format **}
     TITPattern = record
      ALength:Word;
      NRows:Word;
      AReserved:Longint;
     end;


 var
    {** General loader variables **}
    ITFile:File;             { IT-file }
    ITH:TITHeader;           { IT-header }
    ITI:TITNewInstrument;    { One temp. instrument for converting
                               IT instruments to internal format }

    ITInstHeadOffs:Array[1..IT_MaxInstruments] of Longint;
    ITSampHeadOffs:Array[1..IT_MaxSamples] of Longint;
    ITPatDataOffs:Array[0..IT_MaxPatterns] of Longint;
      
    ATempByte:Byte;
    ATempWord:Word;
    ATempUsed:Boolean;
    ATempInt2,
    ATempInt:Longint;

    ATempBufSize:Longint;
    ATempBufData:Pointer;


procedure IT_AllocTempUnpackBuf(ASize:Longint);
begin
 ATempBufSize:= ASize;
 GetMem(ATempBufData, ATempBufSize);
end;

procedure IT_DeallocTempUnpackBuf;
begin
 if (ATempBufData <> NIL) then
    begin
     FreeMem(ATempBufData, ATempBufSize);
     ATempBufData:= NIL;
     ATempBufSize:= 0;
    end;
end;

procedure IT_UnpackPattern(ASourceData:Pointer; ASourceSize:Longint; var ADestPat:TJSSPattern);
 var
    APrevPatMask,
    ALastCHNote, ALastCHInstr, ALastCHVolume,
    ALastCHEffect, ALastCHParam:Array[jsetMinChannel..jsetMaxChannel] of Byte;

    APatPack,
    APatMask:Byte;
    ASourceOffs,
    APatChannel,
    APatRow:Longint;

function jssGetByte:Byte;
begin
 jssGetByte:= PByte(ASourceData)^;
 inc(longint(ASourceData));
 inc(ASourceOffs);
end;

begin
 {** Check data pointers **}
 if (ASourceData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     IT_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern SOURCE pointer NULL.');
     exit;
    end;

 if (ADestPat.AUsed = FALSE) or
    (ADestPat.AData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     IT_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern DEST pointer NULL.');
     exit;
    end;

 {** Init variables **}
 ASourceOffs:= 0;
 APatRow:= 0;
 FillChar(APrevPatMask, SizeOf(APrevPatMask), 0);
 FillChar(ALastCHNote, SizeOf(ALastCHNote), 0);
 FillChar(ALastCHInstr, SizeOf(ALastCHInstr), 0);
 FillChar(ALastCHVolume, SizeOf(ALastCHVolume), 0);
 FillChar(ALastCHEffect, SizeOf(ALastCHEffect), 0);
 FillChar(ALastCHParam, SizeOf(ALastCHParam), 0);

 {** Pattern unpacking loop **}
 while (APatRow < ADestPat.ARows) and
       (ASourceOffs < ASourceSize) do
 begin
  APatPack:= jssGetByte;

  if (APatPack <> 0) then
  begin
   {** Get channel number and rest of data **}
   APatChannel:= (APatPack and 63);
   writeln('apatchannel = ', apatchannel);
   if ((APatPack AND $80) <> 0)
      then APatMask:= jssGetByte
      else APatMask:= APrevPatMask[APatChannel];

   {**** Check if channel is bigger than maxchannels ****}
   with ADestPat.AData^[APatRow, APatChannel] do
   begin
    Param:= 0;
    if ((APatMask and $01) <> 0) then
       begin
        { MASK $01: Read note/octave }
        Note:= jssGetByte;
        ALastCHNote[APatChannel]:= Note;
       end;

    if ((APatMask and $02) <> 0) then
       begin
        { MASK $02: Read instrument }
        Instr:= jssGetByte;
        ALastCHInstr[APatChannel]:= Instr;
       end;

    if ((APatMask and $04) <> 0) then
       begin
        { PACK $04: Read volume/panning value }
        Volume:= jssGetByte;
        ALastCHVolume[APatChannel]:= Volume;
       end;

    if ((APatMask and $08) <> 0) then
       begin
        { MASK $08: Read command/effect and parameter }
        Effect:= jssGetByte;
        Param:= jssGetByte;
        ALastCHEffect[APatChannel]:= Effect;
        ALastCHParam[APatChannel]:= Param;
       end;

    if ((APatMask and $10) <> 0) then
       begin
        { MASK $10: Note == Last note for this channel }
        Note:= ALastCHNote[APatChannel];
       end;

    if ((APatMask and $20) <> 0) then
       begin
        { MASK $20: Instrument == Last instr for this channel }
        Instr:= ALastCHInstr[APatChannel];
       end;

    if ((APatMask and $40) <> 0) then
       begin
        { MASK $40: Volume == Last volume for this channel }
        Volume:= ALastCHVolume[APatChannel];
       end;

    if ((APatMask and $80) <> 0) then
       begin
        { MASK $80: Command/effect+parameter == Last values for this channel }
        Effect:= ALastCHEffect[APatChannel];
        Param:= ALastCHParam[APatChannel];
       end;

   end;
  end else
      inc(APatRow);

 end; { WHILE }
 {** Unpacking OK. **}
end;

function IT_GetVersionStr(AVersion:Word):String;
begin
 IT_GetVersionStr:= ioHexB(HI(AVersion)) +'.'+ ioHexB(LO(AVersion));
end;


begin
 {******************** Check for internal inconsistencies ******************}
 if (IT_MaxChannels > jsetMaxNChannels) or
    (IT_MaxPatterns > jsetMaxPatterns) or
    (IT_MaxInstruments > jsetMaxInstruments) or
    (IT_MaxOrders > jsetMaxOrders) then
    begin
     jssErrorN:= jerrInternalError;
     jssError(jProcName, 'Internal inconsistency in JLOADIT.PAS');
     exit;
    end;

 {******************* Read IT-header and check data in it *****************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Reading and checking IT header...');
 {$endif}
 AStream^.vmGetData(@ITH, sizeof(TITHeader));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading IT header from stream 0x'+ ioPtr2Hex(AStream));
     exit;
    end;

 if (ITH.AIdent <> 'IMPM') then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Not a real IT module.');
     if (ioFailOnError) then exit;
    end;

 if (ITH.CMWTVersion >= $0214) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Compressed IT format is unsupported (v'+ IT_GetVersionStr(ITH.CMWTVersion) +').');
     if (ioFailOnError) then exit;
    end;

 if (ITH.CMWTVersion < $0200) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Old IT format (older than 2.00) is unsupported (v'+ IT_GetVersionStr(ITH.CMWTVersion) +').');
     if (ioFailOnError) then exit;
    end;

 if (ITH.NPatterns > IT_MaxPatterns) then
    begin
     jssWarning(jProcName, 'Number of patterns > '+ ioStrL(IT_MaxPatterns));
     ITH.NPatterns:= IT_MaxPatterns;
    end;

 if (ITH.NOrders > IT_MaxOrders) then
    begin
     jssWarning(jProcName, 'Number of orders > '+ ioStrL(IT_MaxOrders));
     ITH.NOrders:= IT_MaxOrders;
    end;

 if (ITH.NOrders = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of orders was 0');
     if (ioFailOnError) then exit;
    end;

 if (ITH.NPatterns = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of patterns was 0');
     if (ioFailOnError) then exit;
    end;

 if (ITH.NInstruments > IT_MaxInstruments) then
    begin
     jssWarning(jProcName, 'Number of instruments > '+ ioStrL(IT_MaxInstruments));
     ITH.NInstruments:= IT_MaxInstruments;
    end;

 if (ITH.NSamples > IT_MaxSamples) then
    begin
     jssWarning(jProcName, 'Number of samples > '+ ioStrL(IT_MaxSamples));
     ITH.NSamples:= IT_MaxSamples;
    end;

 {************************* Init internal variables ************************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Initializing and allocating IT module...');
 {$endif}
 if (jssAllocateModule(AModule) <> jerrOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssError(jProcName, 'Error allocating new module structure!');
     exit;
    end;

 with AModule^ do
 begin
  {** Convert the Header Info **}
  ModuleType     := jmdtIT;
  jssASCIIZToStr(@ITH.ASongName, ModuleName, sizeof(ITH.ASongName));
  ModuleInfo     := IT_DataString;
  DefSpeed       := ITH.ADefSpeed;
  DefTempo       := ITH.ADefTempo;
  DefGlobalVol   := ITH.AGlobalVol;
  DefMasterVol   := (ITH.AMasterVol and $7f);
  NInstruments   := ITH.NSamples;
  NExtInstruments:= ITH.NInstruments;
  NPatterns      := ITH.NPatterns;
  NOrders        := ITH.NOrders;
  DefFlags       := 0;

  {** Convert the flags **}


  {** Read and convert orders list **}
  for ATempInt:= 0 to (NOrders - 1) do
  begin
   AStream^.vmGetByte;
   if (AStream^.vmStatus <> strOK) then
      begin
       SUP_MAC_FAIL_STREAM
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading orderlist (item #'+
                ioStrL(ATempInt) +'/'+ ioStrL(ITH.NOrders - 1));
       if (ioFailOnError) then exit;
      end;

   case ATempByte of
    $ff: ATempInt2:= mpOrderEnd;
    $fe: ATempInt2:= mpOrderSkip;
    else ATempInt2:= ATempByte;
   end;
   OrderList[ATempInt]:= ATempInt2;
  end;

  {** Read the additional data header offsets **}
  { Extended instrument header offsets }
  AStream^.vmGetData(@ITInstHeadOffs, (SizeOf(Longint) * NExtInstruments));
  if (AStream^.vmStatus <> strOK) then
     begin
       SUP_MAC_FAIL_STREAM
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading instrument header offsets.');
       if (ioFailOnError) then exit;
     end;

  { Sample (normal) instrument header offsets }
  AStream^.vmGetData(@ITSampHeadOffs, (SizeOf(Longint) * NInstruments));
  if (AStream^.vmStatus <> strOK) then
     begin
       SUP_MAC_FAIL_STREAM
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading sample header offsets.');
       if (ioFailOnError) then exit;
     end;

  { Pattern data (header) offsets }
  AStream^.vmGetData(@ITPatDataOffs, (SizeOf(Longint) * NPatterns));
  if (AStream^.vmStatus <> strOK) then
     begin
       SUP_MAC_FAIL_STREAM
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading pattern data offsets.');
       if (ioFailOnError) then exit;
     end;

 end;

 {$ifdef OPT_MODDEBUG}
 writeln;
 {$endif}

 {************************* Read instrument headers ************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading instruments');
 {$endif}
 for ATempInt:= 1 to ITH.NInstruments do
 begin
  {** Seek to the instrument header **}
  {


  }

  {** Read the instrument header **}
  AStream^.vmGetData(@ITI, sizeof(ITI));
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading instrument A-header #'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  {** Allocate extended instrument **}
  if (jssAllocateExtInstrument(AModule^.ExtInstruments[ATempInt]) <> jerrOK) then
     begin
      SUP_MAC_CLOSE_STREAM
      jssError(jProcName, 'Error allocating extinstrument structure #'+ ioStrL(ATempInt));
      exit;
     end;

  with AModule^.ExtInstruments[ATempInt]^ do
  begin
   {** Initialize extended instrument **}
   jssASCIIZToStr(@ITI.ADescription, Description, sizeof(ITI.ADescription));

  end; { WITH }

  {$ifdef OPT_MODDEBUG}
  write('.');
  {$endif}
 end;

 {$ifdef OPT_MODDEBUG}
 writeln;
 {$endif}

 SUP_MAC_CLOSE_STREAM
end;


function jssDetectIT(AStream:PStream; var AExtra:Pointer):Longint;
begin
 AExtra:= NIL;
 
end;

{$endif}
