{
  J Sound System - Protracker Studio (PSM) module loader
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This is loader for the "PSM" module format used in many Epic Megagames's
  games (ehh :) that used Joshua Jensen's sound system.

  I couldn't find a PSM converter able to convert PSM 1.0 to other formats
  and I wanted to play the "Epic Pinball" theme so badly ... :) So, this
  loader + player (jplaypsm) started as a quick hack conversion from S3M.


  BUGS:
  You may notice that all PSM files are not supported. The reason is that
  they are not actually the same format, but there are many versions of the
  PSM format out there.

  - The original PS16 format is totally different from the actual PSM,
  but it is the ancestor of PSM and besides I have the description of the
  format so I might add support for it some day. (I am not sure whether it
  has been ever used anywhere, except in the Jensen's sound system test)

  - And then there is PSM 1.0 format, which is supported by this loader...
  (It was used at least in original "Epic Pinball")

  - The PSM 2.0 format is also different. It has been used in Jazz
  Jackrabbit and One Must Fall 2097.

  - And there might be even more of the PSM versions. I suspect that
  some of Epic's games used something like "1.5" format, but this theory
  remains unproven. Unlike I stated in earlier versions of this text,
  the later version(s) of Epic Pinball might have used the "1.5" version
  of the PSM and not 2.0.

  Btw. If someone has specs for PSM 2.0 or others, send them to me.

  (FYI: The versions other than 1.0 are just theoretical since I don't
  know the real versions. I just know that in history of PSM there have
  been several versions of the format. PSM 1.0 and PS16 are the only ones
  that I have some docs for.)
}


{$ifdef GET_IMPLEMENTATION}
 const
      {** PSM data string **}
      PSM_DataString = 'PSM loader v0.44 (updated 15/3/2001)';

      {** PSM definations **}
      PSM_MaxChannels    = 32;
      PSM_MaxRows        = 64;
      PSM_MaxPatterns    = 255;
      PSM_MaxOrders      = 255;
      PSM_MaxInstruments = 255;

      {** PSM module flags **}

      {** PSM instrument flags **}
      psmIFSynthesized   = $01;   { 0=Digital sampledata, 1=Synth }
      psmIFRESERVED      = $02;   { Reserved }
      psmIF16bit         = $04;   { 16-bit sample }
      psmIFUnsigned      = $08;   { 0=Signed, 1=UNsigned }
      psmIFRawData       = $10;   { 0=DELTA, 1=RAW!! }
      psmIFBidi          = $20;   { Use bidi-loop? }
      psmIFGUSPatch      = $40;   { Sample in GUS patch format (unsupp) }
      psmIFLooped        = $80;   { Looped sample }

 type
     {** PSM-module structure defs **}
     TPSMHeader = record
      Ident:String4;           { Header identificator }
      SongName:Array[1..60] of Char;  { Song name }
      SongType,                { Song type flags
                                 bit0: 0=Module, 1=Song only (no samples)
                                 bit1: 0=3-octave ProTracker, 1=1-5 octave
                                 bitX: RESERVED
                               }
      SongVersion,             { Version of the song, always $10 }
      PatVersion,              { Pattern format version:
                                 0 = 32-channel format
                                 1 = 255-channel format
                               }
      DefSpeed,                { Default speed }
      DefTempo,                { Default BPM   32 - 255}
      MasterVol:Byte;          { Master volume 0  - 255 : 0% - 100% }
      NPlayOrders,             { Number of orders 1 - 255 }
      NOrders,                 { Number of orders }
      NPatterns,               { Number of patterns }
      NInstruments,            { Number of instruments/samples }
      NChannels,               { TOTAL number of channels }
      NCP:Word;                { Number of channels to PLAY }

      OOrders,                 { Fileoffset of Orders }
      OPanPos,                 { --""-- Panning Positions }
      OPatterns,               { --""-- Patterns }
      OInstruments,            { --""-- Instruments }
      OComments,               { --""-- Comments }
      OTotalPatSize:Longint;   { }

      AFiller:Array[1..10] of Longint;
     end;
     { 146 bytes long }


     {** PSM sample instrument header **}
     TPSMInstrument = record
      Filename:Array[1..13] of Char;    { Filename of the sample  }
      Description:Array[1..24] of Char; { Description }
      DataOffs,                { Sample data file offset }
      MemOffs:Longint;         { SHIT-UNUSED-RESERVED-IDIOTISM VALUE }
      SampNum:Word;            { Real # number of sample 1-255 }
      Flags:Byte;              { Sample flags, see psmSFXXXX }
      Length,                  { Sample length }
      LoopS,                   { Sample loop start offset }
      LoopE:Longint;           { Sample loop end offset }
      FineTune,                { Finetuning value }
      Volume:Byte;             { Sample default volume }
      C2BaseSpeed:Word;        { C-2 Base Speed frequency }
     end;
     { 64 bytes long }

     TPSMPattern = record
      Size:Word;
      NRows,
      NChannels:Byte;
     end;
{$endif}


{$ifdef GET_REGISTRATION}
 jssRegisterLoader(jmdtPSM, @jssLoadPSM, @jssDetectPSM, 'PSM', PSM_DataString);
{$endif}


{$ifdef GET_IMPLEMENTATION}
procedure jssLoadPSM(var AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);

 const jProcName = 'jssLoadPSM';

 var
    {** General loader variables **}
    PSMH:TPSMHeader;          { PSM-header }
    PSMI:TPSMInstrument;      { One temp instrument for converting
                                PSM instruments to internal format }
    PSMP:TPSMPattern;         { PSM-pattern header }
    AInstNum:Array[1..PSM_MaxInstruments] of Longint;
    AInstOffs:Array[1..PSM_MaxInstruments] of Longint;
    AInstAttr:Array[1..PSM_MaxInstruments] of Word;
    ATempByte:Byte;
    ADataPtr,
    ATempI,
    ATempInt:Longint;
    ATempUsed:Boolean;
    ATempS4:String4;

    ATempBufSize:Longint;
    ATempBufData:Pointer;


procedure PSM_AllocTempUnpackBuf(ASize:Longint);
begin
 ATempBufSize:= ASize;
 GetMem(ATempBufData, ATempBufSize);
end;

procedure PSM_DeallocTempUnpackBuf;
begin
 if (ATempBufData <> NIL) then
    begin
     FreeMem(ATempBufData, ATempBufSize);
     ATempBufData:= NIL;
     ATempBufSize:= 0;
    end;
end;

procedure PSM_UnpackPattern(ASourceData:Pointer; ASourceSize:Longint; var ADestPat:TJSSPattern);
 var
    ASourceOffs,
    APatChannel,
    APatPacking,
    APatRow:Longint;
    BValue:Byte;

function jssGetByte:Byte;
begin
 jssGetByte:= PByte(ASourceData)^;
 inc(longint(ASourceData));
 inc(ASourceOffs);
end;

begin
 {** Check data pointers **}
 if (ASourceData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     PSM_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern SOURCE pointer NULL.');
     exit;
    end;

 if (ADestPat.AUsed = FALSE) or
    (ADestPat.AData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     PSM_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern DEST pointer NULL.');
     exit;
    end;

 {** Init variables **}
 ASourceOffs:= 0;
 APatRow:= 0;
 BValue:= 0;

 {** Pattern unpacking loop **}
 while (APatRow < ADestPat.ARows) and
       (ASourceOffs < ASourceSize) do
 begin
  BValue:= jssGetByte;

  if (BValue <> 0) then
  begin
   {** Get channel number and packing byte of data **}
   APatChannel:= (BValue and $1f);
   APatPacking:= BValue;

   {**** Check if channel is bigger than maxchannels ****}
   if (APatChannel < PSM_MaxChannels) then
      with ADestPat.AData^[APatRow, (1 + APatChannel)] do
      begin
       Param:= 0;
       if ((APatPacking and $80) <> 0) then
          begin
           { PACK $80: Read note/octave and instrument }
           BValue:= jssGetByte;
           if (BValue > 59)
              then Note:= mpNoteOff      { Notes > 59 are bad in PSM }
              else Note:= (BValue - 1 + 12);  { Just get it :) }

           Instr:= jssGetByte;
          end;

       if ((APatPacking and $40) <> 0) then
          begin
           { PACK $40: Read volume }
           Volume:= jssGetByte;
          end;

       if ((APatPacking and $20) <> 0) then
          begin
           { PACK $20: Read command/effect }
           Effect:= jssGetByte;
           if (Effect < 1) or (Effect > 72)
              then Effect:= mpNoData;

           { Read and handle effect parameter }
           case Effect of
            40:  begin
                  { FIX ME !! Should be done differently, depending on effect .. see docs }
                  Param:= jssGetByte;
                  Param:= jssGetByte;
                  Param:= jssGetByte;
                 end;
            else Param:= jssGetByte;
           end;

writeln('eff: ', Effect, ' :: ', Param);           
          end;
      end;
  end else
      inc(APatRow);

 end; { WHILE }
 {** Unpacking OK. **}
end;


function PSM_GI(AFlag:Word):Boolean;
begin
 PSM_GI:= ((PSMI.Flags and AFlag) = AFlag);
end;

function PSM_DecodeDelta8(SData:Pointer; SSize:Longint):Boolean;
 type PShortInt = ^ShortInt;
 var AValue:ShortInt; APos:Longint;
begin
 {** Check data pointer and size **}
 PSM_DecodeDelta8:= FALSE;
 if (SData = NIL) or
    (SSize = 0) then
    exit;

 {** Decode the data **}
 AValue:= 0;
 for APos:= 1 to SSize do
 begin
  AValue:= (PShortInt(SData)^ + AValue);
  PShortInt(SData)^:= AValue;
  inc(longint(SData));
 end;

 PSM_DecodeDelta8:= TRUE;
end;

function PSM_DecodeDelta16(SData:Pointer; SSize:Longint):Boolean;
 var AValue:Word; APos:Longint;
begin
 {** Check data pointer and size **}
 PSM_DecodeDelta16:= FALSE;
 if (SData = NIL) or
    (SSize = 0) then
    exit;

 {** Decode the data **}
 AValue:= 0;
 for APos:= 1 to SSize do
 begin
  AValue:= (PWord(SData)^ + AValue);
  PByte(SData)^:= AValue;
  inc(longint(SData), 2);
 end;

 PSM_DecodeDelta16:= TRUE;
end;


begin
 {******************** Check for internal inconsistencies ******************}
 if (PSM_MaxChannels > jsetMaxNChannels) or
    (PSM_MaxRows > jsetMaxRows) or
    (PSM_MaxPatterns > jsetMaxPatterns) or
    (PSM_MaxInstruments > jsetMaxInstruments) or
    (PSM_MaxOrders > jsetMaxOrders) then
    begin
     jssErrorN:= jerrInternalError;
     jssError(jProcName, 'Internal inconsistency in JLOADPSM.PAS');
     exit;
    end;


 {******************* Read PSM-header and check data in it *****************}
 {$ifdef DEBUG}
 writeln(' Reading and checking PSM header...');
 {$endif}
 AStream^.vmGetData(@PSMH, sizeof(TPSMHeader));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading PSM header from stream 0x'+ ioPtr2Hex(AStream));
     exit;
    end;

 if (PSMH.Ident <> 'PSM') then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Not a real PSM module.');
     if (ioFailOnError) then exit;
    end;

 if not (PSMH.SongVersion in [$10]) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Unsupported PSM format version 0x'+ ioHexB(PSMH.SongVersion));
     if (ioFailOnError) then exit;
    end;

 if (PSMH.NPatterns > PSM_MaxPatterns) then
    begin
     jssWarning(jProcName, 'Number of patterns > '+ ioStrL(PSM_MaxPatterns) +' ('+ ioStrL(PSMH.NPatterns) +')');
     PSMH.NPatterns:= PSM_MaxPatterns;
    end;

 if (PSMH.NOrders > PSM_MaxOrders) then
    begin
     jssWarning(jProcName, 'Number of orders > '+ ioStrL(PSM_MaxOrders) +' ('+ ioStrL(PSMH.NOrders) +')');
     PSMH.NOrders:= PSM_MaxOrders;
    end;

 if (PSMH.NOrders = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of orders was 0');
     if (ioFailOnError) then exit;
    end;

 if (PSMH.NPatterns = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of patterns was 0');
     if (ioFailOnError) then exit;
    end;

 if (PSMH.NInstruments > PSM_MaxInstruments) then
    begin
     jssWarning(jProcName, 'Number of instruments > '+ ioStrL(PSM_MaxInstruments) +' ('+ ioStrL(PSMH.NInstruments) +')');
     PSMH.NInstruments:= PSM_MaxInstruments;
    end;

 if ((PSMH.SongType AND 1) <> 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, '"Song-only" PSM modules are not supported!');
     if (ioFailOnError) then exit;
    end;

 {************************* Init internal variables ************************}
 {$ifdef DEBUG}
 writeln(' Initializing and allocating PSM module...');
 {$endif}
 if (jssAllocateModule(AModule) <> jerrOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssError(jProcName, 'Error allocating new module structure!');
     exit;
    end;

 with AModule^ do
 begin
  ModuleType    := jmdtPSM;
  jssASCIIZToStr(@PSMH.SongName, ModuleName, sizeof(PSMH.SongName));
  ModuleInfo    := PSM_DataString;
  DefSpeed      := PSMH.DefSpeed;
  DefTempo      := round((PSMH.DefTempo * 5) / 2);
  DefGlobalVol  := (PSMH.MasterVol shr 2);
  NInstruments  := PSMH.NInstruments;
  NPatterns     := PSMH.NPatterns;
  NOrders       := PSMH.NOrders;
  NChannels     := PSMH.NChannels;
  DefFlags      := 0;
 end;

 {**************************** Read orders list ****************************}
 {$ifdef OPT_MODDEBUG}
 write(' Reading orders list');
 {$endif}
 {** Seek to header data and check ident signature **}
 AStream^.vmSeek(PSMH.OOrders - 4);
 AStream^.vmGetData(@ATempS4, sizeof(ATempS4));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_FAIL_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading orders list header block ident signature!');
     if (ioFailOnError) then exit;
    end;

 if (ATempS4 <> 'PORD') then
    jssWarning(jProcName, 'Orders list header block ident signature not found/incorrect.');

 {** Read the instrument headers **}
 for ATempInt:= 0 to (AModule^.NOrders-1) do
 begin
  ATempByte:= AStream^.vmGetByte;
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading orders list item #'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  AModule^.OrderList[ATempInt]:= ATempByte;
 end;

 {************************* Read instrument headers ************************}
 {$ifdef OPT_MODDEBUG}
 write(' Reading instrument headers');
 {$endif}
 {** Seek to header data and check ident signature **}
 AStream^.vmSeek(PSMH.OInstruments - 4);
 AStream^.vmGetData(@ATempS4, sizeof(ATempS4));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_FAIL_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading instrument header block ident signature!');
     if (ioFailOnError) then exit;
    end;

 if (ATempS4 <> 'PSAH') then
    jssWarning(jProcName, 'Instrument header block ident signature not found/incorrect.');

 {** Read the instrument headers **}
 for ATempInt:= 1 to AModule^.NInstruments do
 begin
  {** Read one instrument header **}
  AStream^.vmGetData(@PSMI, sizeof(TPSMInstrument));
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading instrument header #'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  {** Check the instrument **}
  AInstOffs[ATempInt]:= PSMI.DataOffs;
  AInstAttr[ATempInt]:= PSMI.Flags;
  AInstNum[ATempInt] := PSMI.SampNum;
  if (AInstNum[ATempInt] < 1) or (AInstNum[ATempInt] > jsetMaxInstruments) then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'Error in instrument header #'+ ioStrL(ATempInt) +': invalid sample number #'+ ioStrL(AInstNum[ATempInt]));
      if (ioFailOnError) then exit;
     end;

  if (AModule^.Instruments[AInstNum[ATempInt]] <> NIL) then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'Error in instrument header #'+ ioStrL(ATempInt) +': sample #'+ ioStrL(AInstNum[ATempInt]) +' already in use!');
      if (ioFailOnError) then exit;
     end;

  if (PSM_GI(psmIFSynthesized)) or
     (PSM_GI(psmIFGUSPatch)) then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'Unsupported instrument type (Synth/GUS patch) #'+ ioStrL(ATempInt) +' ['+ ioStrL(AInstNum[ATempInt]) +'].');
      if (ioFailOnError) then exit;
     end;

  ATempUsed:= ((PSMI.Length > 0) and (PSMI.DataOffs > 0));

  {** Allocate instrument and init **}
  if (jssAllocateInstrument(AModule^.Instruments[AInstNum[ATempInt]]) <> jerrOK) then
     begin
      SUP_MAC_CLOSE_STREAM
      jssError(jProcName, 'Error allocating instrument structure #'+ ioStrL(ATempInt) +' [#'+ ioStrL(AInstNum[ATempInt]) +'].');
      exit;
     end;

  with AModule^.Instruments[AInstNum[ATempInt]]^ do
  begin
   {** Initialize information **}
   Used        := ATempUsed;
   jssASCIIZToStr(@PSMI.Filename, Filename, sizeof(PSMI.Filename));
   jssASCIIZToStr(@PSMI.Description, Description, sizeof(PSMI.Description));
   Size        := PSMI.Length;
   LStart      := PSMI.LoopS;
   LEnd        := PSMI.LoopE;
   Volume      := PSMI.Volume;

   if (Volume > mpMaxVol) then
      Volume:= mpMaxVol;

   C4BaseSpeed := (PSMI.C2BaseSpeed * 4);
   EFineTune   := (PSMI.FineTune and $f);
   Data        := NIL;
   Flags       := 0;

   {** Check and initialize flags **}
   if (PSM_GI(psmIF16bit)) then
      begin
       writeln('16-bit INSTRUEWER! HELP ME!');
       Flags:= (Flags OR jsf16bit);
       ByteSize:= (Size * 2);
      end else
       ByteSize:= Size;

   if (PSM_GI(psmIFlooped)) then
      Flags:= (Flags OR jsfLooped);

   if (PSM_GI(psmIFBidi)) then
      Flags:= (Flags OR jsfBidi);

  end; { WITH }
  {$ifdef OPT_MODDEBUG}
  write('.');
  {$endif}
 end;

 {*************************** Read instrument data *************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading instrument data');
 {$endif}
 for ATempInt:= 1 to AModule^.NInstruments do
 begin
 ATempI:= AInstNum[ATempInt];

 if (AModule^.Instruments[ATempI] <> NIL) and
    (AModule^.Instruments[ATempI]^.Used) then
 with AModule^.Instruments[ATempI]^ do
 begin
  {** CHECK instrument data **}
  if (Size = 0) then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'Instrument sampledata size was 0 in #'+ ioStrL(ATempInt) +' [#'+ ioStrL(ATempI) +'].');
      if (ioFailOnError) then exit;
     end;

  if ((Flags AND jsfLooped) <> 0) then
     begin
     {** Check looped sample information **}
     if (LStart > LEnd) then
        begin
         SUP_MAC_FAIL_STREAM
         jssError(jProcName, 'Instrument sample loop-start > loop-start in #'+ ioStrL(ATempInt) +' [#'+ ioStrL(ATempI) +'].');
         if (ioFailOnError) then exit;
        end;

     if (LStart = LEnd) then
        begin
         SUP_MAC_FAIL_STREAM
         jssError(jProcName, 'Instrument sample loop-start == loop-start in #'+ ioStrL(ATempInt) +' [#'+ ioStrL(ATempI) +'].');
         if (ioFailOnError) then exit;
        end;
     end; { IF (Looped) }

  {** Compute correct fileposition **}
  ADataPtr:= AInstOffs[ATempInt];

  if (ADataPtr = 0) then
     begin
       SUP_MAC_FAIL_STREAM
       jssError(jProcName, 'No sampledata for instrument #'+ ioStrL(ATempInt) +' [#'+ ioStrL(ATempI) +'].');
       if (ioFailOnError) then exit;
     end;

  if (ADataPtr > AStream^.vmGetSize) then
     begin
       SUP_MAC_FAIL_STREAM
       jssError(jProcName, 'Unexpected end of file reading instrument sampledata #'+ ioStrL(ATempInt) +' [#'+ ioStrL(ATempI) +'].');
       if (ioFailOnError) then exit;
     end;

  {** Allocate the memory **}
  GetMem(Data, ByteSize);
  if (Data = NIL) then
     begin
      SUP_MAC_CLOSE_STREAM
      jssErrorN:= jerrMemAllocError;
      jssError(jProcName, 'Memory allocation error for sample #'+ ioStrL(ATempInt));
      exit;
     end;

  {** Read the sampledata **}
  AStream^.vmSeek(ADataPtr);
  AStream^.vmGetData(Data, ByteSize);
  if (AStream^.vmStatus <> strOK) then
     begin
        SUP_MAC_FAIL_STREAM
        jssErrorN:= jerrFileReadError;
        jssError(jProcName, 'Error reading sample data for instrument #'+ ioStrL(ATempInt) +' [#'+ ioStrL(ATempI) +'].');
        if (ioFailOnError) then exit;
     end;

  {** Convert the sample data **}
  if ((AInstAttr[ATempInt] AND psmIFRawData) = 0) then
     begin
      {** Data is DELTA-encoded **}
      if ((AInstAttr[ATempInt] AND psmIF16bit) = 0)
         then PSM_DecodeDelta8(Data, Size)
         else PSM_DecodeDelta16(Data, Size);
     end;

  if ((AInstAttr[ATempInt] AND psmIFUnSigned) = 0) then
     begin
      {** Data is SIGNED **}
      if ((AInstAttr[ATempInt] AND psmIF16bit) = 0)
         then jssFlipSampleSign8(Data, Size)
         else jssFlipSampleSign16(Data, Size);
     end;

  {$ifdef OPT_MODDEBUG}
  write('.');
  {$endif}
 end; { WITH }
 end; { FOR }

 {***************************** Read patterns ******************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading pattern data');
 {$endif}
 {** Seek to data and check ident signature **}
 AStream^.vmSeek(PSMH.OPatterns - 4);
 AStream^.vmGetData(@ATempS4, sizeof(ATempS4));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_FAIL_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading pattern block ident signature!');
     if (ioFailOnError) then exit;
    end;

 if (ATempS4 <> 'PPAT') then
    jssWarning(jProcName, 'Pattern block ident signature not found/incorrect.');

 {** Read and unpack patterns **}
 for ATempInt:= 0 to (AModule^.NPatterns - 1) do
 begin
   {** Read the pattern data header **}
   AStream^.vmGetData(@PSMP, sizeof(PSMP));
   if (AStream^.vmStatus <> strOK) then
      begin
       SUP_MAC_FAIL_STREAM
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading pattern header for #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
      end;

   {** Read the packed pattern data from file **}
   PSM_AllocTempUnpackBuf(PSMP.Size - SizeOf(PSMP));

   AStream^.vmGetData(ATempBufData, ATempBufSize);
   if (AStream^.vmStatus <> strOK) then
      begin
       SUP_MAC_FAIL_STREAM
       PSM_DeallocTempUnpackBuf;
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading pattern data for #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
      end;

   {** Allocate new pattern and unpack patterndata **}
   if (jssAllocatePattern(AModule^.Patterns[ATempInt], PSMP.NRows) <> jerrOK) then
      begin
       SUP_MAC_CLOSE_STREAM
       PSM_DeallocTempUnpackBuf;
       jssError(jProcName, 'Error allocating pattern data for #'+ ioStrL(ATempInt));
       exit;
      end;

   PSM_UnpackPattern(ATempBufData, ATempBufSize, AModule^.Patterns[ATempInt]);

   PSM_DeallocTempUnpackBuf;

   {$ifdef OPT_MODDEBUG}
   write('.');
   {$endif}
 end;

 {$ifdef OPT_MODDEBUG}
 writeln;
 {$endif}

 SUP_MAC_CLOSE_STREAM
end;


function jssDetectPSM(AStream:PStream; var AExtra:Pointer):Longint;
 var AResult:Longint; PSMH:TPSMHeader;
begin
 AExtra:= NIL;
 AResult:= -1;
 
 AStream^.vmGetData(@PSMH, sizeof(TPSMHeader));
 if (AStream^.vmStatus <> strOK) then exit(-2);

 if (PSMH.Ident = 'PSM') then
    inc(AResult, 2);
    
 if (PSMH.SongVersion = $10) then
    inc(AResult);

 jssDetectPSM:= AResult;
end;

{$endif}
