{
  J Sound System - Poly Tracker (PTM) module loader
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  Self explanatory.

  Bugs: Does not support all PTM versions. Only v2.03 is supported.
}


{$ifdef GET_IMPLEMENTATION}
 const
      {** PTM data string **}
      PTM_DataString = 'PTM loader v0.83 (updated 17/8/2000)';

      {** PTM definations **}
      PTM_MaxChannels    = 32;
      PTM_MaxRows        = 64;
      PTM_MaxPatterns    = 128;
      PTM_MaxOrders      = 255;
      PTM_MaxInstruments = 128;

      PTM_Alignment      = 16;

      {** PTM instrument type/flags **}
      ptmifNoSample      = $00;   { No sample }
      ptmifSample        = $01;   { Normal sample }
      ptmifFM            = $02;   { OPL2 / OPL3 instrument }
      ptmifMIDI          = $03;   { Midi instrument }

      ptmifLooped        = $04;   { 0=NO loop, 1=Looped }
      ptmifBiDiLoop      = $08;   { 0=normal loop, 1=bidi (pingpong) loop }
      ptmif16bit         = $10;   { 16-bit sample }


 type
     {** PTM-header structure **}
     PTMHeader = record
      Songname:String28;          { Songname padded with zeros $00 }
      UNUSED1a:Byte;              { Value is always $1a = 26 }
      Version:Word;               { Version number, currently $0203 }
      RESERVED1:Byte;             { Set to $00 }

      NOrders,                    { Number of orders 0..256 }
      NInstruments,               { Number of instruments 1..255 }
      NPatterns,                  { Number of patterns 1..128 }
      NChannels,                  { Number of channels 1..32 }
      Flags,                      { Module flags }
      RESERVED2:Word;             { Set to $0000 }

      PTMFIdent:String4;          { "PTMF" identificator }

      RESERVED3:Array[1..16] of Byte;  { RESERVED, NOT USED }

      ChSettings:Array[1..32] of Byte; { Channel panning settings }
      OrderList:Array[0..255] of Byte; { Orders table }

      PatSeg:Array[0..127] of Word;    { Pattern segment table }
     end;


     {** PTM Instrument structure **}
     PTMInstrument = record
      Flags:Byte;                 { Instrument flags }
      Filename:String12;          { Filename. ASCIIZ }
      Volume:Byte;                { Default volume }
      C4BaseSpeed:Word;           { C-4 base speed Hz }

      SampleSeg:Word;             { PT INTERNAL USE }
      FileOffset,                 { File offset of data }

      Size,                       { Size of sampledata }
      LStart,                     { Start offset of LOOP }
      LEnd,                       { End offset of LOOP }

      GUSBegin,                   { PT INTERNAL USE }
      GUSLStart,                  { PT INTERNAL USE }
      GUSLEnd:Longint;            { PT INTERNAL USE }
      GUSLoopFlags:Byte;          { PT INTERNAL USE }
      RESERVED0:Byte;             { RESERVED, SET TO 0 }

      Description:String28;       { Instrument description. ASCIIZ }
      BlockDesc:String4;          { "PTMS" }
     end;
{$endif}


{$ifdef GET_REGISTRATION}
 jssRegisterLoader(jmdtPTM, @jssLoadPTM, @jssDetectPTM, 'PTM', PTM_DataString);
{$endif}


{$ifdef GET_IMPLEMENTATION}
procedure jssLoadPTM(var AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);

 const jProcName = 'jssLoadPTM';

 var
    {** General loader variables **}
    PTMH:PTMHeader;           { PTM-header }
    PTMI:PTMInstrument;       { One temp. instrument for converting
                                PTM instruments to internal format }
    AFirstInst:Boolean;
    AFirstInstDataOffs,
    ASaveFileOffs,
    ADataPtr,
    ATempLong,
    ATempInt:Longint;
    ATempUsed:Boolean;

    ATempBufSize:Longint;
    ATempBufData:Pointer;


procedure PTM_AllocTempUnpackBuf(ASize:Longint);
begin
 ATempBufSize:= ASize;
 GetMem(ATempBufData, ATempBufSize);
end;

procedure PTM_DeallocTempUnpackBuf;
begin
 if (ATempBufData <> NIL) then
    begin
     FreeMem(ATempBufData, ATempBufSize);
     ATempBufData:= NIL;
     ATempBufSize:= 0;
    end;
end;

procedure PTM_UnpackPattern(ASourceData:Pointer; ASourceSize:Longint; var ADestPat:TJSSPattern);
 var
    ASourceOffs,
    APatChannel,
    APatPacking,
    APatRow:Longint;
    BValue:Byte;

function jssGetByte:Byte;
begin
 jssGetByte:= PByte(ASourceData)^;
 inc(longint(ASourceData));
 inc(ASourceOffs);
end;

begin
 {** Check data pointers **}
 if (ASourceData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     PTM_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern SOURCE pointer NULL.');
     exit;
    end;

 if (ADestPat.AUsed = FALSE) or
    (ADestPat.AData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     PTM_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern DEST pointer NULL.');
     exit;
    end;

 {** Init variables **}
 ASourceOffs:= 0;
 APatRow:= 0;
 BValue:= 0;

 {** Pattern unpacking loop **}
 while (APatRow < ADestPat.ARows) and
       (ASourceOffs < ASourceSize) do
 begin
  BValue:= jssGetByte;

  if (BValue <> 0) then
  begin
   {** Get channel number and packing byte of data **}
   APatChannel:= (BValue and $1f);
   APatPacking:= BValue;

   {**** Check if channel is bigger than maxchannels ****}
   if (APatChannel < PTM_MaxChannels) then
      with ADestPat.AData^[APatRow, (1 + APatChannel)] do
      begin
       if ((APatPacking and $20) <> 0) then
          begin
           { PACK $20: Read note/octave and instrument }
           Note:= jssGetByte;
           Instr:= jssGetByte;
          end;

       if ((APatPacking and $40) <> 0) then
          begin
           { PACK $40: Read command/effect and parameter }
           Effect:= jssGetByte;
           Param:= jssGetByte;
          end;

       if ((APatPacking and $80) <> 0) then
          begin
           { PACK $80: Read volume}
           Volume:= jssGetByte;
          end;
      end else

      begin
       {**** IGNORE EXTRA CHANNELS ****}
       if ((APatPacking and $20) <> 0) then
          begin
           { PACK $20: Read note/octave and instrument }
           jssGetByte;
           jssGetByte;
          end;

       if ((APatPacking and $40) <> 0) then
          begin
           { PACK $40: Read command-effect and parameter }
           jssGetByte;
           jssGetByte;
          end;

       if ((APatPacking and $80) <> 0) then
          begin
           { PACK $80: Read volume}
           jssGetByte;
          end;
      end;
  end else
      inc(APatRow);

 end; { WHILE }
 {** Unpacking OK. **}
end;


function PTM_GI(AFlag:Word):Boolean;
begin
 PTM_GI:= ((PTMI.Flags and AFlag) = AFlag);
end;


begin
 {******************** Check for internal inconsistencies ******************}
 if (PTM_MaxChannels > jsetMaxNChannels) or
    (PTM_MaxRows > jsetMaxRows) or
    (PTM_MaxPatterns > jsetMaxPatterns) or
    (PTM_MaxInstruments > jsetMaxInstruments) or
    (PTM_MaxOrders > jsetMaxOrders) then
    begin
     jssErrorN:= jerrInternalError;
     jssError(jProcName, 'Internal inconsistency in JLOADPTM.PAS');
     exit;
    end;

 {******************* Read PTM-header and check data in it *****************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Reading and checking PTM header...');
 {$endif}
 AStream^.vmGetData(@PTMH, sizeof(PTMHeader));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading PTM header from stream 0x'+ ioPtr2Hex(AStream));
     exit;
    end;

 if (PTMH.PTMFIdent <> 'PTMF') then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Not a real PTM module stream 0x'+ ioPtr2Hex(AStream));
     if (ioFailOnError) then exit;
    end;

 if (PTMH.Version <> $0203) then
    jssWarning(jProcName, 'Unsupported PTM version 0x'+ ioHexW(PTMH.Version));

 if (PTMH.NPatterns > PTM_MaxPatterns) then
    begin
     jssWarning(jProcName, 'Number of patterns > '+ ioStrL(PTM_MaxPatterns) +' ('+ ioStrL(PTMH.NPatterns) +')');
     PTMH.NPatterns:= PTM_MaxPatterns;
    end;

 if (PTMH.NOrders > PTM_MaxOrders) then
    begin
     jssWarning(jProcName, 'Number of orders > '+ ioStrL(PTM_MaxOrders) +' ('+ ioStrL(PTMH.NOrders) +')');
     PTMH.NOrders:= PTM_MaxOrders;
    end;

 if (PTMH.NOrders = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of orders was 0');
     if (ioFailOnError) then exit;
    end;

 if (PTMH.NPatterns = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of patterns was 0');
     if (ioFailOnError) then exit;
    end;

 if (PTMH.NInstruments > PTM_MaxInstruments) then
    begin
     jssWarning(jProcName, 'Number of instruments > '+ ioStrL(PTM_MaxInstruments) +' ('+ ioStrL(PTMH.NInstruments) +')');
     PTMH.NInstruments:= PTM_MaxInstruments;
    end;

 if (PTMH.NChannels > PTM_MaxChannels) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Too many channels ('+ ioStrL(PTMH.NChannels) +'), maximum is '+ ioStrL(PTM_MaxChannels));
     if (ioFailOnError) then exit;
    end;

 if (PTMH.Flags <> 0) then
    jssWarning(jProcName, 'Modulefile contains unknown flag settings!');

 {************************* Init internal variables ************************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Initializing and allocating PTM module...');
 {$endif}
 if (jssAllocateModule(AModule) <> jerrOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssError(jProcName, 'Error allocating new module structure!');
     exit;
    end;

 with AModule^ do
 begin
  {** Convert the Header Info **}
  ModuleType    := jmdtPTM;
  jssASCIIZToStr(@PTMH.SongName, ModuleName, sizeof(PTMH.SongName));
  ModuleInfo    := PTM_DataString;
  DefSpeed      := 6;
  DefTempo      := 125;
  DefGlobalVol  := 64;
  DefMasterVol  := $7f;
  NInstruments  := PTMH.NInstruments;
  NPatterns     := PTMH.NPatterns;
  NOrders       := PTMH.NOrders;
  NChannels     := PTMH.NChannels;
  DefFlags      := 0;

  {** Convert the Flags **}
  { No flags are supported in version 2.03 of PTM format }

  {** Convert the default pannings **}
  for ATempInt:= 1 to PTM_MaxChannels do
  case PTMH.ChSettings[ATempInt] of
   01..06,
   08..14: DefPanning[ATempInt]:= (PTMH.ChSettings[ATempInt] * $11);
   00: DefPanning[ATempInt]:= jchPanLeft;
   07: DefPanning[ATempInt]:= jchPanMiddle;
   15: DefPanning[ATempInt]:= jchPanRight;
   else DefPanning[ATempInt]:= jchPanMiddle;
  end;

  {** Convert the orders list **}
  for ATempInt:= 0 to (NOrders-1) do
      begin
       ATempLong:= PTMH.OrderList[ATempInt];
       case ATempLong of
        $ff: ATempLong:= mpOrderEnd;
        $fe: ATempLong:= mpOrderSkip;
       end;
       OrderList[ATempInt]:= ATempLong;
      end;
 end;

 {***************************** Read Instruments ***************************}
 {$ifdef OPT_MODDEBUG}
 write(' Reading instruments');
 {$endif}
 AFirstInst:= TRUE;
 AFirstInstDataOffs:= 0;
 ASaveFileOffs:= AStream^.vmGetPos;

 for ATempInt:= 1 to PTMH.NInstruments do
 begin
  {** Read one instrument header **}
  AStream^.vmSeek(ASaveFileOffs);
  AStream^.vmGetData(@PTMI, sizeof(PTMInstrument));
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading instrument header #'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  ASaveFileOffs:= AStream^.vmGetPos;

  {** Check the instrument **}
  if (PTMI.BlockDesc <> 'PTMS') and
     (PTMI.BlockDesc <> #0#0#0#0) then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'Not a compatible instrument #'+ ioStrL(ATempInt)
                           +': "'+ PTMI.BlockDesc +'"');
      if (ioFailOnError) then exit;
     end;

  case (PTMI.Flags and 3) of
    ptmifFM,
    ptmifMIDI: begin
                SUP_MAC_FAIL_STREAM
                jssError(jProcName, 'FM and MIDI instruments are not supported #'+ ioStrL(ATempInt));
                if (ioFailOnError) then exit;
               end;
    ptmifSample:   ATempUsed:= TRUE;
    ptmifNoSample: ATempUsed:= FALSE;
  end;

  {** Allocate and init **}
  if (jssAllocateInstrument(AModule^.Instruments[ATempInt]) <> jerrOK) then
     begin
      SUP_MAC_CLOSE_STREAM
      jssError(jProcName, 'Error allocating instrument structure #'+ ioStrL(ATempInt));
      exit;
     end;

  with AModule^.Instruments[ATempInt]^ do
  begin
   {** Initialize information **}
   Used        := ATempUsed;
   jssASCIIZToStr(@PTMI.Filename, Filename, sizeof(PTMI.Filename));
   Size        := PTMI.Size;
   LStart      := PTMI.LStart;
   LEnd        := PTMI.LEnd;
   Volume      := PTMI.Volume;
   C4BaseSpeed := PTMI.C4BaseSpeed;
   jssASCIIZToStr(@PTMI.Description, Description, sizeof(PTMI.Description));
   Data        := NIL;
   Flags       := 0;

   {** Initialize flags **}
   if (Volume > mpMaxVol) then
      Volume:= mpMaxVol;

   if (PTM_GI(ptmif16bit)) then
      begin
       Flags:= (Flags OR jsf16bit);
       ByteSize:= (Size * 2);
      end else
       ByteSize:= Size;

   if (PTM_GI(ptmifLooped)) then
      Flags:= (Flags OR jsfLooped);

   if (PTM_GI(ptmifBIDILoop)) then
      Flags:= (Flags OR jsfBiDi);

   {** Load instrument data, if there is any... **}
   if (Used) then
   begin
    {** Check information **}
    if (Size = 0) then
       begin
        SUP_MAC_FAIL_STREAM
        jssError(jProcName, 'Instrument sampledata size was 0 in #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
       end;

    if ((Flags and jsfLooped) <> 0) then
    begin
     {** Check looped sample information **}
     if (LStart > LEnd) then
        begin
         SUP_MAC_FAIL_STREAM
         jssError(jProcName, 'Instrument sample loop-start > loop-start in #'+ ioStrL(ATempInt));
         if (ioFailOnError) then exit;
        end;

     if (LStart = LEnd) then
        begin
         SUP_MAC_FAIL_STREAM
         jssError(jProcName, 'Instrument sample loop-start == loop-start in #'+ ioStrL(ATempInt));
         if (ioFailOnError) then exit;
        end;
    end;

    {** Compute correct fileposition **}
    ADataPtr:= PTMI.FileOffset;
    if (AFirstInst) then
       begin
        AFirstInstDataOffs:= ADataPtr;
        AFirstInst:= FALSE;
       end;

    if (ADataPtr = 0) then
       begin
        SUP_MAC_FAIL_STREAM
        jssError(jProcName, 'No sampledata for instrument #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
       end;

    if (ADataPtr > AStream^.vmGetSize) then
       begin
        SUP_MAC_FAIL_STREAM
        jssError(jProcName, 'Unexpected end of file reading instrument sampledata #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
       end;

    {** Allocate the memory **}
    GetMem(Data, ByteSize);
    if (Data = NIL) then
       begin
        SUP_MAC_CLOSE_STREAM
        jssErrorN:= jerrMemAllocError;
        jssError(jProcName, 'Memory allocation error for sample #'+ ioStrL(ATempInt));
        exit;
       end;

    {** Read the sampledata **}
    AStream^.vmSeek(ADataPtr);
    AStream^.vmGetData(Data, ByteSize);
    if (AStream^.vmStatus <> strOK) then
       begin
        SUP_MAC_FAIL_STREAM
        jssErrorN:= jerrFileReadError;
        jssError(jProcName, 'Error reading sample data for instrument #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
       end;

    {** Check if we have to flip sampledata sign **}
    if (jssDecodeDelta8(Data, Size) = FALSE) then
       begin
        SUP_MAC_FAIL_STREAM
        jssError(jProcName, 'Error delta-decoding sample data for instrument #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
       end;

   end; { IF (Used) }
  end;  { WITH AModule^.Instruments[ATempInt]^ DO }
  {$ifdef OPT_MODDEBUG}
  write('.');
  {$endif}
 end;


 {***************************** Read patterns ******************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading pattern data');
 {$endif}

 {** Read and unpack patterns **}
 for ATempInt:= 0 to (PTMH.NPatterns - 1) do
 if (PTMH.PatSeg[ATempInt] <> 0) then
 begin
   {** Go to right fileoffset **}
   ADataPtr:= (longint(PTMH.PatSeg[ATempInt]) * PTM_Alignment);
   AStream^.vmSeek(ADataPtr);

   {** Calculate size for pattern data **}
   if (ATempInt < (PTMH.NPatterns - 1))
      then ATempLong:= (longint(PTMH.PatSeg[ATempInt+1]) * PTM_Alignment)
      else ATempLong:= AFirstInstDataOffs;

   ATempLong:= (ATempLong - ADataPtr);

   {** Read the packed pattern data from file **}
   PTM_AllocTempUnpackBuf(ATempLong);

   AStream^.vmGetData(ATempBufData, ATempBufSize);
   if (AStream^.vmStatus <> strOK) then
      begin
       if (ioFailOnError) then
          begin
           SUP_MAC_CLOSE_STREAM
           PTM_DeallocTempUnpackBuf;
          end;
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading pattern data for #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
      end;

   {** Allocate new pattern and unpack patterndata **}
   if (jssAllocatePattern(AModule^.Patterns[ATempInt], PTM_MaxRows) <> jerrOK) then
      begin
       SUP_MAC_CLOSE_STREAM
       PTM_DeallocTempUnpackBuf;
       jssError(jProcName, 'Error allocating pattern data.');
       exit;
      end;

   PTM_UnpackPattern(ATempBufData, ATempBufSize, AModule^.Patterns[ATempInt]);

   PTM_DeallocTempUnpackBuf;
   {$ifdef OPT_MODDEBUG}
   write('.');
   {$endif}
 end;

 {$ifdef OPT_MODDEBUG}
 writeln;
 {$endif}

 SUP_MAC_CLOSE_STREAM
end;


function jssDetectPTM(AStream:PStream; var AExtra:Pointer):Longint;
 var AResult:Longint; PTMH:PTMHeader;
begin
 AExtra:= NIL;
 AResult:= -1;

 AStream^.vmGetData(@PTMH, sizeof(PTMHeader));
 if (AStream^.vmStatus <> strOK) then exit(-2);

 if (PTMH.PTMFIdent = 'PTMF') then
    inc(AResult, 2);

 if (PTMH.UnUsed1A = $1a) then
    inc(AResult);

 if (PTMH.Version = $0203) then
    inc(AResult);

 jssDetectPTM:= AResult;
end;

{$endif}
