{
  J Sound System - Scream Tracker 3 (S3M) module loader
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  Self explanatory.
}

{$ifdef GET_IMPLEMENTATION}
 const
      {** S3M data string **}
      S3M_DataString = 'S3M loader v1.13 (updated 15/3/2001)';

      {** S3M definations **}
      S3M_MaxChannels    = 32;
      S3M_MaxRows        = 64;
      S3M_MaxPatterns    = 100;
      S3M_MaxOrders      = 255;
      S3M_MaxInstruments = 99;
      S3M_Alignment      = 16;

      {** S3M module flags **}
      s3mMFST2Vibrato    = 1;    { ST2 vibrato }
      s3mMFST2Tempo      = 2;    { ST2 tempo command }
      s3mMFAmigaSlides   = 4;    { Amiga-styled slides }
      s3mMFVol0Opt       = 8;    { Volume0 optimizations }
      s3mMFAmigaLimits   = 16;   { Use Amiga ProTracker limits }
      s3mMFFilter        = 32;   { Filter enable }
      s3mMFST300Slides   = 64;   { Use ST3.00 erratic volumeslides }
      s3mMFCustomData    = 128;  { Special custom data in file, pointed by
                                   the 'special' pointer in header }

      {** S3M instrument flags **}
      s3mIFLooped        = $01;   { Looped sample }
      s3mIFStereo        = $02;   { Stereo sample }
      s3mIF16bit         = $04;   { 16-bit sample }


 type
     {** S3M-module structure defs **}
     PPtr     = Word;

     {** S3M-header **}
     TS3MHeader = record
      Songname:String28;          { Songname padded with zeros $00 }
      UNUSED1a:Byte;              { Value is always $1a }
      SubType:Byte;               { $10 = normal S3M, $11 = S3M-extended }

      UNUSED0:Word;               { For future expansion }

      NOrders,                    { Number of orders }
      NInstruments,               { Number of instruments }
      NPatterns,                  { Number of patterns }
      Flags,                      { Flags: Force Amiga limits,
                                    Vol0 optimizations, etc. }
      TrackerVersion,             { Tracker tag and version }
      FFInfo:Word;                { Fileformat information:
                                    1 = SIGNED samples
                                    2 = UNSIGNED samples
                                    Others are RESERVED
                                   Normally this value is 2.
                                  }
      BlockIdent:String4;         { Should be "SCRM" }
      GlobalVol,                  { Global volume }
      Speed,                      { Speed factor }
      Tempo,                      { Tempo }
      MasterVol,                  { Master volume and Stereo FLAG }
      UltraClick,                 { Ultrasound click removal FLAG }
      Panning:Byte;               { Channel panning FLAG }

      UNUSED1, UNUSED2:Longint;   { Reserved for future expansions }

      SpecialData:PPtr;           { Parapointer to specialdata }

      ChSettings:Array[1..S3M_MaxChannels] of Byte;
     end;


     {** S3M Instrument **}
     TS3MInstrument = record
{00 } InstType:Byte;                    { Type of instrument }

{01 } Filename:String12;                { Filename. ASCIIZ }

{0D } PP1:Byte;                         { Parapointer to sampledata }
      PP2:Word;

{10 } Size,                             { Size of sampledata }
      LStart,                           { Start offset of LOOP }
      LEnd:Longint;                     { End offset of LOOP }

      Volume,                           { Default playing volume }
      UNUSED0,                          { NOT USED }
      Packing,                          { Packing }
      Flags:Byte;                       { Instrument flags }

      C4BaseSpeed:Word;                 { C-4 base speed Hz }
      UNUSED1:Word;                     { UNUSED high word of C4BaseSpeed }

      UNUSED2:Array[1..12] of Byte;     { Some Scream Tracker internal data }
                                        { is stored here, but we don't need }
                                        { them, so we ignore them. }

      Description:String28;             { Instrument description. ASCIIZ }
      BlockDesc:String4;                { "    ", "SCRS", "SCRI" }
     end;
{$endif}


{$ifdef GET_REGISTRATION}
 jssRegisterLoader(jmdtS3M, @jssLoadS3M, @jssDetectS3M, 'S3M', S3M_DataString);
{$endif}


{$ifdef GET_IMPLEMENTATION}
procedure jssLoadS3M(var AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);

 const jProcName = 'jssLoadS3M';

 var
    {** General loader variables **}
    S3MH:TS3MHeader;          { S3M-header }
    S3MI:TS3MInstrument;      { One temp. instrument for converting
                                S3M instruments to internal format }
    S3MInstDataPos:Array[1..S3M_MaxInstruments] of Longint;
    PPInst:Array[1..S3M_MaxInstruments] of PPtr;
    PPPat:Array[0..S3M_MaxPatterns] of PPtr;
    ATempByte:Byte;
    ATempWord:Word;
    ATempUsed:Boolean;
    ADataPtr,
    ATempInt, ATempInt2:Longint;

    ATempBufSize:Longint;
    ATempBufData:Pointer;

procedure S3M_AllocTempUnpackBuf(ASize:Longint);
begin
 ATempBufSize:= ASize;
 GetMem(ATempBufData, ATempBufSize);
end;

procedure S3M_DeallocTempUnpackBuf;
begin
 if (ATempBufData <> NIL) then
    begin
     FreeMem(ATempBufData, ATempBufSize);
     ATempBufData:= NIL;
     ATempBufSize:= 0;
    end;
end;

procedure S3M_UnpackPattern(ASourceData:Pointer; ASourceSize:Longint; var ADestPat:TJSSPattern);
 var
    ASourceOffs,
    APatChannel,
    APatPacking,
    APatRow:Longint;
    BValue:Byte;

function jssGetByte:Byte;
begin
 jssGetByte:= PByte(ASourceData)^;
 inc(longint(ASourceData));
 inc(ASourceOffs);
end;

begin
 {** Check data pointers **}
 if (ASourceData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM;
     S3M_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern SOURCE pointer NULL.');
     exit;
    end;

 if (ADestPat.AUsed = FALSE) or
    (ADestPat.AData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM;
     S3M_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern DEST pointer NULL.');
     exit;
    end;

 {** Init variables **}
 ASourceOffs:= 0;
 APatRow:= 0;
 BValue:= 0;

 {** Pattern unpacking loop **}
 while (APatRow < ADestPat.ARows) and
       (ASourceOffs < ASourceSize) do
 begin
  BValue:= jssGetByte;

  if (BValue <> 0) then
  begin
   {** Get channel number and packing byte of data **}
   APatChannel:= (BValue and $1f);
   APatPacking:= BValue;

   {**** Check if channel is bigger than maxchannels ****}
   if (APatChannel < S3M_MaxChannels) then
      with ADestPat.AData^[APatRow, (1 + APatChannel)] do
      begin
       Param:= 0;
       if ((APatPacking and $20) <> 0) then
          begin
           { PACK $20: Read note/octave and instrument }
           BValue:= jssGetByte;
           if (BValue = $ff) then   { Check if it is a NO-NOTE ($ff) or }
              Note:= mpNoData else
           if (BValue = $fe) then
              Note:= mpNoteOff else
              Note:= ((BValue shr 4) * 12) + (BValue and $0f);

           Instr:= jssGetByte;
          end;

       if ((APatPacking and $40) <> 0) then
          begin
           { PACK $40: Read volume}
           Volume:= jssGetByte;
          end;

       if ((APatPacking and $80) <> 0) then
          begin
           { PACK $80: Read command/effect and parameter }
           Effect:= jssGetByte;
           Param:= jssGetByte;
          end;
      end else

      begin
       {**** IGNORE EXTRA CHANNELS ****}
       if ((APatPacking and $20) <> 0) then
          begin
           { PACK $20: Read note/octave and instrument }
           jssGetByte;
           jssGetByte;
          end;

       if ((APatPacking and $40) <> 0) then
          begin
           { PACK $40: Read volume}
           jssGetByte;
          end;

       if ((APatPacking and $80) <> 0) then
          begin
           { PACK $80: Read command-effect and parameter }
           jssGetByte;
           jssGetByte;
          end;
      end;
  end else
      inc(APatRow);

 end; { WHILE }
 {** Unpacking OK. **}
end;


function S3M_GetFlags(AFlag:Word):Boolean;
begin
 S3M_GetFlags:= ((S3MH.Flags and AFlag) = AFlag);
end;


function S3M_GI(AFlag:Word):Boolean;
begin
 S3M_GI:= ((S3MI.Flags and AFlag) = AFlag);
end;


begin
 {******************** Check for internal inconsistencies ******************}
 if (S3M_MaxChannels > jsetMaxNChannels) or
    (S3M_MaxRows > jsetMaxRows) or
    (S3M_MaxPatterns > jsetMaxPatterns) or
    (S3M_MaxInstruments > jsetMaxInstruments) or
    (S3M_MaxOrders > jsetMaxOrders) then
    begin
     jssErrorN:= jerrInternalError;
     jssError(jProcName, 'Internal inconsistency in JLOADS3M.PAS');
     exit;
    end;

 {******************* Read S3M-header and check data in it *****************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Reading and checking S3M header...');
 {$endif}
 AStream^.vmGetData(@S3MH, sizeof(TS3MHeader));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading S3M header from stream 0x'+ ioPtr2Hex(AStream));
     exit;
    end;

 if (S3MH.BlockIdent <> 'SCRM') then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Not a real S3M module stream 0x'+ ioPtr2Hex(AStream));
     if (ioFailOnError) then exit;
    end;

 if (S3MH.SubType <> $10) then
    jssWarning(jProcName, 'Unknown S3M subtype 0x'+ ioHexB(S3MH.SubType));

 if (S3MH.NPatterns > S3M_MaxPatterns) then
    begin
     jssWarning(jProcName, 'Number of patterns > '+ ioStrL(S3M_MaxPatterns) +' ('+ ioStrL(S3MH.NPatterns) +')');
     S3MH.NPatterns:= S3M_MaxPatterns;
    end;

 if (S3MH.NOrders > S3M_MaxOrders) then
    begin
     jssWarning(jProcName, 'Number of orders > '+ ioStrL(S3M_MaxOrders) +' ('+ ioStrL(S3MH.NOrders) +')');
     S3MH.NOrders:= S3M_MaxOrders;
    end;

 if (S3MH.NOrders = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of orders was 0');
     if (ioFailOnError) then exit;
    end;

 if (S3MH.NPatterns = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of patterns was 0');
     if (ioFailOnError) then exit;
    end;

 if (S3MH.NInstruments > S3M_MaxInstruments) then
    begin
     jssWarning(jProcName, 'Number of instruments > '+ ioStrL(S3M_MaxInstruments)  +' ('+ ioStrL(S3MH.NInstruments) +')');
     S3MH.NInstruments:= S3M_MaxInstruments;
    end;

 if (S3MH.FFInfo <> 2) and
    (S3MH.FFInfo <> 1) then
    jssWarning(jProcName, 'FFInfo field in header was not 1 or 2 ('+ ioStrL(S3MH.FFInfo)+').');

 if (S3MH.TrackerVersion = $1300)
    then S3MH.Flags:= (S3MH.Flags OR s3mMFST300Slides);


 {************************* Init internal variables ************************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Initializing and allocating S3M module...');
 {$endif}
 if (jssAllocateModule(AModule) <> jerrOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssError(jProcName, 'Error allocating new module structure!');
     exit;
    end;

 with AModule^ do
 begin
  {** Convert the Header Info **}
  ModuleType    := jmdtS3M;
  jssASCIIZToStr(@S3MH.SongName, ModuleName, sizeof(S3MH.SongName));
  ModuleInfo    := S3M_DataString;
  DefSpeed      := S3MH.Speed;
  DefTempo      := S3MH.Tempo;
  DefGlobalVol  := S3MH.GlobalVol;
  DefMasterVol  := (S3MH.MasterVol and $7f);
  NInstruments  := S3MH.NInstruments;
  NPatterns     := S3MH.NPatterns;
  NOrders       := S3MH.NOrders;
  DefFlags      := 0;

  {** Convert the flags **}
  if (S3M_GetFlags(s3mMFAmigaSlides))
     then jssSetFlags(AModule, jmdfAmigaPeriods);

  if (S3M_GetFlags(s3mMFAmigaLimits))
     then jssSetFlags(AModule, jmdfAmigaLimits);

  if (S3M_GetFlags(s3mMFST300Slides))
     then jssSetFlags(AModule, jmdfST300Slides);

  if (S3MH.MasterVol and $80 <> 0)
     then jssSetFlags(AModule, jmdfStereo);

  {** Convert the channel settings **}
  for ATempInt:= 1 to S3M_MaxChannels do
      begin
       {** Set panning information **}
       if (S3MH.ChSettings[ATempInt] and $8 = 0)
          then DefPanning[ATempInt]:= jchPanLeft   { bit 4 = 0, pan channel to LEFT }
          else DefPanning[ATempInt]:= jchPanRight; { bit 4 = 1, pan channel to RIGHT }

       {** Find last used channel **}
       if (S3MH.ChSettings[ATempInt] and $80 = 0)
          then NChannels:= ATempInt;
      end;
 end;

 {************************* Read pattern orderlist *************************}
 for ATempInt:= 0 to (AModule^.NOrders - 1) do
 begin
  ATempByte:= AStream^.vmGetByte;
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading orderlist (item #'+
                    ioStrL(ATempInt) +'/'+ ioStrL(S3MH.NOrders - 1));
      if (ioFailOnError) then exit;
     end;

  case ATempByte of
   $ff: ATempInt2:= mpOrderEnd;
   $fe: ATempInt2:= mpOrderSkip;
   else ATempInt2:= ATempByte;
  end;

  AModule^.OrderList[ATempInt]:= ATempInt2;
 end;

 {********* Read parapointers to instrument headers and patterns ***********}
 {$ifdef OPT_MODDEBUG}
 writeln(' Reading parapointers:');
 write  ('  Instruments #', S3MH.NInstruments);
 {$endif}
 for ATempInt:= 1 to S3MH.NInstruments do
     begin
      AStream^.vmGetData(@PPInst[ATempInt], 2);
      if (AStream^.vmStatus <> strOK) then
         begin
          SUP_MAC_FAIL_STREAM
          jssErrorN:= jerrFileReadError;
          jssError(jProcName, 'Error reading instrument parapointer #'+
                        ioStrL(ATempInt) +'/'+ ioStrL(S3MH.NInstruments));
          if (ioFailOnError) then exit;
         end;

      {$ifdef OPT_MODDEBUG}
      write('.');
      {$endif}
     end;

 {$ifdef OPT_MODDEBUG}
 writeln;
 write  ('  Patterns #', S3MH.NPatterns);
 {$endif}
 for ATempInt:= 0 to (S3MH.NPatterns - 1) do
     begin
      AStream^.vmGetData(@PPPat[ATempInt], 2);
      if (AStream^.vmStatus <> strOK) then
         begin
          SUP_MAC_FAIL_STREAM
          jssError(jProcName, 'Error reading pattern parapointer #'+
                     ioStrL(ATempInt) +'/'+ ioStrL(S3MH.NPatterns - 1));
          if (ioFailOnError) then exit;
         end;
      {$ifdef OPT_MODDEBUG}
      write('.');
      {$endif}
     end;
 {$ifdef OPT_MODDEBUG}
 writeln;
 {$endif}

 {************ Read the default channel panning info, if needed ************}
 {! If the "panning" flag in S3M header is 252 ($FC... hehe) the default
    panning information is supplied. We now read it if it is there.
 }
 if (S3MH.Panning = $FC) then
 with AModule^ do
 begin
  {$ifdef OPT_MODDEBUG}
  writeln(' Reading channel panning information...');
  {$endif}

  for ATempInt:= 1 to S3M_MaxChannels do
      begin
       {** Get one panning position **}
       AStream^.vmGetByte;
       if (AStream^.vmStatus <> strOK) then
          begin
           SUP_MAC_FAIL_STREAM
           jssErrorN:= jerrFileReadError;
           jssError(jProcName, 'Error reading channel panning information!');
           if (ioFailOnError) then exit;
          end;

       ATempByte:= (ATempByte AND $0F); { We must mask of the upper 4 bits! }
       case ATempByte of
        01..06,
        08..14: DefPanning[ATempInt]:= (ATempByte * $11);
        00:     DefPanning[ATempInt]:= jchPanLeft;
        07:     DefPanning[ATempInt]:= jchPanMiddle;
        15:     DefPanning[ATempInt]:= jchPanRight;
        else    DefPanning[ATempInt]:= jchPanMiddle;
       end;
      end;
 end;

 {************************* Read instrument headers ************************}
 {$ifdef OPT_MODDEBUG}
 write(' Reading instrument headers');
 {$endif}

 for ATempInt:= 1 to AModule^.NInstruments do
 begin
  {** Read one instrument header **}
  ADataPtr:= (longint(PPInst[ATempInt]) * S3M_Alignment);
  AStream^.vmSeek(ADataPtr);
  AStream^.vmGetData(@S3MI, sizeof(TS3MInstrument));
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading instrument header #'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  {** Check the instrument **}
  if (S3MI.BlockDesc = 'SCRI') then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'FM-melody SCRI instruments (#'+ ioStrL(ATempInt) +') are not supported!');
      if (ioFailOnError) then exit;
     end;

  if (S3MI.BlockDesc <> 'SCRS') and
     (S3MI.BlockDesc <> #0#0#0#0) then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'Not a compatible instrument #'+ ioStrL(ATempInt)
                           +': "'+ S3MI.BlockDesc +'"');
      if (ioFailOnError) then exit;
     end;

  case S3MI.InstType of
   0: ATempUsed:= FALSE;
   1: ATempUsed:= TRUE;
   2, 3: begin
          SUP_MAC_FAIL_STREAM
          jssError(jProcName, 'AdLib instruments not supported #'+ ioStrL(ATempInt));
          if (ioFailOnError) then exit;
         end;
   else  begin
          SUP_MAC_FAIL_STREAM
          jssError(jProcName, 'Unknown S3M instrument type ('+ ioHexW(S3MI.InstType) +') in #'+ ioStrL(ATempInt));
          if (ioFailOnError) then exit;
         end;
  end;

  S3MInstDataPos[ATempInt]:= ( (longint(S3MI.PP1) SHL 16) + longint(S3MI.PP2)) shl 4;

  {** Allocate instrument and init **}
  if (jssAllocateInstrument(AModule^.Instruments[ATempInt]) <> jerrOK) then
     begin
      SUP_MAC_CLOSE_STREAM
      jssError(jProcName, 'Error allocating instrument structure #'+ ioStrL(ATempInt));
      exit;
     end;

  with AModule^.Instruments[ATempInt]^ do
  begin
   {** Initialize information **}
   Used        := ATempUsed;
   jssASCIIZToStr(@S3MI.Filename, Filename, sizeof(S3MI.Filename));
   Size        := S3MI.Size;
   LStart      := S3MI.LStart;
   LEnd        := S3MI.LEnd;
   Volume      := S3MI.Volume;
   if (Volume > mpMaxVol) then
      Volume:= mpMaxVol;
      
   C4BaseSpeed := S3MI.C4BaseSpeed;
   jssASCIIZToStr(@S3MI.Description, Description, sizeof(S3MI.Description));
   Data        := NIL;
   Flags       := 0;

   {** Check and initialize flags **}
   if (S3M_GI(s3mIFstereo)) then
      begin
       SUP_MAC_FAIL_STREAM
       jssError(jProcName, 'Stereo samples are not supported #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
      end;

   if (S3M_GI(s3mIF16bit)) then
      begin
       Flags:= (Flags OR jsf16bit);
       ByteSize:= (Size * 2);
      end else
       ByteSize:= Size;

   if (S3M_GI(s3mIFlooped)) then
      Flags:= (Flags OR jsfLooped);

  end; { WITH }
  {$ifdef OPT_MODDEBUG}
  write('.');
  {$endif}
 end;

 {*************************** Read instrument data *************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading instrument data');
 {$endif}
 for ATempInt:= 1 to AModule^.NInstruments do
 begin
 if (AModule^.Instruments[ATempInt] <> NIL) and
    (AModule^.Instruments[ATempInt]^.Used) then
 with AModule^.Instruments[ATempInt]^ do
 begin
  {** CHECK instrument data **}
  if (Size = 0) then
     begin
      SUP_MAC_FAIL_STREAM
      jssError(jProcName, 'Instrument sampledata size was 0 in #'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  if ((Flags AND jsfLooped) <> 0) then
     begin
     {** Check looped sample information **}
     if (LStart > LEnd) then
        begin
         SUP_MAC_FAIL_STREAM
         jssError(jProcName, 'Instrument sample loop-start > loop-start in #'+ ioStrL(ATempInt));
         if (ioFailOnError) then exit;
        end;

     if (LStart = LEnd) then
        begin
         SUP_MAC_FAIL_STREAM
         jssError(jProcName, 'Instrument sample loop-start == loop-start in #'+ ioStrL(ATempInt));
         if (ioFailOnError) then exit;
        end;
     end; { IF (Looped) }

  {** Compute correct fileposition **}
  ADataPtr:= S3MInstDataPos[ATempInt];

  if (ADataPtr = 0) then
     begin
       SUP_MAC_FAIL_STREAM
       jssError(jProcName, 'No sampledata for instrument #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
     end;

  if (ADataPtr > AStream^.vmGetSize) then
     begin
       SUP_MAC_FAIL_STREAM
       jssError(jProcName, 'Unexpected end of file reading instrument sampledata #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
     end;

  {** Allocate the memory **}
  GetMem(Data, ByteSize);
  if (Data = NIL) then
     begin
        SUP_MAC_CLOSE_STREAM
        jssErrorN:= jerrMemAllocError;
        jssError(jProcName, 'Memory allocation error for sample #'+ ioStrL(ATempInt));
        exit;
     end;


  {** Read the sampledata **}
  AStream^.vmSeek(ADataPtr);
  AStream^.vmGetData(Data, ByteSize);
  if (AStream^.vmStatus <> strOK) then
     begin
        SUP_MAC_FAIL_STREAM
        jssErrorN:= jerrFileReadError;
        jssError(jProcName, 'Error reading sample data for instrument #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
     end;

  {** Check if we have to flip sampledata sign **}
  { If (FFInfo = 1), the sampledata is SIGNED so we flip it.
    This is QUITE rare case, although..
    If FFInfo = 2, the sampledata is UNSIGNED.
    }
  if (S3MH.FFInfo = 1) then
     jssFlipSampleSign8(Data, ByteSize);

  {$ifdef OPT_MODDEBUG}
  write('.');
  {$endif}
 end; { WITH }
 end; { FOR }

 {***************************** Read patterns ******************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading pattern data');
 {$endif}

 {** Read and unpack patterns **}
 for ATempInt:= 0 to (AModule^.NPatterns - 1) do
 if (PPPat[ATempInt] <> 0) then
 begin
   {** Go to right fileoffset **}
   ADataPtr:= (longint(PPPat[ATempInt]) * S3M_Alignment);
   AStream^.vmSeek(ADataPtr);

   {** Read the packed pattern data from file **}
   AStream^.vmGetData(@ATempWord, sizeof(ATempWord));
   if (AStream^.vmStatus <> strOK) then
      begin
       SUP_MAC_FAIL_STREAM
       S3M_DeallocTempUnpackBuf;
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading pattern SIZE-data for #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
      end;

   S3M_AllocTempUnpackBuf(ATempWord);

   AStream^.vmGetData(ATempBufData, ATempBufSize);
   if (AStream^.vmStatus <> strOK) then
      begin
       SUP_MAC_FAIL_STREAM
       S3M_DeallocTempUnpackBuf;
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading pattern data for #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
      end;

   {** Allocate new pattern and unpack patterndata **}
   if (jssAllocatePattern(AModule^.Patterns[ATempInt], S3M_MaxRows) <> jerrOK) then
      begin
       SUP_MAC_CLOSE_STREAM
       S3M_DeallocTempUnpackBuf;
       jssError(jProcName, 'Error allocating pattern data.');
       exit;
      end;

   S3M_UnpackPattern(ATempBufData, ATempBufSize, AModule^.Patterns[ATempInt]);

   S3M_DeallocTempUnpackBuf;
   {$ifdef OPT_MODDEBUG}
   write('.');
   {$endif}
 end;

 {$ifdef OPT_MODDEBUG}
 writeln;
 {$endif}

 SUP_MAC_CLOSE_STREAM
end;


function jssDetectS3M(AStream:PStream; var AExtra:Pointer):Longint;
 var AResult:Longint; S3MH:TS3MHeader;
begin
 AExtra:= NIL;
 AResult:= -1;

 AStream^.vmGetData(@S3MH, sizeof(TS3MHeader));
 if (AStream^.vmStatus <> strOK) then exit(-2);

 if (S3MH.BlockIdent = 'SCRM') then
    inc(AResult, 2);
    
 if (S3MH.UnUsed1A = $1A) then
    inc(AResult);

 if (S3MH.SubType = $10) then
    inc(AResult);

 jssDetectS3M:= AResult;
end;

{$endif}
