{
  J Sound System - Fast Tracker ][ (XM) module loader
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  Self explanatory.

  TO DO:
  * Add support for 1.02/1.03 XM-format versions.
    (Not very useful, but if it's not too hard, then I'll do it)

  * The instrument loading is bit messy, maybe clean it up a little.
  
  Bugs: - None known.
}

{$ifdef GET_IMPLEMENTATION}
 const
      {** XM data string **}
      XM_DataString = 'XM loader v0.97 (updated 12/3/2001)';

      {** XM definations **}
      XM_MaxChannels    = 32;
      XM_MaxRows        = 64;
      XM_MaxPatterns    = 256;
      XM_MaxOrders      = 255;
      XM_MaxInstruments = 128;
      XM_MaxInstSamples = 16;
      XM_MaxEnvPoints   = 12;
      XM_MaxNotes       = 96;

 type
     {** XM-header **}
     TXMHeader = record
      IDText:String17;           { XM header ID "Extended Module: " }
      SongName:String20;         { Module song name }
      UnUsed1A:Byte;             { Always 0x1a }
      TrackerName:String20;      { Name of the tracker }
      Version:Word;              { Version number (should be 0x0104) }
      HSize:Longint;             { Module header size, FROM THIS POINT!! }

      NOrders,                   { Number of orders }
      DefRestartPos,             { Default restart position }
      NChannels,                 { Number of channels }
      NPatterns,                 { Number of patterns }
      NInstruments,              { Number of instruments }
      Flags,                     { Module flags:
                                   bit0: 0 = Amiga frequency table
                                         1 = Linear frequency table
                                 }
      DefSpeed,                  { Default speed }
      DefTempo:Word;             { Default tempo }

      OrderList:Array[0..255] of Byte;
     end;

     {** XM Instrument Header **}
     TXMInstrument1 = record
      HSize:Longint;             { Instrument header size (SEE DOCS!!) }
      InstName:String22;         { Instrument name/description }
      InstType:Byte;             { Instrument type }
      NSamples:Word;             { Number of samples }
     end;

     TXMEnvelopePoint = record
      AFrame, AValue:Word;
     end;

     TXMEnvelopePoints = Array[1..XM_MaxEnvPoints] of TXMEnvelopePoint;

     TXMInstrument2 = record
      HSize:Longint;
      SNumForNotes:Array[1..XM_MaxNotes] of Byte;
      VolumeENVPoints,
      PanningENVPoints:TXMEnvelopePoints;
      NVolumePoints,
      NPanningPoints,

      AVolumeSustainPoint,
      AVolumeLoopStart,
      AVolumeLoopEnd,

      APanningSustainPoint,
      APanningLoopStart,
      APanningLoopEnd,

      AVolumeType,
      APanningType,

      AVibratoType,
      AVibratoSweep,
      AVibratoDepth,
      AVibratoRate:Byte;

      AFadeout:Word;
      ARESERVED:Word;
     end;

     {** XM Sample Header **}
     TXMSample = record
      ASize:Longint;
      ALoopStart,
      ALoopLength:Longint;
      AVolume:Byte;
      AFineTune:ShortInt;
      AType,
      APanning:Byte;
      ARelNote:ShortInt;
      ARESERVED:Byte;
      ASampleName:String22;
     end;

     {** XM pattern data header **}
     TXMPattern = record
      HSize:Longint;
      APacking:Byte;
      NRows:Word;
      ASize:Word;
     end;
{$endif}


{$ifdef GET_REGISTRATION}
 jssRegisterLoader(jmdtXM, @jssLoadXM, @jssDetectXM, 'XM', XM_DataString);
{$endif}


{$ifdef GET_IMPLEMENTATION}
procedure jssLoadXM(var AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);

 const jProcName = 'jssLoadXM';

 var
    {** General loader variables **}
    XMH:TXMHeader;           { XM-header }
    XMI1:TXMInstrument1;     { Primary XM instrument header }
    XMI2:TXMInstrument2;     { Secondary XM instrument header }
    XMP:TXMPattern;
    XMConvTable:Array[0..XM_MaxInstruments] of Longint;

    ATempInt, ATempInt2:Longint;
    ATempBufSize:Longint;
    ATempBufData:Pointer;


procedure XM_AllocTempUnpackBuf(ASize:Longint);
begin
 ATempBufSize:= ASize;
 GetMem(ATempBufData, ATempBufSize);
end;

procedure XM_DeallocTempUnpackBuf;
begin
 if (ATempBufData <> NIL) then
    begin
     FreeMem(ATempBufData, ATempBufSize);
     ATempBufData:= NIL;
     ATempBufSize:= 0;
    end;
end;

procedure XM_UnpackPattern(ASourceData:Pointer; ASourceSize:Longint; var ADestPat:TJSSPattern);
 var
    ASourceOffs,
    APatChannel,
    APatPacking,
    APatRow:Longint;

function jssGetByte:Byte;
begin
 jssGetByte:= PByte(ASourceData)^;
 inc(longint(ASourceData));
 inc(ASourceOffs);
end;

begin
 {** Check data pointers **}
 if (ASourceData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     XM_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern SOURCE pointer NULL.');
     exit;
    end;

 if (ADestPat.AUsed = FALSE) or
    (ADestPat.AData = NIL) then
    begin
     SUP_MAC_CLOSE_STREAM
     XM_DeallocTempUnpackBuf;
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Error unpacking pattern. Pattern DEST pointer NULL.');
     exit;
    end;

 {** Init variables **}
 ASourceOffs:= 0;
 APatRow:= 0;

 {** Pattern unpacking loop **}
 while (APatRow < ADestPat.ARows) and
       (ASourceOffs < ASourceSize) do
 begin
  {** Get channel number and packing byte of data **}
  for APatChannel:= 1 to AModule^.NChannels do
      with ADestPat.AData^[APatRow, APatChannel] do
      begin
       Param:= 0;
       APatPacking:= jssGetByte;
       if ((APatPacking and $80) <> 0) then
       begin
       if ((APatPacking and $01) <> 0) then
          begin
           { PACK $01: Read note }
           Note:= jssGetByte;
           if (Note < 1) or
              (Note > 97) then
              Note:= mpNoData else
           if (Note = 97) then
              Note:= mpNoteOff;
          end;

       if ((APatPacking and $02) <> 0) then
          begin
           { PACK $02: Read instrument }
           Instr:= jssGetByte;
          end;

       if ((APatPacking and $04) <> 0) then
          begin
           { PACK $04: Read volume }
           Volume:= jssGetByte;
           if (Volume < $10)
              then Volume:= mpNoData
              else Volume:= (Volume - $10);
          end;

       if ((APatPacking and $08) <> 0) then
          begin
           { PACK $08: Read effect }
           Effect:= jssGetByte;
          end;

       if ((APatPacking and $10) <> 0) then
          begin
           { PACK $10: Read effect parameter }
           Param:= jssGetByte;

           { If no effect with param, use 0xy }
           if (Effect = mpNoData)
              then Effect:= 0;
          end;
       end else
       begin
       {** All data available **}
       Note:= (APatPacking and $7f);
       if (Note < 1) or 
          (Note > 97) then
          Note:= mpNoData else
       if (Note = 97) then 
          Note:= mpNoteOff;
 
       Instr:= jssGetByte;
       Volume:= jssGetByte;
       if (Volume = 0)
          then Volume:= mpNoData
          else Volume:= (Volume - $10);
       Effect:= jssGetByte;
       Param:= jssGetByte;
       end;
      end;

  inc(APatRow);
 end; { WHILE }
 {** Unpacking OK. **}
end;


function XM_LoadInstSamples:Longint;
 var
    ANSamples,
    ATempI:Longint;
    XMS:TXMSample;
begin
 {** Now read the sample headers **}
 ANSamples:= 0;
 for ATempI:= 1 to XMI1.NSamples do
 begin
  {** Read the header data **}
  AStream^.vmGetData(@XMS, sizeof(TXMSample));
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading inst.sample header #'+ ioStrL(ATempI) +'/'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  { Allocate a normal JSS instrument and convert the XM-format SAMPLE header into that! }
  if (XMS.ASize > 0) then
  begin
  { Update the XMConvTable[] }
  XMConvTable[ATempI - 1]:= 1;

  { Allocate the sample }
  inc(ANSamples);
  if (jssAllocateInstrument(AModule^.Instruments[AModule^.NInstruments + ANSamples]) <> jerrOK) then
     begin
      SUP_MAC_CLOSE_STREAM
      jssError(jProcName, 'Error allocating sampleinst structure #'+ ioStrL(ANSamples) +'/'+ ioStrL(ATempInt));
      exit;
     end;

  with AModule^.Instruments[AModule^.NInstruments + ANSamples]^ do
       begin
        {** Convert the header data **}
        Used        := TRUE;
        Size        := XMS.ASize;
        LStart      := XMS.ALoopStart;
        LEnd        := (XMS.ALoopStart + XMS.ALoopLength);
        Volume      := XMS.AVolume;
        if (Volume > mpMaxVol) then
           Volume:= mpMaxVol;
           
        ERelNote    := XMS.ARelNote;
        EFineTune   := XMS.AFineTune;
        EPanning    := XMS.APanning;

        jssASCIIZToStr(@XMS.ASampleName, Description, sizeof(XMS.ASampleName));

	{** Convert the flags **}
	case (XMS.AType AND 3) of
         0: Flags:= 0;
         1: Flags:= jsfLooped;
         2: Flags:= (jsfLooped OR jsfBidi);
	end;

        if ((XMS.AType AND $10) <> 0) then
           begin
            {** 16-bit sample: convert info **}
            Flags:= (Flags OR jsf16bit);
            ByteSize:= Size;
            Size:= (Size SHR 1);
            LStart:= (LStart SHR 1);
            LEnd:= (LEnd SHR 1);
           end else
            {** 8-bit sample: just copy info **}
            ByteSize:= Size;

        {** Allocate memory for sample data **}
        GetMem(Data, ByteSize);
        if (Data = NIL) then
           begin
            SUP_MAC_CLOSE_STREAM
            jssErrorN:= jerrMemAllocError;
            jssError(jProcName, 'Memory allocation error for sample #'+ ioStrL(ATempInt));
            exit;
           end;
       end; { WITH }
  end;
 end; { FOR }

 {** And then read the sample data **}
 for ATempI:= 1 to ANSamples do
 if (AModule^.Instruments[AModule^.NInstruments + ATempI] <> NIL) then
 with AModule^.Instruments[AModule^.NInstruments + ATempI]^ do
 begin
  {** Read the sample data **}
  AStream^.vmGetData(Data, ByteSize);
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading sample data #'+ ioStrL(ATempI) +'/'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  {** Decode the delta-encoding to normal data **}
  if (jssGetANYFlag(Flags, jsf16bit))
     then jssDecodeDelta16(Data, Size)
     else jssDecodeDelta8(Data, Size);
 end;

 XM_LoadInstSamples:= ANSamples;
end;

 var
    ATempL, ATempL2, ATempL3:Longint;

begin
 {******************** Check for internal inconsistencies ******************}
 if (XM_MaxChannels > jsetMaxNChannels) or
    (XM_MaxRows > jsetMaxRows) or
    (XM_MaxPatterns > jsetMaxPatterns) or
    (XM_MaxInstruments > jsetMaxInstruments) or
    (XM_MaxOrders > jsetMaxOrders) then
    begin
     jssErrorN:= jerrInternalError;
     jssError(jProcName, 'Internal inconsistency in JLOADXM.PAS');
     exit;
    end;

 {******************* Read XM-header and check data in it *****************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Reading and checking XM header...');
 {$endif}
 AStream^.vmGetData(@XMH, sizeof(TXMHeader));
 if (AStream^.vmStatus <> strOK) then
    begin
     SUP_MAC_CLOSE_STREAM
     jssErrorN:= jerrFileReadError;
     jssError(jProcName, 'Error reading XM header from stream 0x'+ ioPtr2Hex(AStream));
     exit;
    end;

 if (XMH.IDText <> 'Extended Module: ') then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Not a real XM module stream 0x'+ ioPtr2Hex(AStream));
     if (ioFailOnError) then exit;
    end;

 if (XMH.Version <> $0104) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Unsupported XM module version 0x'+ ioHexW(XMH.Version) +' in stream 0x'+ ioPtr2Hex(AStream));
     if (ioFailOnError) then exit;
    end;

 if (XMH.NOrders > XM_MaxOrders) then
    begin
     jssWarning(jProcName, 'Number of orders > '+ ioStrL(XM_MaxOrders) +' ('+ ioStrL(XMH.NOrders) +')');
     XMH.NOrders:= XM_MaxOrders;
    end;

 if (XMH.NOrders = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of orders was 0 in stream 0x'+ ioPtr2Hex(AStream));
     if (ioFailOnError) then exit;
    end;

 if (XMH.NPatterns > XM_MaxPatterns) then
    begin
     jssWarning(jProcName, 'Number of patterns > '+ ioStrL(XM_MaxPatterns) +' ('+ ioStrL(XMH.NPatterns) +') in stream 0x'+ ioPtr2Hex(AStream));
     XMH.NPatterns:= XM_MaxPatterns;
    end;

 if (XMH.NPatterns = 0) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Number of patterns was 0 in stream 0x'+ ioPtr2Hex(AStream));
     if (ioFailOnError) then exit;
    end;

 if (XMH.NChannels = 0) or
    (XMH.NChannels > XM_MaxChannels) then
    begin
     SUP_MAC_FAIL_STREAM
     jssError(jProcName, 'Invalid number of channels ('+ ioStrL(XMH.NChannels) +') in stream 0x'+ ioPtr2Hex(AStream));
     if (ioFailOnError) then exit;
    end;

 if (XMH.NInstruments > XM_MaxInstruments) then
    begin
     jssWarning(jProcName, 'Number of instruments > '+ ioStrL(XM_MaxInstruments) +' ('+ ioStrL(XMH.NInstruments) +') in stream 0x'+ ioPtr2Hex(AStream));
     XMH.NInstruments:= XM_MaxInstruments;
    end;


 {************************* Init internal variables ************************}
 {$ifdef OPT_MODDEBUG}
 writeln(' Initializing and allocating XM module...');
 {$endif}
 if (jssAllocateModule(AModule) <> jerrOK) then
    begin
     dispose(AStream, vmClose);
     jssError(jProcName, 'Error allocating new module structure!');
     exit;
    end;

 with AModule^ do
 begin
  {** Convert the Header Info **}
  ModuleType    := jmdtXM;
  jssASCIIZToStr(@XMH.SongName, ModuleName, sizeof(XMH.SongName));
  ModuleInfo    := XM_DataString;
  DefSpeed      := XMH.DefSpeed;
  DefTempo      := XMH.DefTempo;
  DefGlobalVol  := 48;
  DefMasterVol  := $50;
  NExtInstruments:= XMH.NInstruments;
  NInstruments  := 0;
  NPatterns     := XMH.NPatterns;
  NOrders       := XMH.NOrders;
  NChannels     := XMH.NChannels;
  DefFlags      := (jmdfStereo OR jmdfFT2Replay);
  DefRestartPos := XMH.DefRestartPos;
  
  if ((XMH.Flags AND 1) = 0) then
     jssSetFlags(AModule, jmdfAmigaPeriods);

  {** Setup the default pannings **}
  for ATempInt:= 1 to XM_MaxChannels do
      DefPanning[ATempInt]:= jchPanMiddle;

  {** Convert the song orders list **}
  { Replace unexisting patterns (N > NPatterns) with
    pattern number jsetMaxPatterns, which is an empty
    and allocated below.
  }
  for ATempInt:= 0 to (NOrders - 1) do
      begin
       ATempInt2:= XMH.OrderList[ATempInt];
       if (ATempInt2 > NPatterns)
          then ATempInt:= jsetMaxPatterns;
       OrderList[ATempInt]:= ATempInt2;
      end;
 end;

 {***************************** Read patterns ******************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading pattern data');
 {$endif}

 {** Read and unpack patterns **}
 for ATempInt:= 0 to (AModule^.NPatterns - 1) do
 begin
   {** Read the packed pattern data from file **}
   AStream^.vmGetData(@XMP, sizeof(XMP));
   if (AStream^.vmStatus <> strOK) then
      begin
       SUP_MAC_FAIL_STREAM
       XM_DeallocTempUnpackBuf;
       jssErrorN:= jerrFileReadError;
       jssError(jProcName, 'Error reading pattern header for #'+ ioStrL(ATempInt));
       if (ioFailOnError) then exit;
      end;

   if (XMP.APacking <> 0) then
      begin
        SUP_MAC_FAIL_STREAM
        jssErrorN:= jerrFileReadError;
        jssError(jProcName, 'Unsupported pattern packing method in #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
      end;

   if (XMP.NRows = 0) then
      begin
        SUP_MAC_FAIL_STREAM
        jssErrorN:= jerrFileReadError;
        jssError(jProcName, 'Number of rows zero in pattern #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
      end;

   if (XMP.ASize > 0) then
   begin
    XM_AllocTempUnpackBuf(XMP.ASize);

    AStream^.vmGetData(ATempBufData, ATempBufSize);
    if (AStream^.vmStatus <> strOK) then
       begin
        SUP_MAC_FAIL_STREAM
        XM_DeallocTempUnpackBuf;
        jssErrorN:= jerrFileReadError;
        jssError(jProcName, 'Error reading pattern data for #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
       end;

    {** Allocate new pattern and unpack patterndata **}
    if (jssAllocatePattern(AModule^.Patterns[ATempInt], XMP.NRows) <> jerrOK) then
       begin
        dispose(AStream, vmClose);
        XM_DeallocTempUnpackBuf;
        jssError(jProcName, 'Error allocating pattern data for #'+ ioStrL(ATempInt));
        exit;
       end;

    XM_UnpackPattern(ATempBufData, ATempBufSize, AModule^.Patterns[ATempInt]);

    XM_DeallocTempUnpackBuf;
   end;
   {$ifdef OPT_MODDEBUG}
   write('.');
   {$endif}
 end;

 {** Allocate the empty pattern **}
 ATempInt:= jsetMaxPatterns;
 if (jssAllocatePattern(AModule^.Patterns[ATempInt], XMP.NRows) <> jerrOK) then
    begin
     dispose(AStream, vmClose);
     jssError(jProcName, 'Error allocating pattern data for #'+ ioStrL(ATempInt));
     exit;
    end;

 {************************* Read instrument headers ************************}
 {$ifdef OPT_MODDEBUG}
 writeln;
 write(' Reading instruments');
 {$endif}
 for ATempInt:= 1 to XMH.NInstruments do
 begin
  {** Read the main instrument header **}
  AStream^.vmGetData(@XMI1, sizeof(TXMInstrument1));
  if (AStream^.vmStatus <> strOK) then
     begin
      SUP_MAC_FAIL_STREAM
      jssErrorN:= jerrFileReadError;
      jssError(jProcName, 'Error reading instrument A-header #'+ ioStrL(ATempInt));
      if (ioFailOnError) then exit;
     end;

  {** Allocate extended instrument **}
  if (jssAllocateExtInstrument(AModule^.ExtInstruments[ATempInt]) <> jerrOK) then
     begin
      dispose(AStream, vmClose);
      jssError(jProcName, 'Error allocating extinstrument structure #'+ ioStrL(ATempInt));
      exit;
     end;

  with AModule^.ExtInstruments[ATempInt]^ do
  begin
   {** Initialize extended instrument **}
   Used     := (XMI1.NSamples > 0);
   jssASCIIZToStr(@XMI1.InstName, Description, sizeof(XMI1.InstName));

   if (Used) then
   begin
    {** Yes! Read the second instrument header **}
    AStream^.vmGetData(@XMI2, sizeof(TXMInstrument2));
    if (AStream^.vmStatus <> strOK) then
       begin
        SUP_MAC_FAIL_STREAM
        jssErrorN:= jerrFileReadError;
        jssError(jProcName, 'Error reading instrument B-header #'+ ioStrL(ATempInt));
        if (ioFailOnError) then exit;
       end;

    {** Skip the extra data after 2nd header **}
    ATempL:= (SizeOf(TXMInstrument1) + SizeOf(TXMInstrument2));
    if (XMI1.HSize > ATempL) then
       begin
        ATempL:= (XMI1.HSize - ATempL);
        AStream^.vmSeek(AStream^.vmGetPos + ATempL);
       end;

    {** Convert all extended instrument information **}
    { Initialize the SNumForNotes conversion table }
    for ATempL:= 0 to XM_MaxInstruments do
        XMConvTable[ATempL]:= -1;

    { Now read the actual samples ("normal" instruments) }
    NSamples:= XM_LoadInstSamples;

    { The new conversion method is:
      1) Find the used samples.
      2) Make a conversion table for internal numbers.
      3) Convert the values in SNumForNotes to these.
      Why? I didn't previously notice/remember this little fact (hehe :)
      that samples in XM instrument can be in "non-continuos" order,
      for example if only samples 1, 3 and 7 are used! The old algothrim
      did not take this into account so for example ANGEL.XM (by Maggot)
      sounded twisted :)
    }
    { Then, re-format and process the conversion table }
    ATempL3:= 1;
    for ATempL2:= 0 to XM_MaxInstruments do
        begin
         if (XMConvTable[ATempL2] > 0) then
            begin
             ATempL:= (ATempL3 + AModule^.NInstruments);
             if (AModule^.Instruments[ATempL] <> NIL) and
                (AModule^.Instruments[ATempL]^.Used) then
                begin
                 XMConvTable[ATempL2]:= ATempL;
                 inc(ATempL3);
                end else
                 XMConvTable[ATempL2]:= -1;
            end;
        end;

    { Apply the new values to the SNumForNotes values }
    for ATempL:= 1 to XM_MaxNotes do
        begin
         ATempL2:= XMI2.SNumForNotes[ATempL];
         if (ATempL2 >= 0) and
            (ATempL2 <= XMI1.NSamples) then
            SNumForNotes[ATempL]:= XMConvTable[ATempL2];
        end;

    { Convert envelope information }
    for ATempL:= 1 to XM_MaxEnvPoints do
        begin
         with AVolumeEnvelope.APoints[ATempL] do
              begin
               AValue:= XMI2.VolumeENVPoints[ATempL].AValue;
               AFrame:= XMI2.VolumeENVPoints[ATempL].AFrame;
              end;

         with APanningEnvelope.APoints[ATempL] do
              begin
               AValue:= XMI2.PanningENVPoints[ATempL].AValue;
               AFrame:= XMI2.PanningENVPoints[ATempL].AFrame;
              end;
        end;

    { Process and convert rest of envelope information }
    with AVolumeEnvelope do
         begin
          AFlags  := XMI2.AVolumeType;
          NPoints := XMI2.NVolumePoints;
          ALoopS  := (XMI2.AVolumeLoopStart + 1);
          ALoopE  := (XMI2.AVolumeLoopEnd + 1);
          ASustain:= (XMI2.AVolumeSustainPoint + 1);
         end;

    with APanningEnvelope do
         begin
          AFlags  := XMI2.APanningType;
          NPoints := XMI2.NPanningPoints;
          ALoopS  := (XMI2.APanningLoopStart + 1);
          ALoopE  := (XMI2.APanningLoopEnd + 1);
          ASustain:= (XMI2.APanningSustainPoint + 1);
         end;

    { Convert rest of info }
    AVibratoType  := XMI2.AVibratoType;
    AVibratoSweep := XMI2.AVibratoSweep;
    AVibratoDepth := XMI2.AVibratoDepth;
    AVibratoRate  := XMI2.AVibratoRate;
    AFadeOut      := XMI2.AFadeOut;
    
    inc(AModule^.NInstruments, NSamples);
   end else { IF (Used) }
   begin
    {** We may STILL need to skip extra data after 1st instr. header **}
    ATempL:= SizeOf(TXMInstrument1);
    if (XMI1.HSize > ATempL) then
       begin
        ATempL:= (XMI1.HSize - ATempL);
        AStream^.vmSeek(AStream^.vmGetPos + ATempL);
       end;
   end;

  end; { WITH }

  {$ifdef OPT_MODDEBUG}
  write('.');
  {$endif}
 end;

 {$ifdef OPT_MODDEBUG}
 writeln;
 {$endif}

 SUP_MAC_CLOSE_STREAM
end;


function jssDetectXM(AStream:PStream; var AExtra:Pointer):Longint;
 var AResult:Longint; XMH:TXMHeader;
begin
 AExtra:= NIL;
 AResult:= -1;

 AStream^.vmGetData(@XMH, sizeof(TXMHeader));
 if (AStream^.vmStatus <> strOK) then exit(-2);

 if (XMH.IDText = 'Extended Module: ') then
    inc(AResult, 2);
    
 if (XMH.UnUsed1A = $1A) then
    inc(AResult);

 if (XMH.Version = $0104) then
    inc(AResult);

 jssDetectXM:= AResult;
end;

{$endif}
