{
  J Sound System - General Module Handling System
  (C) Copyright 1999-2001 Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This is the main file of J Sound System module handling subsystem.
}
{$I ../jssdef.pas}

{** Translate and define some macros **}
{*$define OPT_CLOSE_STREAMS}

unit JSSMod;

interface

 uses
     BaseAPI,
     JSS;

 const
      {** Max data size definations **}
      jsetMaxRows        = 256;    { Max number of rows }
      jsetMaxPatterns    = 260;    { Max patterns }
      jsetMaxInstruments = 260;    { Max instruments }
      jsetMaxOrders      = 260;    { Max patterns in orderlist }
      jsetMinNote        = 0;      { Smallest note number }
      jsetMaxNote        = ((11 * 12) - 1);
      jsetMaxEnvPoints   = 64;     { Max number of envelope points }

      {** Instrument Envelope flags **}
      jenvfUsed          = $01;    { Envelope is used }
      jenvfSustain       = $02;    { Envelope has a sustain point (XM)
                                     or it has a sustain loop (IT) }
      jenvfLooped	 = $04;    { Envelope is looped }

      {** Some general stuff **}
      mpNoData           = -1;
      mpNoteOff          = -2;
      mpOrderEnd         = -1;
      mpOrderSkip        = -2;

      mpMinVol           = 0;
      mpMaxVol           = 64;

      {** Effect-table for MOD and compatibles **}
      mpMODEffectTable:Array[0..35] of Char =
      '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';

{$define GET_CONSTANTS}
{$I jssmod.inc}
{$undef GET_CONSTANTS}

      {** General module flags **}
      jmdfAmigaPeriods   = $00000001;  { Use non-linear periods (Amiga) }
      jmdfAmigaLimits    = $00000002;  { Use Amiga-styled value limits }
      jmdfStereo         = $00000008;  { Use stereo output if possible }

      jmdfFT2Replay      = $00000010;  { Use FT2 replaying bugs/features }
      jmdfST300Slides    = $00000020;  { Use Scream Tracker 3.00 slides }
      jmdfByteLStart     = $00000040;  { LStart in BYTES instead of WORDS (MOD only) }


 type
     {** General typedefs **}
     String17 = array[1..17] of Char;
     String20 = array[1..20] of Char;
     String22 = array[1..22] of Char;
     String26 = array[1..26] of Char;
     String28 = array[1..28] of Char;
     String12 = array[1..12] of Char;

     {********************* GLOBAL MODULE DEFS *******************}
     jStrFName = String[15];
     jStrDesc  = String[40];


     {** Internal instrument structure **}
     PJSSInstrument = ^TJSSInstrument;
     TJSSInstrument = record
      Used:Boolean;            { Is instrument DATA used? }
      Filename:jStrFName;      { DOS filename }
      Size,                    { Size of sampledata in UNITS }
      LStart,                  { Start offset of LOOP }
      LEnd,                    { End offset of LOOP }
      Volume,                  { Default playing volume }
      Flags,                   { Sample flags (see jsfXXXX constants) }
      C4BaseSpeed:Longint;     { Sample C-4 speed in Hz }
      Description:jStrDesc;    { Sample description }
      Data:Pointer;            { Pointer to instrument sampledata }
      ByteSize:Longint;        { Size of sampledata in BYTES }

      ERelNote,                { EXTENDED RelativeNote value }
      EFineTune:Longint;       { EXTENDED FineTune value }
      EPanning:Byte;           { EXTENDED Panning Position value }
     end;


     {** Envelope point structure **}
     TJSSEnvelopePoint = record
      AFrame, AValue:Longint;
     end;


     {** Envelope structure **}
     TJSSEnvelope = record
      AFlags,                  { Envelope type (see jenvXXXX flags) }
      NPoints,                 { Number of envelope points }
      ASustain,                { Sustain point }
      ALoopS,                  { Loop Start point }
      ALoopE:Byte;             { Loop End point }
      APoints:Array[0..jsetMaxEnvPoints] of TJSSEnvelopePoint;
     end;


     {** Extended instrument structure **}
     PJSSExtInstrument = ^TJSSExtInstrument;
     TJSSExtInstrument = record
      Used:Boolean;            { Is instrument DATA used? }
      Description:jStrDesc;    { Instrument description }
      NSamples:Longint;        { Number of samples used }
      SNumForNotes:Array[jsetMinNote..jsetMaxNote] of Longint;

      AVolumeEnvelope,         { Volume and panning envelopes }
      APanningEnvelope:TJSSEnvelope;

      AVibratoType,            { Type of instrument auto-vibrato }
      AVibratoSweep,           { Vibrato sweep value }
      AVibratoDepth,           { Vibrato depth }
      AVibratoRate:Longint;    { Vibrato rate (speed) }

      AFadeOut:Longint;        { Volume fadeout value }
     end;


     {** Internal pattern structures **}
     TJSSNote = record
      Note,                    { Note number }
      Instr,                   { Instrument number }
      Volume,                  { Volume value }
      Effect,                  { Effect command }
      Param:Integer;           { Effect parameter }
     end;


     PJSSPatternData = ^TJSSPatternData;
     TJSSPatternData = array[0..jsetMaxRows, jsetMinChannel..jsetMaxChannel] of TJSSNote;

     TJSSPattern = record
      AUsed:Boolean;           { Is this pattern used? (i.e. there is data) }
      ARows:Longint;           { Rows in this pattern. Size of data is
                                 always (Rows * jsetPatRowSize) bytes.
                               }
      AData:PJSSPatternData;   { Pattern data. NIL if none / unused }
     end;


     {** Module object structure **}
     PJSSModule = ^TJSSModule;
     TJSSModule = record
      ModuleType:Longint;      { Type of module: MOD, S3M, XM, etc.. }
      ModuleName:String;       { Name of module }
      ModuleInfo:String;       { Information string }

      DefSpeed,                { Default speed }
      DefTempo,                { Default tempo }
      DefGlobalVol,            { Global volume }
      DefMasterVol,            { Master volume }
      DefFlags,                { General module flags }
      DefRestartPos,           { Default re-start position in orderlist }

      IntFlags,                { INTERNAL player/loader specific flags }
      IntVersion,              { INTERNAL version tag }

      NChannels,               { Channels in module }
      NInstruments,            { Number of instruments }
      NExtInstruments,         { Number of extended instruments }
      NPatterns,               { Number of patterns }
      NOrders:Longint;         { Number of orders in orderlist }

      { Module default channel panning settings }
      DefPanning:Array[jsetMinChannel..jsetMaxChannel] of Byte;

      { Module orderlist: List of pattern numbers in playing order. }
      OrderList:Array[0..jsetMaxOrders] of Longint;

      { Module patterns: Pointers to the pattern data. }
      Patterns:Array[0..jsetMaxPatterns] of TJSSPattern;

      { Module instruments: Pointers to instrument structures. }
      Instruments:Array[1..jsetMaxInstruments] of PJSSInstrument;

      { Module extinstruments }
      ExtInstruments:Array[1..jsetMaxInstruments] of PJSSExtInstrument;
     end;

     {** Loader function typedefs **}
     TJSSModuleLoader   = procedure(var AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);
     TJSSModuleDetector = function(AStream:PStream; var AExtra:Pointer):Longint;

 const
      {** Pattern row size in bytes **}
      jsetPatRowSize     = (SizeOf(TJSSNote) * jsetMaxNChannels);


 {*** Module loader routine interfaces ***}
 function  jssRegisterLoader(VType:Longint; VLoader:TJSSModuleLoader; VDetector:TJSSModuleDetector; VName, VInfo:String):Longint;
 function  jssGetLoaders:Longint;
 function  jssGetLoaderInfo(ALoader:Longint):String;
 function  jssGetLoaderName(ALoader:Longint):String;
 function  jssGetLoaderType(ALoader:Longint):Longint;
 function  jssGetLoaderFromType(AType:Longint):Longint;

 procedure jssLoadModule(ALoader:Longint; AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);
 function  jssDetectModule(AStream:PStream; var AExtra:Pointer):Longint;
 function  jssDisposeModule(var AModule:PJSSModule):Longint;
 function  jssAllocateModule(var AModule:PJSSModule):Longint;
 function  jssAllocatePattern(var APattern:TJSSPattern; ARows:Longint):Longint;
 function  jssAllocateInstrument(var AInstr:PJSSInstrument):Longint;
 function  jssAllocateExtInstrument(var AInstr:PJSSExtInstrument):Longint;
 function  jssGetFlags(AModule:PJSSModule; AFlag:Longint):Boolean;

{$define GET_INTERFACE}
{$I jssmod.inc}
{$undef GET_INTERFACE}

implementation

{$PackRecords 1}

{****************************************************************************}
{*                                                                          *}
{* General basic routines for loaders...                                    *}
{*                                                                          *}
{****************************************************************************}
{ Writes a given sample to file as raw format data. Good for debugging.
}
function jssWriteSampleFile(AFilename:String; AData:Pointer; ASize:Longint):Boolean;
 var AFile:File;
begin
 {** Open/create the file **}
 jssWriteSampleFile:= FALSE;
 if (ioCreateFile(AFile, AFilename) <> 0) then
    begin
     jssErrorN:= jerrFileOpenError;
     jssError(jssName, 'Error opening sample file "'+ AFilename +'".');
     exit;
    end;

 {** Write data to the file **}
 blockwrite(AFile, AData^, ASize);
 if (IOResult <> 0) then
    begin
     jssErrorN:= jerrFileWriteError;
     jssError(jssName, 'Error writing to sample file "'+ AFilename +'".');
     exit;
    end;

 close(AFile);
 jssWriteSampleFile:= TRUE;
end;


{ Flips sign of 8-bit sample at given pointer with size given in BYTES.
  Returns TRUE if succesfull and FALSE if conversion failed.
}
function jssFlipSampleSign8(SData:Pointer; SSize:Longint):Boolean;
begin
 {** Check data pointer and size **}
 jssFlipSampleSign8:= FALSE;
 if (SData = NIL) or
    (SSize = 0) then
    exit;

 {** Decode the data **}
 while (SSize > 0) do
 begin
  PByte(SData)^:= (PByte(SData)^ XOR $80);
  inc(longint(SData));
  dec(SSize);
 end;

 jssFlipSampleSign8:= TRUE;
end;


{ Flips sign of 16-bit sample at given pointer with size given in WORDS.
  Returns TRUE if succesfull and FALSE if conversion failed.
}
function jssFlipSampleSign16(SData:Pointer; SSize:Longint):Boolean;
begin
 {** Check data pointer and size **}
 jssFlipSampleSign16:= FALSE;
 if (SData = NIL) or
    (SSize = 0) then
    exit;

 {** Decode the data **}
 while (SSize > 0) do
 begin
  PWord(SData)^:= (PWord(SData)^ XOR $8000);
  inc(longint(SData), 2);
  dec(SSize);
 end;

 jssFlipSampleSign16:= TRUE;
end;


{ Decodes a delta coded 8-bit waveform sample to a normal unsigned sample.
  Returns TRUE if succesfull and FALSE if conversion failed.
}
function jssDecodeDelta8(SData:Pointer; SSize:Longint):Boolean;
 var AValue:Byte; APos:Longint;
begin
 {** Check data pointer and size **}
 jssDecodeDelta8:= FALSE;
 if (SData = NIL) or
    (SSize = 0) then
    exit;

 {** Decode the data **}
 AValue:= 0;
 for APos:= 1 to SSize do
 begin
  AValue:= (PByte(SData)^ + AValue);
  PByte(SData)^:= (AValue XOR $80);
  inc(longint(SData));
 end;

 jssDecodeDelta8:= TRUE;
end;


{ Decodes a delta coded 16-bit waveform sample to a normal unsigned sample.
  Returns TRUE if succesfull and FALSE if conversion failed.
}
function jssDecodeDelta16(SData:Pointer; SSize:Longint):Boolean;
 var AValue:Word; APos:Longint;
begin
 {** Check data pointer and size **}
 jssDecodeDelta16:= FALSE;
 if (SData = NIL) or
    (SSize = 0) then
    exit;

 {** Decode the data **}
 AValue:= 0;
 for APos:= 1 to SSize do
 begin
  AValue:= (PWord(SData)^ + AValue);
  PWord(SData)^:= (AValue XOR $8000);
  inc(longint(SData), 2);
 end;

 jssDecodeDelta16:= TRUE;
end;


{ Converts a given ASCIIZ string to a Pascal "shortstring".
}
procedure jssASCIIZToStr(ASource:Pointer; var ADest:String; SSize:Longint);
 var APos:Longint; AData:Char;
begin
 APos:= 0;
 ADest:= '';
 if (ASource = NIL) then exit;

 AData:= PChar(ASource)^;
 while (AData <> #0) and
       (APos < SSize) do
 begin
  ADest:= ADest + AData;
  inc(longint(ASource));
  inc(APos);
  AData:= PChar(ASource)^;
 end;
end;


{****************************************************************************}
{*                                                                          *}
{* General module routines                                                  *}
{*                                                                          *}
{****************************************************************************}

{ Allocates a new module structure or returns errorvalue if failed.
  Memory is allocated only for the basic structure. Sample- and pattern
  areas must be allocated separately with appropriate routines.
}
function jssAllocateModule(var AModule:PJSSModule):Longint;
 var ATempI:Longint;
begin
 {** Check the pointer **}
 jssAllocateModule:= jerrOK;
 if (AModule <> NIL) then
    begin
     jssErrorN:= jerrNonNULLPointer;
     exit(jssErrorN);
    end;

 {** Check if we have enough memory **}
 New(AModule);
 if (AModule = NIL) then
    begin
     jssErrorN:= jerrMemAllocError;
     exit(jssErrorN);
    end;
 
 with AModule^ do
 begin
  {** Initialize the data **}
  ModuleType   := jmdtUnknown;
  ModuleName   := '';
  DefSpeed     := 0;
  DefTempo     := 0;
  DefGlobalVol := 0;
  DefMasterVol := 0;
  DefFlags     := 0;
  DefRestartPos:= 0;
  IntFlags     := 0;
  IntVersion   := 0;
  
  NInstruments := 0;
  NExtInstruments:= 0;
  NPatterns    := 0;
  NOrders      := 0;

  {** Init default pannings **}
  for ATempI:= jsetMinChannel to jsetMaxChannel do
      DefPanning[ATempI]:= jchPanMiddle;

  {** Init orders list **}
  for ATempI:= 0 to jsetMaxOrders do
      OrderList[ATempI]:= mpOrderEnd;

  {** Initialize patterns **}
  FillChar(Patterns, sizeof(Patterns), 0);

  {** Initialize instruments **}
  FillChar(Instruments, sizeof(Instruments), 0);
  FillChar(ExtInstruments, sizeof(ExtInstruments), 0);
 end;
end;


{ Disposes a given module structure, freeing all memory areas
  that were allocated for it (including patterns, samples, etc.)
}
function jssDisposeModule(var AModule:PJSSModule):Longint;
 var CInt, CTemp:Longint;
begin
 {** Check structure pointer **}
 jssDisposeModule:= jerrOK;
 
 if (AModule = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     exit(jssErrorN);
    end;
    
 {$ifdef OPT_MODDEBUG}
 writeln('jssDisposeModule(0x', ioPtr2Hex(AModule), ', "', AModule^.ModuleName, '")');
 {$endif}

 {** De-allocate all instruments **}
 with AModule^ do
 begin
  {$ifdef OPT_MODDEBUG}
  writeln('* De-allocating instrument data...');
  {$endif}
  for CInt:= 1 to NInstruments do
  if (Instruments[CInt] <> NIL) then
     with Instruments[Cint]^ do
     begin
      {** De-allocate sampledata **}
      {$ifdef OPT_MODDEBUG}
      writeln(ioStrW0(Cint, 3), ': ', Used, ' - "', Description, '", ', ioPtr2Hex(Data),
              ' [', Size, ' (', ByteSize, ') : ', ioHexD(LStart), '-', ioHexD(LEnd), '] vol:',
              ioHexB(Volume), ', flags:', ioHexB(Flags), ', c4f:', C4BaseSpeed);
      {$endif}
      if (Data <> NIL) then
         FreeMem(Data, ByteSize);

      {** De-allocate instrument structure **}
      Dispose(Instruments[CInt]);
     end;

  {$ifdef OPT_MODDEBUG}
  writeln('* De-allocating extended instrument data...');
  {$endif}
  for CInt:= 1 to AModule^.NExtInstruments do
  if (ExtInstruments[CInt] <> NIL) then
     with ExtInstruments[Cint]^ do
     begin
      {** De-allocate sampledata **}
      {$ifdef OPT_MODDEBUG}
      write(ioStrW0(Cint, 3), ': ', Used, ' - "', Description, '" - ', NSamples, ': [');
      for CTemp:= 1 to NSamples do
          write('#', SNumForNotes[CTemp], ', ');
      writeln(']');
      {$endif}

      {** De-allocate instrument structure **}
      Dispose(ExtInstruments[CInt]);
     end;

  {** De-allocate pattern data **}
  {$ifdef OPT_MODDEBUG}
  writeln('* De-allocating pattern data...');
  {$endif}
  for CInt:= 0 to jsetMaxPatterns do
  with Patterns[CInt] do
  begin
   {** De-allocate pattern data **}
   if (AUsed) and
      (AData <> NIL) then
      FreeMem(AData, (jsetPatRowSize * ARows));

   AData   := NIL;
   AUsed   := FALSE;
   ARows   := 0;
  end;
 
 end; { With AModule^ do }

 {$ifdef OPT_MODDEBUG}
 writeln('* De-allocating module structure.');
 {$endif}
 
 dispose(AModule);
 AModule:= NIL;

 {$ifdef OPT_MODDEBUG}
 writeln('* jssDisposeModule(): OK.');
 {$endif}
 {** Everything OK! **}
end;


{ Sets module flags. Used by loaders to easily set flags.
}
procedure jssSetFlags(AModule:PJSSModule; AFlag:Longint);
begin
 AModule^.DefFlags:= (AModule^.DefFlags OR AFlag);
end;


{ Gets module flags. Used by loaders to easily get flags.
}
function jssGetFlags(AModule:PJSSModule; AFlag:Longint):Boolean;
begin
 jssGetFlags:= ((AModule^.DefFlags and AFlag) = AFlag);
end;


{ Allocates and initializes a internal pattern structure.
}
function jssAllocatePattern(var APattern:TJSSPattern; ARows:Longint):Longint;
 var ATempSize:Longint;
begin
 {** Check the pointer **}
 jssAllocatePattern:= jerrOK;
 if (APattern.AUsed) or
    (APattern.AData <> NIL) then
    begin
     jssErrorN:= jerrNonNULLPointer;
     exit(jssErrorN);
    end;

 {** Check the number of rows **}
 if (ARows = 0) then
    begin
     jssErrorN:= jerrInternalError;
     jssError('jssAllocatePattern', 'Invalid number of rows for pattern (0)');
     exit(jssErrorN);
    end;

 {** Check if we have enough memory **}
 ATempSize:= (jsetPatRowSize * ARows);
 APattern.AUsed:= TRUE;
 APattern.ARows:= ARows;
 GetMem(APattern.AData, ATempSize);
 if (APattern.AData = NIL) then
    begin
     jssErrorN:= jerrMemAllocError;
     exit(jssErrorN);
    end;

 {** Initialize data **}
 { Notice! The $FF makes all the data -1 (aka mpNoData),
   BUT ONLY IF THE TYPE OF VARIABLE IS RIGHT (Signed Integer).
 }
 FillChar(APattern.AData^, ATempSize, $FF);
end;


{ Allocates and initializes internal "normal" instrument structure.
}
function jssAllocateInstrument(var AInstr:PJSSInstrument):Longint;
begin
 {** Check the pointer **}
 jssAllocateInstrument:= jerrOK;
 if (AInstr <> NIL) then
    begin
     jssErrorN:= jerrNonNULLPointer;
     exit(jssErrorN);
    end;

 {** Check if we have enough memory **}
 new(AInstr);
 if (AInstr = NIL) then
    begin
     jssErrorN:= jerrMemAllocError;
     exit(jssErrorN);
    end;

 {** Initialize data **}
 with AInstr^ do
 begin
  Used        := FALSE;
  Filename    := '';
  Size        := 0;
  LStart      := 0;
  LEnd        := 0;
  Volume      := 0;
  Flags       := 0;
  C4BaseSpeed := 0;
  Description := '';
  Data        := NIL;
  ByteSize    := 0;
  ERelNote    := 0;
  EFineTune   := 0;
  EPanning    := 0;
 end;
end;


{ Allocates and initializes "extended" instrument structure.
}
function jssAllocateExtInstrument(var AInstr:PJSSExtInstrument):Longint;
begin
 {** Check the pointer **}
 jssAllocateExtInstrument:= jerrOK;
 if (AInstr <> NIL) then
    begin
     jssErrorN:= jerrNonNULLPointer;
     exit(jssErrorN);
    end;

 {** Check if we have enough memory **}
 new(AInstr);
 if (AInstr = NIL) then
    begin
     jssErrorN:= jerrMemAllocError;
     exit(jssErrorN);
    end;

 {** Initialize data **}
 with AInstr^ do
 begin
   Used          := FALSE;
   Description   := '';
   NSamples      := 0;
   FillChar(SNumForNotes, SizeOf(SNumForNotes), 0);
   FillChar(AVolumeEnvelope, SizeOf(TJSSEnvelope), 0);
   FillChar(APanningEnvelope, SizeOf(TJSSEnvelope), 0);
   AVibratoType  := 0;
   AVibratoSweep := 0;
   AVibratoDepth := 0;
   AVibratoRate  := 0;
   AFadeOut      := 0;
 end;
end;


{****************************************************************************}
{*                                                                          *}
{* Rest of the sh*t                                                         *}
{*                                                                          *}
{****************************************************************************}
{$ifdef OPT_CLOSE_STREAMS}
 {$define SUP_MAC_CLOSE_STREAM:= dispose(AStream, vmClose);}
 {$define SUP_MAC_FAIL_STREAM:= if (ioFailOnError) then dispose(AStream, vmClose);}
{$else}
 {$define SUP_MAC_CLOSE_STREAM:= }
 {$define SUP_MAC_FAIL_STREAM:= }
{$endif}

{$define GET_IMPLEMENTATION}
{$I jssmod.inc}
{$undef GET_IMPLEMENTATION}

{ 
 !!NOTICE!! Constant jssMaxModLoaders is defined in "jssmod.inc" 
}

 type
     {** Loader info structure **}
     TJSSModuleLoaderInfo = record
      AType:Longint;           { Type of module (jmdtXXX) loader }      
      AName,                   { String representation of the module type name }
      AInfo:String;            { Info-string provided by loader }
      ALoader:TJSSModuleLoader;{ Function for loading a module }
      ADetector:TJSSModuleDetector; { For autodetecting module }
     end;

 var
    {** Variables for module loader system **}
    jssModNLoaders:Longint;
    jssModLoaders:Array[1..jssMaxModLoaders] of TJSSModuleLoaderInfo;


{ Registers a loader with given information and returns the 'reference number'
  for which was given to the registered loader.
}
function jssRegisterLoader(VType:Longint; VLoader:TJSSModuleLoader; VDetector:TJSSModuleDetector; VName, VInfo:String):Longint;
begin
 {** Try to add the loader **}
 inc(jssModNLoaders);
 if (jssModNLoaders > jssMaxModLoaders) then
    begin
     jssErrorN:= jerrFileOpenError;
     jssError('jssRegisterLoader', 'Module loader registration structure overflow. Too many loaders registered!');
     exit;
    end;

 {** If ok, put the information in. **}
 with jssModLoaders[jssModNLoaders] do
 begin
  AType:= VType;
  AName:= VName;
  AInfo:= VInfo;
  ALoader:= VLoader;
  ADetector:= VDetector;
 end;

 jssRegisterLoader:= jssModNLoaders;
end;


{ Returns the number of loaders compiled into JSS.
  This value can be used with jssGetLoaderInfo().
}
function jssGetLoaders:Longint;
begin
 jssGetLoaders:= jssModNLoaders;
end;


{ Gets loader info string for a given loader reference.
}
function jssGetLoaderInfo(ALoader:Longint):String;
begin
 if (ALoader <= jssModNLoaders) or
    (ALoader >= 1) then
    jssGetLoaderInfo:= jssModLoaders[ALoader].AInfo else
    jssGetLoaderInfo:= 'Invalid loader handle!';
end;


{ Gets loader's module type name string or returns "Unknown"
  if invalid loader (== unknown type)
}
function jssGetLoaderName(ALoader:Longint):String;
begin
 if (ALoader <= jssModNLoaders) AND
    (ALoader >= 1) then
    jssGetLoaderName:= jssModLoaders[ALoader].AName else
    jssGetLoaderName:= 'Unknown';
end;


{ Gets loader's module type constant
}
function jssGetLoaderType(ALoader:Longint):Longint;
begin
 if (ALoader <= jssModNLoaders) AND
    (ALoader >= 1) then
    jssGetLoaderType:= jssModLoaders[ALoader].AType else
    jssGetLoaderType:= jmdtUnknown;
end;


{ Tries to autodetect the module type from given stream.
  Assumes that the stream has been seeked to the beginning
  of the "virtual module file data" or something ;)

  Returns reference to data structure
}
function jssDetectModule(AStream:PStream; var AExtra:Pointer):Longint;
 const jProcName = 'jssDetectModule';
 var 
    ASavedPos:TFileInt;
    ATempP:Pointer;
    AFound:Boolean;
    AResult, AIndex:Longint;
begin
 ASavedPos:= AStream^.vmGetPos;

 AIndex:= 1;
 AFound:= FALSE;
 while (AIndex <= jssModNLoaders) AND
       (AFound = FALSE) do
       begin
        ATempP:= AExtra;
	AResult:= jssModLoaders[AIndex].ADetector(AStream, ATempP);
        AStream^.vmSeek(ASavedPos);

        if (AResult > 0)
        then AFound:= TRUE
        else inc(AIndex);
       end;

 if (AFound) then
    begin
     jssDetectModule:= AIndex;
     AExtra:= ATempP;
    end else
     jssDetectModule:= -1;
end;


{ Tries to load a module from a given stream. A loader reference, data
  stream and other relevant information must be given as arguments.
}
procedure jssLoadModule(ALoader:Longint; AStream:PStream; var AModule:PJSSModule; AExtra:Pointer);
 const jProcName = 'jssLoadModule';
begin
 if (ALoader < 1) or (ALoader > jssModNLoaders) then
    begin
     AModule:= NIL;
     exit;
    end;

 with jssModLoaders[ALoader] do
 ALoader(AStream, AModule, AExtra);
end;


{ Gets a loader reference to a given known module type.
}
function jssGetLoaderFromType(AType:Longint):Longint;
 var AFound:Boolean; AIndex:Longint;
begin
 AIndex:= 1;
 AFound:= FALSE;
 while (AIndex <= jssModNLoaders) AND (AFound = FALSE) do
       if (jssModLoaders[AIndex].AType = AType)
       then AFound:= TRUE
       else inc(AIndex);

 if (AFound)
    then jssGetLoaderFromType:= AIndex
    else jssGetLoaderFromType:= -1;
end;


begin
 {** Initialize info structures **}
 jssModNLoaders:= 0;
 FillChar(jssModLoaders, SizeOf(jssModLoaders), 0);

 {** Include loader registration calls **}
{$define GET_REGISTRATION}
{$I jssmod.inc}
{$undef GET_REGISTRATION}
end.
