{
  J Sound System - MOD Player
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This player is used for playing MOD modules produced by
  Ultimate SoundTracker, ProTracker, SoundTracker, NewTracker,
  NoiseTracker, TakeTracker, Fast Tracker I and all
  similar compatible tracker programs.
}

 const
      mpPTMOD_PitchLimit_Low    = 3424;
      mpPTMOD_PitchLimit_High   = 453;

      mpFT2MOD_PitchLimit_Low   = 56;
      mpFT2MOD_PitchLimit_High  = 27392;


{ Change the pitch of given channel by ADelta.
  Check against MOD format pitch limits.
}
procedure TJP_MODPlayer.mpChangePitch(AChannel, ADelta:Longint);
 var AValue:Longint;
begin
 {** Change the pitch and check it **}
 AValue:= (mpCPitch[AChannel] + ADelta);

{$ifdef GBB}
 if (jssGetFlags(mpModule, jmdfFT2Replay)) then
 begin
  {** Check for FastTracker2-style limits **}
  if (AValue < mpFT2MOD_PitchLimit_Low) then
     begin
      AValue:= 0;
      mpSetNDFlags(AChannel, cdfStop);
     end else   
  if (AValue > mpFT2MOD_PitchLimit_High) then
     begin
      AValue:= 0;
      mpSetNDFlags(AChannel, cdfStop);
     end;
 end else
 begin
  {** Check for ProTracker-style limites, octaves 1-3 **} 
  if (AValue < mpPTMOD_PitchLimit_Low) then
     AValue:= mpPTMOD_PitchLimit_Low else
  if (AValue > mpPTMOD_PitchLimit_High) then
     AValue:= mpPTMOD_PitchLimit_High;
 end;
{$endif}

 {** Set the new pitch **}
 mpCPitch[AChannel]:= AValue;
 mpSetNDFlags(AChannel, cdfNewPitch);
end;



{$ifdef FINALIZED_YES}
 const
      mpFunkTable:Array[0..10] of Longint = 
      (0, 5, 6, 7, 8, 10, 11, 13, 16, 19, 22, 26, 32, 43, 64, 128);
      
procedure TJP_MODPlayer.mpDoInvertLoop(AChannel:Longint);
 var
    AInst:PJSSInstrument;
    AParam, d0, d1:Longint;
begin
 {** Check the parameters and instrument **}
 AInst:= mpGetInstrument(AChannel);
 AParam:= mpInvertLoopParam[AChannel];
 if (AParam = 0) or (AInst = NIL) or (AInst^.Used = FALSE) then exit;

 {** Do the actual effect **}
 { NOTICE: Implementation is straightforward 'translation' from
   ProTracker 2.1A's playercode, which is written in 68k assembler.
   There might be some things like integer size limits that I haven't
   taken into account here, so bugs are possible. :I
 }

 d0:= (mpInvertLoopOffset[AChannel] + mpFunkTable[AParam]);
 if ((d0 AND 7) <> 0) then
 begin
  mpInvertLoopOffset[AChannel]:= 0;
  d0:= AInst^.LStart;
  d1:= AInst^.LEnd - AInst^.LStart;
  d0:= d0 + d1 + d1;
  if (d0 > 1) then
     begin
     end;
 end else
 begin
  mpInvertLoopOffset[AChannel]:= d0;
 end;
 
end;

{$endif} { FINALIZED_YES }


procedure TJP_MODPlayer.mpProcessNewRow(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;

    ATrigger,
    ANewPitch:Boolean;
    AOldPitch,
    ATempL:Longint;
    ATempInst:PJSSInstrument;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 { - WARNING - WARNING - WARNING -

   The logic of this player is very complicated (and getting more so) due to
   differences in original trackers and their versions... Supporting everything
   is impossible, but we try to do what is reasonable.

 }

 {** Update note, sample, volumes, etc. **}
 ANewPitch:= FALSE;
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];
 AEffect:= mpMODEffectTable[ACurrNote.Effect];

 AOldPitch:= 1;

 {** Check for new note **}
 if (ACurrNote.Note <> mpNoData) and
    (ACurrNote.Note <> mpNoteOff) then
    begin
     ANewPitch:= TRUE;
     mpCNote[AChannel]:= ACurrNote.Note;
    end;

 {** Check for new instrument **}
 ATrigger:= FALSE;
 if (ACurrNote.Instr <> mpNoData) and
    (ACurrNote.Instr >= 1) and
    (ACurrNote.Instr <= mpModule^.NInstruments) then
    begin
     ATrigger:= TRUE;
     mpCInstrument[AChannel]:= ACurrNote.Instr;
    end else
 if (mpModule^.IntFlags = jmdiPTModule) AND
    (mpModule^.IntVersion < $0203) then
    begin
     ATrigger:= TRUE;
     ACurrNote.Instr:= mpCInstrument[AChannel];
    end;

 if (ATrigger) then
    begin
     {** Set new instrument **}
     ATempInst:= mpModule^.Instruments[ACurrNote.Instr];
     if (ATempInst <> NIL) and
        (ATempInst^.Used) then
        begin
         {** New instrument, set it **}
         ANewPitch:= TRUE;
         mpCVolume[AChannel]:= ATempInst^.Volume;
         mpCInstrument[AChannel]:= ACurrNote.Instr;

         { Original Ultimate SoundTracker had looped samples
           starting from the loop-start offset, unlike
           the later trackers. 
         }
         if ((mpModule^.IntFlags = jmdiUSTModule) AND
            ((ATempInst^.Flags AND jsfLooped) = jsfLooped))
            then mpCPosition[AChannel]:= ATempInst^.LStart
            else mpCPosition[AChannel]:= 0;

         mpSetNDFlags(AChannel, (cdfNewInstr OR cdfNewPos OR cdfNewVolume));
        end;
    end;

 {** Set a new pitch **}
 if (ANewPitch) then
    begin
     {** Save the old values for Tone Portamento effect **}
     AOldPitch:= mpCPitch[AChannel];

     ATempInst:= mpGetInstrument(AChannel);

     if (ATempInst <> NIL) and
        (ATempInst^.Used) then
        begin
         { Compute the new pitch.
           -- Original Ultimate SoundTracker did NOT
           have preset finetunes for instruments.
         }
         if (mpModule^.IntFlags <> jmdiUSTModule) then
         begin
          if (ATempInst^.C4BaseSpeed > 0)
          then mpCPitch[AChannel]:= round((8363 * single(mpPeriodTable[mpCNote[AChannel]])) / ATempInst^.C4BaseSpeed)
          else mpCPitch[AChannel]:= 0;
         end else
          mpCPitch[AChannel]:= mpPeriodTable[mpCNote[AChannel]];

         mpSetNDFlags(AChannel, cdfNewPitch);
        end;
    end;



 {** Now process the effects **}
 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 case AEffect of
  '3': { 3xy = Porta To Note }
       begin
        if (AParam <> 0) then
           mpLastPortaToNoteParam[AChannel]:= AParam;

        if (ACurrNote.Note <> mpNoData) then
           begin
            mpLastPortaToNotePitch[AChannel]:= mpCPitch[AChannel];
            mpCPitch[AChannel]:= AOldPitch;
            mpUnSetNDFlags(AChannel, (cdfNewPitch OR cdfNewInstr OR cdfNewPos));
           end;
{        mpUnSetNDFlags(AChannel,  (cdfNewInstr OR cdfNewPos));
}
       end;

  '4': { 4xy = Vibrato }
       begin
        if (AParamX <> 0) then
           mpVibratoSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpVibratoDepth[AChannel]:= AParamY;

        if ((mpVibratoWC[AChannel] and $4) = 0)
           then mpVibratoPos[AChannel]:= 0;
       end;

  '7': { 7xy = Tremolo }
       begin
        if (AParamX <> 0) then
           mpTremoloSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpTremoloDepth[AChannel]:= AParamY;

        if ((mpTremoloWC[AChannel] and $4) = 0)
           then mpTremoloPos[AChannel]:= 0;
       end;

  '9': { 9xx = Set Sample Offset }
       begin
        if (AParam <> 0)
           then mpLastSampleOffset[AChannel]:= AParam
           else patMakeParams(mpLastSampleOffset[AChannel]);

        ATempInst:= mpModule^.Instruments[mpCInstrument[AChannel]];
        if (ATempInst <> NIL) then
           begin
            ATempL:= (longint(AParam) shl 8);
            if (ATempL > ATempInst^.Size) then
               ATempL:= ATempInst^.Size;

            mpCPosition[AChannel]:= ATempL;
            mpSetNDFlags(AChannel, cdfNewPos);
           end;
       end;

  'B': { Bxx = Pattern Jump }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        mpNewOrder:= AParam;
        mpNewOrderGot:= TRUE;
        mpJumpFlag:= TRUE;
       end;

  'C': { Cxx = Set Volume }
       mpSetVolume(AChannel, AParam);

  'D': { Dxx = Pattern Break }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        {** Compute the new row **}
        mpNewRow:= ((AParamX * 10) + AParamY);
        if (mpNewRow >= mpModule^.Patterns[mpPattern].ARows)
           then mpNewRow:= 0;

        mpNewRowGot:= TRUE;

        {** Now we do some tricky tests **}
        if (mpBreakFlag = FALSE) and
           (mpJumpFlag = FALSE) then
           begin
            mpNewOrder:= (mpOrder + 1);
            mpNewOrderGot:= TRUE;
           end;

        mpBreakFlag:= TRUE;
       end;

  'E': { Exy = Special Effects }
       case AParamX of
       { $00:   E0x - Set filter (NOT SUPPORTED) }
         $01: { E1x - Fine Portamento Up }
              mpChangePitch(AChannel, -(AParamY shl 2));

         $02: { E2x - Fine Portamento Down }
              mpChangePitch(AChannel, (AParamY shl 2));

         $03: { E3x - Glissando Control (NOT SUPPORTED) }
              jssWarning('TJP_MODPlayer', 'Glissando Control effect used!');

         $04: { E4x - Set Vibrato waveform }
              mpVibratoWC[AChannel]:= (AParamY and $f);

         $05: { E5x - Set Finetune } ;

         $06: { E6x - Set Pattern Loop }
              mpDoPatternLoop(AChannel, AParamY);

         $07: { E7x - Set Tremolo waveform }
              mpTremoloWC[AChannel]:= (AParamY and $f);

         $08: { E8x - Set Pan Position }
              begin
               {** Scale the panpos to internal values **}
               mpCPanning[AChannel]:= (AParamY shl 4);
               mpSetNDFlags(AChannel, cdfNewPanPos);
              end;

         $0A: { EAx - Fine Volumeslide Up }
              mpChangeVolume(AChannel, AParamY);

         $0B: { EBx - Fine Volumeslide Down }
              mpChangeVolume(AChannel, -AParamY);

       { $0C:   ECx - Set Note Cut (NOT PROCESSED IN TICK0) }

         $0D: { EDx - Set Note Delay }
              begin
               { Save the ND-flags, then clear them }
               mpSaveNDFlags[AChannel]:= mpCNewDataFlags[AChannel];
               mpCNewDataFlags[AChannel]:= 0;
              end;

         $0E: { EEx - Set Pattern Delay }
              mpPatternDelay:= AParamY;

{$ifdef FINALIZED_YES}
         $0F: { EFx - Invert Loop }
              begin
               mpInvLoopParam[AChannel]:= AParamY;
               jssWarning('TJP_MODPlayer', 'Invert Loop effect used!');
              end;
{$endif}
       end;

  'F': { Fxy = Set Speed / Tempo }
       if (AParam = 0) then mpStop { Stop with speed 0 }
          else
       if (AParam < $20)
          then mpSpeed:= AParam
          else mpSetNewTempo(AParam);
 end;
end;


procedure TJP_MODPlayer.mpProcessEffects(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;
    ATempInst:PJSSInstrument;
    ATempL:Longint;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 {** Translate the effect and params **}
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];

 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 AEffect:= mpMODEffectTable[ACurrNote.Effect];

 {** Process the effects now **}
 case AEffect of
  '0': { 0xy = Arpeggio }
       begin
        {** Compute arpeggio case **}
        ATempInst:= mpModule^.Instruments[mpCInstrument[AChannel]];

        if (ATempInst <> NIL) and
           (ATempInst^.Used) and
           (ATempInst^.C4BaseSpeed > 0) then
           begin
            {** Add the arpeggio value **}
            ATempL:= mpCNote[AChannel];
            case (mpTick MOD 3) of
             1: inc(ATempL, AParamX);
             2: inc(ATempL, AParamY);
            end;

            mpCPitch[AChannel]:= round((8363 * single(mpPeriodTable[ATempL])) / ATempInst^.C4BaseSpeed);

            mpSetNDFlags(AChannel, cdfNewPitch);
           end;
       end;

  '1': { 1xy = Portamento Up }
       mpChangePitch(AChannel, -(AParam shl 2));

  '2': { 2xy = Portamento Down }
       mpChangePitch(AChannel, (AParam shl 2));

  '3': { 3xy = Porta To Note }
       mpDoPortamento(AChannel);

  '4': { 4xy = Vibrato }
       mpDoVibrato(AChannel, 5);

  '5': { 5xy = Portamento + Volume Slide }
       begin
        {** Do the portamento **}
	mpDoPortamento(AChannel);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, AParam);
       end;


  '6': { 6xy = Vibrato + Volume Slide }
       begin
        {** Do the vibrato **}
        mpDoVibrato(AChannel, 5);

        {** Do the volume slide **}
        { MOD does not support remembering of parameter in this effect! }
        mpDoVolumeSlide(AChannel, AParam);
       end;

  '7': { 7xy = Tremolo }
       mpDoTremolo(AChannel);


  'A': { Axy = Volume slide }
       mpDoVolumeSlide(AChannel, AParam);

  'E': { Exy = Special Effects }
       case AParamX of
         $0C: { ECx - Set Note Cut }
              if (mpTick = AParamY) then
                 mpSetVolume(AChannel, jsetMinVol);

         $0D: { EDx - Set Note Delay }
              if (mpTick = AParamY) then
                 mpCNewDataFlags[AChannel]:= mpSaveNDFlags[AChannel];

{$ifdef FINALIZED_YES}
         $0F: { EFx - Invert Loop }
              mpDoInvertLoop(AChannel);
{$endif}
       end;
 end;
end;

