{
  J Sound System - PSM Player
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This player is used for playing Pro Tracker Studio (16) modules.

  NOTICES:
  Unfortunately, only little information is available about this format.
  Additionally the major problem is that the tracker is not publicly
  available (as far as I know), so no tests for compability have
  been done.

  - Currently the effects have been implemented in S3M style.

  - PSM implements a quite unusual offset-setting command:
    it has a 24-bit parameter (3 bytes) which can be used
    to set position 100% accurately, unlike in other formats.
    
    This is NOT supported, however. (Due to limitations in
    the internal JSS structures)
      
}
procedure TJP_PSMPlayer.mpProcessNewRow(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Byte;
    ANewPitch:Boolean;
    AOldPitch,
    ATempL:Longint;
    ATempInst:PJSSInstrument;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 {** Update note, sample, volumes, etc. **}
 ANewPitch:= FALSE;
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];
 AEffect:= ACurrNote.Effect;

 {$I jplay-i1.pas}

 {** Set New Volume **}
 if (ACurrNote.Volume <> mpNoData) then
    mpSetVolume(AChannel, ACurrNote.Volume);

 {** Now process the effects **}
 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 case AEffect of
  1,2,
  3,4: { Dxy = [Fine] Volume Slide Up/Down }
       begin
        {** Handle the normal volume slides **}
        if (AParam <> 0)
           then mpLastVolSlideParam[AChannel]:= AParam
           else patMakeParams(mpLastVolSlideParam[AChannel]);

        {** Handle the "fine"-type effects **}
        if (AEffect = 1) then
           mpChangeVolume(AChannel, AParam) else
        if (AEffect = 3) then
           mpChangeVolume(AChannel, -AParam);
       end;

  10,11,
  12,13: { Exx = [Fine] Portamento Up/Down }
       begin
        {** Make the effect parameters **}
        if (AParam <> 0)
           then mpLastPortaParam[AChannel]:= AParam
           else patMakeParams(mpLastPortaParam[AChannel]);

        {** Handle the "fine"-type effects **}
        if (AEffect = 10) then
           mpChangePitch(AChannel, -(AParamY shl 2)) else
        if (AEffect = 12) then
           mpChangePitch(AChannel, (AParam shl 2));
       end;


  14:  { Gxx = Tone Portamento }
       begin
        if (AParam <> 0) then
           mpLastPortaToNoteParam[AChannel]:= AParam;

        if (ACurrNote.Note <> mpNoData) then
           begin
            mpLastPortaToNotePitch[AChannel]:= mpCPitch[AChannel];
            mpCPitch[AChannel]:= AOldPitch;
            mpUnSetNDFlags(AChannel, cdfNewPitch);
           end;

	mpUnSetNDFlags(AChannel, (cdfNewInstr OR cdfNewPos));
       end;

  'H', { Hxy = Vibrato }
  'U': { Uxy = Fine Vibrato }
       begin
        if (AParam <> 0) then
           begin
            mpVibratoSpeed[AChannel]:= AParamX;
            mpVibratoDepth[AChannel]:= AParamY;
           end;

        if ((mpVibratoWC[AChannel] and $4) = 0)
           then mpVibratoPos[AChannel]:= 0;
       end;

  'I': { Ixy = Tremor }
       if (AParam <> 0) then
          mpLastTremorParam[AChannel]:= AParam;

  21,22, { Kxy = Vibrato + Volume slide }
  16,17: { Lxy = Portamento + Volume Slide }
       begin
        if (AParam <> 0) then
           mpLastVolSlideParam[AChannel]:= AParam;
       end;

  40: { Oxx = Set Sample Offset }
       begin
        {
        if (AParam <> 0)
           then mpLastSampleOffset[AChannel]:= AParam
           else patMakeParams(mpLastSampleOffset[AChannel]);
        }
       end;

  30: { Rxy = Tremolo }
       begin
        if (AParam <> 0) then
           begin
            mpTremoloSpeed[AChannel]:= AParamX;
            mpTremoloDepth[AChannel]:= AParamY;
           end;

        if ((mpTremoloWC[AChannel] and $4) = 0)
           then mpTremoloPos[AChannel]:= 0;
       end;


  21: { S3x - Set Vibrato waveform }
      mpVibratoWC[AChannel]:= AParamX;

  31: { S4x - Set Tremolo waveform }
      mpTremoloWC[AChannel]:= AParamX;

  72: { S8x - Set Pan Position }
      begin
       {** Scale the panpos to internal values **}
       mpCPanning[AChannel]:= (AParamY shl 4);
       mpSetNDFlags(AChannel, cdfNewPanPos);
      end;


  43: { SDx - Set Note Delay }
      begin
       { Save the ND-flags, then clear them }
       mpSaveNDFlags[AChannel]:= mpCNewDataFlags[AChannel];
       mpCNewDataFlags[AChannel]:= cdfNone;
      end;


  50: { Bxx = Pattern Jump }
      if ((mpOptions AND moptAllowJumps) <> 0) then
      begin
        mpNewOrder:= AParam;
        mpNewOrderGot:= TRUE;
        mpJumpFlag:= TRUE;
      end;

  51: { Cxx = Pattern Break }
      if ((mpOptions AND moptAllowJumps) <> 0) then
      begin
        {** Compute the new row **}
        mpNewRow:= ((AParamX * 10) + AParamY);
        if (mpNewRow >= mpModule^.Patterns[mpPattern].ARows)
           then mpNewRow:= 0;

        mpNewRowGot:= TRUE;

        {** Now we do some tricky tests **}
        if (mpBreakFlag = FALSE) and
           (mpJumpFlag = FALSE) then
           begin
            mpNewOrder:= (mpOrder + 1);
            mpNewOrderGot:= TRUE;
           end;

        mpBreakFlag:= TRUE;
      end;

  52: { SBx - Set Pattern Loop }
      mpDoPatternLoop(AChannel, AParamY);

  53: { SEx - Set Pattern Delay }
      mpPatternDelay:= AParam;


  60: { Axx = Set Speed }
      mpSpeed:= AParam;

  61: { Txx = Set Tempo }
      begin
       if (AParam < $20) then
          AParam:= $20;

       mpSetNewTempo(AParam);
      end;

 end;

 mpRetrigNDFlags[AChannel]:= mpCNewDataFlags[AChannel];
end;


procedure TJP_PSMPlayer.mpProcessEffects(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Byte;
    ATempInst:PJSSInstrument;
    ATempL:Longint;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 {** Translate the effect and params **}
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];

 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 AEffect:= ACurrNote.Effect;

 {** Process the effects now **}
 case AEffect of
  2:   { Dxy = Volume slide Up }
       mpChangeVolume(AChannel, mpLastVolSlideParam[AChannel]);

  4:   { Dxy = Volume slide Down }
       mpChangeVolume(AChannel, -mpLastVolSlideParam[AChannel]);


  11:  { Fxy = Portamento Up }
       mpChangePitch(AChannel, -(longint(mpLastPortaParam[AChannel]) shl 2));

  13:  { Exy = Portamento Down }
       mpChangePitch(AChannel, (longint(mpLastPortaParam[AChannel]) shl 2));


  14:  { Gxx = Tone Portamento }
       mpDoPortamento(AChannel);

  16:  { Lxy = Portamento + Volume Slide Up }
       begin
        {** Do the portamento **}
        mpDoPortamento(AChannel);

        {** Do the volume slide Up **}
        mpChangeVolume(AChannel, mpLastVolSlideParam[AChannel]);
       end;

  17:  { Lxy = Portamento + Volume Slide Down }
       begin
        {** Do the portamento **}
        mpDoPortamento(AChannel);

        {** Do the volume slide Down **}
        mpChangeVolume(AChannel, -mpLastVolSlideParam[AChannel]);
       end;


  20:  { Hxy = Vibrato }
       mpDoVibrato(AChannel, 5);

  30:  { Rxy = Tremolo }
       mpDoTremolo(AChannel);

  42:  { SCx - Set Note Cut }
       if (mpTick = AParamY) then
          mpSetVolume(AChannel, jsetMinVol);

  43:  { SDx - Set Note Delay }
       if (mpTick = AParamY) then
          mpCNewDataFlags[AChannel]:= mpSaveNDFlags[AChannel];

  70: { Jxy = Arpeggio }
       begin
        {** Compute arpeggio case **}
        ATempInst:= mpModule^.Instruments[mpCInstrument[AChannel]];

        if (ATempInst <> NIL) and
           (ATempInst^.Used) and
           (ATempInst^.C4BaseSpeed > 0) then
           begin
            {** Add the arpeggio value **}
            ATempL:= mpCNote[AChannel];
            case (mpTick MOD 3) of
             1: inc(ATempL, AParamX);
             2: inc(ATempL, AParamY);
            end;

            mpCPitch[AChannel]:= round((8363 * single(mpPeriodTable[ATempL])) / ATempInst^.C4BaseSpeed);

            mpSetNDFlags(AChannel, cdfNewPitch);
           end;
       end;

  'K': { Kxy = Vibrato + Volume Slide }
       begin
        {** Do the vibrato **}
        mpDoVibrato(AChannel, 5);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);
       end;

  'Q': { Qxy = Retrig + Volume Slide }
       begin
        patMakeParams(mpLastRetrigParam[AChannel]);
        if ((mpTick MOD AParamY) = 0) then
        begin
        {** Do the retrig **}
        mpCNewDataFlags[AChannel]:= mpRetrigNDFlags[AChannel];

        {** Do the volume slide **}
        AParam:= mpCVolume[AChannel];
        case AParamX of
         $1: dec(AParam, 1);
         $2: dec(AParam, 2);
         $3: dec(AParam, 4);
         $4: dec(AParam, 8);
         $5: dec(AParam, 16);
         $6: AParam:= ((AParam shl 1) div 3);
         $7: AParam:= (AParam shr 1);

         $9: inc(AParam, 1);
         $a: inc(AParam, 2);
         $b: inc(AParam, 4);
         $c: inc(AParam, 8);
         $d: inc(AParam, 16);
         $e: AParam:= ((AParam * 3) shr 1);
         $f: AParam:= (AParam shl 1);
        end;

        mpSetVolume(AChannel, AParam);
        end;
       end;

 end;
end;


