{
  J Sound System - PTM Player
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This player is used for playing Poly Tracker PTM modules.
}


procedure TJP_PTMPlayer.mpProcessNewRow(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;
    ANewPitch:Boolean;
    AOldPitch,
    ATempL:Longint;
    ATempInst:PJSSInstrument;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 {** Update note, sample, volumes, etc. **}
 ANewPitch:= FALSE;
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];
 AEffect:= mpMODEffectTable[ACurrNote.Effect];

 {$I jplay-i1.pas}

 {** Set New Volume **}
 if (ACurrNote.Volume <> mpNoData) then
    mpSetVolume(AChannel, ACurrNote.Volume);

 {** Now process the effects **}
 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 case AEffect of
  '1': { Exx = Portamento Down }
       begin
        {** Make the effect parameters **}
        if (AParam <> 0)
           then mpLastPortaParam[AChannel]:= AParam
           else patMakeParams(mpLastPortaParam[AChannel]);

        case AParamX of
         $E:  { Extra Fine Slide [1*AParamY slide] }
              mpChangePitch(AChannel, AParamY);

         $F:  { Fine Slide [4*AParamY slide] }
              mpChangePitch(AChannel, (AParamY shl 2));
        end;
       end;

  '2': { Fxx = Portamento Up }
       begin
        {** Make the effect parameters **}
        if (AParam <> 0)
           then mpLastPortaParam[AChannel]:= AParam
           else patMakeParams(mpLastPortaParam[AChannel]);

        case AParamX of
         $E:  { Extra Fine Slide [1*AParamY slide] }
              mpChangePitch(AChannel, -AParamY);

         $F:  { Fine Slide [4*AParamY slide] }
              mpChangePitch(AChannel, -(AParamY shl 2));
        end;
       end;

  '3': { 3xy = Porta To Note }
       begin
        if (AParam <> 0) then
           mpLastPortaToNoteParam[AChannel]:= AParam;

        if (ACurrNote.Note <> mpNoData) then
           begin
            mpLastPortaToNotePitch[AChannel]:= mpCPitch[AChannel];
            mpCPitch[AChannel]:= AOldPitch;
            mpUnSetNDFlags(AChannel, (cdfNewPitch OR cdfNewInstr OR cdfNewPos));
           end;
       end;

  '4': { 4xy = Vibrato }
       begin
        if (AParamX <> 0) then
           mpVibratoSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpVibratoDepth[AChannel]:= AParamY;

        if ((mpVibratoWC[AChannel] and $4) = 0)
           then mpVibratoPos[AChannel]:= 0;
       end;

  '5', { 5xy = Portamento + Volume Slide }
  '6': { 6xy = Vibrato + Volume slide }
       if (AParam <> 0) then
          mpLastVolSlideParam[AChannel]:= AParam;

  '7': { 7xy = Tremolo }
       begin
        if (AParamX <> 0) then
           mpTremoloSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpTremoloDepth[AChannel]:= AParamY;

        if ((mpTremoloWC[AChannel] and $4) = 0)
           then mpTremoloPos[AChannel]:= 0;
       end;

  '9': { 9xx = Set Sample Offset }
       begin
        if (AParam <> 0)
           then mpLastSampleOffset[AChannel]:= AParam
           else patMakeParams(mpLastSampleOffset[AChannel]);

        ATempInst:= mpModule^.Instruments[mpCInstrument[AChannel]];
        if (ATempInst <> NIL) then
           begin
            ATempL:= (longint(AParam) shl 8);
            if (ATempL > ATempInst^.Size) then
               ATempL:= ATempInst^.Size;

            if ((ATempInst^.Flags AND jsfLooped) <> 0) and
               (ATempL > ATempInst^.LEnd) then
               ATempL:= ATempInst^.LEnd;

            mpCPosition[AChannel]:= ATempL;
            mpSetNDFlags(AChannel, cdfNewPos);
           end;
       end;

  'A': { Axy = Volume Slide }
       begin
        if (AParam <> 0)
           then mpLastVolSlideParam[AChannel]:= AParam
           else patMakeParams(mpLastVolSlideParam[AChannel]);

        if (AParamY = $F) then
           mpChangeVolume(AChannel, AParamX) else

        if (AParamX = $F) then
            mpChangeVolume(AChannel, -AParamY);
       end;

  'B': { Bxx = Pattern Jump }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        mpNewOrder:= AParam;
        mpNewOrderGot:= TRUE;
        mpJumpFlag:= TRUE;
       end;

  'C': { Cxx = Set Volume }
       mpSetVolume(AChannel, AParam);

  'D': { Dxx = Pattern Break }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        {** Compute the new row **}
        mpNewRow:= ((AParamX * 10) + AParamY);
        if (mpNewRow >= mpModule^.Patterns[mpPattern].ARows)
           then mpNewRow:= 0;

        mpNewRowGot:= TRUE;

        {** Now we do some tricky tests **}
        if (mpBreakFlag = FALSE) and
           (mpJumpFlag = FALSE) then
           begin
            mpNewOrder:= (mpOrder + 1);
            mpNewOrderGot:= TRUE;
           end;

        mpBreakFlag:= TRUE;
       end;

  'E': { Exy = Special Effects }
       case AParamX of
       { $00:   E0x - Set filter (NOT SUPPORTED) }
         $01: { E1x - Fine Portamento Up }
              mpChangePitch(AChannel, -(AParamY shl 2));

         $02: { E2x - Fine Portamento Down }
              mpChangePitch(AChannel, (AParamY shl 2));

       { $03:   E3x - Glissando Control (NOT SUPPORTED) }

         $04: { E4x - Set Vibrato waveform }
              mpVibratoWC[AChannel]:= (AParamY and $f);

         $05: { E5x - Set Finetune } ;

         $06: { E6x - Set Pattern Loop }
              mpDoPatternLoop(AChannel, AParamY);

         $07: { E7x - Set Tremolo waveform }
              mpTremoloWC[AChannel]:= (AParamY and $f);

         $08: { E8x - Set Pan Position }
              begin
               {** Scale the panpos to internal values **}
               mpCPanning[AChannel]:= (AParamY shl 4);
               mpSetNDFlags(AChannel, cdfNewPanPos);
              end;

         $0A: { EAx - Fine Volumeslide Up }
              mpChangeVolume(AChannel, AParamY);

         $0B: { EBx - Fine Volumeslide Down }
              mpChangeVolume(AChannel, -AParamY);

       { $0C:   ECx - Set Note Cut (NOT PROCESSED IN TICK0) }

         $0D: { EDx - Set Note Delay }
              begin
               { Save the ND-flags, then clear them }
               mpSaveNDFlags[AChannel]:= mpCNewDataFlags[AChannel];
               mpCNewDataFlags[AChannel]:= 0;
              end;

         $0E: { EEx - Set Pattern Delay }
              mpPatternDelay:= AParamY;

       { $0F:   SFx - Invert Loop (NOT SUPPORTED) }
       end;

  'F': { Fxy = Set Speed / Tempo }
       if (AParam < $20)
          then mpSpeed:= AParam
          else mpSetNewTempo(AParam);

 end;
end;


procedure TJP_PTMPlayer.mpProcessEffects(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;
    ATempInst:PJSSInstrument;
    ATempL:Longint;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 {** Translate the effect and params **}
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];

 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 AEffect:= mpMODEffectTable[ACurrNote.Effect];

 {** Process the effects now **}
 case AEffect of
  '0': { 0xy = Arpeggio }
       begin
        {** Compute arpeggio case **}
        ATempInst:= mpModule^.Instruments[mpCInstrument[AChannel]];

        if (ATempInst <> NIL) and
           (ATempInst^.Used) and
           (ATempInst^.C4BaseSpeed > 0) then
           begin
            {** Add the arpeggio value **}
            ATempL:= mpCNote[AChannel];
            case (mpTick MOD 3) of
             1: inc(ATempL, AParamX);
             2: inc(ATempL, AParamY);
            end;

            mpCPitch[AChannel]:= round((8363 * single(mpPeriodTable[ATempL])) / ATempInst^.C4BaseSpeed);

            mpSetNDFlags(AChannel, cdfNewPitch);
           end;
       end;

  '1': { 1xy = Portamento Down }
       begin
        patMakeParams(mpLastPortaParam[AChannel]);
        if (AParam < $E0) then
           mpChangePitch(AChannel, (longint(AParam) shl 2));
       end;

  '2': { 2xy = Portamento Up }
       begin
        patMakeParams(mpLastPortaParam[AChannel]);
        if (AParam < $E0) then
           mpChangePitch(AChannel, -(longint(AParam) shl 2));
       end;

  '3': { 3xy = Porta To Note }
       mpDoPortamento(AChannel);

  '4': { 4xy = Vibrato }
       mpDoVibrato(AChannel, 5);

  '5': { 5xy = Portamento + Volume Slide }
       begin
        {** Do the portamento **}
        mpDoPortamento(AChannel);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);
       end;

  '6': { 6xy = Vibrato + Volume Slide }
       begin
        {** Do the vibrato **}
        mpDoVibrato(AChannel, 5);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);
       end;

  '7': { 7xy = Tremolo }
       mpDoTremolo(AChannel);

  'A': { Axy = Volume slide }
       mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);

  'E': { Exy = Special Effects }
       case AParamX of
         $0C: { ECx - Set Note Cut }
              if (mpTick = AParamY) then
                 mpSetVolume(AChannel, jsetMinVol);

         $0D: { EDx - Set Note Delay }
              if (mpTick = AParamY) then
                 mpCNewDataFlags[AChannel]:= mpSaveNDFlags[AChannel];
       end;
 end;
end;


