{
  J Sound System - S3M Player
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This player is used for playing Scream Tracker III S3M modules.
}
 const
      mpS3M_PitchLimit_Low    = 56;
      mpS3M_PitchLimit_High   = 27392;


{ Change the pitch of given channel by ADelta.
  Check against S3M format pitch limits.
}
procedure TJP_S3MPlayer.mpChangePitch(AChannel, ADelta:Longint);
 var AValue:Longint;
begin
 {** Change the pitch and check it **}
 AValue:= (mpCPitch[AChannel] + ADelta);

 if (AValue < mpS3M_PitchLimit_Low) then
    AValue:= mpS3M_PitchLimit_Low else   

 if (AValue > mpS3M_PitchLimit_High) then
    AValue:= mpS3M_PitchLimit_High;

 {** Set the new pitch **}
 mpCPitch[AChannel]:= AValue;
 mpSetNDFlags(AChannel, cdfNewPitch);
end;


procedure TJP_S3MPlayer.mpChangePitchWithLimits(AChannel, ADelta:Longint);
 var AValue:Longint;
begin
 {** Change the pitch and check it **}
 AValue:= (mpCPitch[AChannel] + ADelta);

 if (AValue < mpS3M_PitchLimit_Low) then
    mpSetNDFlags(AChannel, cdfStop) else

 if (AValue > mpS3M_PitchLimit_High) then
    mpSetNDFlags(AChannel, cdfStop);

 {** Set the new pitch **}
 mpCPitch[AChannel]:= AValue;
 mpSetNDFlags(AChannel, cdfNewPitch);
end;


procedure TJP_S3MPlayer.mpProcessNewRow(AChannel:Longint);

 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;
    ANewPitch:Boolean;
    AOldPitch,
    ATempL:Longint;
    ATempInst:PJSSInstrument;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 {** Update note, sample, volumes, etc. **}
 ANewPitch:= FALSE;
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];
 AEffect:= char(64 + ACurrNote.Effect);

 {$I jplay-i1.pas}


 {** Set New Volume **}
 if (ACurrNote.Volume <> mpNoData) then
    mpSetVolume(AChannel, ACurrNote.Volume);

 {** Now process the effects **}
 if (ACurrNote.Effect = mpNoData) then exit;

 if (ACurrNote.Param = mpNoData) then
    ACurrNote.Param:= 0;

 patMakeParams(ACurrNote.Param);

 case AEffect of
  'A': { Axx = Set Speed }
       if (AParam > 0) then
          mpSpeed:= AParam;

  'B': { Bxx = Pattern Jump }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        mpNewOrder:= AParam;
        mpNewOrderGot:= TRUE;
        mpJumpFlag:= TRUE;
       end;

  'C': { Cxx = Pattern Break }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        {** Compute the new row **}
        mpNewRow:= ((AParamX * 10) + AParamY);
        if (mpNewRow >= mpModule^.Patterns[mpPattern].ARows)
           then mpNewRow:= 0;

        mpNewRowGot:= TRUE;

        {** Now we do some tricky tests **}
        if (mpBreakFlag = FALSE) and
           (mpJumpFlag = FALSE) then
           begin
            mpNewOrder:= (mpOrder + 1);
            mpNewOrderGot:= TRUE;
           end;

        mpBreakFlag:= TRUE;
       end;

  'D': { Dxy = Volume Slide }
       begin
        {** Handle the normal volume slides **}
        if (AParam <> 0)
           then mpLastVolSlideParam[AChannel]:= AParam
           else patMakeParams(mpLastVolSlideParam[AChannel]);

        if (AParamY = $F) then
           mpChangeVolume(AChannel, AParamX) else

        if (AParamX = $F) then
           mpChangeVolume(AChannel, -AParamY);

        {** Handle the buggy ST3.00 "fast volume slides" **}
        if (jssGetFlags(mpModule, jmdfST300Slides)) then
           mpDoVolumeSlide(AChannel, AParam);
       end;

  'E': { Exx = Portamento Down }
       begin
        {** Make the effect parameters **}
        if (AParam <> 0)
           then mpLastPortaParam[AChannel]:= AParam
           else patMakeParams(mpLastPortaParam[AChannel]);

        case AParamX of
         $E:  { Extra Fine Slide [1*AParamY slide] }
              mpChangePitchWithLimits(AChannel, AParamY);

         $F:  { Fine Slide [4*AParamY slide] }
              mpChangePitchWithLimits(AChannel, (AParamY * 4));
        end;
       end;

  'F': { Fxx = Portamento Up }
       begin
        {** Make the effect parameters **}
        if (AParam <> 0)
           then mpLastPortaParam[AChannel]:= AParam
           else patMakeParams(mpLastPortaParam[AChannel]);

        case AParamX of
         $E:  { Extra Fine Slide [1*AParamY slide] }
              mpChangePitchWithLimits(AChannel, -AParamY);

         $F:  { Fine Slide [4*AParamY slide] }
              mpChangePitchWithLimits(AChannel, -(AParamY * 4));
        end;
       end;

  'G': { Gxx = Tone Portamento }
       begin
        if (AParam <> 0) then
           mpLastPortaToNoteParam[AChannel]:= AParam;

        if (ACurrNote.Note <> mpNoData) then
           begin
            mpLastPortaToNotePitch[AChannel]:= mpCPitch[AChannel];
            mpCPitch[AChannel]:= AOldPitch;
            mpUnSetNDFlags(AChannel, cdfNewPitch);
           end;

	mpUnSetNDFlags(AChannel, (cdfNewInstr OR cdfNewPos));
       end;

  'H', { Hxy = Vibrato }
  'U': { Uxy = Fine Vibrato }
       begin
        if (AParamX <> 0) then
           mpVibratoSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpVibratoDepth[AChannel]:= AParamY;

        if ((mpVibratoWC[AChannel] and $4) = 0)
           then mpVibratoPos[AChannel]:= 0;
       end;

  'I': { Ixy = Tremor }
       if (AParam <> 0) then
          mpLastTremorParam[AChannel]:= AParam;

  'K', { Kxy = Vibrato + Volume slide }
  'L': { Lxy = Portamento + Volume Slide }
       begin
        if (AParam <> 0) then
           mpLastVolSlideParam[AChannel]:= AParam;
       end;

  'O': { Oxx = Set Sample Offset }
       begin
        if (AParam <> 0)
           then mpLastSampleOffset[AChannel]:= AParam
           else patMakeParams(mpLastSampleOffset[AChannel]);

        ATempInst:= mpModule^.Instruments[mpCInstrument[AChannel]];
        if (ATempInst <> NIL) then
           begin
            {** Check few things here **}
            ATempL:= (longint(AParam) shl 8);
            if (ATempL > ATempInst^.Size) then
               begin
                { ST3 does not play a note which gets offset > size.
                  Instrument is also ignored. But VOLUME is set!
                }
                mpUnSetNDFlags(AChannel, cdfNewPitch OR cdfNewInstr);
               end else

               begin
                { Everything is OK. Set the new position. }
                mpCPosition[AChannel]:= ATempL;
                mpSetNDFlags(AChannel, cdfNewPos);
               end;
           end;
       end;

  'R': { Rxy = Tremolo }
       begin
        if (AParamX <> 0) then
           mpTremoloSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpTremoloDepth[AChannel]:= AParamY;

        if ((mpTremoloWC[AChannel] and $4) = 0)
           then mpTremoloPos[AChannel]:= 0;
       end;

  'S': { Sxy = Special Commands }
       case AParamX of
       { $00: S0x - Set filter (NOT SUPPORTED) }
       { $01: S1x - Set glissando (NOT SUPPORTED) }
         $02: { S2x - Set Finetune }
              begin

              end;

         $03: { S3x - Set Vibrato waveform }
              mpVibratoWC[AChannel]:= (AParamY and $f);

         $04: { S4x - Set Tremolo waveform }
              mpTremoloWC[AChannel]:= (AParamY and $f);

         $08: { S8x - Set Pan Position }
              begin
               {** Scale the panpos to internal values **}
               mpCPanning[AChannel]:= (AParamY * 16);
               mpSetNDFlags(AChannel, cdfNewPanPos);
              end;

         $0A: { SAx - Stereo Control }
              begin
               {** Convert and scale the panpos to internal values **}
               if (AParamY > 7)
                  then ATempL:= (AParamY - 8)
                  else ATempL:= (AParamY + 8);

               mpCPanning[AChannel]:= (ATempL * 16);
               mpSetNDFlags(AChannel, cdfNewPanPos);
              end;

         $0B: { SBx - Set Pattern Loop }
              mpDoPatternLoop(AChannel, AParamY);

         $0D: { SDx - Set Note Delay }
              begin
               { Save the ND-flags, then clear them }
               mpSaveNDFlags[AChannel]:= mpCNewDataFlags[AChannel];
               mpCNewDataFlags[AChannel]:= cdfNone;
              end;


         $0E: { SEx - Set Pattern Delay }
              mpPatternDelay:= AParamY;

         $0F: { SFx - Funk Repeat (NOT SUPPORTED) } ;
       end;

  'T': { Txx = Set Tempo }
       begin
        if (AParam < $20) then
           AParam:= $20;

        mpSetNewTempo(AParam);
       end;

  'V': { Vxx = Set Global Volume }
       begin
   { FIXME FIX ME: fix this }
        mpDevice^.vmSetGlobalVol(AParam);
       end;

  'J': ;       
  'Q': writeln('!! WARNING: Buggy effect: ', ACurrNote.Effect, '/', AEffect, ioHexB(AParam), ' [row=', mpRow, ', pat=', mpPattern, ', chan=', AChannel, ']');

  else writeln('!! WARNING: Unsupported effect: ', ACurrNote.Effect, '/', AEffect, ioHexB(AParam), ' [row=', mpRow, ', pat=', mpPattern, ', chan=', AChannel, ']');
 end;

 mpRetrigNDFlags[AChannel]:= mpCNewDataFlags[AChannel];
end;


procedure TJP_S3MPlayer.mpProcessEffects(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;
    ATempInst:PJSSInstrument;
    ATempL:Longint;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

begin
 {** Translate the effect and params **}
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];

 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 AEffect:= char(64 + ACurrNote.Effect);

 {** Process the effects now **}
 case AEffect of
  'D': { Dxy = Volume slide }
       mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);

  'E': { Exy = Portamento Down }
       begin
        patMakeParams(mpLastPortaParam[AChannel]);
        if (AParam < $E0) then
           mpChangePitchWithLimits(AChannel, (longint(AParam) * 4));
       end;

  'F': { Fxy = Portamento Up }
       begin
        patMakeParams(mpLastPortaParam[AChannel]);
        if (AParam < $E0) then
           mpChangePitchWithLimits(AChannel, -(longint(AParam) * 4));
       end;

  'G': { Gxx = Tone Portamento }
       mpDoPortamento(AChannel);

  'H': { Hxy = Vibrato }
       mpDoVibrato(AChannel, 5);

  'I': { Ixy = Tremor }
       begin
        patMakeParams(mpLastTremorParam[AChannel]);
        inc(AParamX);
        inc(AParamY);
        AParam:= (mpTremorCount[AChannel] mod (AParamX + AParamY));
        if (AParam < AParamX)
           then mpSetVolume(AChannel, mpSVolume[AChannel])
           else mpSetVolume(AChannel, 0);

        inc(AParam);
        mpTremorCount[AChannel]:= AParam;
       end;

  'J': { Jxy = Arpeggio }
       begin
        {** Compute arpeggio case **}
        ATempInst:= mpModule^.Instruments[mpCInstrument[AChannel]];

        if (ATempInst <> NIL) and
           (ATempInst^.Used) and
           (ATempInst^.C4BaseSpeed > 0) then
           begin
            {** Add the arpeggio value **}
            ATempL:= mpCNote[AChannel];
            case (mpTick MOD 3) of
             1: inc(ATempL, AParamX);
             2: inc(ATempL, AParamY);
            end;

            mpCPitch[AChannel]:= round((8363 * single(mpPeriodTable[ATempL])) / ATempInst^.C4BaseSpeed);

            mpSetNDFlags(AChannel, cdfNewPitch);
           end;
       end;

  'K': { Kxy = Vibrato + Volume Slide }
       begin
        {** Do the vibrato **}
        mpDoVibrato(AChannel, 5);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);
       end;

  'L': { Lxy = Portamento + Volume Slide }
       begin
        {** Do the portamento **}
        mpDoPortamento(AChannel);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);
       end;

{ FIXME FIX ME: Qxy does NOT work right, the retriggering should be done differently! }
  'Q': { Qxy = Retrig + Volume Slide }
       begin
        patMakeParams(mpLastRetrigParam[AChannel]);
        
        if (AParamY > 0) AND
           ((mpTick MOD AParamY) = 0) then
        begin
         {** Do the retrig **}
         mpCNewDataFlags[AChannel]:= mpRetrigNDFlags[AChannel];

         {** Do the volume slide **}
         AParam:= mpCVolume[AChannel];
         case AParamX of
          $1: dec(AParam, 1);
          $2: dec(AParam, 2);
          $3: dec(AParam, 4);
          $4: dec(AParam, 8);
          $5: dec(AParam, 16);
          $6: AParam:= ((AParam * 2) DIV 3);
          $7: AParam:= (AParam DIV 2);

          $9: inc(AParam, 1);
          $a: inc(AParam, 2);
          $b: inc(AParam, 4);
          $c: inc(AParam, 8);
          $d: inc(AParam, 16);
          $e: AParam:= ((AParam * 3) DIV 2);
          $f: AParam:= (AParam DIV 2);
         end;

         mpSetVolume(AChannel, AParam);
        end;
       end;

  'R': { Rxy = Tremolo }
       mpDoTremolo(AChannel);

  'S': { Sxy = Special commands }
       case AParamX of
         $0C: { SCx - Set Note Cut }
              if (mpTick = AParamY) then
                 mpSetVolume(AChannel, jsetMinVol);

         $0D: { SDx - Set Note Delay }
              if (mpTick = AParamY) then
                 begin
                  mpCNewDataFlags[AChannel]:= mpSaveNDFlags[AChannel];
                 end;
       end;

  'U': { Uxy = Fine Vibrato }
       mpDoVibrato(AChannel, 7)

 end;
end;


