{
  J Sound System - XM Player
  (C) Copyright Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This player is used for playing Fast Tracker II XM modules.

  NOTICE: This is still _PARTIALLY_ experimental, so there is
          quite a lot of debug-code included here.

  TODO:
 * Fix / implement the Amiga-styled periods.

 * Fix and/or check out the implementations of vibrato and tremolo.

 * Implement other possibly missing effects.

 * Lxx implementation may be buggy. Needs more investigation.

 * 0xy is bugging, don't know why.
}


 const
      mpXMAmigaPeriodTab:Array[0..(12*8)-1] of Word = (
      907,900,894,887,881,875,868,862,856,850,844,838,832,826,820,814,
      808,802,796,791,785,779,774,768,762,757,752,746,741,736,730,725,
      720,715,709,704,699,694,689,684,678,675,670,665,660,655,651,646,
      640,636,632,628,623,619,614,610,604,601,597,592,588,584,580,575,
      570,567,563,559,555,551,547,543,538,535,532,528,524,520,516,513,
      508,505,502,498,494,491,487,484,480,477,474,470,467,463,460,457);


procedure TJP_XMPlayer.mpInitPlay;
 var AIndex:Longint;
begin
 {** Call inherited mpInitPlay **}
 inherited mpInitPlay;

 {** Initialize XM-specific stuff **}
 for AIndex:= jsetMinChannel to jsetMaxChannel do
     begin
      mpLastFineVolumeslideUpParam[AIndex]       := 0;
      mpLastFineVolumeslideDownParam[AIndex]     := 0;

      mpLastExtraFinePortamentoUpParam[AIndex]   := 0;
      mpLastExtraFinePortamentoDownParam[AIndex] := 0;

      mpLastFinePortamentoUpParam[AIndex]        := 0;
      mpLastFinePortamentoDownParam[AIndex]      := 0;
     end;
end;


function mpGetAmigaPeriod(IFinalNote:Longint; IFineTune:Single):Single;
 var ATempI:Longint; ATempFi,  ATempS1, ATempS2:Single;
begin
 IFinalNote:= IFinalNote - 12;
 ATempI:= ((IFinalNote MOD 12) * 8);

 ATempFi:= (IFineTune + 32) / 32;

 ATempS1:= mpXMAmigaPeriodTab[ATempI];
 ATempS2:= mpXMAmigaPeriodTab[ATempI];

{
 write('mpgeta(', IFinalNote, ', ', IFineTune:1:0, '): i=', ATempI, ', FI=', ATempFi:1:3, ', S1=', ATempS1:1:0, ', S2=',ATempS2:1:0);

 ATempFi:= (ATempS1 *  * (16 / power(2, (IFinalNote / 12)));

 writeln(', result=', ATempFi:1:4);
}
 mpGetAmigaPeriod:= ATempFi;
end;


function mpGetLinearPeriod(IFinalNote, IFineTune:Single):Single;
begin
 mpGetLinearPeriod:= (7680 - (IFinalNote * 64) - (IFineTune / 2));
end;


procedure TJP_XMPlayer.mpCSetFreq(AChannel, AValue:Longint);
begin

 if (jssGetFlags(mpModule, jmdfAmigaPeriods)) then 
    begin
     if (AValue > 0) then
        mpDevice^.vmSetFreq(mpChanTable[AChannel], round((8363 * 1712) / single(AValue)) );
    end else

     mpDevice^.vmSetFreq(mpChanTable[AChannel], round(8363 * power(2, ((4608 - single(AValue)) / 768) ) ));
end;


procedure TJP_XMPlayer.mpCSetVolume(AChannel, AVolume:Longint);
begin
 { AVolume: 0..64
   FadeOut: 0..65536
   Envelop: 0..64
   Output : 0..255

   The 16.2 value is a fiddle to get it below 255. 
 }

 mpDevice^.vmSetVolume(mpChanTable[AChannel], 
 round((((mpCFadeOutVol[AChannel] * mpCEnvelopeVol[AChannel]) / 65536) * AVolume) / 16.12));
end;


procedure TJP_XMPlayer.mpCSetPanning(AChannel, APanning:Longint);
 var AFinalPan:Longint;
begin
 AFinalPan:= APanning + ( ((mpCEnvelopePan[AChannel]-32) * (128 - abs(APanning-128))) div 32);
 mpDevice^.vmSetPanning(mpChanTable[AChannel], AFinalPan);
end;


procedure TJP_XMPlayer.mpProcessExtendedInstrument(AChannel:Longint);

function mpFindPoint(AEnv:TJSSEnvelope; AFindFrame:Longint):Longint;
 var AFound:Boolean; AResult:Longint;
begin
 { Because there is ACTUALLY no other "easy" way to know 'between which env.points'
   we are on frame #n, we must perform a search of the point-table here.
 }
{$define ALWAYS_THE_SURE_THING}
{$ifdef ALWAYS_THE_SURE_THING}
 with AEnv do
 begin
  AResult:= 1;
  AFound:= FALSE;
  while (AResult < NPoints) and (AFound = FALSE) do
  begin
   if (APoints[AResult].AFrame <= AFindFrame) and
      (APoints[AResult+1].AFrame >= AFindFrame)
      then AFound:= TRUE
      else inc(AResult);
  end;
 end;

{$else}
 with AEnv do
 begin
  AResult:= (NPoints - 1);
  AFound:= FALSE;
  
  while ((AResult > 0) AND (AFound = FALSE)) do
  begin
   if (APoints[AResult].AFrame <= AFindFrame)
      then AFound:= TRUE
      else dec(AResult);
  end;
  
 end;
{$endif}

 { Return point number if found, else -1 to indicate failure }
 if (AFound)
    then mpFindPoint:= AResult
    else mpFindPoint:= -1;
end;


procedure mpExecEnvelope(AEnv:TJSSEnvelope; AKeyOff:Boolean;
                         var AFrames:Longint; var AExec:Boolean;
                         var AResult:Longint);
 var
    ACurrPoint:Longint;
    AP1, AP2:TJSSEnvelopePoint;
    ADelta:Single;
begin
 with AEnv do
 begin
   {** OK, find the current point based on frame # **}
   ACurrPoint:= mpFindPoint(AEnv, AFrames);

   {** Then check if the envelope has ended **}
   if ((AFlags AND jenvfLooped) = 0) and (ACurrPoint < 0) then
      begin
       AExec:= FALSE;
       exit;
      end;

   {** Do the envelope looping here, if needed **}
   if ((AFlags AND jenvfLooped) <> 0) and
       (AFrames >= APoints[ALoopE].AFrame) then
      begin
       ACurrPoint:= ALoopS;
       AFrames:= APoints[ALoopS].AFrame;
      end;

   {** If current point is OK, then process the envelope **}
   if (ACurrPoint >= 0) then
      begin
       {** Now we interpolate the value for given frame **}
       AP1:= APoints[ACurrPoint];
       AP2:= APoints[ACurrPoint + 1];

       ADelta:= (AP2.AFrame - AP1.AFrame);
       if (ADelta > 0)
          then ADelta:= ((AFrames - AP1.AFrame) / ADelta)
          else ADelta:= 0;

       AResult:= round(AP1.AValue + ((AP2.AValue - AP1.AValue) * ADelta));
      end;

   {** The frame counter IS processed even if the envelope is not! **}
   if ((AFlags AND jenvfSustain) <> 0) and
       (ACurrPoint = ASustain) and
       (APoints[ACurrPoint].AFrame = APoints[ASustain].AFrame) then
       begin
        if (AKeyOff) then
           inc(AFrames);
       end else
        inc(AFrames);
 end;
end;

 var
    ATempL:Longint;
    ATempEInst:PJSSExtInstrument;

begin
 {** Get the instrument for envelope data **}
 ATempEInst:= mpGetExtInstrument(AChannel);
 if (ATempEInst = NIL) or
    (ATempEInst^.Used = FALSE)
    then exit;

 {** Process the autovibrato **}
 with ATempEInst^ do
 begin
 {
  if (AVibrato
}
 end;

 {** Process the Volume Envelope **}
 if ((ATempEInst^.AVolumeEnvelope.AFlags AND jenvfUsed) = 0) then
    begin
     {** If the envelope is not used, set max volume **}
     mpCEnvelopeVol[AChannel]:= mpMaxVol;
     mpSetNDFlags(AChannel, cdfNewVolume);
    end else
 if (mpCEnvVol_Exec[AChannel]) then
    begin
     {** Process the instrument volume fadeout **}
     if (mpCKeyOff[AChannel]) and
        (mpCFadeOutVol[AChannel] > 0) and
        (ATempEInst^.AFadeOut > 0) then
        begin
         ATempL:= (mpCFadeOutVol[AChannel] - ATempEInst^.AFadeOut);
         if (ATempL < 0) then ATempL:= 0;
         mpCFadeOutVol[AChannel]:= ATempL;
        end;

     {** Execute the volume envelope **}
     mpExecEnvelope(ATempEInst^.AVolumeEnvelope, mpCKeyOff[AChannel],
                    mpCEnvVol_Frames[AChannel], mpCEnvVol_Exec[AChannel],
                    mpCEnvelopeVol[AChannel]);

     mpSetNDFlags(AChannel, cdfNewVolume);
    end;

 {** Process the Panning Envelope **}
 if ((ATempEInst^.APanningEnvelope.AFlags AND jenvfUsed) = 0) then
    begin
     {** If the envelope is not used, set middle panning **}
     mpCEnvelopePan[AChannel]:= jchPanMiddle;
     mpSetNDFlags(AChannel, cdfNewPanPos);
    end else
 if (mpCEnvPan_Exec[AChannel]) then
    begin
     {** Execute the panning envelope **}
     mpExecEnvelope(ATempEInst^.APanningEnvelope, mpCKeyOff[AChannel],
                    mpCEnvPan_Frames[AChannel], mpCEnvPan_Exec[AChannel],
                    mpCEnvelopePan[AChannel]);

     mpSetNDFlags(AChannel, cdfNewPanPos);
    end;
end;


procedure TJP_XMPlayer.mpProcessNewRow(AChannel:Longint);

 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;
    ANewPitch:Boolean;
    AOldPitch,
    ATempL:Longint;
    ATempNInst:PJSSInstrument;
    ATempEInst:PJSSExtInstrument;
    AFinalNote:Longint;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

procedure patProcessEffects;
begin
 {** Process the effects **}
 case AEffect of
  '0': ;
  
  'W': ; { Used widely in demo-music as MIDAS Sound System sync-command }
  'Q': ; { SoundTracker/OpenCP: Qxx = Set LP filter resonance   }
  'Z': ; { SoundTracker/OpenCP: Zxx = Set LP filter cutoff freq }

  '1',
  '2': { 1xy = Portamento Up, 2xy = Portamento Down : IMPL.VERIFIED }
       if (AParam <> 0) then
          mpLastPortaParam[AChannel]:= AParam;

  '3': { 3xy = Porta To Note }
       begin
        if (AParam <> 0) then
           mpLastPortaToNoteParam[AChannel]:= AParam;

        if (ACurrNote.Note <> mpNoData) then
           begin
            mpLastPortaToNotePitch[AChannel]:= mpCPitch[AChannel];
            mpCPitch[AChannel]:= AOldPitch;
            mpUnSetNDFlags(AChannel, (cdfNewPitch OR cdfNewInstr OR cdfNewPanPos OR cdfNewPos));
           end;
       end;

  '4': { 4xy = Vibrato : IMPL.VERIFIED }
       begin
        if (AParamX <> 0) then
           mpVibratoSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpVibratoDepth[AChannel]:= AParamY;

        if ((mpVibratoWC[AChannel] and $4) = 0)
           then mpVibratoPos[AChannel]:= 0;
       end;

  '5', { 5xy = Portamento + Volume Slide }
  '6': { 6xy = Vibrato + Volume slide }
       if (AParam <> 0) then
          mpLastVolSlideParam[AChannel]:= AParam;

  '7': { 7xy = Tremolo }
       begin
        if (AParamX <> 0) then
           mpTremoloSpeed[AChannel]:= AParamX;

        if (AParamY <> 0) then
           mpTremoloDepth[AChannel]:= AParamY;

        if ((mpTremoloWC[AChannel] and $4) = 0)
           then mpTremoloPos[AChannel]:= 0;
       end;

  '8': { 8xx = Set Panning }
       begin
        writeln('!! WARNING: Set Panning used  [ 8', ioHexB(AParam), ', ', AParam, ' ]');
       end;

  '9': { 9xx = Set Sample Offset : IMPL.VERIFIED }
       begin
        if (AParam <> 0)
           then mpLastSampleOffset[AChannel]:= AParam
           else patMakeParams(mpLastSampleOffset[AChannel]);

        ATempNInst:= mpGetInstrument(AChannel);
        if (ATempNInst <> NIL) and
           (ATempNInst^.Used) then
           begin
            ATempL:= (longint(AParam) SHL 8);
            if (ATempL > ATempNInst^.Size) then
               ATempL:= ATempNInst^.Size;

            mpCPosition[AChannel]:= ATempL;
            mpSetNDFlags(AChannel, cdfNewPos);
           end;
       end;

  'A': { Axy = Volume Slide : IMPL.VERIFIED }
       if (AParam <> 0)
          then mpLastVolSlideParam[AChannel]:= AParam;

  'B': { Bxx = Pattern Jump : IMPL.VERIFIED }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        mpNewOrder:= AParam;
        mpNewOrderGot:= TRUE;
        mpJumpFlag:= TRUE;

        mpLastPatLoopRow:= 0;
       end;

  'C': { Cxx = Set Volume : IMPL.VERIFIED }
       mpSetVolume(AChannel, AParam);

  'D': { Dxx = Pattern Break : IMPL.VERIFIED }
       if ((mpOptions AND moptAllowJumps) <> 0) then
       begin
        {** Compute the new row **}
        mpNewRow:= ((AParamX * 10) + AParamY);
        if (mpNewRow >= mpModule^.Patterns[mpPattern].ARows)
           then mpNewRow:= 0;

        mpNewRowGot:= TRUE;

        {** Now we do some tricky tests **}
        if (mpBreakFlag = FALSE) and
           (mpJumpFlag = FALSE) then
           begin
            mpNewOrder:= (mpOrder + 1);
            mpNewOrderGot:= TRUE;
           end;

        mpBreakFlag:= TRUE;
       end;

  'E': { Exy = Special Effects }
       case AParamX of
       { $00:   E0x - Set filter (NOT SUPPORTED) }

         $01: { E1x - Fine Portamento Up }
              begin
               if (AParamY <> 0) then
                  mpLastFinePortamentoUpParam[AChannel]:= AParamY;

               mpChangePitch(AChannel, -(longint(mpLastFinePortamentoUpParam[AChannel]) * 4));
              end;

         $02: { E2x - Fine Portamento Down }
              begin
               if (AParamY <> 0) then
                  mpLastFinePortamentoDownParam[AChannel]:= AParamY;

               mpChangePitch(AChannel, (longint(mpLastFinePortamentoDownParam[AChannel]) * 4));
              end;

       { $03:   E3x - Glissando Control (NOT SUPPORTED) }

         $04: { E4x - Set Vibrato waveform }
              mpVibratoWC[AChannel]:= (AParamY and $f);

       { $05:  E5x - Set Finetune }

         $06: { E6x - Set Pattern Loop }
              mpDoPatternLoop(AChannel, AParamY);

         $07: { E7x - Set Tremolo waveform }
              mpTremoloWC[AChannel]:= (AParamY and $f);

         $08: { E8x - Set Pan Position }
              begin
               {** Scale the panpos to internal values **}
               mpCPanning[AChannel]:= (AParamY * 16);
               mpSetNDFlags(AChannel, cdfNewPanPos);
              end;

         $09: { E9x - Retrig note }
              begin
               
               writeln('!! WARNING: Retrig Note (E9x) used  [ E9', AParamY, ' ]');
              end;

         $0A: { EAx - Fine Volumeslide Up }
              begin
               if (AParamY <> 0) then
                  mpLastFineVolumeslideUpParam[AChannel]:= AParamY;

               mpChangeVolume(AChannel, mpLastFineVolumeslideUpParam[AChannel]);
              end;

         $0B: { EBx - Fine Volumeslide Down }
              begin
               if (AParamY <> 0) then
                  mpLastFineVolumeslideDownParam[AChannel]:= AParamY;

               mpChangeVolume(AChannel, -mpLastFineVolumeslideDownParam[AChannel]);
              end;

       { $0C:   ECx - Set Note Cut (NOT PROCESSED IN TICK0) }

         $0D: { EDx - Set Note Delay : IMPL.VERIFIED }
              begin
               { Save the ND-flags, then clear them }
               mpSaveNDFlags[AChannel]:= mpCNewDataFlags[AChannel];
               mpCNewDataFlags[AChannel]:= 0;
              end;


         $0E: { EEx - Set Pattern Delay : IMPL.VERIFIED }
              mpPatternDelay:= AParamY;

       { $0F:   SFx - Invert Loop (NOT SUPPORTED) }
       end;

  'F': { Fxy = Set Speed / Tempo : IMPL.VERIFIED }
       if (AParam > 0) then
       if (AParam < $20)
          then mpSpeed:= AParam
          else mpSetNewTempo(AParam);

  'G': { Gxx = Global Volume }
       begin
        writeln('!! WARNING: Set Global Volume used  [ G', ioHexB(AParam), ', ', AParam, ' ]');
{
        mpGlobalVol:= AParam;
        mpNewGlobalVol:= TRUE;
}
       end;

  'H': { Hxx = Global Volume Slide }
       begin
        writeln('!! WARNING: Global Volume Slide used  [ H', ioHexB(AParam), ', ', AParam, ' ]');
       end;

  'K': { Kxx = Key-off (Same as key-off note) }
       mpCKeyOff[AChannel]:= TRUE;

  'L': { Lxx = Set Envelope Position }
       begin
        writeln('!! WARNING: Set Env Position used  [ L', ioHexB(AParam), ', ', AParam, ' ]');
        mpCEnvPan_Frames[AChannel] := AParam;
        mpCEnvVol_Frames[AChannel] := AParam;
        mpCEnvPan_Exec[AChannel]   := TRUE;
        mpCEnvVol_Exec[AChannel]   := TRUE;
       end;

  'R': { Rxy = Multi Retrig note }
       begin
        writeln('!! WARNING: Multi Retrig Note used  [ R', ioHexB(AParam), ', ', AParam, ' ]');
       end;

  'T': { Txy = Tremor }
       if (AParam <> 0) then
          mpLastTremorParam[AChannel]:= AParam;

  'X': { Xxy = Extra Fine Portamento }
       case AParamX of
         $01: { X1y - Extra Fine Portamento Up }
              begin
               if (AParamY <> 0) then
                  mpLastExtraFinePortamentoUpParam[AChannel]:= AParamY;

               mpChangePitch(AChannel, -mpLastExtraFinePortamentoUpParam[AChannel]);
              end;

         $02: { X2y - Extra Fine Portamento Down }
              begin
               if (AParamY <> 0) then
                  mpLastExtraFinePortamentoDownParam[AChannel]:= AParamY;

               mpChangePitch(AChannel, mpLastExtraFinePortamentoDownParam[AChannel]);
              end;
       end;

  else writeln('!! WARNING: Unsupported effect: ', ACurrNote.Effect, '/', AEffect, ioHexB(AParam), ' [row=', mpRow, ', pat=', mpPattern, ', chan=', AChannel, ']');
 end;
end;


begin
 {** Update note, sample, volumes, etc. **}
 ANewPitch:= FALSE;
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];
 AEffect:= mpMODEffectTable[ACurrNote.Effect];

 {** Check for new note/key-off here **}
 if (ACurrNote.Note = mpNoteOff) then
    begin
     {** Key-Off --> note released **}
     mpCKeyOff[AChannel]:= TRUE;
    end else
 if ((ACurrNote.Note >= 0) AND (ACurrNote.Note <= 96)) then
    begin
     {** New note was set **}
     ANewPitch:= TRUE;
     mpCNote[AChannel]:= ACurrNote.Note;
     mpCKeyOff[AChannel]:= FALSE;
     mpCFadeOutVol[AChannel]:= 65536;
    end;

 {** Check for new instrument **}
 if (ACurrNote.Instr <> mpNoData) and
    (ACurrNote.Instr >= 1) and
    (ACurrNote.Instr <= mpModule^.NExtInstruments) then
    begin
     {** Set new instrument **}
     ATempEInst:= mpModule^.ExtInstruments[ACurrNote.Instr];

     if (ATempEInst <> NIL) and (ATempEInst^.Used) then
        begin
         {** Check the sample instrument **}
         ATempL:= ATempEInst^.SNumForNotes[mpCNote[AChannel]];

         if (ATempL > 0) and (ATempL <= mpModule^.NInstruments) then
            begin
            ATempNInst:= mpModule^.Instruments[ATempL];

            if (ATempNInst <> NIL) and (ATempNInst^.Used) then
               begin
                {** Set the new sample-instrument **}
                mpCExtInstrument[AChannel]:= ACurrNote.Instr;
                mpCVolume[AChannel]:= ATempNInst^.Volume;
                mpCPanning[AChannel]:= ATempNInst^.EPanning;

                if (ATempL <> mpCInstrument[AChannel]) or
                   (ANewPitch) then
                   begin
                    ANewPitch:= TRUE;
                    mpCInstrument[AChannel]:= ATempL;
                    mpCPosition[AChannel]:= 0;
                    mpSetNDFlags(AChannel, (cdfNewInstr OR cdfNewPanPos OR cdfNewVolume OR cdfNewPos));
                   end else
                    mpSetNDFlags(AChannel, (cdfNewPanPos OR cdfNewVolume));
               end;
            end;
        end;
    end;

 {** Check for new instrument pitch **}
 if (ANewPitch) then
    begin
     {** Process the new pitch **}
     ATempEInst:= mpGetExtInstrument(AChannel);

     if (ATempEInst <> NIL) and
        (ATempEInst^.Used) then
     begin
      {** Reset the envelopes **}
      mpResetEnvelopes(AChannel);

      {** Get the new sample-instrument record **}
      ATempL:= ATempEInst^.SNumForNotes[mpCNote[AChannel]];

      if (ATempL > 0) and (ATempL <= mpModule^.NInstruments)
         then ATempNInst:= mpModule^.Instruments[ATempL]
         else ATempNInst:= NIL;

      {** Now check and init new instrument & frequency **}
      if (ATempNInst <> NIL) and
         (ATempNInst^.Used) then
         begin
          {** Save the old values for Tone Portamento effect **}
          AOldPitch:= mpCPitch[AChannel];

          {** Compute the new pitch **}
          AFinalNote:= (mpCNote[AChannel] + ATempNInst^.ERelNote - 1);
          if (AFinalNote < 0) then
             AFinalNote:= 0 else
          if (AFinalNote > 119) then
             AFinalNote:= 119;

          if (jssGetFlags(mpModule, jmdfAmigaPeriods)) 
             then mpCPitch[AChannel]:= round(mpGetAmigaPeriod(AFinalNote, ATempNInst^.EFineTune))
             else mpCPitch[AChannel]:= round(mpGetLinearPeriod(AFinalNote, ATempNInst^.EFineTune)); 

          mpSetNDFlags(AChannel, cdfNewPitch);
         end;
     end;
    end;

 {** Process the envelopes and autovibrato **}
 mpProcessExtendedInstrument(AChannel);

 {** ..then process the Volume Column effects **}
 patMakeParams(ACurrNote.Volume);
 case AParamX of
   $0..$4: mpSetVolume(AChannel, ACurrNote.Volume);

   $7:     { Dx = Fine Volumeslide Down : IMPL.VERIFIED }
           mpChangeVolume(AChannel, -AParamY);

   $8:     { Ux = Fine Volumeslide Up : IMPL.VERIFIED }
           mpChangeVolume(AChannel, AParamY);

   $9:     { Sx = Set vibrato speed : IMPL.VERIFIED }
           mpVibratoSpeed[AChannel]:= AParamY;

   $a:     { Vx = Vibrato : IMPL.VERIFIED }
           if (AParamY <> 0) then
              mpVibratoDepth[AChannel]:= AParamY;

   $e:     { Mx = Porta To Note : IMPL.VERIFIED }
           begin
            if (AParamY <> 0) then
               mpLastPortaToNoteParam[AChannel]:= AParamY;

            if (ACurrNote.Note <> mpNoData) then
               begin
                mpLastPortaToNotePitch[AChannel]:= mpCPitch[AChannel];
                mpCPitch[AChannel]:= AOldPitch;
                mpUnSetNDFlags(AChannel, (cdfNewPitch OR cdfNewInstr OR cdfNewPanPos OR cdfNewPos));
               end;
           end;
 end;

 {** ...And finally process the Normal effects **}
 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 AEffect:= mpMODEffectTable[ACurrNote.Effect];

 patProcessEffects;
end;


procedure TJP_XMPlayer.mpProcessEffects(AChannel:Longint);
 var
    ACurrNote:TJSSNote;
    AParam, AParamX, AParamY:Longint;
    AEffect:Char;
    ATempNInst:PJSSInstrument;
    ATempL,
    AFinalNote:Longint;

procedure patMakeParams(ANewParam:Byte);
begin
 AParam:= ANewParam;
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);
end;

procedure patProcessEffects;
begin
 {** Process the effects **}
 case AEffect of
  '0': { 0xy = Arpeggio }
       begin
        {** Compute arpeggio case **}
        ATempNInst:= mpGetInstrument(AChannel);

        if (ATempNInst <> NIL) and
           (ATempNInst^.Used) then
           begin
            {** Add the arpeggio value **}
            ATempL:= mpCNote[AChannel];
            case (mpTick MOD 3) of
             1: inc(ATempL, AParamX);
             2: inc(ATempL, AParamY);
            end;

            AFinalNote:= (ATempL + ATempNInst^.ERelNote);

{
            if (jssGetFlags(mpModule, jmdfAmigaPeriods)) 
               then mpCPitch[AChannel]:= round(mpGetAmigaPeriod(AFinalNote, ATempNInst^.EFineTune))
               else} mpCPitch[AChannel]:= round(mpGetLinearPeriod(AFinalNote, ATempNInst^.EFineTune));

            mpSetNDFlags(AChannel, cdfNewPitch);
           end;
       end;

  '1': { 1xy = Portamento Up }
       if (mpLastPortaParam[AChannel] > 0) then
          mpChangePitch(AChannel, -(mpLastPortaParam[AChannel] * 4));

  '2': { 2xy = Portamento Down }
       if (mpLastPortaParam[AChannel] > 0) then
          mpChangePitch(AChannel, (mpLastPortaParam[AChannel] * 4));

  '3': { 3xy = Porta To Note }
       mpDoPortamento(AChannel);

  '4': { 4xy = Vibrato }
       mpDoVibrato(AChannel, 5);

  '5': { 5xy = Portamento + Volume Slide }
       begin
        {** Do the portamento **}
        mpDoPortamento(AChannel);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);
       end;

  '6': { 6xy = Vibrato + Volume Slide }
       begin
        {** Do the vibrato **}
        mpDoVibrato(AChannel, 5);

        {** Do the volume slide **}
        mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);
       end;

  '7': { 7xy = Tremolo }
       mpDoTremolo(AChannel);

  'A': { Axy = Volume slide }
       mpDoVolumeSlide(AChannel, mpLastVolSlideParam[AChannel]);

  'E': { Exy = Special Effects }
       case AParamX of
         $0C: { ECx - Set Note Cut }
              if (mpTick = AParamY) then
                 mpSetVolume(AChannel, jsetMinVol);

         $0D: { EDx - Set Note Delay }
              if (mpTick = AParamY) then
                 mpCNewDataFlags[AChannel]:= mpSaveNDFlags[AChannel];
       end;

  'T': { Txy = Tremor }
       begin
        patMakeParams(mpLastTremorParam[AChannel]);
        inc(AParamX);
        inc(AParamY);
        AParam:= (mpTremorCount[AChannel] mod (AParamX + AParamY));
        if (AParam < AParamX)
           then mpSetVolume(AChannel, mpSVolume[AChannel])
           else mpSetVolume(AChannel, 0);

        inc(AParam);
        mpTremorCount[AChannel]:= AParam;
       end;

 end;
end;

begin
 {** Process the envelopes and autovibrato **}
 mpProcessExtendedInstrument(AChannel);

 {** ... then process the Volume Column effects **}
 ACurrNote:= mpModule^.Patterns[mpPattern].AData^[mpRow, AChannel];
 patMakeParams(ACurrNote.Volume);
 case AParamX of
   $5: { -x = Volumeslide Down : IMPL.VERIFIED }
       mpChangeVolume(AChannel, -AParamY);

   $6: { +x = Volumeslide Down : IMPL.VERIFIED }
       mpChangeVolume(AChannel, AParamY);

   $a: { Vx = Vibrato : IMPL.VERIFIED }
        mpDoVibrato(AChannel, 5);

   $e: { Mx = Porta To Note : IMPL.VERIFIED }
       mpDoPortamento(AChannel);
 end;

 {** ...And finally process the Normal effects **}
 if (ACurrNote.Effect = mpNoData) then exit;

 patMakeParams(ACurrNote.Param);

 AEffect:= mpMODEffectTable[ACurrNote.Effect];

 patProcessEffects;
end;


