{
  J Sound System - Module Playing Routines
  (C) Copyright 1999-2001 Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:
  This unit contains the JSS module player routines. The player system
  is naturally implemented in OOP way, so adding new players is 
  (relatively) easy.

  Player code for specific formats are contained in JPLAY*.PAS files.
}
{$I ../jssdef.pas}
unit JSSPlr;

interface

 uses BaseAPI,
      JSS,
      JSSMod;

 const
      {** Playing option-flags **}
      moptLoopModule    = $00000001;
      moptAllowJumps    = $00000002;

      {** Channel New-Data flags **}
      { These flags inform the main playing routine for new arrived data
        from the module-specific players.
        The child objects for TJSSModulePlayer class contain just methods
        mpProcessNewRow() and mpProcessEffects() [what they do, is
        self-explanatory ;) ] while the origin class methods handle
        all other functions! This greatly simplifies adding new formats!

        Even the most complex formats can be handled this way (like IT)
        just by setting these flags (cdfXXXX) and passing simple data!
      }
      cdfNONE           = $00;  { Set Nothing }
      cdfNewInstr       = $01;  { Set new Instrument }
      cdfNewPitch       = $02;  { Set new Pitch }
      cdfNewPos         = $04;  { Set new Position }
      cdfNewVolume      = $08;  { Set new Volume }
      cdfNewPanPos      = $10;  { Set new Panning Position }
      cdfStop           = $80;  { Stop channel playing }

      {** Some generals **}
      mpBaseDivisor     = 14317056;

      mpSineTable:array[0..31] of Longint = (
        0, 24, 49, 74, 97,120,141,161,
      180,197,212,224,235,244,250,253,
      255,253,250,244,235,224,212,197,
      180,161,141,120, 97, 74, 49, 24);

      mpRandTable:array[0..31] of Longint = (
      6, 56, 64, 11, 250, 167, 34, 4, 78, 208,
      0, 33, 91, 47, 185, 103, 228, 67, 101, 2,
      250,180,97, 24, 141,253,74, 36, 42, 79, 14, 12);


{$I jplay-i2.pas}


 type
     {** Basic module player object **}
     PJSSModulePlayer = ^TJSSModulePlayer;
     TJSSModulePlayer = object(TJSSCallBack)

mpcnt:Longint;

      {** General settings **}
      mpChannel,		 { Currently processed channel }
      mpTempo,                   { Current Tempo }
      mpSpeed,                   { Speed }
      mpTick,                    { Tick }
      mpOrder,                   { Order }
      mpPattern,                 { Pattern }
      mpRow,                     { Row }
      mpOptions:Longint;         { Playing options }
      mpPlaying:Boolean;         { Are we playing currently? }

      {** Efektien parametrit jne data **}
      mpNewOrder,                { NEW order number, if mpNewOrderGot TRUE }
      mpNewRow:Longint;          { NEW row number, if mpNewRowGot TRUE }
      mpNewOrderGot,             { TRUE if new order has been set }
      mpNewRowGot,               { TRUE if new row has been set }

      mpJumpFlag,                { Pattern jump flag }
      mpBreakFlag:Boolean;       { Pattern break flag }
      mpPatternDelay:Longint;    { PatternDelay tick-counter }

      mpPatLoopRow,              { PatternLoop starting row }
      mpPatLoopCount:TULongint;  { PatternLoop loop-counter }

      mpLastPortaParam:TUByte;   { Last porta effect parameter }
      mpLastPortaToNoteParam:TUByte;   { Porta-To-Note last parameter }
      mpLastPortaToNotePitch:TULongint;{ Porta-To-Note pitch }

      mpVibratoPos:TUShortInt;    { Current vibrato waveform position }
      mpVibratoSpeed,             { Current vibrato speed }
      mpVibratoDepth,             { Current vibrato depth }
      mpVibratoWC:TUByte;         { Current vibrato wave control }

      mpTremoloPos:TUShortInt;    { Current tremolo waveform position }
      mpTremoloSpeed,             { Current tremolo speed }
      mpTremoloDepth,             { Current tremolo depth }
      mpTremoloWC:TUByte;         { Current tremolo wave control }

      mpLastTremorParam,
      mpTremorCount,
      mpLastSampleOffset,
      mpLastRetrigParam,
      mpLastVolSlideParam:TUByte;

      mpLastPatLoopRow:Longint;  { Latest set pattern loop row (any channel) }

      mpRetrigNDFlags,           { For Retrig effect }
      mpSaveNDFlags:TUByte;      { For NoteDelay effect }

      {** General player channel defs **}
      mpCNewDataFlags:TUByte;    { New data flags, eli jos jotain uutta
                                   dataa pitisi asettaa sounddevicelle,
                                   tll on sitten tarvittava lippu ON.}
      mpCNote,                   { Current notes }
      mpCInstrument,             { Instruments }
      mpCPitch,                  { Pitches (NOT frequencies!) }
      mpCPosition:TULongint;     { Sample positions }
      mpCVolume:TUByte;          { Volumes }
      mpCPanning:TUByte;         { Panning positions }

      {** Saved values **}
      mpSVolume:TUByte;          { Volumes }

      {** Current module & sound device **}
      mpChanHandle:Longint;      { Channel allocation handle }
      mpChanTable:TJSSChannelTable; { Channel translation table }
      mpModule:PJSSModule;       { Current module of this player }
      mpDevice:PJSSSoundDevice;  { "Owner" sound device. This pointer is
                                   used to allocate channels and interact
                                   with the device. }

      {** General routines **}
      procedure mpSetNDFlags(AChannel:Longint; AFlags:Byte); virtual;
      procedure mpUnSetNDFlags(AChannel:Longint; AFlags:Byte); virtual;
      function  mpGetNDFlags(AChannel:Longint; AFlags:Byte):Boolean; virtual;

      function  mpGetInstrument(AChannel:Longint):PJSSInstrument; 

      procedure mpSetNewOrder(AOrder:Longint); virtual;
      procedure mpSetNewTempo(ATempo:Longint); virtual;
      procedure mpChangePitch(AChannel, ADelta:Longint); virtual;
      procedure mpChangeVolume(AChannel, ADelta:Longint); virtual;
      procedure mpSetVolume(AChannel, AValue:Longint); virtual;

      procedure mpCSetFreq(AChannel, AValue:Longint); virtual;
      procedure mpCSetVolume(AChannel, AVolume:Longint); virtual;
      procedure mpCSetPanning(AChannel, APanning:Longint); virtual;
      procedure mpDoPortamento(AChannel:Longint); virtual;
      procedure mpDoTremolo(AChannel:Longint); virtual;
      procedure mpDoVibrato(AChannel, AShiftValue:Longint); virtual;
      procedure mpDoVolumeSlide(AChannel, AParam:Longint); virtual;
      procedure mpDoPatternLoop(AChannel, AParamY:Longint); virtual;

      procedure mpProcessNewRow(AChannel:Longint); virtual;
      procedure mpProcessEffects(AChannel:Longint); virtual;
      procedure vmExec; virtual;


      procedure mpInitChannels; virtual;
      procedure mpClearChannels; virtual;
      
      procedure mpInitPlay; virtual;
      procedure mpPlayOrder(AStartOrder:Longint); virtual;
      procedure mpPlayPattern(APattern:Longint); virtual;
      procedure mpStop; virtual;
      procedure mpResume; virtual;

      procedure mpSetOptions(AOptions:Cardinal); virtual;
      procedure mpSetOption(AOptions:Cardinal); virtual;
      procedure mpSetModule(AModule:PJSSModule); virtual;

      constructor vmInit(ADevice:PJSSSoundDevice);
      destructor vmClose; virtual;
     end;


     {** Extended module player for XM, IT, etc **}
     PJSSExtendedPlayer = ^TJSSExtendedPlayer;
     TJSSExtendedPlayer = object(TJSSModulePlayer)
      {** Some variables **}
      mpCExtInstrument:TULongint;  { Extended Instruments }
      mpCFadeOutVol,
      mpCEnvelopePan,
      mpCEnvelopeVol:TULongint;
      mpCAutoVib_Frames,
      mpCEnvPan_Frames,
      mpCEnvVol_Frames:TULongint;
      mpCEnvPan_Exec,
      mpCEnvVol_Exec,
      mpCKeyOff:TUBoolean;

      {** The methods **}
      function  mpGetExtInstrument(AChannel:Longint):PJSSExtInstrument; 
      procedure mpResetEnvelopes(AChannel:Longint); virtual;
      procedure mpInitPlay; virtual;
     end;


{$define GET_INTERFACE}
{$I jssplr.inc}
{$undef GET_INTERFACE}

implementation

{ DOS/GO32V2: Start of code to be locked }
{$ifdef PLATFORM_DOS}
procedure jssLOCKCODE_START; begin end;
{$endif}



function TJSSModulePlayer.mpGetInstrument(AChannel:Longint):PJSSInstrument; 
 var ATempL:Longint;
begin
 ATempL:= mpCInstrument[AChannel];
 if (ATempL >= 1) AND (ATempL <= jsetMaxInstruments)
    then mpGetInstrument:= mpModule^.Instruments[ATempL]
    else mpGetInstrument:= NIL;
end;


function TJSSExtendedPlayer.mpGetExtInstrument(AChannel:Longint):PJSSExtInstrument;
 var ATempL:Longint;
begin
 ATempL:= mpCExtInstrument[AChannel];
 if (ATempL >= 1) AND (ATempL <= jsetMaxInstruments)
    then mpGetExtInstrument:= mpModule^.ExtInstruments[ATempL]
    else mpGetExtInstrument:= NIL;
end;


{****************************************************************************}
{*                                                                          *}
{* Base module player - General routines                                    *}
{*                                                                          *}
{****************************************************************************}
constructor TJSSModulePlayer.vmInit(ADevice:PJSSSoundDevice);
begin
 {** Initialize the player **}
 inherited vmInit;
 
 if (ADevice = NIL) then
    begin
     jssError('TJSSModulePlayer.mpInit', 'Given sound device was NULL!');
     exit;
    end;

 mpDevice       := ADevice;
 mpChanHandle   := jsetNotSet;
 mpPlaying      := FALSE;
 mpModule       := NIL;
end;


destructor TJSSModulePlayer.vmClose;
begin
 {** Deinitialize player **}
 mpStop;
 
 { We don't close the device }
 mpDevice:= NIL;

 {** Call the inherited vmClose **}
 inherited vmClose;
end;


procedure TJSSModulePlayer.mpSetOptions(AOptions:Cardinal);
begin
 {** Set player options **}
 mpOptions:= AOptions;
end;


procedure TJSSModulePlayer.mpSetOption(AOptions:Cardinal);
begin
 {** Set player options **}
 mpOptions:= (mpOptions OR AOptions);
end;


procedure TJSSModulePlayer.mpSetModule(AModule:PJSSModule);
begin
 jssMsg('TJSSModulePlayer.mpSetModule', 'Setting module 0x'+ ioPtr2Hex(AModule));
 mpModule:= AModule;
end;


{ Stops the currently allocated channels
  and de-allocates them.
}
procedure TJSSModulePlayer.mpClearChannels;
 var ATempInt:Longint;
begin
 if (mpModule <> NIL) AND
    (mpDevice <> NIL) AND
    (mpChanHandle <> jsetNotSet) then
    begin
     {** Stop currently playing channels **}
     for ATempInt:= 1 to mpModule^.NChannels do
         mpDevice^.vmClearChannel(mpChanTable[ATempInt]);

     {** De-allocate old channels **}
     mpDevice^.vmDeallocChannels(mpChanHandle);
    end;
end;


{ Allocate channels for player and initialize them (also gets 
  the channel translation table for allocated channels)
}
procedure TJSSModulePlayer.mpInitChannels;
 var ATempInt:Longint;
begin
 if (mpModule <> NIL) AND
    (mpDevice <> NIL) then
    begin
     {** Allocate channels and get translation table **}
     mpChanHandle:= mpDevice^.vmAllocChannels(mpModule^.NChannels);
     mpDevice^.vmGetChannelTransTable(mpChanHandle, mpChanTable);

     {** Clear the new channels **}
     for ATempInt:= 1 to mpModule^.NChannels do
         mpDevice^.vmClearChannel(mpChanTable[ATempInt]);
    end;
end;


{ (Re-)Initializes the player's basic structures.
  ALSO stops currently playing module!
}
procedure TJSSModulePlayer.mpInitPlay;
 var AIndex:Longint;
begin

mpcnt:= 0;

 {** Stop if already playing **}
 mpStop;
 
 {** Initialize general variables **}
 jssMsg('TJSSModulePlayer.mpInitPlay', 'Initializing new play sequence...');
 mpPatternDelay   := 0;
 mpNewRowGot      := FALSE;
 mpNewOrderGot    := FALSE;

 mpRow            := 0;
 mpTick           := -1;
 mpOrder          := 0;
 mpPattern        := 0;

 mpLastPatLoopRow := 0;

 mpInitChannels;

 {** Initialize channel data **}
 for AIndex:= jsetMinChannel to jsetMaxChannel do
     begin
      {** General channel variables **}
      mpCNote[AIndex]         := mpNoData;
      mpCInstrument[AIndex]   := mpNoData;
      mpCPitch[AIndex]        := 0;
      mpCVolume[AIndex]       := 0;
      mpCPanning[AIndex]      := jchPanMiddle;
     
      {** Initialize effect vars **}
      mpPatLoopRow[AIndex]    := 0;
      mpPatLoopCount[AIndex]  := 0;

      mpLastPortaParam[AIndex]:= 0;
      mpLastPortaToNoteParam[AIndex]:= 0;
      mpLastPortaToNotePitch[AIndex]:= 0;

      mpVibratoPos[AIndex]    := 0;
      mpVibratoSpeed[AIndex]  := 0;
      mpVibratoDepth[AIndex]  := 0;
      mpVibratoWC[AIndex]     := 0;

      mpTremoloPos[AIndex]    := 0;
      mpTremoloSpeed[AIndex]  := 0;
      mpTremoloDepth[AIndex]  := 0;
      mpTremoloWC[AIndex]     := 0;

      mpLastTremorParam[AIndex]  := 0;
      mpTremorCount[AIndex]      := 0;

      mpLastSampleOffset[AIndex] := 0;
      mpLastRetrigParam[AIndex]  := 0;
      mpLastVolSlideParam[AIndex]:= 0;

      mpSaveNDFlags[AIndex]      := 0;
     end;
end;


{ Starts playing module from a given ORDER.
}
procedure TJSSModulePlayer.mpPlayOrder(AStartOrder:Longint);
 const jProcName = 'TJSSModulePlayer.mpPlayOrder';
begin
 {** Do some checkings **}
{$ifdef OPT_DEBUG}
 if (mpModule = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Module pointer was NULL!');
     exit;
    end;

 if (mpDevice = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Device pointer was NULL!');
     exit;
    end;
{$endif}

 if (AStartOrder < 0) or
    (AStartOrder >= mpModule^.NOrders) then
    begin
     jssError(jProcName, 'Invalid playing startorder given.');
     exit;
    end;

 {** Initialize playing **}
 mpInitPlay;

 mpSetNewOrder(AStartOrder);

 if (mpOrder = -1) then
    begin
     jssError(jProcName, 'Could not start playing from given order!');
     exit;
    end;
    
 {** (Re-)Initialize callback **}
 if (mpPlaying) then
    begin
     {** If we are already playing **}
{$ifdef OPT_DEBUG}
     if (SCallBackSet = FALSE) then
        jssWarning(jProcName, 'Callback NOT set, but playing is ON?');
{$endif}
     mpSpeed:= mpModule^.DefSpeed;
     mpSetNewTempo(mpModule^.DefTempo)
    end else
    begin
     {** If NOT playing, install callback **}
{$ifdef OPT_DEBUG}
     if (SCallBackSet) then
        jssWarning(jProcName, 'Callback already set, but not playing?');
{$endif}
     
     mpDevice^.vmInstallCallBack(@Self);
     mpSpeed:= mpModule^.DefSpeed;
     mpSetNewTempo(mpModule^.DefTempo)
    end;
    
 mpPlaying:= TRUE;
end;


{ Plays a given PATTERN, then jumps to order 1
  (0 is the FIRST order) (NOTICE THAT WHEN USING THIS!!)
}
procedure TJSSModulePlayer.mpPlayPattern(APattern:Longint);
 const jProcName = 'TJSSModulePlayer.mpPlayPattern';
begin
 {** Do some checkings **}
{$ifdef OPT_DEBUG}
 if (mpModule = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Module pointer was NULL!');
     exit;
    end;

 if (mpDevice = NIL) then
    begin
     jssErrorN:= jerrNULLPointer;
     jssError(jProcName, 'Device pointer was NULL!');
     exit;
    end;
{$endif}

 {** Start playing a PATTERN **}
 mpInitPlay;

 {** Check the pattern number **}
 mpPattern:= APattern;
 if (mpPattern < 0) or
    (mpPattern > mpModule^.NPatterns) then
    begin
     jssError(jProcName, 'Invalid pattern number given #'+ ioStrL(mpPattern));
     exit;
    end;

 with mpModule^.Patterns[mpPattern] do
 if (AUsed = FALSE) or
    (AData = NIL) then
    begin
     jssError(jProcName, 'Invalid pattern number given #'+ ioStrL(mpPattern));
     exit;
    end;

 {** (Re-)Initialize callback **}
 if (mpPlaying) then
    begin
     {** If we are already playing **}
{$ifdef OPT_DEBUG}
     if (SCallBackSet = FALSE) then
        jssWarning(jProcName, 'Callback NOT set, but playing is ON?');
{$endif}
     mpSpeed:= mpModule^.DefSpeed;
     mpSetNewTempo(mpModule^.DefTempo)
    end else
    begin
     {** If NOT playing, install callback **}
{$ifdef OPT_DEBUG}
     if (SCallBackSet) then
        jssWarning(jProcName, 'Callback already set, but not playing?');
{$endif}
     
     mpDevice^.vmInstallCallBack(@Self);

     mpSetNewTempo(mpModule^.DefTempo)
    end;
    
 mpPlaying:= TRUE;
end;


{ Clears channels, stops playing, removes callback.
  All other data stays untouched. Use mpResume()
  to continue playing!
}
procedure TJSSModulePlayer.mpStop;
begin
 {** Clear and free the channels **}
 mpClearChannels;
 
 {** Stop the playing **}
 if (SCallBackSet) then
    mpDevice^.vmRemoveCallBack(SID);
    
 mpPlaying:= FALSE;
end;


{ Initializes channels, installs callback and
  continues playing where it was left (if mpStop()
  was used)
}
procedure TJSSModulePlayer.mpResume;
begin
 {** Initialize the callback **}
 mpClearChannels; { Just to be sure... }
 mpInitChannels;

 mpDevice^.vmInstallCallBack(@Self);

 vmSetFreq((mpTempo * 2) / 5);

 mpPlaying:= TRUE;
end;


{****************************************************************************}
{*                                                                          *}
{* Base module player - Playing routines                                    *}
{*                                                                          *}
{****************************************************************************}
{ Sets new order using given value as reference.
  Jumps over skip-points and invalid values, loops
  to first order if enabled.
}
procedure TJSSModulePlayer.mpSetNewOrder(AOrder:Longint);
 var AOrderOK:Boolean;
begin
 mpOrder:= AOrder;
 repeat
  AOrderOK:= TRUE;
  mpPattern:= mpModule^.OrderList[mpOrder];
  if (mpPattern = mpOrderSkip) or
     (mpModule^.Patterns[mpPattern].AData = NIL) or
     (mpModule^.Patterns[mpPattern].AUsed = FALSE) then
     begin
      inc(mpOrder);
      AOrderOK:= FALSE;
     end else

  if ((mpPattern = mpOrderEnd) OR
      (mpOrder >= mpModule^.NOrders)) then
     begin
      if ((mpOptions AND moptLoopModule) <> 0) then 
         begin
          mpOrder:= mpModule^.DefRestartPos;
          AOrderOK:= FALSE;
         end else
         begin
          mpStop;
          mpOrder:= -1;
         end;
     end;
 until (AOrderOK);
end;


{ Set new tempo-value of the player.
}
procedure TJSSModulePlayer.mpSetNewTempo(ATempo:Longint);
begin
 mpTempo:= ATempo;
 vmSetFreq((mpTempo * 2) / 5);
end;


{ Set value(s) of given New Data-flags.
}
procedure TJSSModulePlayer.mpSetNDFlags(AChannel:Longint; AFlags:Byte);
begin
 mpCNewDataFlags[AChannel]:= (mpCNewDataFlags[AChannel] OR AFlags);
end;


{ UNset value(s) of given New Data-flags.
}
procedure TJSSModulePlayer.mpUnSetNDFlags(AChannel:Longint; AFlags:Byte);
begin
 mpCNewDataFlags[AChannel]:= (mpCNewDataFlags[AChannel] AND (NOT (AFlags)));
end;


{ Get value of given New Data-flags. Returns TRUE if matches.
}
function  TJSSModulePlayer.mpGetNDFlags(AChannel:Longint; AFlags:Byte):Boolean;
begin
 mpGetNDFlags:= ((mpCNewDataFlags[AChannel] AND AFlags) = AFlags)
end;


{ Change volume of given channel by ADelta.
}
procedure TJSSModulePlayer.mpChangeVolume(AChannel, ADelta:Longint);
 var AValue:Longint;
begin
 {** Change the value and check it **}
 AValue:= (longint(mpCVolume[AChannel]) + ADelta);

 if (AValue < mpMinVol) then
    AValue:= mpMinVol else

 if (AValue > mpMaxVol) then
    AValue:= mpMaxVol;

 {** Set the volume **}
 mpCVolume[AChannel]:= AValue;
 mpSetNDFlags(AChannel, cdfNewVolume);
end;


{ Change the pitch of given channel by ADelta.
}
procedure TJSSModulePlayer.mpChangePitch(AChannel, ADelta:Longint);
 var AValue:Longint;
begin
 {** Change the pitch and check it **}
 AValue:= (mpCPitch[AChannel] + ADelta);
 if (AValue < 0) then
    AValue:= 0;

 {** Set the new pitch **}
 mpCPitch[AChannel]:= AValue;
 mpSetNDFlags(AChannel, cdfNewPitch);
end;


{ Set volume for given module channel.
}
procedure TJSSModulePlayer.mpSetVolume(AChannel, AValue:Longint);
begin
 {** Check the value of volume **}
 if (AValue < mpMinVol) then
    AValue:= mpMinVol else

 if (AValue > mpMaxVol) then
    AValue:= mpMaxVol;

 {** Set the volume **}
 mpCVolume[AChannel]:= AValue;
 mpSetNDFlags(AChannel, cdfNewVolume);
end;


{ Set frequency (pitch is given as argument) for a device channel.
  Channel is translated to device channels by this routine.
}
procedure TJSSModulePlayer.mpCSetFreq(AChannel, AValue:Longint);
begin
 {** Set frequency if pitch > 0 **}
 if (AValue > 0) then
    mpDevice^.vmSetFreq(mpChanTable[AChannel], (mpBaseDivisor div AValue));
end;


{ Set the volume for given module channel.
  Channel is translated to device channels by this routine.
}
procedure TJSSModulePlayer.mpCSetVolume(AChannel, AVolume:Longint);
begin
 { Scale 64 into maximum of 255 }
 mpDevice^.vmSetVolume(mpChanTable[AChannel], round(AVolume * 3.97));
end;


{ Set panning position for given module channel. Channel
  is translated to device channels by this routine.
}
procedure TJSSModulePlayer.mpCSetPanning(AChannel, APanning:Longint);
begin
 mpDevice^.vmSetPanning(mpChanTable[AChannel], APanning);
end;


{ Do a note portamento (pitch slide) effect for given module channel.
}
procedure TJSSModulePlayer.mpDoPortamento(AChannel:Longint);
begin
 {** Check for zero parameter **}
 if (mpLastPortaToNoteParam[AChannel] = 0) then
    begin
     mpSetNDFlags(AChannel, cdfNewPitch);
     exit;
    end;

 {** Do the Note Portamento **}
 { Slide the pitch of channel/sample to the destination value
   with speed of (mpLastPortaToNoteParam[] * 4) and stop when it equals.
 }
 if (mpCPitch[AChannel] <> mpLastPortaToNotePitch[AChannel]) then
    begin
     if (mpCPitch[AChannel] < mpLastPortaToNotePitch[AChannel]) then
        begin
         {** INCREASE pitch UP **}
         mpChangePitch(AChannel, (longint(mpLastPortaToNoteParam[AChannel]) shl 2));
         if (mpCPitch[AChannel] > mpLastPortaToNotePitch[AChannel])
            then mpCPitch[AChannel]:= mpLastPortaToNotePitch[AChannel];
        end else
        begin
         {** DECREASE pitch DOWN **}
         mpChangePitch(AChannel, -(longint(mpLastPortaToNoteParam[AChannel]) shl 2));
         if (mpCPitch[AChannel] < mpLastPortaToNotePitch[AChannel])
            then mpCPitch[AChannel]:= mpLastPortaToNotePitch[AChannel];
        end;
    end;
end;


{ Do a tremolo effect for given module channel.
}
procedure TJSSModulePlayer.mpDoTremolo(AChannel:Longint);
 var ATempL, ATDelta:Word;
begin
 {** Check parameters **}
 if (mpTremoloDepth[AChannel] = 0) or
    (mpTremoloSpeed[AChannel] = 0) then exit;

 {** Get position of tremolo waveform **}
 ATempL:= (mpTremoloPos[AChannel] and 31);

 {** Compute the delta value of the volume **}
 case (mpTremoloWC[AChannel] and 3) of
  0: ATDelta:= mpSineTable[ATempL];      { look up sine table }

  1: begin
      ATempL:= (ATempL shl 3);           { ramp down }
      if (mpTremoloPos[AChannel] < 0)
         then ATempL:= (255 - ATempL);
      ATDelta:= ATempL;
     end;

  2: ATDelta:= 255;                      { square }

  3: ATDelta:= mpRandTable[ATempL];      { random }
 end;

 ATDelta:= ((ATDelta * mpTremoloDepth[AChannel]) shr 6);

 {** Set the new volume **}
 if (mpTremoloPos[AChannel] >= 0) then
    begin
     if ((mpCVolume[AChannel] + ATDelta) > mpMaxVol)
        then mpCSetVolume(AChannel, mpMaxVol)
        else mpCSetVolume(AChannel, mpCVolume[AChannel] + ATDelta);
    end else
    begin
     if ((mpCVolume[AChannel] - ATDelta) < mpMinVol)
        then mpCSetVolume(AChannel, mpMinVol)
        else mpCSetVolume(AChannel, mpCVolume[AChannel] - ATDelta);
    end;

 {** Advance the tremolo waveform position **}
 mpTremoloPos[AChannel]:= (mpTremoloPos[AChannel] + mpTremoloSpeed[AChannel]);

 if (mpTremoloPos[AChannel] > 31)
    then mpTremoloPos[AChannel]:= -32;
end;


{ Do a vibrato effect for given module channel.
}
procedure TJSSModulePlayer.mpDoVibrato(AChannel, AShiftValue:Longint);
 var ATempL, ATDelta:Word;
begin
 {** Check parameters **}
 if (mpVibratoDepth[AChannel] = 0) or
    (mpVibratoSpeed[AChannel] = 0) then exit;

 {** Get position of vibrato waveform **}
 ATempL:= (mpVibratoPos[AChannel] and 31);

 {** Compute the delta value of the pitch **}
 case (mpVibratoWC[AChannel] and 3) of
  0: ATDelta:= mpSineTable[ATempL];      { look up sine table }

  1: begin
      ATempL:= (ATempL shl 3);           { ramp down }
      if (mpVibratoPos[AChannel] < 0)
         then ATempL:= (255 - ATempL);
      ATDelta:= ATempL;
     end;

  2: ATDelta:= 255;                      { square }

  3: ATDelta:= mpRandTable[ATempL];      { random }
 end;

 ATDelta:= ((ATDelta * mpVibratoDepth[AChannel]) shr AShiftValue);

 {** Set the new frequency **}
 if (mpVibratoPos[AChannel] >= 0)
    then mpCSetFreq(AChannel, mpCPitch[AChannel] + ATDelta)
    else mpCSetFreq(AChannel, mpCPitch[AChannel] - ATDelta);

 {** Advance the vibrato waveform position **}
 mpVibratoPos[AChannel]:= (mpVibratoPos[AChannel] + mpVibratoSpeed[AChannel]);

 if (mpVibratoPos[AChannel] > 31)
    then mpVibratoPos[AChannel]:= -32;
end;


{ Do a volume slide effect for given module channel.
}
procedure TJSSModulePlayer.mpDoVolumeSlide(AChannel, AParam:Longint);
 var AParamX, AParamY:Longint;
begin
 {** Make parameters **}
 AParamX:= ((AParam SHR 4) AND $F);
 AParamY:= (AParam AND $F);

 {** Do Volume Slide **}
 if (AParamY = 0) then
    mpChangeVolume(AChannel, AParamX);

 if (AParamX = 0) then
    mpChangeVolume(AChannel, -AParamY);
end;


{ Execute a pattern loop effect/command for given module channel.

  This routine works for most of the supported formats, as they
  use the 'standard' implementation ascending from MOD. However,
  here is included also a slightly kludgy implementation of the
  FT2 patloop bug.
}
procedure TJSSModulePlayer.mpDoPatternLoop(AChannel, AParamY:Longint);
begin
 {** Check what we need to do... **}
 if (AParamY > 0) then
    begin
     {** SBx/E6x loops <x> times **}
     if (mpPatLoopCount[AChannel] = 1) then
        mpPatLoopCount[AChannel]:= 0 else
        begin
         {** Check if we need to set the count **}
         if (mpPatLoopCount[AChannel] = 0) then
            mpPatLoopCount[AChannel]:= (AParamY + 1);

         {** Loop to the specified row **}
         dec(mpPatLoopCount[AChannel]);
         mpNewRow:= mpPatLoopRow[AChannel];
         mpNewRowGot:= TRUE;
        end;
    end else
    begin
     {** SB0/E60 sets the loop start point **}
     { Just set the row-to-loop to current row }
     mpPatLoopRow[AChannel]:= mpRow;

     { This is here for the infamous FT2 patloop bug }
     mpLastPatLoopRow:= mpRow;
    end;
end;


procedure TJSSModulePlayer.mpProcessNewRow(AChannel:Longint);
begin
 jssError('TJSSModulePlayer.mpProcessNewRow', 'Originator method called.');
end;


procedure TJSSModulePlayer.mpProcessEffects(AChannel:Longint);
begin
 jssError('TJSSModulePlayer.mpProcessEffects', 'Originator method called.');
end;


{ This is the main processing callback-loop of a module player.
  It processes the ticks, calling the needed mpProcessNewRow()
  and mpProcessEffects() methods for processing the module playing.

  This routine is almost totally format independent. (At least currently,
  but there DO exist formats that work differently and may need another
  instance of this kind of routine.)
}
procedure TJSSModulePlayer.vmExec;
 var
    AInstrument:PJSSInstrument;
    ANewDataFlags:Byte;
begin
 {** Sound playing callback **}
 { This procedure is called by the sound device.
   The main module playing system is here.
 }
inc(mpcnt);

 {** Clear the channel new data flags **}
 mpJumpFlag      := FALSE;
 mpBreakFlag     := FALSE;
 FillChar(mpCNewDataFlags, SizeOf(mpCNewDataFlags), cdfNONE);
 Move(mpCVolume, mpSVolume, SizeOf(mpSVolume));

 {** Check for init-tick **}
 if (mpTick < 0) then
    begin
     {** Init pattern **}
     if (mpOrder <> -1) then
        mpSetNewOrder(mpOrder);

     mpRow       := 0;
     mpNewRowGot := TRUE;
     mpTick      := mpSpeed;
    end;

 {** Check if playing has stopped **}
 if (mpPlaying = FALSE) then exit;

{$ifdef OPT_PLRDEBUG}
 writeln('?? ccnt=', mpCnt, ', row=', mpRow, ' : tick=', mpTick, ' : ord=', mpOrder, ' : pat=', mpPattern);
{$endif}

 {** Update the ticker **}
 inc(mpTick);
 if (mpTick >= mpSpeed) then
    begin
     {** Re-init tick counter **}
     mpTick:= 0;

     {** Check pattern delay **}
     if (mpPatternDelay > 0) then
        dec(mpPatternDelay) else
        begin
         {** New pattern row **}
         if (mpNewRowGot) then
            begin
             mpRow:= mpNewRow;
             mpNewRowGot:= FALSE;
            end else
             inc(mpRow);

         if (mpRow >= mpModule^.Patterns[mpPattern].ARows) then
            begin
             {** Go to next order **}
             if (mpOrder <> -1) then
                mpSetNewOrder(mpOrder + 1);

             if (jssGetFlags(mpModule, jmdfFT2Replay))
                then mpRow:= mpLastPatLoopRow
                else mpRow:= 0;
            end;

         if (mpNewOrderGot) then
            begin
             mpSetNewOrder(mpNewOrder);
             mpNewOrderGot:= FALSE;
            end;

         {** Check if playing has stopped **}
         if (mpPlaying = FALSE) then exit;

         {** TICK 0: Process the new row **}
         for mpChannel:= jsetMinChannel to mpModule^.NChannels do
             mpProcessNewRow(mpChannel);

        end; { IF mpPatternDelay }
    end else { IF mpTick }
    begin
     {** Implement FT2's pattern delay effect: Don't update effects while on patDELAY! **}
     if (jssGetFlags(mpModule, jmdfFT2Replay) = FALSE) OR
        ((jssGetFlags(mpModule, jmdfFT2Replay)) AND (mpPatternDelay <= 0)) then
        begin
         {** TICK n: Process the effects **}
         for mpChannel:= jsetMinChannel to mpModule^.NChannels do
            mpProcessEffects(mpChannel);
        end;
    end;

 {** Check if playing has stopped **}
 if (mpPlaying = FALSE) then exit;


 {** Update the player data to sound device **}
 for mpChannel:= jsetMinChannel to mpModule^.NChannels do
 if (mpCNewDataFlags[mpChannel] <> cdfNone) then
 begin
  ANewDataFlags:= mpCNewDataFlags[mpChannel];

  { The sample is about to stop }
  if ((ANewDataFlags AND cdfStop) <> 0) then
     mpDevice^.vmStop(mpChanTable[mpChannel])

  else

  begin
  if ((ANewDataFlags AND cdfNewInstr) <> 0) then
     begin
      {** Set new instrument sample **}
      AInstrument:= mpModule^.Instruments[mpCInstrument[mpChannel]];

      with AInstrument^ do
      mpDevice^.vmSetSample(mpChanTable[mpChannel], Data, Size, LStart, LEnd, Flags);

      mpDevice^.vmPlay(mpChanTable[mpChannel]);
     end;

  { New sample pitch }
  if ((ANewDataFlags AND cdfNewPitch) <> 0) then
     mpCSetFreq(mpChannel, mpCPitch[mpChannel]);

  { New sample offset position }
  if ((ANewDataFlags AND cdfNewPos) <> 0) then
     mpDevice^.vmSetPosition(mpChanTable[mpChannel], mpCPosition[mpChannel]);

  { New sample volume }
  if ((ANewDataFlags AND cdfNewVolume) <> 0) then
     mpCSetVolume(mpChannel, mpCVolume[mpChannel]);

  { New sample panning position }
  if ((ANewDataFlags AND cdfNewPanPos) <> 0) then
     mpCSetPanning(mpChannel, mpCPanning[mpChannel]);

  end; { if (NOT cfgStop) }

{$ifdef OPT_PLRDEBUG}
  {** Some debugging stuff here **}
  writeln('%% chan #', mpChannel, ':');
  if ((ANewDataFlags AND cdfNewInstr) <> 0) then
     with AInstrument^ do
     writeln(' - inst : 0x', ioPtr2Hex(Data), ', ', Size, ', [', LStart, '-', LEnd, '], 0x', ioHexB(Flags));

  if ((ANewDataFlags AND cdfNewPitch) <> 0) then
     writeln(' - speed: ', mpCPitch[mpChannel]);

  if ((ANewDataFlags AND cdfNewPos) <> 0) then
     writeln(' - pos  : ', mpCPosition[mpChannel]);

  if ((ANewDataFlags AND cdfNewVolume) <> 0) then
     writeln(' - vol  : ', mpCVolume[mpChannel]);

  if ((ANewDataFlags AND cdfStop) <> 0) then
     writeln(' - stop');
  writeln('%% ---------------');
{$endif}
 end;
end;


{****************************************************************************}
{*                                                                          *}
{* Extended module player routines                                          *}
{*                                                                          *}
{****************************************************************************}
{ Resets the envelopes for given channel.
}
procedure TJSSExtendedPlayer.mpResetEnvelopes(AChannel:Longint);
begin
 {** Resets the envelope processing for given channel **}
 mpCEnvPan_Frames[AChannel] := 0;
 mpCEnvVol_Frames[AChannel] := 0;
 mpCEnvPan_Exec[AChannel]   := TRUE;
 mpCEnvVol_Exec[AChannel]   := TRUE;
end;


{ Extra inits for Extended module player.
}
procedure TJSSExtendedPlayer.mpInitPlay;
 var AIndex:Longint;
begin
 {** Call inherited mpInitPlay **}
 inherited mpInitPlay;

 {** Initialize extended structures **}
 for AIndex:= jsetMinChannel to jsetMaxChannel do
     begin
      mpCExtInstrument[AIndex] := mpNoData;
      mpCEnvelopePan[AIndex]   := jchPanMiddle;
      mpCEnvelopeVol[AIndex]   := 0;
      mpCFadeOutVol[AIndex]    := 0;

      mpCAutoVib_Frames[AIndex]:= 0;
      mpCEnvPan_Frames[AIndex] := 0;
      mpCEnvVol_Frames[AIndex] := 0;
      
      mpCEnvPan_Exec[AIndex]   := FALSE;
      mpCEnvVol_Exec[AIndex]   := FALSE;
      mpCKeyOff[AIndex]        := FALSE;
     end;
end;


{****************************************************************************}
{*                                                                          *}
{* Specific module player routines                                          *}
{*                                                                          *}
{****************************************************************************}
{$define GET_IMPLEMENTATION}
{$I jssplr.inc}
{$undef GET_IMPLEMENTATION}

{$ifdef PLATFORM_DOS}
{ DOS/GO32V2: End of code to be locked }
procedure jssLOCKCODE_END; begin end;
{$endif}

end.
