{
  J Sound System - Sample handling/loading subsystem
  (C) Copyright 2000 Tecnic Software productions (TNSP)
  Programmed and designed by Matti "ccr" Hamalainen
  PLEASE READ FILE "README" FOR CONTACTING INFORMATION!

  LICENSE:
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation (should
  be found in file COPYING in main/root directory of J Sound
  System distribution package); either version 2.1 of the
  License, or any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
                
  You should have received a copy of the GNU Lesser General Public
  License (in file COPYING) along with this library; if not,
  write to the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA  02111-1307  USA.

  DESCRIPTION:

  Contains routines for loading sample data and streams from different formats.
}
{$I ../jssdef.pas}
unit JSSSamp;

interface

 uses BaseAPI,
      JSS;

 type
     {** A JSS sample structure **}
     PJSSSample = ^TJSSSample;
     TJSSSample = record
      AName:String40;
      ASize:Cardinal;
      AFlags:Longint;
      AData:Pointer;
     end;
     
     {** A JSS sample-stream object **}
     PJSSSampleStream = ^TJSSSampleStream;
     TJSSSampleStream = object
      SStream:PStream;
      SFlags:Longint;
      
      function vmGetAudioData(ASize:Longint; ABuffer:Pointer):Longint; virtual;
      constructor vmInit(AStream:PStream);
      destructor  vmClose; virtual;
     end;


     { RIFF WAVE stream }
     TJSSWAVStream = object(TJSSSampleStream)
      function vmGetAudioData(ASize:Longint; ABuffer:Pointer):Longint; virtual;
      constructor vmInit(AStream:PStream);
      destructor  vmClose; virtual;
     end;


     { MPEG1 audio layer 3 stream }
     TJSSMP3Stream = object(TJSSSampleStream)
      function vmGetAudioData(ASize:Longint; ABuffer:Pointer):Longint; virtual;
      constructor vmInit(AStream:PStream);
      destructor  vmClose; virtual;
     end;
     

 function  jssDisposeSample(var ASample:PJSSSample):Longint;
 function  jssSampLoadWAV(AStream:PStream; var ASample:PJSSSample):Longint;
 function  jssSampLoadRAW(AStream:PStream; ASize, AFlags:Longint; var ASample:PJSSSample):Longint;
 function  jssSampLoadIFF(AStream:PStream; var ASample:PJSSSample):Longint;  
 function  jssSampLoadMP3(AStream:PStream; var ASample:PJSSSample):Longint;


implementation


function jssDisposeSample(var ASample:PJSSSample):Longint;
begin
end;


{* 
  convert endianess from Motorola HI-LO format to Intel LO-HI format
*}
function vmConvEndianess(AValue:Word):Word;
begin
 vmConvEndianess:= swap(AValue);
end;


function vmConvEndianess(AValue:Integer):Integer;
begin
 vmConvEndianess:= swap(AValue);
end;


function vmConvEndianess(AValue:Longint):Longint;
 var AResult:Longint;
begin
 TDWordRec(AResult).HI:= swap(TDWordRec(AValue).LO); 
 TDWordRec(AResult).LO:= swap(TDWordRec(AValue).HI); 
 vmConvEndianess:= AResult;
end;



{* 
  RIFF WAVE format aka Windows WAV-file loader 
*}

 type
     TRIFFChunk = record
      ckName:String4;
      ckSize:Longint;
     end;

function jssLoadWAV(AStream:PStream; var ASample:PJSSSample):Longint;
begin
 {** Check that the stream is valid **}
 if (ioStreamValid(AStream) = FALSE) then
    jssError('fukken lamer');

 {** Check that it is a RIFF file **}
 AStream^.vmGetData();
 
end;


{*
  RAW format sample file loader
  format flags tell what kind of data we are talking about
  size is in UNITS (meaning: actual data length=ASize*UNITS,
  where UNITS are: 16/8-bit <-> stereo/mono)
*}
function jssLoadRAW(AStream:PStream; ASize, AFlags:Longint; var ASample:PJSSSample):Longint;
begin
 {** Check that the stream is valid **}
 if (ioStreamValid(AStream) = FALSE) then
    jssError('fukken');

 {** Ok, Initialize the sample structure **}
 if (jssNewSample(ASample, AFlags, ASize) = FALSE) then
    jssError('fukk');

 AStream^.vmGetData(ASample^.AData, ASample^.AByteSize);
end;


{*
  Amiga IFF format loader
  (much like RIFF WAVE since RIFF is a "Microsoftified version of IFF" ;)
*}
function jssLoadIFF(AStream:PStream; var ASample:PJSSSample):Longint;
begin
 {** Check that the stream is valid **}
 if (ioStreamValid(AStream) = FALSE) then
    jssError('fukken lamer');
    
 {** Check that it is a IFF file **}
 AStream^.
end;


{$ifdef SUP_MP3_FORMAT}
{*
  MPEG1 audio layer 3 loader
  Uses mpglib (for Pascal, ported from C by me ;)
  by Michael Hipp to load a MP3 format stream to a sample.
  Great for long samples and maybe in some 64k intros!!
*}
function jssLoadMP3(AStream:PStream; var ASample:PJSSSample):Longint;
begin
{
 mpglib_InitMP3();
 mpglib_DecodeMP3();
 mpglib_ExitMP3();
}
end;
{$endif}

end.
