{
  JDiSKP - Disk player utility for JSS
  Programmed and designed by Matti "ccr" Hamalainen.

  This program was the first "player" written for JSS to
  test the module-players and stuff. It has been improved
  after that quite a bit, but bugs may exist etc.

  NOTICE:
  This should be combined into jPL some day...
}
program JDiSKP;

 uses BaseAPI,
      JSS,
      JSSMod,
      JSSPlr;

 const
      vmProgName    = 'jDiSKP';          { Program name }
      vmVersion     = '0.90';
      vmMessage     = vmProgName +' v'+ vmVersion + ioCopyright;

      vmBits:Array[Boolean] of String[2] = ('8', '16');
      vmStereo:Array[Boolean] of String[6] = ('Mono', 'Stereo');
      vmInterpol:Array[Boolean] of String[15] = ('Normal', 'Interpolated');
      vmYesNo:Array[Boolean] of String[3] = ('No', 'Yes');

      vmNoteTable:Array[0..11] of String[2] =
      ('C-', 'C#', 'D-',
       'D#', 'E-', 'F-',
       'F#', 'G-', 'G#',
       'A-', 'A#', 'B-');

      SET_ModuleName:String  = '';     { Default module filename }
      SET_OutputFile:String  = 'result';
      SET_DebugMode:Boolean  = FALSE;
      SET_StartOrder:Longint = 0;
      SET_OutputTime:Longint = 60;
      SET_Opt_Loop:Boolean   = TRUE;
      SET_Opt_Jumps:Boolean  = TRUE;
      SET_MuteChannelsS:Longint = 0;

 var
    SET_MuteChannels:Array[jsetMinChannel..jsetMaxChannel] of Boolean;
    ATempStr:PStream;
    AModule:PJSSModule;
    AModuleLoader:Longint;
    APlayer:PJSSModulePlayer;
    ASaveExitProc:Pointer;
    ADevice:PJD_DiskWriter;
    AConfig:TJSSConfigFile;
    AExtra:Pointer;
    I:Longint;


procedure jplParseCommandline;

procedure wmsg(AStr:String);
begin
 writeln('! ', AStr);
end;

procedure wopt(AStr:String; ASetting:Boolean);
begin
 writeln('! ', AStr, ': ', vmYesNo[ASetting]);
end;

procedure jplParseOption(AOption:String);
 var
    ATempS:String;
    ATempI, APos, ACode:Longint;
    ASetting:Boolean;
begin
 ASetting:= (AOption[length(AOption)] <> '-');

 case AOption[2] of
  'd': SET_DebugMode:= ASetting;

  '1': AConfig.SOutMode16bit:= TRUE;
  '8': AConfig.SOutMode16bit:= FALSE;
  's': AConfig.SOutModeStereo:= TRUE;
  'm': AConfig.SOutModeStereo:= FALSE;
  'f': begin
        val(ioCopyFrom(AOption,3), AConfig.SOutFreq, ACode);
        if (ACode < 0) then
           ioInternalError(vmProgName, 'Invalid output frequency option "'+ AOption +'".');
        wmsg('Output frequency: '+ ioStrL(AConfig.SOutFreq) +' Hz.');
       end;

  'j': begin
        SET_Opt_Jumps:= ASetting;
        wopt('Allow jump-commands', SET_Opt_Jumps);
       end;

  'l': begin
        SET_Opt_Loop:= ASetting;
        wopt('Allow looping of module', SET_Opt_Loop);
       end;

  'o': begin
        val(ioCopyFrom(AOption,3), SET_StartOrder, ACode);
        if (ACode < 0) then
        ioInternalError(vmProgName, 'Invalid starting order option "'+ AOption +'".');
        wmsg('Starting order: '+ ioStrL(SET_StartOrder));
       end;

  't': begin
        val(ioCopyFrom(AOption,3), SET_OutputTime, ACode);
        if (ACode < 0) then
        ioInternalError(vmProgName, 'Invalid output time option "'+ AOption +'".');
        wmsg('Output time: '+ ioStrL(SET_OutputTime) +' seconds.');
       end;

  'w': begin
        SET_OutputFile:= ioParseFilename(ioCopyFrom(AOption,3), SET_OutputFile, '.raw');
        wmsg('Output file: "'+ SET_OutputFile +'"');
       end;

  'c': begin
        { Check the suboption }
        case AOption[3] of
         'o': SET_MuteChannelsS:= 1; 
         'm': SET_MuteChannelsS:= 2;
         else ioInternalError(vmProgName, 'Invalid sub-option for -c<X>[channels] ('+ AOption +')');         
        end;
        
        { Parse the given channel numbers }
        APos:= 4;
        while (APos <= length(AOption)) do
        begin
         ATempS:= ioGetNextItem(AOption, APos, [',', ':']);
         val(ATempS, ATempI, ACode);
        
         if (ACode < 0) or (ATempI < jsetMinChannel) or (ATempI > jsetMaxChannel) then
            ioInternalError(vmProgName, 'Invalid channel option "'+ AOption +'".');

         SET_MuteChannels[ATempI]:= TRUE;         
         inc(APos);
        end;
       end;

  else ioInternalError(vmProgName, 'Unknown option "'+ AOption +'".');
 end;
end;

 var
    AParameter:String;
    NParameter:Integer;

begin
 NParameter:= 1;
 while (NParameter <= ParamCount) do
 begin
  AParameter:= ParamStr(NParameter);
  if (AParameter[1] <> '-') then
     begin
      if (SET_ModuleName = '') then
         SET_ModuleName:= AParameter else
         ioInternalError(vmProgName, 'Too many filename-parameters in commandline!');
     end else
      jplParseOption(AParameter);

  inc(NParameter);
 end;
end;

procedure jplShowNote(APattern:TJSSPattern; ARow, AChannel:Longint);
begin
  if (APattern.AUsed = FALSE) or (APattern.AData = NIL) then exit;

  with APattern.AData^[ARow, AChannel] do
  begin
   case Note of
    mpNoData:  write('... ');
    mpNoteOff: write('--- ');
    else       write(vmNoteTable[Note mod 12], ioStrW(Note div 12), ' ');
   end;

   if (Instr <> $ff)
      then write(ioStrW0(Instr, 2), ' ')
      else write('.. ');

   if (Volume <> $ff)
      then write(ioStrW0(Volume, 2), ' ')
      else write('.. ');

   if (Effect <> $ff)
      then write(char(64 + Effect))
      else write('.');

   if (Effect <> $ff)
      then write(ioHexB(Param))
      else write('..');

   write('|');
 end;
end;


procedure jplExitProc;
begin
 {** Check for error **}
 if (ExitCode <> 0) or
    (SET_DebugMode) then
    begin
     writeln('o ExitCode: ', ExitCode);
     writeln('o General error information:');
     writeln('  JSS errorcode: ', jssErrorN, ' [0x', ioHexW(jssErrorN), ']');
     if (APlayer <> NIL) then
        with APlayer^ do
	begin
        writeln('  Order=', mpOrder, ', pat=', mpPattern, ', row=', mpRow,
                ', chan=', mpChannel, ' [tick=', mpTick, '/', mpSpeed, ' : tempo=', mpTempo, ']');

	if (AModule <> NIL) AND (mpPattern >= 0) then
	begin
	 write('  TrackNote: ');
	 jplShowNote(AModule^.Patterns[mpPattern], mpRow, mpChannel);
	 writeln;
	end;
	end;

     if (ADevice <> NIL) then
        with ADevice^ do
        begin
        writeln('  Device=', SDevice, ', baseport=0x', ioHexW(SBase),
                ', irq=', SIRQ, ', dma=', SDMA);
        writeln('  Freq=', SOutFreq, ' Hz, gvol=', SGlobalVol, ', chn=', SNChannels);
        end;
    end;

 {** Do some de-initializations **}
 if (APlayer <> NIL) then
    begin
     writeln('o Close player.');
     dispose(APlayer, vmClose);
     APlayer:= NIL;
    end;

 if (ADevice <> NIL) then
    begin
     writeln('o Close sound device.');
     dispose(ADevice, vmClose);
     ADevice:= NIL;
    end;

 if (AModule <> NIL) then
    begin
     writeln('o Dispose module.');
     jssDisposeModule(AModule);
     AModule:= NIL;
    end;

 writeln('o jssClose(): ', jssClose);

 if (ioClose <> ioerrOK) then
    writeln('o Error closing libBaseAPI!');

 ExitProc:= ASaveExitProc;
end;

procedure jplMsgProc(AFrom, AWhat:String);
begin
 writeln('- ', AFrom, ': ', AWhat);
end;

procedure jplSMsgProc(AFrom, AWhat:String);
begin
 write('- ', AFrom, ': ', AWhat);
end;


procedure jplWarningProc(AFrom, AMessage:String);
begin
 writeln('- Warning: ', AFrom, ': ', AMessage);
end;


procedure jplErrorProc(AFrom, AMessage:String);
begin
 writeln(AFrom, ': ', AMessage);
 write('JSS: Last errorcode #', jssErrorN);
 halt;
end;

 var ATempI:Longint;

begin
 {** Initialize program **}
 ASaveExitProc:= ExitProc;
 ExitProc:= @jplExitProc;
 jssMsg:= @jplMsgProc;
 jssMsgS:= @jplSMsgProc;
 jssError:= @jplErrorProc;
 jssWarning:= @jplWarningProc;
 AModule:= NIL;
 APlayer:= NIL;
 ADevice:= NIL;

 for ATempI:= jsetMinChannel to jsetMaxChannel do
     SET_MuteChannels[ATempI]:= FALSE;
     
 {** Set the predefined configuration **}
 with AConfig do
 begin
  SDevice         := sdDiskWriter;
  SBase           := -1;
  SIRQ            := -1;
  SlDMA           := -1;
  ShDMA           := -1;
  SOutMode16bit   := TRUE;
  SOutModeStereo  := FALSE;
  SOutFreq        := 44100;
  SMixInterpolate := FALSE;
 end;


 {** Show the help & startup texts **}
 writeln;
 writeln(vmMessage);
 writeln(jssMessage);
 if (ParamCount = 0) then
    begin
     writeln('Usage: ', ParamStr(0), ' [options] <module[.ext]>');
     writeln('Available module loaders:');
     for i:= 1 to jssGetLoaders do
         writeln('  ', jssGetLoaderInfo(i));

     writeln('Options:');
     writeln('-d      : Debugging mode.');
     writeln('-16/-8  : Set output resolution to 16/8 bits.');
     writeln('-s/-m   : Set output to Stereo or Mono.');
     writeln('-f<x>   : Set output/mixing frequency.');
     writeln('-i      : Use interpolated mixing.');
     writeln('-l[+/-] : Enable/disable looping of module.');
     writeln('-j[+/-] : Enable/disable jump commands (pat jump/pat break).');
     writeln('-o<x>   : Set start order.');
     writeln('-t<x>   : Set output time in Seconds to <x>.');
     writeln('-w<x>   : Set output file name to <x>.');
     writeln('-co<x>  : Output only specified channel numbers.');
     writeln('          Multiple channels can be specified: -cox,x,...,x');
     writeln('-cm<x>  : Mute specified channel numbers. Opposite of -co.');
     writeln('          Multiple channels can be specified: -cmx,x,...,x');
     halt;
    end;

{
 if (ioInit <> ioerrOK) then halt(255);
}

 {** Parse the commandline **}
 jplParseCommandline;

 {** Initialize J Sound System **}
 writeln('o jssInit(): ', jssInit);
 with AConfig do
 begin
  writeln('  ', vmBits[SOutMode16bit], '-bit ', vmStereo[SOutModeStereo], ' output.');
  writeln('  ', vmInterpol[SMixInterpolate], ' mixing with ', SOutFreq, ' Hz frequency.');
 end;

 {** Load the module file **}
 writeln('o Opening module file stream: ', SET_ModuleName);

 { Buffered version disabled until the reliability of PStream class has been proven clearly enough!
 ATempStr:= new(PBufferedFileStream, vmOpen(SET_ModuleName, ioOpen_RdOnly, 128*1024));
 }
 ATempStr:= new(PFileStream, vmOpen(SET_ModuleName, ioOpen_RdOnly, TRUE));
 if (ATempStr = NIL) then
    ioInternalError(vmProgName, 'Error allocating/opening stream.');

 AModuleLoader:= jssDetectModule(ATempStr, AExtra);
 writeln('o Autodetected: ', jssGetLoaderName(AModuleLoader));

 writeln('o Loading module ...');
 jssLoadModule(AModuleLoader, ATempStr, AModule, AExtra);
 if (AModule = NIL) then
    ioInternalError(vmProgName, 'Error loading module.');

 dispose(ATempStr, vmClose);
  

 {** Initialize sound device **}
 writeln('o Initializing sound device...');
 new(ADevice, vmInit(AConfig, SET_OutputFile));

 {** Allocate and initialize player **}
 writeln('o Initializing player...');

 case jssGetLoaderType(AModuleLoader) of

  jmdtS3M: APlayer:= new(PJP_S3MPlayer, vmInit(ADevice));
{
  jmdtPSM: APlayer:= new(PJP_PSMPlayer, vmInit(ADevice));
}
  jmdtMOD: APlayer:= new(PJP_MODPlayer, vmInit(ADevice));

  jmdtPTM: APlayer:= new(PJP_PTMPlayer, vmInit(ADevice));

  jmdtXM:  APlayer:= new(PJP_XMPlayer, vmInit(ADevice));

  else ioInternalError(vmProgName, 'Format is known and loaded, but there is NO PLAYER for it. :)');
 end;



 {** Start playing **}
 writeln('o PLAYING... Please wait...');
 APlayer^.mpSetModule(AModule);
 APlayer^.mpSetOptions(moptAllowJumps OR moptLoopModule);
 APlayer^.mpPlayOrder(SET_StartOrder); 

 {** Disable channels **}
 case SET_MuteChannelsS of
  0: ;
  
  1: begin
      for ATempI:= jsetMinChannel to jsetMaxChannel do
          ADevice^.vmMuteChannel(ATempI, SET_MuteChannels[ATempI]);
     end;

  2: begin
      for ATempI:= jsetMinChannel to jsetMaxChannel do
          ADevice^.vmMuteChannel(ATempI, (NOT (SET_MuteChannels[ATempI])));
     end;

  else ioInternalError(vmProgName, 'Internal error #foobar.');
 end;
 
 { Now execute the disk playing device }
 ADevice^.vmSetGlobalVol(60);

 ADevice^.vmExecute(SET_OutputTime);

 writeln('o OK.');
end.

