{$PackRecords 1}
program jTestMod;

 uses BaseAPI,
      JSS,
      JSSMod;

 const
      vmProgName = 'jTestMod';

      vmMinChan:Longint = 1;
      vmMaxChan:Longint = 32;

      vmYesNo:Array[Boolean] of String[3] = ('No ', 'Yes');

      vmNoteTable:Array[0..11] of String[2] =
      ('C-', 'C#', 'D-',
       'D#', 'E-', 'F-',
       'F#', 'G-', 'G#',
       'A-', 'A#', 'B-');


procedure vmShowPatternITS3M(APattern:TJSSPattern);
 var AChannel, ARow:Longint;
begin
 {** Check the given pattern **}
 if (APattern.AUsed = FALSE) or
    (APattern.AData = NIL)
    then exit;

 {** Show the pattern **}
 for ARow:= 0 to (APattern.ARows-1) do
 begin
  write(ioStrL0(ARow, 2), ': ');

  for AChannel:= vmMinChan to vmMaxChan do
  with APattern.AData^[ARow, AChannel] do
  begin
   case Note of
    mpNoData:  write('... ');
    mpNoteOff: write('--- ');
    else       write(vmNoteTable[Note mod 12], ioStrW(Note div 12), ' ')
   end;

   if (Instr <> mpNoData)
      then write(ioStrW0(Instr, 2), ' ')
      else write('.. ');

   if (Volume <> mpNoData)
      then write(ioStrW0(Volume, 2), ' ')
      else write('.. ');

   if (Effect <> mpNoData)
      then write(char(64 + Effect))
      else write('.');

   if (Param <> mpNoData)
      then write(ioHexB(Param))
      else write('..');

   write('|');
  end;

  writeln;
 end;
end;


procedure vmShowPatternXMMOD(APattern:TJSSPattern);
 var AChannel, ARow:Longint;
begin
 {** Check the given pattern **}
 if (APattern.AUsed = FALSE) or
    (APattern.AData = NIL)
    then exit;

 {** Show the pattern **}
 for ARow:= 0 to (APattern.ARows-1) do
 begin
  write(ioHexB(ARow), ': ');

  for AChannel:= vmMinChan to vmMaxChan do
  with APattern.AData^[ARow, AChannel] do
  begin
   case Note of
    mpNoData:  write('... ');
    mpNoteOff: write('=== ');
    else       write(vmNoteTable[Note mod 12], ioStrW(Note div 12), ' ')
   end;

   if (Instr <> mpNoData)
      then write(ioHexB(Instr), ' ')
      else write('.. ');

   case Volume of
    $00..$40: write(ioHexB(Volume));

    $50..$5f: write('-', ioHexDigits[(Volume AND $F)]);
    $60..$6f: write('+', ioHexDigits[(Volume AND $F)]);

    $70..$7f: write('/', ioHexDigits[(Volume AND $F)]);
    $80..$8f: write('\', ioHexDigits[(Volume AND $F)]);

    $90..$9f: write('s', ioHexDigits[(Volume AND $F)]);
    $a0..$af: write('v', ioHexDigits[(Volume AND $F)]);

    $b0..$bf: write('p', ioHexDigits[(Volume AND $F)]);

    $c0..$cf: write('<', ioHexDigits[(Volume AND $F)]);
    $d0..$df: write('>', ioHexDigits[(Volume AND $F)]);

    $e0..$ef: write('m', ioHexDigits[(Volume AND $F)]);

    mpNoData: write('..');
    else      write('??');
   end;
   write(' ');

   if (Effect <> mpNoData)
      then write(mpMODEffectTable[Effect])
      else write('.');

   if (Param <> mpNoData)
      then write(ioHexB(Param))
      else write('..');

   write('|');
  end;

  writeln;
 end;
end;


procedure vmShowEnvelope(APad:String; AEnv:TJSSEnvelope);
 var AIndex:Longint; AStr:String;
begin
 with AEnv do
 begin
  {** Interprete the envelope flags **}
  AStr:= '';
  if ((AFlags AND jenvfUsed) <> 0) then AStr:= AStr + 'Used, ';
  if ((AFlags AND jenvfSustain) <> 0) then AStr:= AStr + 'Sustain, ';
  if ((AFlags AND jenvfLooped) <> 0) then AStr:= AStr + 'Looped, ';
  AStr:= copy(AStr, 1, length(AStr)-2);

  {** Show the information **}
  writeln(APad, 'Flags...........: ', AFlags, ' [', AStr, ']');
  writeln(APad, 'NumPoints.......: ', NPoints);
  writeln(APad, 'Sustain point...: ', ASustain);
  writeln(APad, 'Loop Start point: ', ALoopS);
  writeln(APad, 'Loop End point..: ', ALoopE);
  write  (APad, 'Points..........: ');
  for AIndex:= 1 to NPoints do
      with APoints[AIndex] do
      begin
       write(ioStrL(AFrame), ':', ioStrl(AValue));
       if (AIndex < NPoints)
          then write(', ')
          else writeln;
      end;
 end;
end;


function vmOrderStr(NOrder, AOrder:Longint):String;
 var AStr:String;
begin
 case AOrder of
  mpOrderEnd: AStr:= '---';
  mpOrderSkip: AStr:= '+++';
  else AStr:= ioStrL0(AOrder, 3);
 end;
 vmOrderStr:= '#'+ ioStrL0(NOrder, 3) +': '+ AStr;
end;


 var
    AExtra,
    ASaveExitProc:Pointer;
    AModLoader,
    AModType:Longint;
    AModName:String;
    AModule:PJSSModule;
    AShowPatInfo,
    ANoInfo:Boolean;
    AStream:PStream;
    I:Longint;


procedure TestExitProc;
begin
 if (AModule <> NIL) then
    begin
    writeln('o Disposing module.');
    jssDisposeModule(AModule);
    end;

 if (AStream <> NIL) then
    dispose(AStream, vmClose);

 ExitProc:= ASaveExitProc;
end;


procedure TestWarningProc(AFrom, AMessage:String);
begin
 writeln('- ', AFrom, ': ', AMessage);
end;


procedure TestErrorProc(AFrom, AMessage:String);
begin
 writeln(AFrom, ': ', AMessage);
 write('JSS: Last errorcode [');
 case jssErrorN of
  jerrOK:                write('OK');
  jerrError:             write('General error');
  jerrInvalidArgument:   write('Invalid argument');
  jerrDivisionByZero:    write('Division by zero');
  jerrShutdownFailure:   write('General shutdown failure');
  jerrInitFailure:       write('General init failure');

  jerrFileOpenError:     write('File open error');
  jerrFileCreateError:   write('File creation error');
  jerrFileReadError:     write('File read error');
  jerrFileWriteError:    write('File write error');

  jerrDMAChannelNotDefined: write('DMA channel not defined');
  jerrInvalidDMAChannel:    write('Invalid DMA channel');
  jerrInvalidMixingFreq:    write('Invalid mixing frequency');

  jerrNULLPointer:       write('NULL structure/data pointer');
  jerrNonNULLPointer:    write('Non-NULL structure/data pointer');
  jerrMemAllocError:     write('Memory allocation error');
  else                   write('Unknown error code');
 end;
 writeln(']');
end;


begin
 ASaveExitProc:= ExitProc;
 ExitProc:= @TestExitProc;
 jssError:= @TestErrorProc;
 jssError:= @TestWarningProc;
 ioFailOnError:= FALSE;
 AModule:= NIL;
 AModName:= '';
 AModType:= 0;
 AShowPatInfo:= TRUE;
 ANoInfo:= FALSE;
 AStream:= NIL;

 writeln;
 writeln(vmProgName + ioCopyright);
 writeln(jssMessage);

 {** Ksittele parametrit **}
 if (ParamCount = 0) then
    begin
     writeln('Usage: testmod <filename.ext> [option]');
     writeln('-f : No pattern information.');
     writeln('-n : To just test the module loading.');
     writeln('Available loaders: (compiled in)');
     for I:= 1 to jssGetLoaders do
         writeln('  ', jssGetLoaderInfo(I));
     halt;
    end;

 if (ParamCount > 1) then
 if (ParamStr(2) = '-n') then ANoInfo:= TRUE else
 if (ParamStr(2) = '-f') then AShowPatInfo:= FALSE;

 {** Lataa moduuli **}
 AModName:= ioParseFilename(ParamStr(1), '', '.mod');
 writeln('o SOURCE modulefile: ', AModName);
 AStream:= new(PFileStream, vmOpen(AModName, ioOpen_RdOnly, TRUE));

 writeln('o Detecting SOURCE module type...');
 AModLoader:= jssDetectModule(AStream, AExtra);
 AModType:= jssGetLoaderType(AModLoader);

 writeln('o Loading ', jssGetLoaderName(AModLoader), '-format module...');
 jssLoadModule(AModLoader, AStream, AModule, AExtra);

 if (AModule = NIL) then
    ioInternalError(vmProgName, 'Could not load the module!');


 {** Nyt jotain infoa moduulista **}
 if (ANoInfo = FALSE) then
 with AModule^ do
 begin
  {** Show general module information **}
  writeln;
  writeln('General Module Information');
  writeln('==========================');
  writeln('Moduletype.........: ', ModuleType);
  writeln('Module name........: "', ModuleName, '"');
  writeln('Loader info........: ', ModuleInfo);
  writeln('Default Speed......: ', DefSpeed);
  writeln('Default Tempo......: ', DefTempo);
  writeln('Default GlobalVol..: ', DefGlobalVol);
  writeln('Default MasterVol..: ', DefMasterVol);
  
  {** Show flags information **}
  writeln('Default Flags......: ', DefFlags);

  if ((DefFlags AND jmdfAmigaPeriods) <> 0) then
     write('[AmigaPeriods] ');

  if ((DefFlags AND jmdfAmigaLimits) <> 0) then
     write('[AmigaLimits] ');
  
  if ((DefFlags AND jmdfStereo) <> 0) then
     write('[Stereo] ') else
     write('[Mono] ');
     
  if ((DefFlags AND jmdfFT2Replay) <> 0) then
     write('[FT2 Replay Emulation] ');
     
  writeln;
  
  {** More general information **}
  writeln('Channels...........: ', NChannels);
  writeln('Instruments........: ', NInstruments);
  writeln('Ext. Instruments...: ', NExtInstruments);
  writeln('Patterns...........: ', NPatterns);
  writeln('Orders.............: ', NOrders);

  {** Show orders-list contents **}
  writeln('Orders list:');
  for i:= 0 to ((NOrders-1) shr 1) do
  begin
   writeln(vmOrderStr(i, OrderList[i]), '    | ',
           vmOrderStr(i + 1 + ((NOrders-1) SHR 1), OrderList[i + 1 + ((NOrders-1) SHR 1)]));
  end;
 end;

 if (AShowPatInfo) and
    (ANoInfo = FALSE) then
 begin
  {** Show patterns information **}
  writeln;
  writeln('Pattern Data');
  writeln('============');
  for i:= 0 to (AModule^.NPatterns-1) do
      begin
       with AModule^.Patterns[i] do
       writeln('Pattern #', i, ' [USED: ', AUsed, ', ROWS: ', ARows, ', DATA: ', ioPtr2Hex(AData), ']');

       case AModType of
        jmdtS3M, jmdtPSM, jmdtIT: vmShowPatternITS3M(AModule^.Patterns[i]);
        jmdtPTM, jmdtMOD, jmdtXM: vmShowPatternXMMOD(AModule^.Patterns[i]);
       end;
      end;
 end;

 if (ANoInfo = FALSE) then
 begin
  {** Show extended instrument information **}
  writeln;
  writeln('Extended instruments');
  writeln('====================');
  for I:= 1 to AModule^.NExtInstruments do
  if (AModule^.ExtInstruments[I] <> NIL) then
     with AModule^.ExtInstruments[I]^ do
     begin
      writeln('INSTRUMENT #', ioStrW0(i, 3), ': "', Description, '"');
      if (Used) then
      begin
      writeln('NumSamples....: ', NSamples);
      writeln('Vib.type......: ', AVibratoType);
      writeln('Vib.sweep.....: ', AVibratoSweep);
      writeln('Vib.depth.....: ', AVibratoDepth);
      writeln('Vib.rate......: ', AVibratoRate);
      writeln('Volume FadeOut: ', AFadeOut);
      writeln('Volume Envelope information:');
      vmShowEnvelope(' - ', AVolumeEnvelope);
      writeln;
      writeln('Panning Envelope information:');
      vmShowEnvelope(' - ', APanningEnvelope);
      writeln;
      end;
     end;

  {** Show normal instrument information **}
  writeln('Normal instruments');
  writeln('==================');
  for I:= 1 to AModule^.NInstruments do
  if (AModule^.Instruments[I] <> NIL) then
     with AModule^.Instruments[I]^ do
     begin
      writeln(' #', ioStrL0(I, 3), ': ', vmYesNo[Used], ', "', Description, '", "', Filename, '"');
      writeln(' data:', ioPtr2Hex(Data), ', size:', ioHexD(Size), ' [', ioHexD(ByteSize), '], loop:', ioHexD(LStart), '-', ioHexD(LEnd));
      writeln(' size:', Size, ' [', ByteSize, '], loop:', LStart, '-', LEnd);
      writeln(' vol:', Volume, ', flags:', ioHexB(Flags), ', c4spd:', C4BaseSpeed, ', relnote:', ERelNote, ', ftune:', EFineTune, ', pan:', EPanning);
      writeln;
     end;
 end;
end.
